/************************************************************************/
/*       CellBlkMatrix: FDATA Block Matrix procedures for cells  	*/
/************************************************************************/
/* SCCS INFO @(#)CellBlkMatrix.c 1.2	last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 94/08/19 
*/
/* set of non-communicating matrix routines used for DBLAS.
 * These routines can be run on cells separately, but the operations
 * assume block-cyclic storage, eg. stride block copy routines.
 */
#include "CellBlkMatrix.h"
#include "FDAssert.h"  
#include "MachineParams.h"  
#include "MatDim.h"

#include "CellMatCopy.h"
/********************* rectangular block copy routines  *******************/ 

void BlkMatCopy(Add, M, N, A, alpha, B,
		RB0, RB, RBIncA, RBIncB, CB0, CB, CBIncA, CBIncB)
  int Add, M, N; FDATA *alpha; pLocMat A, B;
  int RB0, RB, RBIncA, RBIncB, CB0, CB, CBIncA, CBIncB;
{ int i, j;  int BlkDim;
  void (*CopyFn) STDC_PARAMS((int, int, pLocMat, FDATA *, pLocMat));
  FDATA *Ax = pLM_ADR(A), *Bx = pLM_ADR(B);
  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%dx%d,a%d %d %d,%s,a%d %d %d,%d %d %d|%d,%d %d %d|%d)",
     Add, M, N, FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A), pFD2Str(alpha), 
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), 
     RB0, RB, RBIncA, RBIncB, CB0, CB, CBIncA, CBIncB),
    pLM_BlkIncOK(M, N, B, RB0, RB, RBIncB, CB0, CB, CBIncB) &&
    pLM_BlkIncOK(M, N, A, RB0, RB, RBIncA, CB0, CB, CBIncA),
    0, "BlkMatCopy", 3, TRACE_COMPUTE
  ); 
  TRACE_LEVEL_SHIFT(1);			/* suppress traces in MatCopy()*/

  CopyFn = Add? MatAdd: MatCopy;
       
  BlkDim = 3;  				/* optimize for unit block stride;
  					   typically, at least 1 will be unit */
  if (M+RB0<=RB || (RBIncA==1 && RBIncB==1)) { 
    RB0 = 0; RB = 1;			RBIncA = RBIncB = 1;
    BlkDim-=2;
  }
  if (N+CB0<=CB || (CBIncA==1 && CBIncB==1)) { 
    CB0 = 0; CB = 1;			CBIncA = CBIncB = 1;	
    BlkDim--;
  }
  if (BlkDim == 0)			/* degenerate case of block copy */
    (*CopyFn)( M, N, A, alpha, B );        
  else {
/* rationale: for performance, need to minimize strides (by Blk,Blkj = 1)
 * while maintaining a sufficient primary vector length (=max(m, n)). 
 * The case BlkDim = {1, 2} (see above) is most important for this purpose. 
 * The macro USE_BLK_CPY() should achieve this except for the following cases:
 *	- if (RIncA,RIncB=1): RB just exceeds MIN_VECSIZE, but RB still small
 *	  and M/RB >> RB 		[primary vector length a bit too small] 
 *	- if (CIncA,CIncB=1, n large): RB < MIN_VECSIZE, but RB>=4 (say),
 *	  ie. RB is large enough for a secondary vector length 
 *					[could get more TLB misses]
 * and for the corresponding column-wise cases.
 */
    LocMat A_, B_;			/* matrices of expanded stride */
    int RIncA=pLM_RInc(A), 		CIncA=pLM_CInc(A),
        RIncB=pLM_RInc(B), 		CIncB=pLM_CInc(B);
    int Blki, Blkj; 			int ni, nj; 
    int VSRIncA, VSCIncA, 		VSRIncB, VSCIncB;
    
#define GET_STRIDE_BLK(Inc, B, BInc, RInc) 				\
    (Inc) = FAST_MUL(B, BInc); (Inc) = FAST_MUL2(Inc, RInc);
/* decide whether to copy as M/B blocks, or as B stride-vectors of length M/B */
#define USE_BLK_CPY(M, B, SVStride) 					      \
	 (((B)>=MIN_VECSIZE)  ||	/* block stride is large enough */    \
	  ((M)<=FAST_MUL1(B, B)) ||	/* block size>=size of stride-vector*/\
	  LM_InvalStride(SVStride))	/*SVStride is too large!!*/

    GET_STRIDE_BLK(VSRIncA, RB, RBIncA, RIncA);
    GET_STRIDE_BLK(VSRIncB, RB, RBIncB, RIncB);
    Blki = USE_BLK_CPY(M, RB, MAX(VSRIncA, VSRIncB));

    GET_STRIDE_BLK(VSCIncA, CB, CBIncA, CIncA);
    GET_STRIDE_BLK(VSCIncB, CB, CBIncB, CIncB);  
    Blkj = USE_BLK_CPY(N, CB, MAX(VSCIncA, VSCIncB));

    A = pMkLocMat(&A_, Ax, Blki? RIncA: VSRIncA, Blkj? CIncA: VSCIncA);
    B = pMkLocMat(&B_, Bx, Blki? RIncB: VSRIncB, Blkj? CIncB: VSCIncB);

    ni = Blki? FAST_DVD(M+RB0+RB-1, RB): RB;
    nj = Blkj? FAST_DVD(N+CB0+CB-1, CB): CB;
             
    for (i=0; i < ni; i++) {
      int Ni = N, 			CB0i = CB0;
      int m; 
      if (Blki) 
        m = MIN(M, RB - RB0), 		M -= m;			
      else {
	m = FAST_DVD(M-i + RB-1, RB);
        if (i+RB0 == RB) 
          Ax += VSRIncA - RIncA*RB, 	Bx += VSRIncB - RIncB*RB;
      }
      if (m==0) break;
      pLM_ADR(A) = Ax; 			pLM_ADR(B) = Bx;	
      for (j=0; j < nj; j++) {
        int n;
        if (Blkj) 
	  n = MIN(Ni, CB - CB0i), 	Ni -= n;
        else {
          n = FAST_DVD(N-j + CB-1, CB);	if (n==0) break;
          if (j+CB0==CB) 
            pLM_ADR(A) += VSCIncA - CIncA*CB,
            pLM_ADR(B) += VSCIncB - CIncB*CB;
        }
        if (n==0) break;
        (*CopyFn)( m, n, A, alpha, B );
        if (Blkj)
          pLM_ADR(A) += VSCIncA - FAST_MULZ(CB0i, CIncA),
	  pLM_ADR(B) += VSCIncB - FAST_MULZ(CB0i, CIncB),	CB0i = 0;
	else
          pLM_ADR(A) += CIncA, 		pLM_ADR(B) += CIncB;		    
      } /* for (j...) */
      if (Blki) 
 	Ax += VSRIncA - FAST_MULZ(RB0, RIncA), 	
	Bx += VSRIncB - FAST_MULZ(RB0, RIncB),	RB0 = 0;  
      else
	Ax += RIncA,			Bx += RIncB;
    } /* for (i...) */ 
  }
  TRACE_LEVEL_SHIFT(-1);
  TRACE_EXIT;
} /*BlkMatCopy()*/
 

