/************************************************************************/
/*       	CellBlkMatrix: FDATA Matrix procedures for cells   */
/************************************************************************/
/* SCCS INFO @(#)CellBlkMatrix.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 93/03/23 
*/
/* set of non-communicating matrix routines used for DBLAS.
 * These routines can be run on cells separately, but the operations
 * assume block-cyclic storage, eg. stride block copy routines.
 * See CellMatrix.h for description of matrix strides.
 * A block-stride is defined by:
 *       B0 (offset within 1st block),
 *       B  (width of block)
 *       BInc (block-stride)
 * See BlkIncOK() for contraints in the stride abd block parameters.
 */
#ifndef __CellBlkMatrix_INCL__
#define __CellBlkMatrix_INCL__  

#include "FDATASetUp.h"
#include "LocMat.h"
#include "MatDim.h"

#define LM_BlkIncOK(M, N, A, RB0, RB, RBInc, CB0, CB, CBInc) 		\
	BlkIncOK(M, N, LM_RInc(A), LM_CInc(A), RB0, RB, RBInc, CB0, CB, CBInc) 

#define pLM_BlkIncOK(M, N, A, RB0, RB, RBInc, CB0, CB, CBInc) 		\
	BlkIncOK(M, N, pLM_RInc(A), pLM_CInc(A), RB0, RB, RBInc, CB0, CB, CBInc) 


#define BlkMatStridePack(Add, M, N, A, alpha, B, RB0, RB, RBInc, CB0, CB, CBInc)\
	BlkMatCopy(Add, M, N, A, alpha, B, RB0, RB, RBInc, 1, CB0, CB, CBInc, 1)

#define BlkMatStrideUnpack(Add, M, N, A, alpha, B, RB0, RB, RBInc, CB0, CB, CBInc)\
	BlkMatCopy(Add, M, N, A, alpha, B, RB0, RB, 1, RBInc, CB0, CB, 1, CBInc)


#ifdef CELLBLAS		/* bypass this module; assume RBInc* = CBInc* = 1) */

#define BlkIncOK(M, N, RInc, CInc, RB0, RB, RBInc, CB0, CB, CBInc) 	1

#define BlkMatCopy(Add, M, N, A, alpha, B, RB0, RB, RBIncA, RBIncB,
                  CB0, CB, CBIncA, CBIncB)
        (((Add)? MatAdd: MatCopy)(M, N, A, alpha, B))          

#else

# define BlkIncOK(M, N, RInc, CInc, RB0, RB, RBInc, CB0, CB, CBInc) 	\
        /* ensures no 2 elts of block-strided matrix map to self */     \
        (0<=RB0 && RB0<RB  &&  0<=CB0 && CB0<CB && 0<RBInc && 0<CBInc &&\
         (RInc > L2G(N-1, CB0, CB, 0, CBInc)*CInc ||                  	\
          CInc > L2G(M-1, RB0, RB, 0, RBInc)*RInc ))

# if FDATATYPE==SINGLE
#   define BlkMatCopy		DSBlkMatCopy
# elif FDATATYPE==DOUBLE
#   define BlkMatCopy		DDBlkMatCopy
# elif FDATATYPE==COMPLEX
#   define BlkMatCopy		DCBlkMatCopy
# else
#   define BlkMatCopy		DZBlkMatCopy
# endif

void BlkMatCopy STDC_PARAMS( 
	(int Add, int M, int N, pLocMat A, FDATA *alpha, pLocMat B,
	 int RB0, int RB, int RBIncA, int RBIncB, 
	 int CB0, int CB, int CBIncA, int CBIncB)
);
/* pre:  LM_BlkIncOK(M, N, A, RB0, RB, RBIncA, CB0, CB, CBIncA) &&  
 *       LM_BlkIncOK(M, N, B, RB0, RB, RBIncB, CB0, CB, CBIncB)
 * post: B[iB,jB] = (Add)*B[iB,jB] + (*alpha)*A[iA,jA], 0<=i<M, 0<=j<N
 *       where iA=L2G(i, RB0, RB, 0, RBIncA), jA = L2G(j, RB0, RB, 0, RBIncA)
 *       where iB=L2G(i, RB0, RB, 0, RBIncB), jB = L2G(j, RB0, RB, 0, RBIncB)
 */

#endif /* CELLBLAS */

#endif /*__CellBlkMatrix_INCL__*/
