/************************************************************************/
/*       CellMatMult: FDATA cell matrix multiply procedure		*/
/************************************************************************/
/* SCCS INFO @(#)CellMatMult.c 1.3         last modified P. Strazdins, 99/05/31
   first created: P. Strazdins 96/01/16 
*/
/* provides a cell MM routine interfacing to the BLASInnerLoops kernels.
 * This procedure is suitable for scalr-based architectures like the SPARC
 *
 * compilation: BLASInnerLoops procedures turned on.
 *		if macro MM_VANILLA defined, vanilla version of MatMult() is 
 *		invoked, and BLASInnerLoops module is not referenced.
 */
/*#define MM_VANILLA*/
 
#include "CellMatrix.h"
#include "BLASInnerLoops.h"

#include "FDAssert.h"

void MatMult(M, N, K, alpha, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta;  pLocMat A, B, C;
/* pre:  pLM_IncOKUnit(M, K, A) && pLM_IncOKUnit(K, N, B) &&
 *       pLM_IncOKUnit(M, N, C) && 0 <= M, N, K,
 *       !MatOverlap(M, K, A, M, N, C) && !MatOverlap(K, N, B, M, N, C)
 * post: C := C + (*alpha)A*B;  A is MxK, B is KxN, C is MxN;
 */
{ register int j, n, LdC;  FDATA betaf;
  register int RIncA=pLM_RInc(A), CIncA=pLM_CInc(A), 
	       RIncB=pLM_RInc(B), CIncB=pLM_CInc(B);
  FDATA *Ax = pLM_ADR(A), *Bx= pLM_ADR(B), *Cx = pLM_ADR(C);
  int i, m;			int M0 = M, N0 = N;
  int bounds[3];		/* stride bounds for matrices */

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%dx%d,%s,a%d %d %d,a%d %d %d,%s,a%d %d %d)",
     M, N, K, pFD2Str(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A),
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD2Str(beta),
     FD_ADR(pLM_ADR(C)), pLM_RInc(C), pLM_CInc(C)),
    (N==0 || pLM_IncOKUnit(M, K, A)) && (M==0 || pLM_IncOKUnit(K, N, B)) &&
    pLM_IncOKUnit(M, N, C) && 0<=M && 0<=N && 0<=K, 
    MatOverlap(M, K, A, M, N, C) + 2*MatOverlap(K, N, B, M, N, C),
    "MatMult", 3, TRACE_COMPUTE
  );

  if (alpha!=NULL && FD_EQ(*alpha, FD_0)) {
    MatScale(M, N, beta, C);
    TRACE_RETURN;
  }  
#if defined(MM_VANILLA) || (FDATATYPE>=COMPLEX)
# include "FDATAOps.h"
  { int k;
  if (alpha==NULL) alpha = &FD_1; 
  betaf = beta==NULL? FD_1: *beta;
  for (i = 0; i < M; i++) {
    FDATA *b = pLM_ADR(B), *cx = Cx;
    for (j=0; j < N; j++) {   
      FDATA *ax=Ax, *bx=b; FDATA c = FD_0;
      for (k=0; k < K; k++) {
        FD_MULSET(c, +=, *ax, *bx);
        ax += pLM_CInc(A); bx += pLM_RInc(B);
      }
      FD_MULSET(*cx, =, *cx, betaf);
      FD_MULSET(*cx, +=, c, *alpha);
      b += pLM_CInc(B);  cx += pLM_CInc(C);
    } /* for (j...) */
    Cx += pLM_RInc(C); Ax+= pLM_RInc(A);
  } /* for (i...) */

#if 0				/* include to enforce linking in of inner loops code */
  { int bounds[] = {1, 1, 1}; 
  VecMatMul(0, 0, bounds, pLM_ADR(A), pLM_ADR(B), pLM_ADR(C));
  Axpy1(0, pLM_ADR(A),  pLM_ADR(B), 1, pLM_ADR(C), 1);
  }
#endif  
  }     
#else

  if (!pFD_UNITY(alpha) && K>0) {
    betaf = beta==NULL? FD_1: *beta;	
    FD_DVDSET(betaf, betaf, *alpha);
    beta = &betaf;
  }  
  if (!pFD_UNITY(beta))
    MatScale(M, N, beta, C);
  
  LdC = MAX(pLM_RInc(C), pLM_CInc(C));
  
  if (pLM_RInc(C) == 1) {	/* transpose computation to make CIncC=1 */
    SWAP(FDATAPtr, Ax, Bx);
    SWAP(int, M, N);	
    SWAP(int, RIncA, CIncB);	SWAP(int, CIncA, RIncB);	
  }
    	
  if (K >= MIN_UPDATE_RECT_K) { 
#define RECT_i(a)	(tAB==3? (a RECT_X): (a RECT_Y))
#define RECT_j(a)	(tAB==3? (a RECT_Y): (a RECT_X))
    int tAB = (RIncA==1) + 2*(RIncB==1);
    int di = RECT_i( ),
    	dj = RECT_j( ),
    	Ainc= RECT_i(RIncA *), Binc=RECT_j(CIncB *), Cinc=RECT_i(LdC *); 
    FDATA *b;
         	                      
    /* base multiply on updating each di x dj sub-block;
       remaining cols/rows of C updated using matrix-vector / vector-matrix mult*/ 

    m = M - RECT_i(M %);  n = N - RECT_j(N %); bounds[2] = LdC; 
    if (n < N) {
      b = Bx + (n==0? 0: n*CIncB);	
      if (RIncA==1) {      	
        bounds[0] = RIncB; bounds[1] = CIncA; 
      } else {
	bounds[0] = RIncA; bounds[1] = RIncB; 
      }
      for (j=n; j < N; j++) {
        if (RIncA==1) {
          VecMatMul(K, M, bounds, b, Ax, Cx+j);
        } else {
	  MatVecMul(M, K, bounds, Ax, b, Cx+j);
        }
        b += CIncB;
      }
      N = n;
    }
    bounds[0] = RIncA==1? CIncA: RIncA; bounds[1] = RIncB==1? CIncB: RIncB; 
    if (tAB == 3) {
      SWAP(int, bounds[0], bounds[1]);
    }
#if 0
{ extern int mpl_cellid    ;
if (m>0 && n>0 /*&& mpl_cellid==1*/ )
printf("%d", tAB);       
    }
#endif
    for (i=0; i < m; i+=di) {
      b = Bx;
      for (j=0; j < n; j+=dj) {
      	if (tAB==1) {
#if 1
      	  UpdateRectTr1(K, bounds, Ax, b, Cx+j);
#else
	  int i1, j1, k;
	  for (i1=0; i1 < di; i1++)
	    for (j1=0; j1 < dj; j1++)
	      for (k=0; k < K; k++)
	        *(Cx+i1*bounds[2]+j1+j) += *(Ax + k*bounds[0] + i1) *
	        			   *(b +  k*bounds[1] + j1);
          if (bounds[1] < 0)
      	    UpdateRectTr1(0, bounds, Ax, b, Cx+j);
#endif      	  
      	} else if (tAB==0) {
#if 1
      	  UpdateRect(K, 0, bounds, Ax, b, Cx+j);  
#else
	  int i1, j1, k;
	  for (i1=0; i1 < di; i1++)
	    for (j1=0; j1 < dj; j1++)
	      for (k=0; k < K; k++)
	        *(Cx+i1*bounds[2]+j1+j) += *(Ax + i1*bounds[0] + k) *
	        			   *(b +  k*bounds[1] + j1);
	  if (bounds[1] < 0)
	    UpdateRect(0, 0, bounds, Ax, b, Cx+j);	        			   
#endif
      	  
      	} else if (tAB==2) {
#if 1
      	  UpdateRectTr2(K, bounds, Ax, b, Cx+j);
#else
	  int i1, j1, k;
	  for (i1=0; i1 < di; i1++)
	    for (j1=0; j1 < dj; j1++)
	      for (k=0; k < K; k++)
	        *(Cx+i1*bounds[2]+j1+j) += *(Ax + i1*bounds[0] + k) *
	        			   *(b +  j1*bounds[1] + k);
          if (bounds[1] < 0) 
      	    UpdateRectTr2(0, bounds, Ax, b, Cx+j);
#endif      	  
      	} else {
#if 1     		
      	  UpdateRect(K, 1, bounds, b, Ax, Cx+j);
#else
	  int i1, j1, k;
	  for (i1=0; i1 < di; i1++)
	    for (j1=0; j1 < dj; j1++)
	      for (k=0; k < K; k++)
	        *(Cx+j1*bounds[2]+i1+j) += *(Ax + k*bounds[1] + j1) *
	        			   *(b +  i1*bounds[0] + k);
      	  
#endif
      	}  
        b += Binc;                	  
      }
      Cx += Cinc; Ax+= Ainc;
    }
    if (m < M) {			/* A, C now at [m,0] */
      bounds[2] = 1;      
      if (RIncB == 1) {
        bounds[0] = CIncB; bounds[1] = CIncA;
      } else { 
        bounds[0] = CIncA; bounds[1] = RIncB; 
      }
      for (i=m; i < M; i++) {
        if (RIncB == 1) {
          MatVecMul(N, K, bounds, Bx, Ax, Cx);
        } else {
          VecMatMul(K, N, bounds, Ax, Bx, Cx); 
        }
        Cx += LdC; Ax += RIncA;
      }
      M = m;
    }   
  } else { 	/* base multiply on rank 1 updates (outer products) */
    bounds[0] = RIncA; bounds[1] = CIncB; bounds[2] = LdC;
    for (j=0; j < K; j++) {	
      Rank1Update(M, N, bounds, Ax, Bx, Cx);
      Ax += CIncA; Bx += RIncB;
    }
  }
  
  if (!pFD_UNITY(alpha) && K>0) 
    MatScale(M0, N0, alpha, C);
#endif /*MM_VANILLA*/
  TRACE_EXIT;
} /* MatMult() */

