/************************************************************************/
/*       	CellMatrix: FDATA Matrix procedures for cells    	*/
/************************************************************************/
/* SCCS INFO @(#)CellMatScale.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 93/03/23 
*/
/* Set of support matrix routines used for DBLAS, independent of distributed
 * memory issues.
 * Efficiency: the use of the following techniques for scalar arhcitectures 
 * are used (if BLAS and VANILLA not defined):
 *  	1. use of a local *FDATA pointer (eg ax) to access a column of a matrix;
 *	   incrementing by last dimension avoids expensive int multiply
 *	2. for procedures scaling by FDATA *alpha, the macro
 *		FD_ALPHA_REGS(*alpha)
 *	   sets up fpu registers (as many as FDATA needs) to store *alpha
 *	   then, the other FD_ALPHA... macros use these registers
 *	   to perform fast scaling
 *	3. the matrices are MxN; these procedures typically iterate
 *	   over these dimensions; if M_INNER(...), the order of the loops
 *	   are interchanged
 *	4. 4 way loop unrolling and software pipelining where appropriate   
 *    generally, the SPARC C compiler does well with these optimizations.
 * compilation: if macro VANILLA defined, `vanilla' versions of the routines
 *              are used (slower, but useful for debugging or vectorization.
 *              if macro BLAS defined, Fortran BLAS is used whenever
 *              possible
 */

#include "BLASOptions.h"
#include "FDATAOps.h"
#include "FDAssert.h"

#include "CellMatrix.h"
#include "CellMatScale.h"

#include "MachineParams.h"
	
/********************** matrix versions of level 1 routines *****************/
static FD_BASE Zero = 0.0;	/* force gcc to use SPARC FPU regs - faster!*/	
static FD_BASE Minus1 = -1.0;	

void MatZero(M, N, A)
  int M, N; pLocMat A;
{ int i, RIncA=pLM_RInc(A), CIncA=pLM_CInc(A); FDATA *Ax = pLM_ADR(A);

  COMMON_ASSERT_TR(
    (TraceStr, "(%dx%d,a%d %d %d)", M, N, FD_ADR(Ax), RIncA, CIncA),
    IncOK(M, N, RIncA, CIncA), 0, "MatZero", 4, TRACE_COMPUTE
  );
  if M_INNER(M, N, RIncA, CIncA) { /* interchange so larger index on inner loop*/
      SWAP(int, M, N); SWAP(int, RIncA, CIncA);
  }
#if defined(VANILLA) || (!defined(BLAS) && FDATATYPE>=COMPLEX)
  for (i = 0; i < M; i++) {
    int j; register FDATA *ax = Ax;	  	  
    for (j = 0; j < N; j++) {
      FD_SET_REAL(*ax, 0.0);
      FD_SET_IMAG(*ax, 0.0);
      ax += CIncA;
    }
    Ax += RIncA;
  } /* for (i...) */
#elif defined(BLAS)
    i=0;
    if (M==0 || N==0) {		/* avoid possible illegal strides in FTN_GEMM()*/
      return;
    }
    if (RIncA==1) {
      if (CIncA==1 && N==1)
        CIncA = M;	    
      FTN_GEMM(C2F_CHAR("N"), C2F_CHAR("T"), &M, &N, &i, &FD_0, 
      		Ax, &M, Ax, &N, &FD_0, Ax, &CIncA);
    } else if (CIncA==1)
      FTN_GEMM(C2F_CHAR("N"), C2F_CHAR("T"), &N, &M, &i, &FD_0, 
      		Ax, &N, Ax, &M, &FD_0, Ax, &RIncA);
    else    
      for (i=0; i < M; i++) {
      	/* bug in the netlib blas: if A[i,j]=NaN before call, A[i,j]=NaN after.
      	   ie. its tries to mulitply A[i,j] by 0.0 */
        FTN_SCAL(&N, &FD_0, Ax, &CIncA);      
        Ax += RIncA;
      } /* for (i...) */
#else    
  {
  register int CIncA2=-CIncA-CIncA, CIncA3=-CIncA, CIncA4=-CIncA2-CIncA2;
  register int n=4*(N/4), j; 
  for (i = 0; i < M; i++) {
    register FDATA *ax = Ax; register FD_BASE z =  Zero*Minus1;
    
    for (j=-n; j < 0; j+= 4) {       /*4 way loop unrolling*/
      FD_SET_REAL(*(ax), z);		FD_SET_IMAG(*(ax), 0.0);
      FD_SET_REAL(*(ax+CIncA), z);	FD_SET_IMAG(*(ax+CIncA), 0.0);
      ax += CIncA4;	
      FD_SET_REAL(*(ax+CIncA2), z);	FD_SET_IMAG(*(ax+CIncA2), 0.0);
      FD_SET_REAL(*(ax+CIncA3), z);	FD_SET_IMAG(*(ax+CIncA3), 0.0);
    }       
    for (j=n; j < N; j++) {
      FD_SET_REAL(*ax, 0.0);      	FD_SET_IMAG(*ax, 0.0);
      ax += CIncA;
    }
    Ax += RIncA;
  } /* for (i...)*/
  } 
#endif /*VANILLA*/
} /* MatZero() */


void MatScale(M, N, alpha, C)
  int M, N; FDATA *alpha; pLocMat C; 
{ int i; int RIncC=pLM_RInc(C), CIncC=pLM_CInc(C); FDATA *Cx = pLM_ADR(C);
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%d,%s,a%d %d %d)", 
     M, N, pFD2Str(alpha), FD_ADR(Cx), RIncC, CIncC),
     IncOK(M, N, RIncC, CIncC), 0, "MatScale", 3, TRACE_COMPUTE
  );

  if (pFD_UNITY(alpha)) {			/* fast exit*/
    TRACE_RETURN;
  } else if (FD_EQ(*alpha, FD_0))		/* *alpha==0; fpu can be avoided */
    MatZero(M, N, C);
  else {
    if (M_INNER(M, N, RIncC, CIncC)) {/* interchange so larger dim. on inner loop*/
      SWAP(int, M, N);     SWAP(int, RIncC, CIncC);
    }
#if defined(VANILLA) 
    for (i=0; i < M; i++) {
      FD_register a = pFD_VAL(alpha);
      register FDATA *cx = Cx; int j;	
      for (j=0; j < N; j++) {
        FD_MULSET(*cx, =, a, *cx);
        cx += CIncC;
      }  
      Cx += RIncC;
    } /* for (i...) */
#elif defined(BLAS)
    for (i=0; i < M; i++) {
      FTN_SCAL(&N, alpha, Cx, &CIncC);      
      Cx += RIncC;
    } /* for (i...) */
#else
    {
    register int j, n, CIncC2, CIncC3, CIncC4, CIncC5, CIncC6;
    FD_register t0, t1, t2, t3, c0, c1, c2, c3; 	
    FD_register a = pFD_VAL(alpha);
     
    CIncC2=CIncC+CIncC; CIncC3=CIncC2+CIncC; CIncC4=CIncC2+CIncC2;
    CIncC5=CIncC3+CIncC2; CIncC6=CIncC4+CIncC2;
    n = 4*(N/4);

    for (i=0; i < M; i++) {
      register FDATA *cx = Cx;	
#ifndef ULTRA
      if (n >= 4) {
                t0 = *(cx);    
        j=4;
                t1 = *(cx+CIncC);
                                FD_MULSET(c0, =, a, t0);
                t2 = *(cx+CIncC2);      
                                FD_MULSET(c1, =, a, t1);
        if (j < n) do {         	
                t3 = *(cx+CIncC3);
                                FD_MULSET(c2, =, a, t2);
                                                        *(cx) = c0;
                t0 = *(cx+CIncC4);    
                                FD_MULSET(c3, =, a, t3);
                                                        *(cx+CIncC) = c1;
          j += 4;                                        
                t1 = *(cx+CIncC5);
                                FD_MULSET(c0, =, a, t0);
                                                        *(cx+CIncC2) = c2;
                t2 = *(cx+CIncC6);
                                FD_MULSET(c1, =, a, t1);
                                                        *(cx+CIncC3) = c3;
	  cx += CIncC4;		
        } while (j < n);
		t3 = *(cx+CIncC3);
				FD_MULSET(c2, =, a, t2);
							*(cx) = c0;
				FD_MULSET(c3, =, a, t3);
							*(cx+CIncC) = c1;
							*(cx+CIncC2) = c2;
							*(cx+CIncC3) = c3;
	  cx += CIncC4;		
      } /* if (n...) */  
#else
      if (n >= 4) {
      		FDATA *cs; cs = cx - CIncC4;
                t0 = *(cx);    
        j=4;
                t1 = *(cx+CIncC);
                t2 = *(cx+CIncC2);      
        if (j < n) do {         	
                t3 = *(cx+CIncC3);
          cs += CIncC4;      
                                FD_MULSET(c0, =, a, t0);
                                                        *(cs) = c0;
                t0 = *(cx+CIncC4);    
                                FD_MULSET(c1, =, a, t1);
	  cx += CIncC4;		
                                                        *(cs+CIncC) = c1;
                t1 = *(cx+CIncC);
          j += 4;                                        
                                FD_MULSET(c2, =, a, t2);
                                                        *(cs+CIncC2) = c2;
                t2 = *(cx+CIncC2);
                                FD_MULSET(c3, =, a, t3);
                                                        *(cs+CIncC3) = c3;
        } while (j < n);
		t3 = *(cx+CIncC3);
	cs += CIncC4;
                                FD_MULSET(c0, =, a, t0);
							*(cs) = c0;
                                FD_MULSET(c1, =, a, t1);
							*(cs+CIncC) = c1;
				FD_MULSET(c2, =, a, t2);
							*(cs+CIncC2) = c2;
				FD_MULSET(c3, =, a, t3);
							*(cs+CIncC3) = c3;
	  cx += CIncC4;		
     } /* if (n...) */  
#endif      
      for (j=n; j < N; j++) {
	  	t0 = *(cx); 	FD_MULSET(c0, =, a, t0);	*(cx) = c0;
	  cx += CIncC;
      } 
      Cx += RIncC;
    } /* for (i...)*/
    }
#endif /*VANILLA*/
  }
  TRACE_EXIT;
} /* MatScale() */
            
