#define INLINE_BLAS
/************************************************************************/
/*       	CellMatrix: FDATA Matrix procedures for cells    	*/
/************************************************************************/
/* SCCS INFO @(#)CellMatrix.h 1.5         last modified P. Strazdins, 99/05/28
   first created: P. Strazdins 93/03/23 
*/
/* Set of support matrix routines used for DBLAS, independent of distributed
 * memory issues.
 * Most of these can be implemented via level 1 BLAS, although for convenience
 * and efficiency, they are implemented as proper matrix operations,
 * using local matrix dimensions and strides.
 * These refer to local cell view of matrices only with generalized row-column 
 * strides (see LocMat.h).
 *
 * CPP macros special to this file:
 *	OLD_PARTMM:	if defined, use old block-packed partitioning for matrix multiply
 */
 
/*#define OLD_PARTMM*/

#ifndef __CellMatrix_INCL__
#define __CellMatrix_INCL__  

#include "FDATASetUp.h"
#include "LocMat.h"

#include "CellMatCopy.h"
#include "CellMatScale.h"
#include "FDMalloc.h"

/*********************  exported symbols  ***********************************/

#ifndef CELLBLAS

#if FDATATYPE==SINGLE
#define PackMMOpdsP	DSPackMMOpdsP
#define PackMMAlphaB	DSPackMMAlphaB
#define MatPartMult	DSMatPartMult   
#define MatMult		DSMatMult   
#elif FDATATYPE==DOUBLE
#define PackMMOpdsP	DDPackMMOpdsP
#define PackMMAlphaB	DDPackMMAlphaB
#define MatPartMult	DDMatPartMult   
#define MatMult		DDMatMult   
#elif FDATATYPE==COMPLEX
#define PackMMOpdsP	DCPackMMOpdsP
#define PackMMAlphaB	DCPackMMAlphaB
#define MatPartMult	DCMatPartMult   
#define MatMult		DCMatMult   
#else
#define PackMMOpdsP	DZPackMMOpdsP
#define PackMMAlphaB	DZPackMMAlphaB
#define MatPartMult	DZMatPartMult   
#define MatMult		DZMatMult   
#endif

#else	/* rename routines to avoid name clash */

#if FDATATYPE==SINGLE
#define PackMMOpdsP	SPackMMOpdsP
#define PackMMAlphaB	SPackMMAlphaB
#define MatPartMult	SMatPartMult   
#define MatMult		SMatMult   
#elif FDATATYPE==DOUBLE
#define PackMMOpdsP	DPackMMOpdsP
#define PackMMAlphaB	DPackMMAlphaB
#define MatPartMult	DMatPartMult   
#define MatMult		DMatMult   
#elif FDATATYPE==COMPLEX
#define PackMMOpdsP	CPackMMOpdsP
#define PackMMAlphaB	CPackMMAlphaB
#define MatPartMult	CMatPartMult   
#define MatMult		CMatMult   
#else
#define PackMMOpdsP	ZPackMMOpdsP
#define PackMMAlphaB	ZPackMMAlphaB
#define MatPartMult	ZMatPartMult   
#define MatMult		ZMatMult   
#endif

#endif /*CELLBLAS*/

#if defined(BLAS) || defined(BLAS_MM)             
  /* define Fortran BLAS procedure names called */
# include "FtnBLAS.h"  
#endif /*BLAS*/

/************************ optimization-related macros ***********************/
	/* IncCost(Inc): (partial order) cost of memory access with stride Inc:	   
	   on memory bank architectures, assumes 1 fastest, odd is same as 2.
	   for even(Inc), more accurate would be:
	   # of least sig bits of Inc that are all 0 */
#define IncCost(Inc) ((LG_MEM_BANKS==1 && (Inc)%2)? 2: (Inc))
 
#define IncCacheClash(N, Inc) 						\
        (((Inc) > 1) && ((N)>1) && (((N)-1)*(Inc) >= CACHESIZE/2))
        /*= cache confict between elements of vector of length N stride Inc;
            more precisely, but less efficiently: 
            ((N)-1)*GCD(Inc, CACHESIZE) >= CACHESIZE*CACHE_ASSOC
          if CACHESIZE represents the area that can be covered by TLB,
          conservative to use CACHESIZE/2 instead of CACHESIZE
          if CACHESIZE=0, same as ((Inc) > 1): this means    
        */
                 
#define M_INNER(M, N, RInc, CInc) 					\
	((N)<=1 || ((M)>=MIN_VECSIZE && IncCost(RInc)<=IncCost(CInc)))
        /* choose vector (M, RInc) over (N, CInc) as the inner vector 	*/
        /* note: IncOK(M, N, RInc, CInc) can be assumed; in this case,
           one of |RInc|,|CInc| must be much greater than the other
           (unless both are unity)  					*/
	
#define CopyEfficient( N ) 						\
	( (N) > 2*MINOR_OUTER_VECSIZE)
	/* = whether efficent to copy MxK A being re-used by a factor of N */ 

#define LM_Contiguous(RC, K, N, A) 					\
	(LM_Inc(1-(RC), A)==1 && ((K)==1 || (N)==LM_Inc(RC, A)))
	/* = whether (RC? K x N: N x K) A is contiguous.		*/


#define OUTER_PARTITION_WIDTH(K) 					\
	((CACHESIZE && (K)>=1)? CACHESIZE/2/(K): OUTER_VECSIZE)

#define LM_PackUseful_(M, N, A) 					\
	PackUseful_(M, K, LM_RInc(A), LM_CInc(A))
	/* whether it is useful to pack (small) matrix A */
	
#define PackUseful_(M, K, RInc, CInc)					\
         (CACHESIZE && ((RInc) >= 2*(K) || (CInc) >= 2*(M)) &&          \
          (IncCacheClash(M, RInc) || IncCacheClash(K, CInc)))        
	
#define LM_MMStridePack(rc, A, M, K)					\
	(MM_INNER_K? 							\
	 !(((K)<=1|| LM_Inc(1-(rc), A)==1) && ((M)<=1 || LM_Inc(rc,A)<2*(K))):\
	 !(((M)<=1|| LM_Inc(rc, A)==1) && ((K)<=1 || LM_Inc(1-(rc),A)<2*(M))))
	/* whether it is useful to pack matrix A, to which will be applied 
	   partitioning */	

/**** the folowing are superseded ************************************/
#define PackUseful(M, K, RInc, CInc, N)					\
	(CopyEfficient(N) && /*(K)*(M)<=CACHESIZE &&*/ 			\
	 ((RInc) >= 2*(K) || (CInc) >= 2*(M)) &&			\
	 (IncCacheClash(M, RInc) || IncCacheClash(K, CInc)))
	/* = whether useful to pack M x K matrix with strides 		*/
	/*   (RInc, CInc) and a reuse factor of N 			*/

#define LM_PackUseful(M, K, A, N) PackUseful(M, K, LM_RInc(A), LM_CInc(A), N)

#define PartitionUseful(M, N, K)					\
	((K>1) && (2*((M)+(N))*(K) >= 3*CACHESIZE))	
	/* = whether (M+N)*K exceeds CACHESIZE sufficiently to justify 	*/ 
	/*   partitioning of computation to be applied 			*/
		


	
/********************** matrix routines *********************************/
    
void MatZero STDC_PARAMS( (int M, int N, pLocMat A) );
/* pre:  pLM_IncOK(M, N, A)
 * post: A[i,j] = 0 for 0<=i<M, 0<=j<N
 */

void MatScale STDC_PARAMS( (int M, int N, FDATA *alpha, pLocMat A ) );
/* pre:  pLM_IncOK(M, N, A)
 *	 A[i,j] = a(i,j)  for 0<=i<M, 0<=j<N and some function a
 * post: A[i,j] = *alpha * a(i,j) for 0<=i<M, 0<=j<N
 * nb. optimized for *alpha == 0.0, 1.0
 */

#ifndef OLD_PARTMM
#define PackMMOpds(M, N, K, alphaf, A, B)			\
	PackMMOpdsP(M, N, N,  K, alphaf, A, B)  
/* pre:  *A (*B) represents an M x K (K x N) matrix
 * post: if it is determined efficient (in context of matrix multiply *A * *B)
 *       & possible to pack *A and/or *B (via PackUseful()),
 *       and *A_pre and/or B*_pre are packed into workspaces  malloced from W,
 *       and A and/or B point to the copies.
 *       In this case, W is returned (for future disposal) and one of
 *       *A or *B will be scaled by **alphaf and alphaf will be set to NULL.
 *       otherwise returns NULL. 
 */
            
FDATA *PackMMOpdsP STDC_PARAMS( 
	(int M, int N, int Np, int K, FDATA **alphaf, pLocMat *A, pLocMat *B)
);

extern FDATA *PackMMAlphaB;		/* indicates scaling factor applied to B
					   in last call to PackMMOpds */

#else

FDATA *PackMMOpds STDC_PARAMS(
        (int M, int N, int K, FDATA **alphaf, pLocMat *A, pLocMat *B)
        );
        
#endif 
       
#if defined(BLAS) || defined(BLAS_MM)
#include "FtnBLAS.h"
#include "FDAssert.h"

#ifdef INLINE_BLAS
#undef MatPartMult
#undef MatMult
#define MatPartMult MatMult
#define MatMult MatMult_f
#endif

static inline void MatMult_f(M, N, K, alpha, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta;  pLocMat A, B, C;
{
  static char Trans[] = "NT";
  int LdA, LdB;                 	char *TrA, *TrB;
  int *m = &M, *n = &N;   
  int LdC = pLM_CInc(C),		tC = (LdC==1);
  if (tC) {                      	/* interchange the computation ... */
    LdC = pLM_RInc(C);
    if (LdC==1) { 			/* ... unless strides are degerenate */
      if (*m > 1) LdC = *m;
      tC = 0;    
    } else {
      SWAP(int *, m, n);      		SWAP(pLocMat, A, B);
    }
  }   
  if (M <= 0 || N<= 0) return;		/* nasty bug fix */
  TrA = TrB = &Trans[0]; 		/*NoTranspose*/
  LdA = pLM_Inc(1-tC, A);		
  if (LdA==1) {
    LdA = pLM_Inc(tC, A);
    if (LdA==1) {			/* degenerate strides: => MIN(*m,K)<=1*/
      if (*m > 1) LdA = *m;		/* can always use NoTranspose */
    } else 
      TrA++; /*Transpose*/;
  }
  LdB = pLM_Inc(1-tC, B);
  if (LdB==1) {
    LdB = pLM_Inc(tC, B);
    if (LdB==1) {
      if (K > 1) LdB = K;
    } else
      TrB++; /*Transpose*/
  }
  if (alpha==NULL) alpha = &FD_1;        if (beta==NULL) beta = &FD_1;
  COMMON_ASSERT_TR(
         (TraceStr, "(%c,%c,%dx%dx%d,%sf,a%d %d,a%d %d,%s,a%d %d)",
          *TrA, *TrB, *m, *n, K, pFD2Str(alpha), FD_ADR(pLM_ADR(A)), LdA,
          FD_ADR(pLM_ADR(B)), LdB, pFD2Str(beta), FD_ADR(pLM_ADR(C)), LdC),
          *m>=0 && *n>=0 && K>=0 && LdA>=1 && LdA>=(*TrA=='N'? *m: K) && 
          LdB>=1 && LdB>=(*TrB=='N'? K: *n) && LdC>=1 && LdC>=*m,
          0, "GEMM", 4, TRACE_COMPUTE);
        
  FTN_GEMM(C2F_CHAR(TrA), C2F_CHAR(TrB), m, n, &K, alpha, pLM_ADR(A),
           &LdA, pLM_ADR(B), &LdB, beta, pLM_ADR(C), &LdC);
} /* MatMult_f() */

#else

void MatPartMult STDC_PARAMS( 
	(int M, int N, int K, FDATA *alpha, pLocMat A, pLocMat B, FDATA *beta, 
	 pLocMat C)
);
/* pre:  pLM_IncOKUnit(M, K, A) && pLM_IncOKUnit(K, N, B) &&
 * 	 pLM_IncOKUnit(M, N, C) && 0 <= M, N, K, 
 *	 !Overlap(M, K, A, M, N, C) && !Overlap(K, N, B, M, N, C)
 * post: C := (*beta)C + (*alpha)A*B;  A is MxK, B is KxN, C is MxN; 
 * 	 applying partitioning methods & packing to reduce cache conflicts
 *	 assumes the underlying MM procedure (MatMult) can not (efficiently)
 *	 perform the scaled computation (*alpha != 1), so scaling is performed
 *	 whenever workspace is allocated.
 */  	

void MatMult STDC_PARAMS(
	(int M, int N, int K, FDATA *alpha, pLocMat A, pLocMat B, FDATA *beta, 
	 pLocMat C)
);
/* pre/post: as for MatPartMult(), except 
 *	(currently) for !defined(VANILLA or BLAS) *alpha = 1.0
 * note: no partitioning methods are applied 	      
 */

#endif

#if defined(BLAS) || defined(BLAS_MM)

static inline void R1Mult(M, N, alpha, A, B, C)					
  int M, N; FDATA *alpha; pLocMat A, B, C;
{ int LdA = pLM_RInc(A), LdB=pLM_CInc(B), LdC;
  if (alpha==NULL) alpha = &FD_1;        
  if (pLM_RInc(C)==1) {
    LdC = pLM_CInc(C);						
    FTN_GER(&M, &N, alpha, pLM_ADR(A), &LdA, pLM_ADR(B), &LdB, pLM_ADR(C), &LdC);
  } else {
    LdC = pLM_RInc(C);	
    FTN_GER(&N, &M, alpha, pLM_ADR(B), &LdB, pLM_ADR(A), &LdA, pLM_ADR(C), &LdC);
  }
} /* R1Mult() */

static inline void VMMult(N, K, alpha, A, B, beta, C)	
  int N, K; FDATA *alpha, *beta; pLocMat A, B, C;
{ int LdA = pLM_CInc(A), LdB, LdC=pLM_CInc(C); char TrB;
  if (alpha==NULL) alpha = &FD_1;        if (beta==NULL) beta = &FD_1;
  if (pLM_RInc(B)==1) {
    LdB = pLM_CInc(B);			TrB = 'T';			
    SWAP(int, N, K);
  } else {
    LdB = pLM_RInc(B);			TrB = 'N';
  }   	  
  FTN_GEMV(C2F_CHAR(&TrB), &N, &K, alpha, pLM_ADR(B), &LdB, pLM_ADR(A),
	   &LdA, beta, pLM_ADR(C), &LdC);
} /* VMMult() */

static inline void AxpyMult(N, alpha, A, B)
  int N; FDATA *alpha;  pLocMat A, B;
{ int IncA = pLM_CInc(A), IncB = pLM_CInc(B);
  if (alpha==NULL) alpha = &FD_1;        
  FTN_AXPY(&N, alpha, pLM_ADR(A), &IncA, pLM_ADR(B), &IncB);
}

static inline void IPMult(v, K, A, B)	
  FDATA *v; int K; pLocMat A, B; 
{ int IncA = pLM_CInc(A), IncB = pLM_RInc(B);   
#if FDATATYPE < COMPLEX
  *v = FTN_DOT(&K, pLM_ADR(A), &IncA,  pLM_ADR(B), &IncB);
#else
/*  FTN_DOT(v, &K, pLM_ADR(A), &IncA,  pLM_ADR(B), &IncB);*/
  char TrB = 'N';			int One = 1, M = One, N = K;  
  if (IncB==1) 
    IncB = pLM_CInc(B), TrB = 'T', M = K, N = 1;
  FTN_GEMV(C2F_CHAR(&TrB), &M, &N, &FD_1, pLM_ADR(B), &IncB, pLM_ADR(A),
           &IncA, &FD_0, v, &One);
             
#endif  
}

#else

#define R1Mult(M, N, alpha, A, B, C) 					\
	MatMult(M, N, 1, alpha, A, B, NULL, C)

#define VMMult(N, K, alpha, A, B, beta, C) 				\
	MatMult(1, N, K, alpha, A, B, beta, C)

# if (FDATATYPE < COMPLEX)
#   include "BLASInnerLoops.h"			/* Axpy1, VecMatMul() */

#   define AxpyMult(N, alpha, A, B)					\
	Axpy1(N, alpha, pLM_ADR(A), pLM_CInc(A), pLM_ADR(B), pLM_CInc(B))

#   define IPMult(v, K, A, B)						\
	{ int bounds[3];						\
  	bounds[0] = pLM_CInc(A), bounds[1] = pLM_RInc(B), bounds[2] = 1;\
  	VecMatMul(K, 1, bounds, pLM_ADR(A), pLM_ADR(B), v); 		\
	} /* IPMult() */ 

# else

#   define AxpyMult(N, alpha, A, B)                                     \
    { LocMat la; LM_Init(la, alpha, 1 ,1);				\
    MatMult(1, N, 1, NULL, &la, A, NULL, B); }
        
#   define IPMult(v, K, A, B)                                           \
    { LocMat lv; LM_Init(lv, v, 1 ,1);					\
    MatMult(1, 1, K, NULL, A, B, NULL, &lv); }		

# endif
#endif


#endif /*__CellMatrix_INCL__*/

