 /************************************************************************/
/*  SplitUnpack: matrix copy routines for a packed `split' source	*/
/************************************************************************/
/* SCCS INFO @(#)CellSplitUnpack.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 96/01/30
 */
/* These routines are required for unpacking messages in AP ring buffer;
 * a message from such a buffer can be represented by:
 *	size:	total message size
 *	Ax1:	pointer to start of message
 *	size1:  size of part of message from Ax1
 *	Ax2:	pointer to 2nd half of message (length size-size1)
 * Here size1 < size (and Ax2 != NULL) when ring buffer wrap around occurs.
 * compilation: if DEBUG_SPLIT defined, perform extensive cheking of split messages
 *		if CHECK_VALS defined, abs check values of elts in received 
 *		messages are bounded.
 *              if BLACS_COMP, compile this module separately for AP BLACS
 *              (SplitMatUnpack only is needed)
 * note: for complex datatypes, communication occurs in units of FD_BASE. 
 */

#define FDB_ADR(a) (((int) (a))/sizeof(FD_BASE))

#define CHECK_VALS
#define DEBUG_SPLIT

#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#include "LocMat.h"
#include "CellMatrix.h"

#ifdef BLACS_COMP
# include "CellSplitUnpack.h"
# define static_ 
#else
# define static_ static
#endif
#include "CellBlkMatrix.h"

#define CACHE_LINE_SIZE	16	/* AP1000 line send size */
#if FDATATYPE<COMPLEX
# define FD_BR 		1
# define FD_re_sz	0
#else
# define FD_BR 		2
# define FD_re_sz	sizeof(FD_BASE)
#endif
#define COMM_UNIT 	(CACHE_LINE_SIZE/sizeof(FD_BASE))
#define LG_COMM_UNIT 	(4 - LG_FD_SIZE + FD_BR - 1)

#ifdef DEBUG_SPLIT		/* only for AP1000 */
int cap_wr_limit();
extern int FD_rb_msg_size;	
	/* size of last ring buffer message received minus initial alignment & header */
#endif

#ifdef CHECK_VALS
static int CheckRecvVals(M, N, B)
  int M, N; LocMat B;
{ int i; int CIncB = LM_CInc(B);
  if (MsgMaxAbsVal > 0.0) {
     FD_BASE Max = 0.0;
     for (i=0; i < M; i++) {
       int MxIJ = VecAbsMaxPos(N, LM_Index(B, i, 0), CIncB);
       FD_BASE BIJr = FD_REAL(*LM_Index(B, i, MxIJ));
       FD_BASE BIJi = FD_IMAG(*LM_Index(B, i, MxIJ));
       if (ABS(BIJr) > Max) Max = BIJr;
       if (ABS(BIJi) > Max) Max = BIJi;
     }    	
     if (Max > MsgMaxAbsVal) {
       int j; int EltCnt=8;
       printf("invalid elts (Max=%G): ", Max);
       for (i=0; i<M && EltCnt>0; i++) {      
       	 for (j=0; j<N && EltCnt>0; j++) {
	   FDATA AIJ = *LM_Index(B, i, j);
           FD_BASE BIJr = FD_REAL(*LM_Index(B, i, j));
           FD_BASE BIJi = FD_IMAG(*LM_Index(B, i, j));
	   if (ABS(BIJr) > MsgMaxAbsVal || ABS(BIJi) > MsgMaxAbsVal) {
#if (FDATATYPE < COMPLEX)
	     printf(" %d,%d:%G", i, j, BIJr);
#else
	     printf(" %d,%d:%G,%G", i, j, BIJr, BIJi);
#endif
	     EltCnt--;
	   }  
	 } /* for (j...) */
       } /* for (i...) */
       printf("\n");
       return(0);
     } else
       return(1);	
  } else
    return(1);	 	
} /*CheckVals()*/
#endif /*CHECK_VALS*/

  
static inline int calc_stride_disp(a, m, n, inc)
  int *a; int m, n, inc;
/* pre:  m, n, inc > 0;
 * post: returns the displacement of an mxn array of alignment *a
 *	 which was generated by communication via units of COMM_UNIT;
 *       the source vector had skip%COMMUN_UNIT=s%COMM_UNIT (if s=0, message 
 *	 was packed).
 *	 *a is alignment of mth element of the vector
 *	 note: the alignment of the mth row is included in the displacement
 */
{ int s = inc & (COMM_UNIT-1),
      i,
      size = 0;		
   if (inc == 0)
     return(FAST_MUL1(m,n));   
   for (i=0; i < m; i++) {
     size += ((n + *a + COMM_UNIT-1)/COMM_UNIT)*COMM_UNIT - *a;
     *a = (*a + s) & (COMM_UNIT-1);
     size += *a;
   }
  return (size);
} /* calc_stride_disp() */


static_ void SplitMatUnpack(Add, N, size, s, Ax1, size1, Ax2, alpha, Bx, 
			   RIncB, CIncB)
  FDATA *Ax1, *Ax2, *alpha, *Bx;
  int Add, size, s, size1, N, RIncB, CIncB;
/* pre:  size1 <= size
 * post: let A = strip(s,M,N,concat(Ax1[0..size1-1]:Ax2[0..size-size1])),
 *       M=size/N, a = Add? *alpha: FD_1;
 *       as for (Add? MatCopy: MatAdd)(M, N, {A, N, 1}, a, {Bx, RIncB, CIncB})
 *       strip(a,s,M,N,A) = concat(A[0..N-1]:strip(a',s,M-1,N, A+N1))
 *       where  N1 = ((N + a + COMM_UNIT-1)/COMM_UNIT)*COMM_UNIT - a + a'
 *              a' = (a+s)%COMM_UNIT
 * notes:the strip() operation arises when A has been formed by communication
 *       of a vector of length M, element size N and stride S st. S%COMM_UNIT=s.
 *       with the unit communication being COMM_UNIT*sizeof(FDATA),
 *       surplus words at the either end of each element may be sent
 *       and must be now be ignored.
 */
{ int a = (((int) Ax1) / sizeof(FD_BASE)) % COMM_UNIT;
  unsigned int /*g, lg,*/       /* g = s = GCD(s, CACHE_SIZE), lg = log2(g) */
           Cg, lCg;             /* Cg = COMM_UNIT/g;   lCg = log2(Cg)  */
				/* Cg is period of regular row stride in A */	
  unsigned int CgInc;		/* min. regular stride of A */
                /* note: the rows of the msg are periodic with period Cg */
  int i, M, M1;     			LocMat A, B, Bj;
  int N_ = FD_BR * N, 			size1_ = size1 % FD_BR;	

#ifdef DEBUG_SPLIT
# define TRLEVEL (SplitMsg? 0: 5)
  int raw_size = FD_rb_msg_size / sizeof(FDATA);
  FDATA *EndAx1 = Ax2!=NULL? (FDATA *) cap_wr_limit(): Ax1+raw_size;
  int raw_size1 = ((FD_BASE *) EndAx1) - ((FD_BASE *) Ax1);
  FD_BASE *EndAx2 = ((FD_BASE *) Ax2) + FD_BR * raw_size - raw_size1;
  int OldTrLevel=TraceLevel, 		SplitMsg = (Ax2!=NULL || size1/FD_BR < size);
  if (SplitMsg)
    TraceLevel=CurTraceDepth+1;		/* force tracing */
#else
# define TRLEVEL 5    
#endif 

  size1 /= FD_BR;  
  if (s==0) {
    Cg = 1;				lCg = 0;
    CgInc = N;
  } else {
    int g=1, lg=0;
      while (g < s) {
      g *= 2; 				lg++;
    }
    Cg = COMM_UNIT >> lg; 		lCg = LG_COMM_UNIT - lg;
    CgInc = (((a&(g-1)) + ((a+N_)&(g-1)) + N_ + ((Cg-1)<<lg)) << lCg) / FD_BR;
  }

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%d,%d a%d %d.%d a%d,%s,a%d %d %d)",
     Add, N, size, s, FDB_ADR(Ax1), size1, size1_, FD_ADR(Ax2), pFD2Str(alpha), FD_ADR(Bx), RIncB, CIncB),
     size1<=size && ((size1<size) == (Ax2!=NULL)) && !LM_InvalStride(CgInc) 
     && !LM_InvalStride(RIncB << lCg),
     0, "SplitMatUnpack", TRLEVEL, TRACE_COMMUN
  );


  if (N==0 || size==0) {
#ifdef DEBUG_SPLIT
    if (SplitMsg)
      TraceLevel=OldTrLevel;
#endif
    TRACE_RETURN;
  }

  if (size1 > size)			/* an error, to be trapped later */
    size1 = size; 
  M1 = FAST_DVDEQ(size1, N);
  M = size1==size? M1: size/N;

#ifdef DEBUG_SPLIT
  if (SplitMsg)
    printf("raw_size=%d (s1=%d.%d), N=%d M=%d, M1=%d, Cg=%d, CgInc=%d\n", 
    	    raw_size, raw_size1/FD_BR, raw_size1%FD_BR, N, M, M1, Cg, CgInc);
#endif

  LM_Init(A, Ax1, CgInc, 1);	  LM_Init(B, Bx, RIncB << lCg, CIncB);

  for (i=0; i < Cg; i++) {
    int m = (M1 - i + Cg-1) >> lCg;		if (m==0) break;	
#ifdef DEBUG_SPLIT
    if (m>0 && (((int) LM_Index(A,m-1,N-1)) + FD_re_sz >= (int) EndAx1)) {
        COMMON_ASSERT( 0, 0, "SplitUnpack: Msg1 error", 7, TRACE_COMMUN);
      }      	
#endif
    (!Add? MatCopy: MatAdd)(m, N, &A, alpha, &B); 
    if (m==M) break;				/* quick exit */
    LM_ADR(A) += calc_stride_disp(&a, 1, N_, s) / FD_BR;
    LM_ADR(B) += RIncB;	  	
  } /* for (i...) */

  if (M1 < M) {				/* message is split */
    int dM1;
    size1 = size1 % N;
    a = (((int) Ax1) / sizeof(FD_BASE)) % COMM_UNIT;
#ifdef DEBUG_SPLIT  
    printf("a=%d ", a);
#endif
    dM1 = calc_stride_disp(&a, M1&(Cg-1), N_, s) / FD_BR;
    LM_ADR(A) = Ax1 + dM1 + (M1>>lCg)*CgInc;
#ifdef DEBUG_SPLIT 
    printf("a(M1)=%d, dM1=%d, m1CgInc=%d\n", a, dM1, (M1>>lCg)*CgInc);
#endif
    LM_ADR(B) = Bx + M1*RIncB;		M -= M1;
    
    if (size1+size1_==0) {
      Ax2 = (FDATA *) (((FD_BASE *) Ax2) + (s==0? 0: a));	
      LM_ADR(A) = Ax2;
    } else {
#ifdef DEBUG_SPLIT
      if (((int) (LM_Index(A, 0,size1-1))) + size1_*sizeof(FD_BASE) >= (int) EndAx1) {
      	printf("last Ax1 = a%d, size1=%d.d, end a%d\n", FDB_ADR(LM_ADR(A)), size1, size1_, FDB_ADR(EndAx1));
      	COMMON_ASSERT( 0, 0, "SplitUnpack: Msg1_ error", 7, TRACE_COMMUN);
      }
#endif
      (!Add? MatCopy: MatAdd)(1, size1, &A, alpha, &B);

      if (size1_) { /* split has occured in the middle of a complex element */
        FDATA At, Alpha = pFD_VAL(alpha);
        FD_BASE * Ax2b = (FD_BASE *) Ax2;
        FDATA *Bx = LM_ADR(B) + size1*CIncB;
        FD_SET_REAL(At, *((FD_BASE *) (LM_ADR(A)+size1)));
        FD_SET_IMAG(At, *((FD_BASE *) Ax2));
        if (Add) {
          FD_MULSET(*Bx, +=, Alpha, At);
        } else {      	
          FD_MULSET(*Bx, =, Alpha, At);
        }  	                     	
        Ax2b++; Ax2 = (FDATA *) Ax2b;	size1++;
      }

      LM_ADR(A) = Ax2;
#ifdef DEBUG_SPLIT
      if (((int) LM_Index(A, 0, N-size1-1)) + FD_re_sz >= (int) EndAx2) {
        COMMON_ASSERT( 0, 0, "SplitUnpack: Msg2_ error", 7, TRACE_COMMUN); 	
      }      	
#endif      

      (!Add? MatCopy: MatAdd)(1, N-size1, &A, alpha, pLocSubMat_(&Bj, &B, 0, size1)); 
      LM_ADR(A) = Ax2-size1 + calc_stride_disp(&a, 1, N_, s) / FD_BR;
      LM_ADR(B) += RIncB; 			      M--;
    }  

    for (i=0; i < Cg; i++) {
      int m = (M - i + Cg-1) >> lCg;		if (m==0) break;		
#ifdef DEBUG_SPLIT
      if (m>0 && (((int) LM_Index(A,m-1,N-1)) + FD_re_sz >= (int) EndAx2)) {
        COMMON_ASSERT( 0, 0, "SplitUnpack: Msg2 error", 7, TRACE_COMMUN);
      }      	
#endif
      (!Add? MatCopy: MatAdd)(m, N, &A, alpha, &B); 
      LM_ADR(A) += calc_stride_disp(&a, 1, N_, s) / FD_BR;
      LM_ADR(B) += RIncB;
    } /* for (i...) */
  } /* if (M1 < M) */

#ifdef DEBUG_SPLIT
  if (SplitMsg)
    TraceLevel=OldTrLevel;    
#endif
#ifdef CHECK_VALS
  LM_Init(B, Bx, RIncB, CIncB); 
  if (!CheckRecvVals(size/N, N, B)) {   
      COMMON_ASSERT_TR(
       (TraceStr, "(%d,%d,%d,%d a%d %d a%d,%s,a%d %d %d)",
       Add, N, size, s, FD_ADR(Ax1), size1, FD_ADR(Ax2), pFD2Str(alpha), FD_ADR(Bx), RIncB, CIncB),
       0, 0, "SplitMatUnpack: elts", TRLEVEL, TRACE_COMMUN
     );
  }   
#endif 
  TRACE_EXIT;
} /* SplitMatUnpack() */


#ifndef BLACS_COMP
#define L2G0_(i, B0, B, P) (((i)+(B0))/(B)*(P)*(B) + ((i)+(B0))%(B) - (B0))
			  /*equivalent to: i + (i+B0)/B*B*(P-1)*/


static_ void SplitBlkMatStrideUnpack(Add, N, size, Ax1, size1, Ax2, RIncA, alpha,
			     B, RB0, RB, RBInc, CB0, CB, CBInc) 
  FDATA *Ax1, *Ax2, *alpha; pLocMat B;
  int Add, size1, size, N, RIncA, RB0, RB, RBInc, CB0, CB, CBInc; 
/* pre:   size % N == 0 && RIncA >= N && size >= size1;
 * post:  as for BlkMatStrideUnpack(Add, M, N, {A, RIncA, 1}, &FD_1,
 *                                  B, RB0, RB, RBInc, CB0, CB, CBInc);
 *        where A = concat(Ax1[0..size1-1], Ax2[0..size-size1-1]), M = size/N
 */
{ int m1, m, n1; 
  LocMat A, B0, B_, Bj;
  int size1_ = size1 % FD_BR;
#ifdef DEBUG_SPLIT
# define TRLEVEL (SplitMsg? 0: 5)
  int raw_size = FD_rb_msg_size / sizeof(FDATA);
  FDATA *EndAx1 = Ax2!=NULL? (FDATA *) cap_wr_limit(): Ax1+raw_size;
  int raw_size1 = ((FD_BASE *) EndAx1) - ((FD_BASE *) Ax1);
  FD_BASE *EndAx2 = ((FD_BASE *) Ax2) + FD_BR * raw_size - raw_size1;
  int OldTrLevel=TraceLevel;            
  int SplitMsg = (Ax2!=NULL || size1/FD_BR < size || raw_size1<raw_size);
#else
# define TRLEVEL 5
#endif
#ifdef CHECK_VALS
  B0 = *B;
#endif
#ifdef DEBUG_SPLIT
  if (SplitMsg)
    TraceLevel=CurTraceDepth+1;         /* force tracing */
#endif
  
  size1 /= FD_BR;  
  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%d,a%d,%d.%d,a%d %d,%s,a%d %d %d,%d %d %d,%d %d,%d)",
     Add, N, size, FD_ADR(Ax1), size1, size1_, FD_ADR(Ax2), RIncA, pFD2Str(alpha), 
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), RB0, RB, RBInc, CB0, CB, CBInc),
    RIncA>=N && N>0 && size%N==0 && size>=size1 && ((size1<size) == (Ax2!=NULL)), 0, 
    "SplitBlkMatStrideUnpack", TRLEVEL, TRACE_COMMUN
  );

  if (N==0 || size==0) {
#ifdef DEBUG_SPLIT
    if (SplitMsg)
      TraceLevel=OldTrLevel;
#endif
    TRACE_RETURN;
  }

  m =  FAST_DVDEQ(size, RIncA);
  m1 = size1==size?  m: size1/RIncA;     	

#ifdef DEBUG_SPLIT
  n1 = size1 % RIncA;
  COMMON_ASSERT_TR(
      (TraceStr, "Ax1=a%d %d %x-%x, r_s(1)=%d %d.%d, s(1)=%d %d; N=%d m(1)=%d %d, Ax2=%x-%x", 
       FD_ADR(Ax1), RIncA, Ax1, EndAx1, raw_size, raw_size1/FD_BR, raw_size1%FD_BR, size, size1, N, m, m1, Ax2, EndAx2),
       m1*RIncA+n1==raw_size1/FD_BR && m*RIncA==raw_size, 0, 
      "SplitBlkMatStrideUnpack: size", TRLEVEL, TRACE_COMMUN
  );
#endif

  LM_Init(A, Ax1, RIncA, 1);          
#ifdef DEBUG_SPLIT
  if (m1>0 && (((int) LM_Index(A,m1-1,N-1)) + FD_re_sz >= (int) EndAx1)) {
     COMMON_ASSERT( 0, 0, "BlkSplitUnpack: Msg1 error", 7, TRACE_COMMUN);
  }
#endif  
  if (m1 > 0)     
    BlkMatStrideUnpack(Add, m1, N, &A, alpha, B, RB0, RB, RBInc, CB0, CB, CBInc);  
    
  if (size1==size) {		/* fast exit is possible */
    TRACE_RETURN;
  }
  
  B = pLocSubMat_(&B_, B, L2G0_(m1, RB0, RB, RBInc), 0);
  RB0 = (RB0 + m1)%RB;
  n1 = size1 % RIncA;

  if (n1 > 0 || size1_) {
    int n1a = (n1 > N)? N: n1;
    LM_ADR(A) += m1*RIncA;

#ifdef DEBUG_SPLIT
  COMMON_ASSERT_TR(
      (TraceStr, ": A1=a%d %d, r_s(1)=%d %d, s(1)=%d %d; N=%d m(1)=%d %d", 
       FD_ADR(LM_ADR(A)), RIncA, raw_size, raw_size1, size, size1, n1a, m, m1),
       ((int) LM_Index(A, 0, n1a-1)) + FD_re_sz < (int) EndAx1,
        0, "BlkSplitUnpack: cp1", TRLEVEL, TRACE_COMMUN
  );
#endif
    BlkMatStrideUnpack(Add, 1, n1a, &A, alpha, B, RB0, RB, RBInc, CB0, CB, CBInc);

    if (size1_ && n1a < N) { /* split has occured in the middle of a complex element */
        FDATA At, Alpha = pFD_VAL(alpha);
        FDATA *Bx = LM_Index(*B, 0, L2G0_(n1a, CB0, CB, CBInc));

	if (((int) LM_Index(A, 0, n1a)) >= (int) EndAx1) {
          COMMON_ASSERT(0, 0, "BlkSplitUnpack: cp1_", TRLEVEL, TRACE_COMMUN);
        }
        
        FD_SET_REAL(At, *((FD_BASE *) (LM_ADR(A)+n1a)));
        FD_SET_IMAG(At, *((FD_BASE *) Ax2));
        if (Add) {
          FD_MULSET(*Bx, +=, Alpha, At);
        } else {      	
          FD_MULSET(*Bx, =, Alpha, At);
        }  	                     	
        n1a++;
    }
    if (size1_) {
      FD_BASE * Ax2b = (FD_BASE *) Ax2;
      Ax2b++; Ax2 = (FDATA *) Ax2b;
    }
    
    LM_ADR(A) = Ax2;
#ifdef DEBUG_SPLIT
  COMMON_ASSERT_TR(
      (TraceStr, ": A2=a%d %d, r_s(1)=%d %d, s(1)=%d %d; N=%d m(1)=%d %d", 
       FD_ADR(LM_ADR(A)), RIncA, raw_size, raw_size1, size, size1, N-n1a, m, m1),
       ((int) (LM_Index(A, 0, N-n1a-1))) + FD_re_sz <= (int) EndAx2, 0, 
      "BlkSplitUnpack: cp2", TRLEVEL, TRACE_COMMUN
  );
#endif
    BlkMatStrideUnpack(Add, 1, N-n1a, &A, alpha, 
		       pLocSubMat_(&Bj, B, 0, L2G0_(n1a, CB0, CB, CBInc)),
    		       RB0, RB, RBInc, (CB0+n1a)%CB, CB, CBInc);
    pLocSubMat(B, L2G0_(1, RB0, RB, RBInc), 0);
    RB0 = (RB0 + 1)%RB;     
    LM_ADR(A) += (RIncA-n1-size1_);
    m1++;
  } else
    LM_ADR(A) = Ax2;
    
  if (m > m1) {
#ifdef DEBUG_SPLIT
    COMMON_ASSERT_TR(
      (TraceStr, ": A2=a%d %d, r_s(1)=%d %d, s(1)=%d %d; N=%d m(1)=%d %d", 
       FD_ADR(LM_ADR(A)), RIncA, raw_size, raw_size1, size, size1, N, m, m1),
       ((int) (LM_Index(A, m-m1-1, N-1))) + FD_re_sz < (int) EndAx2, 0,
      "BlkSplitStrideUnpack: cp3", TRLEVEL, TRACE_COMMUN
    );
#endif
    BlkMatStrideUnpack(Add, m-m1, N, &A, alpha, B, RB0, RB, RBInc, CB0, CB, CBInc);
  }

#ifdef CHECK_VALS
  if (RBInc==1 && CBInc==1 && !CheckRecvVals(m, N, B0)) {   
    COMMON_ASSERT_TR(
      (TraceStr, "(%d,%d,%d,a%d,%d,a%d %d,%s,a%d %d %d,%d %d %d,%d %d,%d)",
       Add, N, size, FD_ADR(Ax1), size1, FD_ADR(Ax2), RIncA, pFD2Str(alpha), 
       FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), RB0, RB, RBInc, CB0, CB, CBInc),
       0, 0, "SplitBlkMatStrideUnpack: elts", TRLEVEL, TRACE_COMMUN
    );
   }
#endif
#ifdef DEBUG_SPLIT
    if (SplitMsg)
      TraceLevel=OldTrLevel;
#endif
  TRACE_EXIT;
} /* SplitBlkMatStrideUnpack() */

#endif /*BLACS_COMP*/
