/************************************************************************/
/*       CellTri1Matrix: Cell Tri Input Matrix procedures 		*/
/************************************************************************/
/* SCCS INFO %Z%%M% %I%	last modified P. Strazdins, %E%
   first created: P. Strazdins 97/03 
*/
/* set of non-communication triangular matrix routines used for DBLAS.
 * These routines can be run on cells separately, but the operations
 * assume block matrix storage on a P[0] x P[1] mesh.
 */

#include "CellTri1Matrix.h"
#include "FDATAOps.h"
#include "BLASOptions.h"
#include "FDAssert.h"  

#include "MatDim.h"
#include "DistMat.h"
#include "CellMatrix.h"
#include "FDMalloc.h"
#include "MachineParams.h"

#include "D_BLAS.h"

#define lA DM_Loc(A)

static inline InnerProd(k, alpha, A, IncA, B, IncB, C)
  int k, IncA, IncB; FDATA *alpha, *A, *B, *C;
/* pre:  k>0; 
 * post: *C += pFD_VAL(alpha) * (Sum i: 0<=i<k: A[i]*B[i])
 */ 
{ FDATA c = FD_0;	
  COMMON_ASSERT_TR(
    (TraceStr, "(%d,%s,a%d %d,a%d %d,a%d)",
     k, pFD2Str(alpha), FD_ADR(A), IncA, FD_ADR(B), IncB, FD_ADR(C)),
     1, 0, "InnerProd", 6, TRACE_COMPUTE
  ); 
  do {  	  
    FD_MULSET(c, +=, *A, *B);		k--;	 
    A += IncA;				B += IncB;
  } while (k > 0);
  if (alpha!=NULL)
    FD_MULSET(c, =, c, *alpha);
  FD_ASSIGN(*C, +=, c);
} /* InnerProd() */
                    

void Tri1MatMult(M, N, K, alpha, UpLo, Diag, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta; 
  char UpLo, Diag; DistMat A; pLocMat B, C; 
/* notes:1. di performs a `slicing' of (rows or columns of) A.
 *	 its optimal value depends on the minimum outer dimension for 
 *	 optimal cell GEMM performance (this may be a function of K). 
 *	 Underestimation can yield to low cell GEMM performance;
 *	 overestimation can lead to a higher fraction of small vector-matrix
 *	 computations for the triangular portion of A
 *
 */ 
{ int i;	
  int L = (UpLo==Lower)^DM_Trans(A);	int TrA = DM_Trans(A);
  int UnitDiag = (Diag==Unit);
  pMatDim Di, Dj;			/* points to dim. of A corresp i & j */
  int di = (K < K0_MINOR_OUTER_VECSIZE)? MINOR_OUTER_VECSIZE_K0: 
  					 MINOR_OUTER_VECSIZE;
  LocMat *Ai0, Ai0_, Bk0, Ci0, Aik;    	/* used to prevent side effects *A,..*/ 
  FDATA *WS;				/* used for packing A, B if needed */
  FDATA *oldalpha = alpha,		UnitVal = FD_1;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%dx%d,%s,%c%c %d a%d %d %d,a%d %d %d,%s,a%d %d %d)",
     M, N, K, pFD2Str(alpha), UpLo, Diag, DM_Trans(A), 
     FD_ADR(LM_ADR(lA)), LM_RInc(lA), LM_CInc(lA),
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD2Str(beta), 
     FD_ADR(pLM_ADR(C)), pLM_RInc(C), pLM_CInc(C)),
    (M==0 || pLM_IncOKUnit(K, N, B)) && (K==0 || pLM_IncOKUnit(M, N, C)) 
    && DistMatOK(M, K, A) && 0<=M && 0<=N && 0<=K && UpLoOK(UpLo), 
    0,  "Tri1MatMult", 3, TRACE_COMPUTE
  );  

  if (N==0 || M==0) {             		/* fast exit*/
    TRACE_RETURN;
  }
  if (pFD_UNITY(beta)) {
    if (K==0) {
      TRACE_RETURN;
    }  				
    beta = NULL;
  }
  
  Di = DM_D(TrA, A);	    		Dj = DM_D(1-TrA, A);
  
  LM_Copy(TrA, Ai0_, DM_Loc(A));	LM_Copy(0, Ci0, *C);
  Ai0 = &Ai0_;				
  WS = PackMMOpds(M, N, K, &alpha, &Ai0, &B);
  if (Ai0 != &Ai0_ && !FD_EQ(*pLM_ADR(Ai0), *LM_ADR(Ai0_))) 
    /* Ai0 now scaled by alpha; must scale UnitVal accordingly */	
    UnitVal = pFD_VAL(oldalpha); 
  LM_Copy(0, Bk0, *B);			LM_Copy(0, Aik, *Ai0);
  
  if (L) {  
    for (i=0; i < M; i+= di) {
      FDATA Aii, *xAii;	
      /* A[i..i0Max, 0..j-1] is the rectangular portion in this slice */ 
      int i0, i0Max = MIN(i + di-1, M-1);
      int iG = L2Gd(i, Di),		k = G2Ld(iG+1, Dj);	
      k = MIN(k, K);
    
      xAii = pLM_Index(Ai0, 0, k-1);	
      if (UnitDiag && SubMatP0d(iG, Dj)==0) 
        Aii = *xAii, *xAii = UnitVal;
      MatMult(i0Max-i+1, N, k, alpha, Ai0, B, beta, &Ci0);
      if (UnitDiag && SubMatP0d(iG, Dj)==0)
        *xAii = Aii;
      pLocSubMat_x(&Aik, Ai0, 0, k);	pLocSubMat_x(&Bk0, B, k, 0);
    
      for (i0=i+1; i0 <= i0Max; i0++) {
        /* iterate rowwise over proper triangular portion of slice */	
        int i0G = L2Gd(i0, Di),	     	k0 = G2Ld(i0G+1, Dj);  
        k0 = MIN(k0, K);
        pLocSubMat(&Aik, 1, 0);		pLocSubMat(&Ci0, 1, 0); 
        if (k0 > k) {
          xAii = LM_Index(Aik, 0, k0-k-1);  Aii = *xAii;
          if (UnitDiag && SubMatP0d(i0G, Dj)==0) 
            Aii = *xAii, *xAii = UnitVal;    
          if (N > 1) 	
            MatMult(1, N, k0-k, alpha, &Aik, &Bk0, NULL, &Ci0);
          else  			/* inline to reduce call overheads */
	    InnerProd(k0-k, alpha, LM_ADR(Aik), LM_CInc(Aik), 
	  	      LM_ADR(Bk0), LM_RInc(Bk0), LM_ADR(Ci0));
	  if (UnitDiag && SubMatP0d(i0G, Dj)==0)	            	      
            *xAii = Aii;
        }  
      } /* for (i0...) */ 
      pLocSubMat(Ai0, i0Max-i+1, 0);	pLocSubMat(&Ci0, 1, 0);  
    } /* for (i...) */  	  
  } else {
    for (i=0; i < M; i+= di) {
      FDATA Aii, *xAii;	
      /* A[i..i0Max, j..N-1] is the rectangular portion in this slice */ 
      int i0, i0Max = MIN(i + di-1, M-1);
      int iG = L2Gd(i0Max, Di),		k = G2Ld(iG, Dj);	
      k = MIN(k, K);
    
      pLocSubMat_x(&Aik, Ai0, 0, k);	pLocSubMat_x(&Bk0, B, k, 0);
      xAii = LM_Index(Aik, i0Max-i, 0);	
      if (UnitDiag && SubMatP0d(iG, Dj)==0)
        Aii = *xAii, *xAii = UnitVal;
      MatMult(i0Max-i+1, N, K-k, alpha, &Aik, &Bk0, beta, &Ci0);
      if (UnitDiag && SubMatP0d(iG, Dj)==0)      
        *xAii = Aii;

      for (i0=i; i0 < i0Max; i0++) {
        /* iterate rowwise over proper triangular portion of slice */	
        int i0G = L2Gd(i0, Di),	     	k0 = G2Ld(i0G, Dj);  
        k0 = MIN(k0, K);
        if (k > k0) {
          pLocSubMat_x(&Aik, Ai0, 0, k0);pLocSubMat_x(&Bk0, B, k0, 0);
          xAii = LM_ADR(Aik);  		
          if (UnitDiag && SubMatP0d(i0G, Dj)==0)
            Aii = *xAii, *xAii = UnitVal;
          if (N > 1)	
            MatMult(1, N, k-k0, alpha, &Aik, &Bk0, NULL, &Ci0);
          else  			/* inline to reduce call overheads */
	    InnerProd(k-k0, alpha, LM_ADR(Aik), LM_CInc(Aik), 
	  	      LM_ADR(Bk0), LM_RInc(Bk0), LM_ADR(Ci0));
	  if (UnitDiag && SubMatP0d(i0G, Dj)==0)	      
            *xAii = Aii;
        }
        pLocSubMat(&Ci0, 1, 0);  	pLocSubMat(Ai0, 1, 0);	
      } /* for (i0...) */ 

      pLocSubMat(&Ci0, 1, 0);  		pLocSubMat(Ai0, 1, 0);
    } /* for (i...) */  	    	
  }  
  if (WS != NULL)
    MatFree(WS); 
  TRACE_EXIT; 
} /* Tri1MatMult() */


void Tri1MatPartMult(M, N, K, alpha, UpLo, Diag, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta; 
  char UpLo, Diag; DistMat A; pLocMat B, C; 
{ int k, m;	
  int Np;
  int L = (UpLo==Lower)^DM_Trans(A);	int TrA = DM_Trans(A);
  DistMatStore Akk_;			
  LocMat A0k, Bk0, Ci0;    		/* used to prevent side effects *B,..*/ 
  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%dx%d,%s,%c%c %d a%d %d %d,a%d %d %d,%s,a%d %d %d)",
     M, N, K, pFD2Str(alpha), UpLo, Diag, DM_Trans(A), 
     FD_ADR(LM_ADR(lA)), LM_RInc(lA), LM_CInc(lA),
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD2Str(beta), 
     FD_ADR(pLM_ADR(C)), pLM_RInc(C), pLM_CInc(C)),
    (M==0 || pLM_IncOKUnit(K, N, B)) && (K==0 || pLM_IncOKUnit(M, N, C)) 
    && DistMatOK(M, K, A) && 0<=M && 0<=N && 0<=K && UpLoOK(UpLo), 
    0,  "Tri1MatPartMult", 3, TRACE_COMPUTE
  );  


  Np = OUTER_PARTITION_WIDTH(K);
  if (Np>=MAX(M, N)) {		
    Tri1MatMult(M, N, K, alpha, UpLo, Diag, A, B, beta, C);
    TRACE_RETURN; 
  }
    
  LM_Copy(0, Bk0, *B);			LM_Copy(0, Ci0, *C);
  if (!L) {
    LM_Copy(TrA, A0k, DM_Loc(A));	m=0;
  }  
  k=0; 					
  while (k < K) {
    int dK = K - k; 			int dKG, di;
    Np = OUTER_PARTITION_WIDTH(dK); 
    if (2*dK > 3*INNER_VECSIZE && Np > MAX(M,N)) {
      dK = INNER_VECSIZE;
      Np = OUTER_PARTITION_WIDTH(INNER_VECSIZE);
    }

    dKG = L2Gd(dK, DM_DN(A));    	di = G2Ld(dKG, DM_DM(A)); 
    di = L? MIN(M, di): MIN(M-m, di);
    
    Tri1MatMult(di, N, dK, alpha, UpLo, Diag, A, &Bk0, beta, &Ci0);    
    pLocSubMat(&Ci0, di, 0);		    
    if (L) {
      LocMat Ai0;			LM_Copy(TrA, Ai0, DM_Loc(A));
      M -= di; 
      MatPartMult(M, N, dK, alpha, pLocSubMat(&Ai0, di, 0), &Bk0, beta, &Ci0);
      beta = NULL;			/* now all elts of C have been scaled*/
    } else {
      MatPartMult(m, N, dK, alpha, &A0k, &Bk0, NULL, C);	
      m += di;				pLocSubMat(&A0k, 0, dK);
    }
    pLocSubMat(&Bk0, dK, 0);
    k += dK;
    if (k < K)
      A = SubMat_(&Akk_, A, dKG, dKG);
  } /* while (k...) */
  TRACE_EXIT; 
} /* Tri1MatPartMult() */

