/************************************************************************/
/*       CellTri1Matrix: Cell Tri Input Matrix procedures 		*/
/************************************************************************/
/* SCCS INFO @(#)CellTri1Matrix.h 1.2	last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 97/03 
*/
/* set of non-communication triangular matrix routines used for DBLAS.
 * These routines can be run on cells separately, but the operations
 * assume block matrix storage on a P[0] x P[1] mesh.
 * compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking is turned on
 *
 *              if macro NO_PARTITION defined, partitioning methods
 *              in Tri1MatPartMult() are minimized.
 */

#ifndef __CellTri1Matrix_INCL__
#define __CellTri1Matrix_INCL__

#include "MatDim.h"
#include "DistMat.h"

/********************* exported symbols  ***********************************/
 
#ifndef CELLBLAS
 
#if FDATATYPE==SINGLE
#define Tri1MatMult		DSTri1MatMult
#define Tri1MatPartMult		DSTri1MatPartMult
#elif FDATATYPE==DOUBLE
#define Tri1MatMult             DDTri1MatMult
#define Tri1MatPartMult		DDTri1MatPartMult
#elif FDATATYPE==COMPLEX
#define Tri1MatMult             DCTri1MatMult
#define Tri1MatPartMult		DCTri1MatPartMult
#else
#define Tri1MatMult             DZTri1MatMult
#define Tri1MatPartMult		DZTri1MatPartMult
#endif

#else   /* rename to avoid name clash */

#if FDATATYPE==SINGLE
#define Tri1MatMult		STri1MatMult
#define Tri1MatPartMult		STri1MatPartMult
#elif FDATATYPE==DOUBLE
#define Tri1MatMult             DTri1MatMult
#define Tri1MatPartMult		DTri1MatPartMult
#elif FDATATYPE==COMPLEX
#define Tri1MatMult             CTri1MatMult
#define Tri1MatPartMult		CTri1MatPartMult
#else
#define Tri1MatMult             ZTri1MatMult
#define Tri1MatPartMult		ZTri1MatPartMult
#endif

#endif /* CELLBLAS */


void Tri1MatMult  STDC_PARAMS(
	(int M, int N, int K, FDATA *alpha, char UpLo, char Diag, 
	 DistMat A, pLocMat B, FDATA *beta, pLocMat C)
);
/* pre:  DistMatOK(M, K, A) && pLM_IncOKUnit(K, N, B) &&
 *       pLM_IncOKUnit(M, N, C) && UpLoOK(UpLo) && 0 <= M, N, K &&
 *       DM_Loc(A) is local MxK block triangular matrix of form UpLo
 * post: C = (*alpha)DM_Loc(A)*B +(*beta)*C;  C is MxN
 * notes:generally A would be the cell view of a global M' x M' tri matrix.
 *       in this case,  M:N ~ d_P(DM_DM(A)):d_P(DM_DN(A)), or more precisely:
 *              M' = L2Gd(M, DM_DM(A)) = L2Gd(N, DM_DN(A))
 *	 Note that elt. (0,0) of A must be on a diagonal; M & K can 
 * 	 be set values corresponding to a trapezoidal matrix, 
 *	 provided this holds.
 */

void Tri1MatMult  STDC_PARAMS(
	(int M, int N, int K, FDATA *alpha, char UpLo, char Diag, 
	 DistMat A, pLocMat B, FDATA *beta, pLocMat C)
);
/* pre/post: as for Tri1MatMult()
 * notes:    performs partitioning over K in units of INNER_VECSIZE
 */
#endif /*__CellTri1Matrix_INCL__*/
