/************************************************************************/
/*       CellTriMatrix: Triangular Output Matrix procedures for cells  	*/
/************************************************************************/
/* SCCS INFO @(#)CellTriMatrix.c 1.4	last modified P. Strazdins, 00/08/01
   first created: P. Strazdins 94/08/19 
*/
/* set of non-communication matrix routines used for DBLAS.
 * These routines can be run on cells separately, but the operations
 * assume block matrix storage on a P[0] x P[1] mesh.
 * compilation: if macro NO_PART defined, partitioning methods
 *              in TriMatPartMult() are minimized.
 */
#include "CellTriMatrix.h"
#include "FDATAOps.h"
#include "BLASOptions.h"
#include "FDAssert.h"  

#include "MatDim.h"
#include "CellMatrix.h"
#include "MachineParams.h"

#ifndef OLD_PARTMM
/********************* BLAS-like triangular block routines *******************/
#define lC DM_Loc(C)

#define GetLDiagLen(UnitSlope, i, Di, Dj, j, N) 			\
	(j) = (UnitSlope)? (i)+((UnitSlope)+1)/2: DiagLen(i, 1, Di, Dj); (j) = MIN(j, N)
#define GetLDiagLen_(UnitSlope, i, Di, Dj, j, N) 			\
	(j) = (UnitSlope)? (i)+((UnitSlope)+1)/2: DiagLen_(i, 1, Di, Dj); (j) = MIN(j, N)
#define GetUDiagLen(UnitSlope, i, Di, Dj, j, N) 			\
	(j) = (UnitSlope)? (i)+(1-(UnitSlope))/2: DiagLen(i, 0, Di, Dj); (j) = MIN(j, N)
#define GetUDiagLen_(UnitSlope, i, Di, Dj, j, N) 			\
	(j) = (UnitSlope)? (i)+(1-(UnitSlope))/2: DiagLen_(i, L, Di, Dj); (j) = MIN(j, N)
	
# define DiagLen_(i, L, Di, Dj) G2Ld(L2Gd(i, Di)+(L), Dj)
#if 1
static int DiagLen(i, low, Di, Dj)
  int i; int low; pMatDim Di, Dj;
/* pre:   i>=0; the last i-1 calls DiagLen had params (0,Di,Dj);...;(i-1,Di,Dj)
 * post:  returns j = G2Ld(L2Gd(i, Di)+low, Dj), 
 *	  ie. j is the local length along dimension Dj of local row i 
 *	  for a lower global triangular matrix of dim specifiers {Di, Dj} 
 * notes: reference: A. Petitet's PhD thesis (CS Dept UTK Dec 1996) Ch 2
 *	  L is LCM-table value to access diag r0-s0 of a super-matrix of 
 *	  block offset (0,0);
 *        jB is the offset of all blocks in the strictly lower part of the matrix
 */          
{ int j, dj, p, q, s0; 
  int r, s, r0, Pr, Qs, jB, L; 
  static struct {
    int r0, r, s, Pr, Qs, jB, L;
  } DL;

  if (i==0) {
    r0 = d_B0(Di); 	  	s0 = d_B0(Dj);		
    r = d_B(Di); 		s = d_B(Dj);
    p = d_P0(Di);		q = d_P0(Dj);		
    Pr = d_P(Di)*r;		Qs = d_P(Dj)*s;	 
    L = q*s - s0 - p*r + r0,	jB = 0;	
    while (L <= -s)
      L += Qs, jB += s;
    if (p!=0)
      r0 = 0;
    if (q==0)  
      jB -= s0;
    DL.r0 = r0, DL.r = r, DL.s = s, DL.Pr = Pr, DL.Qs = Qs;
    DL.L = L, DL.jB = jB;
  } else {
    L = DL.L, jB = DL.jB;	
    r = DL.r; s = DL.s;
    r0 = DL.r0; 
    r0++;
    if (r0==r) {
      r0 = 0;	
      Pr = DL.Pr, Qs = DL.Qs;
      L -= Pr;
      while (L <= -s) 
        L += Qs, jB += s;
      DL.L = L, DL.jB = jB, DL.r0 = r0;
    } else
      DL.r0 = r0;
  }
  j = jB;
  dj = -L + r0 + low;		/* horiz. distance to diag elt. */
  if (dj > s) {
     Qs = DL.Qs;	
     if (L + Qs < r) 		/* diag elt. on next block in this cell */
       DL.L = L + Qs, DL.jB = jB + s; 	/* advance for next int'n in r-block */
     else			/* diag elt on another cell; must clip dj */
       dj = s; 
  }        
  if (dj > 0) 
    j += dj;
  
  COMMON_ASSERT_TR(
    	(TraceStr, "(%d,%d,%d %d:%d %d, %d %d:%d %d)=%d (%d) r0,L,jB=%d,%d,%d",
    	i, low, d_B0(Di), d_B(Di), d_P0(Di), d_P(Di), d_B0(Dj), d_B(Dj), d_P0(Dj), 
    	d_P(Dj), j, G2Ld(L2Gd(i, Di)+low, Dj), r0, L, jB), 
        j==G2Ld(L2Gd(i, Di)+low, Dj), 0, "DiagLen", 7, TRACE_COMPUTE
  );      ;  
  return(j);
} /* DiagLen() */ 
#else
# define DiagLen(i, L, Di, Dj) DiagLen_(i, L, Di, Dj)
#endif 

void TriMatScale(M, N, alpha, FormC, C)
  int M, N; FDATA *alpha; int FormC; DistMat C;  
{ register int i;	    		
  int UnitSlope = 0;
  
  int DoTr = LM_Inc(1-DM_Trans(C), DM_Loc(C))!=1,  
  TrC = DoTr^DM_Trans(C),           	L = DoTr^(FormC==Lower);

  pMatDim Di, Dj;			/* points to dim. of C corresp i & j */
  LocMat Ci0;				/* prevent side effects DM_Loc(C) */
  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%d,%s,%c,%d a%d %d %d)", M, N, pFD2Str(alpha), FormC, 
     DM_Trans(C), FD_ADR(LM_ADR(DM_Loc(C))), LM_RInc(DM_Loc(C)), LM_CInc(DM_Loc(C))),
    UpLoOK(FormC) && (!DM_Trans(C)? LM_IncOK(M, N, lC): LM_IncOK(N, M, lC)), 
    0, "TriMatScale", 3, TRACE_COMPUTE
  );
  
  if (pFD_UNITY(alpha) || N==0) {       /* fast exit*/
    TRACE_RETURN;
  }
  if (DoTr)                            /* optimize memory refs on C */
    SWAP(int, M, N);
  LM_Copy(TrC, Ci0, DM_Loc(C));
  Di = DM_D(TrC, C);			Dj = DM_D(1-TrC, C);
  
  if ((d_P(Di)==d_P(Dj) && d_B(Di)==1 && d_B(Dj)==1) ||
       (d_P0(Di)==0 && Blk1d_(M, Di) && d_P0(Dj)==0 && Blk1d_(N, Dj)))  {
    UnitSlope = 1;
    if (L) {
      if (d_P0(Di)<d_P0(Dj))
        UnitSlope = -1;
    } else {
      if (d_P0(Di)> d_P0(Dj)) 
	UnitSlope = -1;
    }	
  }

  TRACE_LEVEL_SHIFT(4);  		/* suppress traces in MatScale() calls*/

  if (L) {
    int j;
    for (i=0; i < M; i++) {
      GetLDiagLen(UnitSlope, i, Di, Dj, j, N);
      MatScale(1, j, alpha, &Ci0);
      pLocSubMat(&Ci0, 1, 0);
    } /* for (i...) */
  } else {
    int jp = 0, j;
    for (i=0; i < M; i++) {
      GetUDiagLen(UnitSlope, i, Di, Dj, j, N);
      pLocSubMat(&Ci0, 0, j-jp);	jp=j;     
      MatScale(1, N-j, alpha, &Ci0);
      pLocSubMat(&Ci0, 1, 0);	
    }
  }  
  TRACE_LEVEL_SHIFT(-4);
  TRACE_EXIT;
} /* TriMatScale() */


static inline void AxpyM(n, alpha, Ax, Bx, IncB, beta, Cx, IncC)
  int n, IncB, IncC; FDATA *alpha, *Ax, *Bx, *beta, *Cx; 
/* axpy-like routine for avoiding MatMult() (BLAS) overheads; assumes n small */
{ FDATA A = *Ax;	
  if (!pFD_UNITY(alpha)) {
    FD_MULSET(A, =, A, *alpha);  
  }  
  if (beta != NULL)
    for (;n>0; n--) {
       FD_MULSET(*Cx, =, *Cx, *beta);
     FD_MULSET(*Cx, +=, A, *Bx);
     Cx += IncC;		Bx += IncB;
    } 
  else 
    for (;n>0; n--) {
     FD_MULSET(*Cx, +=, A, *Bx);
     Cx += IncC;		Bx += IncB;
    } 
} /* AxpyM() */

   
void TriMatMult(M, N, K, alpha, A, B, beta, FormC, C)
  int M, N, K;  FDATA *alpha, *beta; pLocMat A, B; int FormC; DistMat C; 
/* notes: di performs a `slicing' of rows (or columns of) C.
 *	 its optimal value depends on the minimum outer dimension for 
 *	 optimal cell GEMM performance (this may be a function of K). 
 *	 Underestimation can yield to low cell GEMM performance;
 *	 overestimation can lead to a higher fraction of vector-matrix
 *	 computations for the triangular portion of C.
 * 	 Computation may be transposed to ensure unit stride across slice of C.
 */ 
{ int i;	
  pMatDim Di, Dj;			/* points to dim. of C corresp i & j */
  int di = (K < K0_MINOR_OUTER_VECSIZE)? MINOR_OUTER_VECSIZE_K0: 
  					 MINOR_OUTER_VECSIZE;
  int UnitSlope = 0;
  LocMat Ai0, B_, B0j, Ci0, Cij;    	/* used to prevent side effects *A,..*/ 

  int DoTr = LM_Inc(1-DM_Trans(C), DM_Loc(C))!=1,
      TrC = DoTr^DM_Trans(C),		L = DoTr^(FormC==Lower);  

#define TriMatMult_Cond 						\
    (N==0 || pLM_IncOKUnit(M, K, A)) && (M==0 || pLM_IncOKUnit(K, N, B))\
    && (!DM_Trans(C)? LM_IncOKUnit(M, N, lC): LM_IncOKUnit(N, M, lC))	\
    && 0<=M && 0<=N && 0<=K && UpLoOK(FormC)
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%dx%d,%s,a%d %d %d,a%d %d %d,%s,%c,%d a%d %d %d)",
     M, N, K, pFD2Str(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A), 
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD2Str(beta), FormC, 
     DM_Trans(C), FD_ADR(LM_ADR(lC)), LM_RInc(lC), LM_CInc(lC)),
     TriMatMult_Cond, 0,
    "TriMatMult", 3, TRACE_COMPUTE
  );  

  if (N==0) {             		/* fast exit*/
    TRACE_RETURN;
  }
  if (pFD_UNITY(beta) && K==1)		/* a small optimization */
    beta = NULL;

  if (DoTr) {				/* optimize memory refs on C */
    SWAP(int, M, N);	    
    LM_Copy(1, Ai0, *B);    		LM_Copy(1, B_, *A);
    B = &B_;				
  } else {
    LM_Copy(0, Ai0, *A);    	
  }
  LM_Copy(TrC, Ci0, DM_Loc(C));
  Di = DM_D(TrC, C);	    		Dj = DM_D(1-TrC, C);
  LM_Copy(0, B0j, *B);			LM_Copy(0, Cij, Ci0);

  if ((d_P(Di)==d_P(Dj) && d_B(Di)==1 && d_B(Dj)==1) ||
       (d_P0(Di)==0 && Blk1d_(M, Di) && d_P0(Dj)==0 && Blk1d_(N, Dj)))  {
    UnitSlope = 1;
    if (L) {
      if (d_P0(Di)<d_P0(Dj))
        UnitSlope = -1;
    } else {
      if (d_P0(Di)> d_P0(Dj)) 
	UnitSlope = -1;
    }	
    if (K==1 && M<=MIN_VECSIZE)
      di = M;          	/* small mult; use inlined MM to reduce s/w overheads */
  }

  if ((d_P(Di)==d_P(Dj)) && d_P0(Dj)!=d_P0(Di) && di < d_B(Di) &&
  	BlkB0d(0, Di)==0)
    di = d_B(Di);	/* coalesc MM calls by making step size = block size */

  TRACE_LEVEL_SHIFT(4);  		/* suppress traces in MatMult() calls */

  if (L) {  
    for (i=0; i < M; i+= di) {
      /* C'[i..i0Max, 0..jp-1] is the rectangular portion in this slice */ 
      int i0, j, i0Max = MIN(i + di-1, M-1);
      int jp = DiagLen_(i, L, Di, Dj);
      if (jp > 0) {
      	if (K==1 && pFD_UNITY(beta))
      	  R1Mult(i0Max-i+1, jp, alpha, &Ai0, B, &Ci0);
      	else  
          MatMult(i0Max-i+1, jp, K, alpha, &Ai0, B, beta, &Ci0);
        pLocSubMat_x(&Cij, &Ci0, 0, jp);	pLocSubMat_x(&B0j, B, 0, jp);
      }
      for (i0=i; i0<=i0Max; i0++) {
        /* iterate rowwise over triangular portion of slice */	
	GetLDiagLen(UnitSlope, i0, Di, Dj, j, N);
        if (K>1 && j>jp)
          VMMult(j-jp, K, alpha, &Ai0, &B0j, beta, &Cij);
        else if (j>jp) 			/* reduce overheads */
          AxpyM(j-jp, alpha, LM_ADR(Ai0), LM_ADR(B0j), LM_CInc(B0j), beta, 
		LM_ADR(Cij), LM_CInc(Cij));	    
        pLocSubMat(&Cij, 1, 0);		pLocSubMat(&Ai0, 1, 0); 
      } /* for (i0...) */ 
      pLocSubMat(&Ci0, i0Max-i+1, 0);	jp = j;
    } /* for (i...) */  	  
  } else {
    int j = 0;
    LocMat Bjn;				LM_Copy(0, Bjn, B0j);
    for (i=0; i < M; i+= di) {
      /* C'[i..i0Max, jn..N-1] is the rectangular portion in this slice */ 
      int i0, i0Max = MIN(i + di-1, M-1);
      int jn;				
      GetUDiagLen_(UnitSlope, i0Max, Di, Dj, jn, N);/* avoid use GetDiagLenU(); o.w DiagLen() error*/
      if (jn < N) {
        pLocSubMat_x(&Ci0, &Cij, 0, jn-j);  pLocSubMat_x(&Bjn, &B0j, 0, jn-j);
        if (K==1 && pFD_UNITY(beta))
          R1Mult(i0Max-i+1, N-jn, alpha, &Ai0, &Bjn, &Ci0);
        else  
          MatMult(i0Max-i+1, N-jn, K, alpha, &Ai0, &Bjn, beta, &Ci0);
      }
      for (i0=i; i0<=i0Max; i0++) {
        /* iterate rowwise over proper triangular portion of slice */	        
	int jp=j;
        GetUDiagLen(UnitSlope, i0, Di, Dj, j, N);
	pLocSubMat(&B0j, 0, j-jp);	pLocSubMat(&Cij, 0, j-jp);
        if (jn>j && K > 1) 	
          VMMult(jn-j, K, alpha, &Ai0, &B0j, beta, &Cij);
        else if (jn>j)  		/* reduce overheads */
          AxpyM(jn-j, alpha, LM_ADR(Ai0), LM_ADR(B0j), LM_CInc(B0j), beta, 
		LM_ADR(Cij), LM_CInc(Cij));
        pLocSubMat(&Ai0, 1, 0); 	pLocSubMat(&Cij, 1, 0);
      } /* for (i0...) */ 
    } /* for (i...) */  	  
  } 
  TRACE_LEVEL_SHIFT(-4);
  TRACE_EXIT;
} /* TriMatMult() */

void TriMatPartMult(M, N, K, alpha, A, B, beta, FormC, C)
  int M, N, K;  FDATA *alpha, *beta; pLocMat A, B; int FormC; DistMat C;
/* notes:see notes for TriMatMult(). for 2., di should be ~  OUTER_VECSIZE
 *	 for 3., and assuming large M, N, the granularity issue is not
 *	 relevant. Hence 2 versions of partitioning are required, 
 *	 slicing in different dimensions according to the stride of C
 */
{ 
  int DoTr = LM_Inc(MM_INNER_K^DM_Trans(C), DM_Loc(C))!=1,
      TrC = DoTr^DM_Trans(C),		L = DoTr^(FormC==Lower);  
  int k, Np;
  LocMat A_, B_, C_;                    /* used to prevent side effects *A,..*/ 
  pMatDim Di, Dj;			/* points to dim. of C corresp i & j */
  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%dx%d,%s,a%d %d %d,a%d %d %d,%s,%c,%d a%d %d %d)",
     M, N, K, pFD2Str(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A), 
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD2Str(beta), FormC, 
     DM_Trans(C), FD_ADR(LM_ADR(lC)), LM_RInc(lC), LM_CInc(lC)),
     TriMatMult_Cond, 0,
    "TriMatPartMult", 2, TRACE_COMPUTE
  );  
      
  TRACE_LEVEL_SHIFT(2);

/*
printf("C:tr=%d, I=%d %d, DoTr=%d, L=%d, TrC=%d\n", 
       DM_Trans(C), LM_RInc(DM_Loc(C)), LM_CInc(DM_Loc(C)), DoTr, L, TrC);
*/
  
  Np = OUTER_PARTITION_WIDTH(K);
  if (Np>=MAX(M, N)) {                  /* reduce overheads for small comp'n */
    FDATA * WS = PackMMOpdsP(M, N, Np, K, &alpha, &A, &B);
    TriMatMult(M, N, K, alpha, A, B, beta, FormC, C);
    if (WS != NULL)
      MatFree(WS);
    TRACE_LEVEL_SHIFT(-2);		TRACE_RETURN;
  }  
          
  if (DoTr) {  /* transpose computation to achieve unit stride along rows C_ */
    SWAP(int, M, N);     		
    DM_Transpose(C);			FormC = TransForm(FormC);
    LM_Copy(1, A_, *B);     		LM_Copy(1, B_, *A);     
  } else {
    LM_Copy(0, A_, *A); 		LM_Copy(0, B_, *B);
  }
  A = &A_;            			B = &B_;
  LM_Copy(TrC, C_, lC);
  Di = DM_D(TrC, C); 			Dj = DM_D(1-TrC, C);
       
  k=0;
  while (k < K) {
    int dK = K-k, j;
    FDATA *alpha_ = alpha;
    FDATA *WS = NULL;                   int CpyB;
    LocMat Bj;                          LM_Copy(0, Bj, *B);
 
    Np = OUTER_PARTITION_WIDTH(dK);
    if (2*dK > 3*INNER_VECSIZE && Np > MAX(M,N)) {
      dK = INNER_VECSIZE;
      Np = OUTER_PARTITION_WIDTH(INNER_VECSIZE); 
    } 

    WS = PackMMOpdsP(M, N, Np, dK, &alpha_, &A, &B);
    CpyB = pLM_ADR(B)!=LM_ADR(Bj);

    for (j=0; j < N; j+=Np) {
      DistMat_RVAlloc(1);	
      LocMat Ai, Cj;     		pLocMat pBj = CpyB? B: &Bj;
      int n = MIN(Np, N-j),		jG = L2Gd(j, Dj), 	
	  i = G2Ld(jG, Di),		in = G2Ld(L2Gd(j+n, Dj), Di);
      i = MIN(i, M);              	in = MIN(in, M);
      if (j>0 && CpyB)
        MatCopy(dK, n, &Bj, PackMMAlphaB, B);
      pLocSubMat_(&Ai, A, i, 0);
      if (FormC==Lower) {
	TriMatMult(in-i, n, dK, alpha_, &Ai, pBj, beta, 
		   FormC, SubMat(C, jG, jG));	
	MatMult(M-in, n, dK, alpha_, pLocSubMat(&Ai, in-i, 0), pBj, beta, 
	        pLocSubMat_(&Cj, &C_, in, j));
      } else {
	MatMult(i, n, dK, alpha_, A, pBj, beta, pLocSubMat_(&Cj, &C_, 0, j));
	TriMatMult(in-i, n, dK, alpha_, &Ai, pBj, beta, FormC, SubMat(C, jG, jG));
      }
      pLocSubMat(&Bj, 0, n); 
    }
    if (WS != NULL)
      MatFree(WS);        
    beta = NULL;
    A = &A_;                            B = &B_;
    pLocSubMat(A, 0, dK);               pLocSubMat(B, dK, 0);
    k += dK;
  } /* while (k<K) */

  TRACE_LEVEL_SHIFT(-2);
  TRACE_EXIT;  
} /*TriMatPartMult()*/


#else
/************************* old code ***************************************/

#define lC DM_Loc(C)


#if 1
static int DiagLen(i, Di, Dj)
  int i; pMatDim Di, Dj;
/* pre:   i>=0; the last i-1 calls DiagLen had params (0,Di,Dj);...;(i-1,Di,Dj)
 * post:  returns j = G2Ld(L2Gd(i, Di)+1, Dj), 
 *	  ie. j is the local length along dimension Dj of local row i 
 *	  for a lower global triangular matrix of dim specifiers {Di, Dj} 
 * notes: reference: A. Petitet's PhD thesis (CS Dept UTK Dec 1996) Ch 2
 *	  L is LCM-table value to access diag r0-s0 of a super-matrix of 
 *	  block offset (0,0);
 *        jB is the offset of all blocks in the strictly lower part of the matrix
 */          
{ int j, dj, p, q, s0; 
  int r, s, r0, Pr, Qs, jB, L; 
  static struct {
    int r0, r, s, Pr, Qs, jB, L;
  } DL;

  if (i==0) {
    r0 = d_B0(Di); 	  	s0 = d_B0(Dj);		
    r = d_B(Di); 		s = d_B(Dj);
    p = d_P0(Di);		q = d_P0(Dj);		
    Pr = d_P(Di)*r;		Qs = d_P(Dj)*s;	 
    L = q*s - s0 - p*r + r0,	jB = 0;	
    while (L <= -s)
      L += Qs, jB += s;
    if (p!=0)
      r0 = 0;
    if (q==0)  
      jB -= s0;
    DL.r0 = r0, DL.r = r, DL.s = s, DL.Pr = Pr, DL.Qs = Qs;
    DL.L = L, DL.jB = jB;
  } else {
    L = DL.L, jB = DL.jB;	
    r = DL.r; s = DL.s;
    r0 = DL.r0; 
    r0++;
    if (r0==r) {
      r0 = 0;	
      Pr = DL.Pr, Qs = DL.Qs;
      L -= Pr;
      while (L <= -s) 
        L += Qs, jB += s;
      DL.L = L, DL.jB = jB, DL.r0 = r0;
    } else
      DL.r0 = r0;
  }
  j = jB;
  dj = -L + r0 + 1;		/* horiz. distance to diag elt. */
  if (dj > s) {
     Qs = DL.Qs;	
     if (L + Qs < r) 		/* diag elt. on next block in this cell */
       DL.L = L + Qs, DL.jB = jB + s; 	/* advance for next int'n in r-block */
     else			/* diag elt on another cell; must clip dj */
       dj = s; 
  }        
  if (dj > 0) 
    j += dj;
  
  COMMON_ASSERT_TR(
    	(TraceStr, "(%d,%d %d:%d %d, %d %d:%d %d)=%d (%d) r0,L,jB=%d,%d,%d",
    	i, d_B0(Di), d_B(Di), d_P0(Di), d_P(Di), d_B0(Dj), d_B(Dj), d_P0(Dj), 
    	d_P(Dj), j, G2Ld(L2Gd(i, Di)+1, Dj), r0, L, jB), 
        j==G2Ld(L2Gd(i, Di)+1, Dj), 0, "DiagLen", 7, TRACE_COMPUTE
  );      ;  
  return(j);
} /* DiagLen() */ 
#else
# define DiagLen(i, Di, Dj) G2Ld(L2Gd(i, Di)+1, Dj)
#endif 

void TriMatScale(M, N, alpha, FormC, C)
  int M, N; FDATA *alpha; char int; DistMat C;  
{ register int i;	    		
  register int L = (FormC == Lower);	int TrC = 1-(L^DM_Trans(C));
  pMatDim Di, Dj;			/* points to dim. of C corresp i & j */
  LocMat Ci0;				/* prevent side effects DM_Loc(C) */
  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%d,%s,%c,%d a%d %d %d)", M, N, pFD2Str(alpha), FormC, 
     DM_Trans(C), FD_ADR(LM_ADR(DM_Loc(C))), LM_RInc(DM_Loc(C)), LM_CInc(DM_Loc(C))),
    UpLoOK(FormC) && (!DM_Trans(C)? LM_IncOK(M, N, lC): LM_IncOK(N, M, lC)), 
    0, "TriMatScale", 3, TRACE_COMPUTE
  );
  
  if (pFD_UNITY(alpha) || N==0) {       /* fast exit*/
    TRACE_RETURN;
  }
  if (!L) 
    SWAP(int, M, N);
  LM_Copy(TrC, Ci0, DM_Loc(C));
  Di = DM_D(TrC, C);			Dj = DM_D(1-TrC, C);
  

  TRACE_LEVEL_SHIFT(4);  		/* suppress traces in MatScale() calls*/
  for (i=0; i < M; i++) {
    int j = DiagLen(i, Di, Dj);		j = MIN(j, N);
    MatScale(1, j, alpha, &Ci0);
    pLocSubMat(&Ci0, 1, 0);
  } /* for (i...) */
  TRACE_LEVEL_SHIFT(-4);
  TRACE_EXIT;
} /* TriMatScale() */


void TriMatMult(M, N, K, alpha, A, B, beta, FormC, C)
  int M, N, K;  FDATA *alpha, *beta; pLocMat A, B; char int; DistMat C; 
/* notes:2. di performs a `slicing' of (rows or columns of) C.
 *	 its optimal value depends on the minimum outer dimension for 
 *	 optimal cell GEMM performance (this may be a function of K). 
 *	 Underestimation can yield to low cell GEMM performance;
 *	 overestimation can lead to a higher fraction of vector-matrix
 *	 computations for the triangular portion of C.
 *
 */ 
{ int i;	
  int L = (FormC == Lower);		int TrC = 1-(L^DM_Trans(C));
  pMatDim Di, Dj;			/* points to dim. of C corresp i & j */
  int di = (K < K0_MINOR_OUTER_VECSIZE)? MINOR_OUTER_VECSIZE_K0: 
  					 MINOR_OUTER_VECSIZE;
  LocMat Ai0, B_, B0j, Ci0, Cij;    	/* used to prevent side effects *A,..*/ 
  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%dx%d,%s,a%d %d %d,a%d %d %d,%s,%c,%d a%d %d %d)",
     M, N, K, pFD2Str(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A), 
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD2Str(beta), FormC, 
     DM_Trans(C), FD_ADR(LM_ADR(lC)), LM_RInc(lC), LM_CInc(lC)),
    (N==0 || pLM_IncOKUnit(M, K, A)) && (M==0 || pLM_IncOKUnit(K, N, B)) 
    && (!DM_Trans(C)? LM_IncOKUnit(M, N, lC): LM_IncOKUnit(N, M, lC))
    && 0<=M && 0<=N && 0<=K && UpLoOK(FormC), 0,
    "TriMatMult", 3, TRACE_COMPUTE
  );  

  if (N==0) {             		/* fast exit*/
    TRACE_RETURN;
  }
  if (pFD_UNITY(beta) && K==1)		/* a small optimization */
    beta = NULL;

  if (!L) {     	   /* transpose computation for `Upper' triangular */
    SWAP(int, M, N);	    
    LM_Copy(1, Ai0, *B);    		LM_Copy(1, B_, *A);
    B = &B_;
  } else {
    LM_Copy(0, Ai0, *A);    	
  }
  LM_Copy(TrC, Ci0, DM_Loc(C));
  Di = DM_D(TrC, C);	    		Dj = DM_D(1-TrC, C);
  LM_Copy(0, B0j, *B);			LM_Copy(0, Cij, Ci0);

  if (K==1 && M<=MIN_VECSIZE && 
      ((d_P(Di)==d_P(Dj) && d_B(Di)==1 && d_B(Dj)==1) || 
       (d_P0(Di)==0 && Blk1d_(M, Di) && d_P0(Dj)==0 && Blk1d_(N, Dj))) ) {  	
       /* small mult. with slope of 1; inline to reduce s/w overheads */
    for (i=0; i < M; i++) {
      int j = MIN(i+1, N);	
      FDATA *Cx = LM_ADR(Ci0),		*Bx = LM_ADR(B0j);	
      FDATA A = *LM_ADR(Ai0);	
      if (!pFD_UNITY(alpha)) A *= *alpha;      
      for (; j > 0; j--) {
	if (beta != NULL)
  	  FD_MULSET(*Cx, =, *Cx, *beta);
        FD_MULSET(*Cx, +=, A, *Bx);
	Cx += LM_CInc(Ci0);		Bx += LM_CInc(B0j);
      } /* for (j...) */         	
      pLocSubMat(&Ci0, 1, 0);		pLocSubMat(&Ai0, 1, 0);  
    } /* for (i...) */
    TRACE_RETURN;
  }  

  TRACE_LEVEL_SHIFT(4);  		/* suppress traces in MatMult() calls */
  
  for (i=0; i < M; i+= di) {
    /* C'[i..i0Max, 0..j-1] is the rectangular portion in this slice */ 
    int i0, i0Max = MIN(i + di-1, M-1);
    int j = DiagLen(i, Di, Dj);		j = MIN(j, N);
    
    MatMult(i0Max-i+1, j, K, alpha, &Ai0, B, beta, &Ci0);

    pLocSubMat_x(&Cij, &Ci0, 0, j);	pLocSubMat_x(&B0j, B, 0, j);
    for (i0=i+1; i0 <= i0Max; i0++) {
      /* iterate rowwise over proper triangular portion of slice */	
      int j0 = DiagLen(i0, Di, Dj);     j0 = MIN(j0, N);
      pLocSubMat(&Cij, 1, 0);		pLocSubMat(&Ai0, 1, 0); 

      if (j0 > j && K > 1) 	
        MatMult(1, j0-j, K, alpha, &Ai0, &B0j, beta, &Cij);
      else if (j0 > j) { 		/* inline to reduce call overheads */
        FDATA A = *LM_ADR(Ai0);	
        FDATA *Cx = LM_ADR(Cij),	*Bx = LM_ADR(B0j);	
        if (!pFD_UNITY(alpha)) A *= *alpha;
        for (;j < j0; j0--) {
	  if (beta != NULL)
  	    FD_MULSET(*Cx, =, *Cx, *beta);
          FD_MULSET(*Cx, +=, A, *Bx);
	  Cx += LM_CInc(Cij);		Bx += LM_CInc(B0j);
        }          	
      }
    } /* for (i0...) */ 
    pLocSubMat(&Ci0, i0Max-i+1, 0);	pLocSubMat(&Ai0, 1, 0);  
  } /* for (i...) */  	  
  TRACE_LEVEL_SHIFT(-4);
  TRACE_EXIT;
} /* TriMatMult() */

#define FORCE_NOPARTITION (ForceAlgVers>=4 && ForceAlgVers%2==1)
#define FORCE_VERS2	  (ForceAlgVers>=4 && ((ForceAlgVers-4)/2)%2==1)
#define FORCE_VERS1	  (ForceAlgVers>=4 && ((ForceAlgVers-4)/2)%2==0)

void TriMatPartMult(M, N, K, alpha, A, B, beta, FormC, C)
  int M, N, K;  FDATA *alpha, *beta; pLocMat A, B; int FormC; DistMat C;
/* notes:see notes for TriMatMult(). for 2., di should be ~  OUTER_VECSIZE
 *	 for 3., and assuming large M, N, the granularity issue is not
 *	 relevant. Hence 2 versions of partitioning are required, 
 *	 slicing in different dimensions according to the relative sizes of
 *	 their procesor fields.
 */
{ int U = (FormC == Upper);		int TrC = 1-(U^DM_Trans(C));
  int k;
  LocMat A_, B_, C_;                    /* used to prevent side effects *A,..*/ 
  pMatDim Di, Dj;			/* points to dim. of C corresp i & j */
      
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%dx%d,%s,a%d %d %d,a%d %d %d,%s,%c,%d a%d %d %d)",
     M, N, K, pFD2Str(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A), 
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD2Str(beta), FormC, 
     DM_Trans(C), FD_ADR(LM_ADR(lC)), LM_RInc(lC), LM_CInc(lC)),
     (N==0 || pLM_IncOKUnit(M, K, A)) && (M==0 || pLM_IncOKUnit(K, N, B)) 
     && (!DM_Trans(C)? LM_IncOKUnit(M, N, lC): LM_IncOKUnit(N, M, lC))
     && 0<=M && 0<=N && 0<=K && UpLoOK(FormC), 0,
     "TriMatPartMult", 2, TRACE_COMPUTE
  );
  TRACE_LEVEL_SHIFT(2);
  
  if (!PartitionUseful(M, N, K)) {   	/* reduce overhead for small comput`n */
    FDATA * WS = PackMMOpds(M, N, K, &alpha, &A, &B);
    TriMatMult(M, N, K, alpha, A, B, beta, FormC, C);
    if (WS != NULL)
      MatFree(WS);
    TRACE_LEVEL_SHIFT(-2);		TRACE_RETURN;
  }  
          
  LM_Copy(TrC, C_, lC);
  Di = DM_D(TrC, C); 			Dj = DM_D(1-TrC, C);
  if (!U) {             /* transpose computation for `Lower' triangular */
    SWAP(int, M, N);     		FormC = Upper;  	
    LM_Copy(1, A_, *A);     		LM_Copy(1, B_, *B);     
    A = &B_;               	 	B = &A_;   
    DM_Transpose(C);
  }
       
  k=0;
  while (k < K) {
    int dK;
    int DoPartition = 0;
    FDATA *WS = NULL; int Mw, Ldw;      /* partitioning workspace parameters */
    LocMat Aik, Bkj, Cij;

    dK = K - k;    
#ifndef NO_PART
    if (2*dK > 3*INNER_VECSIZE && PartitionUseful(M, N, dK)) 
      dK = INNER_VECSIZE;
#endif

#define DO_VERS1 (FORCE_VERS1 || !FORCE_VERS2 && d_P(Dj)>=d_P(Di))
    Mw = OUTER_VECSIZE; 		DoPartition=0; 		WS = NULL;
#ifndef NO_PART  
    if (!FORCE_NOPARTITION && PartitionUseful(M, N, K)) { 
      int Np = DO_VERS1? N: M; 
    				/*generally = MIN(M, N)*/
      if (CopyEfficient(Np)) {
      /* try major partitioning over M & N only*/
        Ldw = Mw = (CACHESIZE/2) / dK;
        DoPartition = MatMalloc("TriMatPartMult", 0, &WS, CACHESIZE*((Np+Ldw-1)/Ldw));
      }
    }
#endif /*NO_PART*/
            
    if (DO_VERS1) {
      int i;	    
      if (!DoPartition) {             
        for (i=0; i < M; i+=Mw) {
          DistMat_RVAlloc(1);	
          int m = MIN(Mw, M-i),
              iG = L2Gd(i, Di),	j = G2Ld(iG, Dj),
              jm = G2Ld(L2Gd(i+m, Di), Dj);
          pLocSubMat_(&Aik, A, i, k);      
          jm = MIN(jm, N);   
          if (jm > j)            	
            TriMatMult(m, jm-j, dK, alpha, &Aik, pLocSubMat_(&Bkj, B, k, j), 
        	       beta, FormC, SubMat(C, iG, iG));
          if (jm < N)    
            MatPartMult(m, N-jm, dK, alpha, &Aik, pLocSubMat_(&Bkj, B, k, jm), 
        	        beta, pLocSubMat_(&Cij, &C_, i, jm));
        } /* for (i...) */
      } else {
        LocMat WA, WB, WAi, WBj;  
        LM_Init(WA, WS+CACHESIZE/2, 1, Ldw);
        for (i=0; i < M; i+=Mw) {
          /* nb. unless d_B(Dj) ~ Mw, we have jm-j<=m<=Mw since d_P(Dj)>=d_P(j)*/
          int m = MIN(Mw, M-i),				
              iG = L2Gd(i, Di), j = G2Ld(iG, Dj),
              jm = G2Ld(L2Gd(i+m, Di), Dj);
          jm = MIN(jm, N);   
          MatCopy(m, dK, pLocSubMat_(&Aik, A, i, k), alpha, &WA);	
          LM_Init(WB, WS+(j/Mw)*CACHESIZE, Ldw, 1);
          while (j < N) {
            DistMat_RVAlloc(1);	
            int j0 = j%Mw;		/* offset into current WB block */
            int n = MIN(Mw-j0, N-j), 	/* active width of current WB block */
                nt = MAX(0, MIN(n, jm-j));/* width of trapez. MM in this block*/
            if (i==0) {
      	      MatCopy(dK, MIN(Mw, N-j), pLocSubMat_(&Bkj, B, k, j-j0), NULL, &WB);
            }
            if (nt > 0) {		/* a non-empty trapezoidal section */
              int jG = L2Gd(j, Dj),	/* (jG,jG) G2L -> (i+ijm,j+j0)*/
                  ijm = G2Ld(jG, Di)-i;
              pLocSubMat_(&WBj, &WB, 0, j0);
              if (ijm > 0)		/* TriMatMult() cant handle this part */
                MatMult(ijm, nt, dK, NULL, &WA, &WBj, beta, 
                        pLocSubMat_(&Cij, &C_, i, j));
              TriMatMult(m-ijm, nt, dK, NULL, pLocSubMat_(&WAi, &WA, ijm, 0), 
            	         &WBj, beta, FormC, SubMat(C, jG, jG));
            }
            if (n>nt)
      	      MatMult(m, n-nt, dK, NULL, &WA, pLocSubMat_(&WBj, &WB, 0, j0+nt), 
      	    	      beta, pLocSubMat_(&Cij, &C_, i, j+nt)); 
	    LM_ADR(WB) += CACHESIZE;	  
	    j += n;
	  } /* while (j...) */  
        } /* for (i...) */
      } /* else (DoPartition) */ 
    } else {
      int j;
      int iA=0;		     /* A[0..iA-1,j..j+n-1] has been copied into WS*/
      if (!DoPartition) {
        for (j=0; j < N; j+=Mw) {
          DistMat_RVAlloc(1);	
          int n = MIN(Mw, N-j),
              jG = L2Gd(j, Dj), i = G2Ld(jG, Di),
              im = G2Ld(L2Gd(j+n, Dj), Di);
          pLocSubMat_(&Bkj, B, k, j);   
          i = MIN(i, M);		im = MIN(im, M);
          if (i > 0)
            MatPartMult(i, n, dK, alpha, pLocSubMat_(&Aik, A, 0, k), &Bkj, 
              		beta, pLocSubMat_(&Cij, &C_, 0, j));
          if (im > i)
            TriMatMult(im-i, n, dK, alpha, pLocSubMat_(&Aik, A, i, k), &Bkj, 
        	       beta, FormC, SubMat(C, jG, jG));
        } /* for (i...) */
      } else {
        LocMat WA, WB, WAi, WBj;   
        LM_Init(WB, WS+CACHESIZE/2, Ldw, 1);
      
        for (j=0; j < N; j+=Mw) {
          int n = MIN(Mw, N-j),		  i,
              jG = L2Gd(j, Dj), ir = G2Ld(jG, Di),
              in = G2Ld(L2Gd(j+n, Dj), Di);
          ir = MIN(ir, M);		in = MIN(in, M);
    	         
          MatCopy(dK, n, pLocSubMat_(&Bkj, B, k, j), alpha, &WB);
          LM_Init(WA, WS, 1, Ldw);
        
          i = 0;
          while (i < in) {
            DistMat_RVAlloc(1);	
            int m =  MIN(Mw, in-i), 	/* active size of current WA block */
                mr = MAX(0, MIN(m, ir-i));/* width of rect MM in this block */
            if (i+m > iA) {		/* copy i/Mw th block into WA as needed*/
              int m1 = MIN(Mw, M-i);	/* total size of current WA block */
      	      MatCopy(m1, dK, pLocSubMat_(&Aik, A, i, k), NULL, &WA);
      	      iA += m1;
      	    }
      	    if (mr > 0)
      	      MatMult(mr, n, dK, NULL, &WA, &WB, beta, 
      	    	      pLocSubMat_(&Cij, &C_, i, j));
      	    if (m > mr) {		/* TriMatMult() can handle trapezoid*/
      	      int iG = L2Gd(i+mr, Di);
      	      int ji = G2Ld(iG, Dj);	/* (iG,iG) G2L -> (i+mr,ji) */
              TriMatMult(m-mr, n-(ji-j), dK, NULL, 
             	         pLocSubMat_(&WAi, &WA, mr, 0), pLocSubMat_(&WBj, &WB, 0, ji-j), 
        	         beta, FormC, SubMat(C, iG, iG));
            }
	    LM_ADR(WA) += CACHESIZE;
	    i += m;
          } /* while (i...) */
        } /* for (j...) */ 
      } /* else (DoPartition) */
    } /* if-else on choosing aspect */
  
    if (WS != NULL)
      MatFree(WS);        
    beta = NULL;
    k += dK;
  } /* while (k<K) */

  TRACE_LEVEL_SHIFT(-2);
  TRACE_EXIT;  
} /*TriMatPartMult()*/

#endif /* defined(OLD_PARTMM) */

#undef lC


