/************************************************************************/
/*       	CellVector: FDATA Vector procedures for cells    	*/
/************************************************************************/
/* SCCS INFO @(#)CellVector.c 1.4         last modified P. Strazdins, 99/06/02
   first created: P. Strazdins 93/03/23 
*/
/* Set of support vector routines used for DBLAS, independent of distributed
 * memory issues.
 * Efficiency: the use of the following techniques for scalar arhcitectures 
 * are used (if BLAS and VANILLA not defined):
 *    1. 2-4 way loop unrolling and software pipelining where appropriate   
 *       generally, the SPARC C compiler does well with these optimizations.
 * compilation: if macro VANILLA defined, `vanilla' versions of the routines
 *              are used (slower, but useful for debugging or vectorization.
 *              if macro BLAS defined, Fortran BLAS is used whenever
 *              possible
 */

#include <math.h>			/* fabs() */
#if (FDATATYPE < COMPLEX)
# define FD_SETABS(x)	x = (FDATA) fabs((double) x)	
			/* for sparcs, gcc should compile to an fabs instrn */
/*# define FD_SETABS(x) 	if ((x) < 0.0) (x) = -(x);*/
/*# define FD_SETABS(x)	(x) = ((x) < 0.0)? -(x): (x);*/
#else
# define FD_SETABS(x)	FD_REAL(x) = (FD_BASE) fabs((double) FD_REAL(x) ) + \
				     (FD_BASE) fabs((double) FD_IMAG(x) )
#endif

#include "FDATAOps.h"
#include "FDAssert.h"
#include "CellVector.h"

static FD_BASE Zero = 0.0;      /* force gcc to use SPARC FPU regs - faster!*/
static FD_BASE Minus1 = -1.0;


void VecSwap(n, A, IncA, B, IncB)
  int n, IncA, IncB; FDATA A[], B[];
{ FD_register a, b;
  COMMON_ASSERT_TR(
    (TraceStr, "(%d,a%d %d,a%d %d)",n, FD_ADR(A), IncA, FD_ADR(B), IncB),
    1, 0, "VecSwap", 4, TRACE_COMPUTE
  ); 
#ifdef VANILLA
  for (; n > 0; n--) {
    a = *A; 	b=*B;
    *B = a;	*A = b;
    A += IncA; 	B += IncB;      	
  } /* for (n...) */
#elif defined BLAS
  FTN_SWAP(&n, A, &IncA, B, &IncB); 
#else   
# ifndef ULTRA
  {
    register int IncA2=IncA+IncA, IncB2=IncB+IncB;
    FD_register a1, b1; 

    if (n%2==1) {
      a = *A;  b = *B;		      
      *B = a;  *A = b;
      B += IncB; A += IncA;
      n--;            
    }
    for (n=-n; n<0; n+=2) {
      a = *A;		a1 = *(A+IncA);	
      b = *B;		b1 = *(B+IncB); 
      *B = a; 		*(B+IncB) = a1;	
      B += IncB2;
      *A = b;		*(A+IncA) = b1;
      A += IncA2;
    }
  } 
# else 
  { 
    register int IncA2=IncA+IncA,  	IncB2=IncB+IncB;
    register int IncA3=IncA2+IncA,  	IncB3=IncB2+IncB;
    register int IncA4=IncA2+IncA2,  	IncB4=IncB2+IncB2;
    FD_register a1, b1, a2, b2, a3, b3; 
    register FDATA *As;
    int i;

    if (A==B && IncA==IncB) {
      if (n>0)				/* quick exit possible */
        return;
      else {   				/* fool gcc into using FPU registers */
        a = Zero*Minus1, b = a+Zero, a1 = b * Zero, a2 = a1 + Zero,
        a3 = a2 * Zero,  b1 = b+Zero, b2 = b1 * Zero, b3 = b2 + Zero; 	
        for (i=0; i < n; i++)
          a = *A + a3 * b3, *A = a, A += IncA;          
      }    
    }

    for (i=0; i<n%4; i++) {
      a = *A;  b = *B;		      
      *B = a;  *A = b;
      B += IncB; A += IncA;
    }
    As = A - IncA4;
    for (n=-n+i; n<0; n+=4) {
#if 0
      a = *A;		a1 = *(A+IncA);	
      a2 = *(A+IncA2);	a3 = *(A+IncA3);	As += IncA4;	
      b = *B;		b1 = *(B+IncB); 
      b2 = *(B+IncB2);	b3 = *(B+IncB3); 	A += IncA4;
      *B = a; 		*(B+IncB) = a1;	
      *(B+IncB2) = a2; 	*(B+IncB3) = a3;	B += IncB4;
      *As = b;		*(As+IncA) = b1;
      *(As+IncA2) = b2;	*(As+IncA3) = b3;
#else	/* this was found to be faster on UltraSPARC */
      a = *A;		a1 = *(A+IncA);	
      b = *B;		b1 = *(B+IncB); 
      a2 = *(A+IncA2);	a3 = *(A+IncA3);	As += IncA4;	
      b2 = *(B+IncB2);	b3 = *(B+IncB3); 	A += IncA4;
      *B = a; 		*(B+IncB) = a1;	
      *As = b;		*(As+IncA) = b1;
      *(B+IncB2) = a2; 	*(B+IncB3) = a3;	B += IncB4;
      *(As+IncA2) = b2;	*(As+IncA3) = b3;

#endif      
    }
  } 
# endif 
#endif /*VANILLA*/
} /* VecSwap() */

  
int VecAbsMaxPos(m, A, IncA)
  FDATA A[];  int m, IncA;
{ register int i, iM; 			FD_register M0; register FD_BASE Max;
  int m4=4*(m/4);			
  COMMON_ASSERT_TR(
    (TraceStr, "(%d,a%d %d)", m, FD_ADR(A), IncA),
    1, 0, "VecAbsMaxPos", 4, TRACE_COMPUTE
  );
  i=0; iM=m>0? 0: -1; 			M0=FD_0; Max=0.0;
#if !defined(VANILLA) && !defined(BLAS) && (FDATATYPE < COMPLEX) 
# ifndef ULTRA
  if (m4 > 0) {
    register int IncA2 = IncA+IncA, IncA3=IncA2+IncA, IncA4=IncA2+IncA2;  
    FD_register M1, M2, M3;		
    register int iM1;			FD_register Max1;
    M3 = M2 = Max;
    iM1 = iM;				Max1 = Max;
    do {
    	M0 = *A;	
			FD_SETABS(M3);
					if (M2 > Max) { Max = M2; iM=i-2; };
    	M1 = *(A+IncA);	
			FD_SETABS(M0);
					if (M3 > Max1) { Max1 = M3; iM1=i-1; };
    	M2 = *(A+IncA2);
	i += 4;	
			FD_SETABS(M1);
					if (M0 > Max ) { Max = M0;  iM=i-4; };
    	M3 = *(A+IncA3);	
			FD_SETABS(M2);
	A += IncA4; 
					if (M1 > Max1) { Max1 = M1; iM1=i-3; };
    } while (i < m4);	
			FD_SETABS(M3);
					if (M2 > Max) { Max = M2; iM=i-2; };
   					if (M3 > Max1) { Max1 = M3; iM1=i-1; };
    if (Max1 > Max) { Max = Max1; iM = iM1; };	
  } /* if (m4...)*/ 					
# else
#  if 0
  if (m4 > 0) {
    register int IncA2 = IncA+IncA, IncA3=IncA2+IncA, IncA4=IncA2+IncA2;  
    FD_register a0, a1, a2, a3, M1, M2, M3;		
    register int iM1;			FD_register Max1;

    iM1 = iM;
    Max1 = Max;
       			a0 = *A;
 			a1 = *(A+IncA);
        M3 = 0.0; 
    			a2 = *(A+IncA2);
			a3 = *(A+IncA3);					
	A += IncA4; 
    for (m4=-m4+4; m4 < 0; m4+=4) {
    	M0 = fabs(a0);
 			a0 = *A;
					if (M3 > Max1) { Max1 = M3; iM1=i-1; };
    	i+=4;
    	M1 = fabs(a1);	
 			a1 = *(A+IncA);
					if (M0 > Max) { Max = M0; iM=i-4; };
    	M2 = fabs(a2);	
    			a2 = *(A+IncA2);
					if (M1 > Max1) { Max1 = M1; iM1=i-3; };
	M3 = fabs(a3); 				
			a3 = *(A+IncA3);					
	A += IncA4; 
					if (M2 > Max) { Max = M2; iM=i-2; };
    } /* for (m4..,) */
    	M0 = fabs(a0);
					if (M3 > Max1) { Max1 = M3; iM1=i-1; };
    	i+=4;
    	M1 = fabs(a1);	
					if (M0 > Max) { Max = M0; iM=i-4; };

    	M2 = fabs(a2);	
					if (M1 > Max1) { Max1 = M1; iM1=i-3; };
	M3 = fabs(a3); 				
					if (M2 > Max) { Max = M2; iM=i-2; };
					if (M3 > Max1) { Max1 = M3; iM1=i-1; };
    if (Max1 > Max) { Max = Max1; iM = iM1; };	
    
  } /* if (m4...)*/
#  else
  if (m >= 5) {
    register int IncA2 = IncA+IncA, IncA3=IncA2+IncA, IncA4=IncA2+IncA2,
    		 IncA5 = IncA3+IncA2;  
    FD_register a0, a1, a2, a3, a4, M1, M2, M3, M4;		
    register int iM1;			FD_register Max1;
    register int m5;
    
    iM1 = iM;
    Max1 = Max;
       	a0 = *A;
 	a1 = *(A+IncA);
    	a2 = *(A+IncA2);
	a3 = *(A+IncA3);					
	a4 = *(A+IncA4);					
    			M0 = fabs(a0);
	A += IncA5; 
    for (m5=-m+9; m5 < 0; m5+=5) {
 	a0 = *A;
    			M1 = fabs(a1);	
					if (M0 > Max) { Max = M0; iM=i; };
 	a1 = *(A+IncA);
    			M2 = fabs(a2);	
					if (M1 > Max1) { Max1 = M1; iM1=i+1; };
    	a2 = *(A+IncA2);
			M3 = fabs(a3); 				
					if (M2 > Max) { Max = M2; iM=i+2; };
    	i+=5;
	a3 = *(A+IncA3);					
			M4 = fabs(a4);		
					if (M3 > Max1) { Max1 = M3; iM1=i-2; };
	a4 = *(A+IncA4);					
        A += IncA5; 
    			M0 = fabs(a0);
					if (M4 > Max) { Max = M4; iM=i-1; };
    } /* for (m4..,) */
    			M1 = fabs(a1);	
					if (M0 > Max) { Max = M0; iM=i; };
    			M2 = fabs(a2);	
					if (M1 > Max1) { Max1 = M1; iM1=i+1; };
			M3 = fabs(a3); 				
					if (M2 > Max) { Max = M2; iM=i+2; };
    	i+=5;
			M4 = fabs(a4);		
					if (M3 > Max1) { Max1 = M3; iM1=i-2; };
					if (M4 > Max) { Max = M4; iM=i-1; };
    if (Max1 > Max) { Max = Max1; iM = iM1; };	
    
  } /* if (m>= 5)*/
#  endif 
# endif  
#endif 

#if !defined(BLAS)
  for (; i < m; i++) {
    M0 = *A;
			FD_SETABS(M0);
    A += IncA;			
					if (FD_REAL(M0) > Max) { Max = FD_REAL(M0); iM=i; };
  }
  return (iM);
#else
  return( FTN_IAMAX(&m, A, &IncA)-1 );
#endif /*VANILLA*/
} /* VecAbsMaxPos() */ 


FD_BASE VecAbsSum(m, A, IncA)
  FDATA A[];  int m, IncA;
{ register int i; FD_register M0; register FD_BASE Sum;
  int m4=4*(m/4);
  COMMON_ASSERT_TR(
    (TraceStr, "VecAbsSum(%d,a%d %d)", m, FD_ADR(A), IncA),
    1, 0, "VecAbsSum", 4, TRACE_COMPUTE
  );
  Sum=0.0;
#if !defined(VANILLA) && !defined(BLAS) && (FDATATYPE<COMPLEX)
  if (m4 > 0) {
    register int IncA2 = IncA+IncA, IncA3=IncA2+IncA, IncA4=IncA2+IncA2;  
    FD_register M1, M2, M3, Sum1 = 0.0;
    M3 = M2 = 0.0; i=0;   
    do {
    	M0 = *A;	
			FD_SETABS(M3); 
					Sum += M2;
    	M1 = *(A+IncA);	
			FD_SETABS(M0); 
					Sum1 += M3;
    	M2 = *(A+IncA2);
	i += 4;	
			FD_SETABS(M1);
					Sum += M0;
    	M3 = *(A+IncA3);		
			FD_SETABS(M2);
					Sum1 += M1; 
	A += IncA4; 
    } while (i < m4);	
			FD_SETABS(M3);
					Sum += M2;
					Sum1 += M3;
    m -= m4;		Sum += Sum1;
  } /* if (m4...)*/
#endif
#if !defined(BLAS)
  for (i=0; i < m; i++) {
    M0 = *A;
			FD_SETABS(M0); 
					Sum += FD_REAL(M0);
					
    A += IncA;			
  }
  return (Sum);
#else
  return( FTN_ASUM(&m, A, &IncA) );
#endif /*VANILLA*/
} /* VecAbsSum() */ 

