/************************************************************************/
/*         D_ScaleAdd: Distributed BLAS add scaled matrix 		*/
/************************************************************************/
/* SCCS INFO @(#)D_ScaleAdd.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 96/01
 */

#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#include "FDMalloc.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#include "CellMatrix.h"			/* MatScale(), MatCopy/Add(), MatMult() */
#include "FDCommun.h"

#include "DistMatrix.h"			/* TransMat(), RotMat()  */
#include "SpreadMatrix.h"
#include "CellBlkMatrix.h"

void D_ScaleAdd (M, N, alpha, A, beta, C)
  int M; int N; FDATA alpha; DistMat A; FDATA beta; DistMat C;
{ int m = -1, n = -1;  
  int Beta0 = FD_EQ(beta, FD_0);
  int ParamErr=0;                       
  int AlignM, AlignN;
  AlphaBetaDecln();
  
  int PA = d_P(DM_D0(A)),		PC = d_P(DM_D0(C)),
      QA = d_P(DM_D1(A)),		QC = d_P(DM_D1(C));
  int P = MAX(PA, PC),			Q = MAX(QA, QC);            
  int p = MIN(PA, PC),			q = MIN(QA, QC);	

  if (!NoOuterErrCheck) {
    CheckParam(ParamErr, (p==1 || p==P) && (q==1 || q==Q), 0x46);    
    CheckAlign(ParamErr, AlignM, M, DM_Inc(1-Tr(A), A),   DM_DM(A),
               DM_Inc(1-Tr(C), C), DM_DM(C), 0x146);
    if (M<=1) AlignM = 1;           
    CheckAlign(ParamErr, AlignN, N, DM_Inc(Tr(A), A),     DM_DN(A),
               DM_Inc(Tr(C), C), DM_DN(C), 0x246);
    if (N<=1) AlignN = 1;           
               
    if (q==1 && p==1) {
      if (P>1 && Q>1 && ((PA > PC) != (QA > QC)))
        CheckParam(ParamErr, AlignM && AlignN && DM_Trans(A)!=DM_Trans(C), 0x46);
    } else if (q==1) {
      CheckParam(ParamErr, AlignM && (PC==PA || 
      			   (DM_Trans(A)==DM_Trans(C) && 
      			   d_P0(DM_DM(A))==d_P0(DM_DM(C)))), 0x46);
    } else if (p==1) {
      CheckParam(ParamErr, AlignN && (QC==QA || 
      			   (DM_Trans(A)==DM_Trans(C) &&
      			   d_P0(DM_DN(A))==d_P0(DM_DN(C)))), 0x46);
    } else
      CheckParam(ParamErr, AlignM && AlignN, 0x46);
  }

  CommSetUp(P, Q);		
  ERROR_CHECK_TR(
    (TraceStr, "(%dx%d,%s,%s,%s,%s)",
                M, N, pFD2Str(&alpha), DM_TRACE(0,A), pFD2Str(&beta), DM_TRACE(1, C)),
    ParamErr, "ScaleAdd", 1
  );

  AlphaSet(alpha, pAlpha);              BetaSet(beta, pBeta);
  if (DM_Trans(C)) { 
    SWAP(int, M, N);
    A = Trans(A);			C = Trans(C);
  }
  m = G2Ld(M, DM_D0(C));		n = G2Ld(N, DM_D1(C));  
  if ((!Beta0 && !FD_EQ(beta,FD_1)) || FD_EQ(alpha, FD_0)) 	
    MatScale(m, n, pBeta, DM_pLoc(C));
  if (FD_EQ(alpha, FD_0)) {	
    TRACE_RETURN;    
  }

  if (PA==PC && QA==QC) {
    if (DM_Trans(A)==0) {
      if (d_P0(DM_D0(A))==d_P0(DM_D0(C)) && d_P0(DM_D1(A))==d_P0(DM_D1(C)))
        (Beta0? MatCopy: MatAdd)(m, n, DM_pLoc(A), pAlpha, DM_pLoc(C));
      else {
        int D0_dP0 = d_P0(DM_D0(A))-d_P0(DM_D0(C)),
            D1_dP0 = d_P0(DM_D1(A))-d_P0(DM_D1(C));
        int mA = D0_dP0?    G2Ld(M, DM_D0(A)): m,
            m_ = D0_dP0? G2Ld_Mx(M, DM_D0(A)): m,
            nA = D1_dP0?    G2Ld(N, DM_D1(A)): n,
            n_ = D1_dP0? G2Ld_Mx(N, DM_D1(A)): n;
        LocMat lA;			LM_Init(lA, NULL, n_, 1);
        MatMalloc_F("ScaleAdd", 1, &LM_ADR(lA), m_*n_);    	
        MatCopy(mA, nA, DM_pLoc(A), pAlpha, &lA);
        RotMat(0, D0_dP0, m_, &UnitDim, n_, LM_ADR(lA));
        RotMat(1, D1_dP0, m_, &UnitDim, n_, LM_ADR(lA));
        (Beta0? MatCopy: MatAdd)(m, n, &lA, NULL, DM_pLoc(C));     
        MatFree(LM_ADR(lA));
      }    
    } else if (M >= N) {			/* partition along longest dim. */
      LM_Trans(1-DM_Trans(A), DM_Loc(A));	LM_Trans(1-DM_Trans(C),  DM_Loc(C));
      TransMat(1-DM_Trans(A), 1-Beta0, N, M, DM_pLoc(A), DM_DN(A), DM_DM(A), 1, 
 	       pAlpha, DM_pLoc(C), d_P0(DM_DM(C)), d_P0(DM_DN(C)), 1);
      LM_Trans(1-DM_Trans(A), DM_Loc(A));	LM_Trans(1-DM_Trans(C),  DM_Loc(C));
    } else {
      LM_Trans(DM_Trans(A), DM_Loc(A));	LM_Trans(DM_Trans(C), DM_Loc(C));
      TransMat(DM_Trans(A), 1-Beta0, M, N, DM_pLoc(A), DM_DM(A), DM_DN(A), 1, 
 	       pAlpha, DM_pLoc(C), d_P0(DM_DN(C)), d_P0(DM_DM(C)), 1); 
      LM_Trans(DM_Trans(A), DM_Loc(A));	LM_Trans(DM_Trans(C), DM_Loc(C));
    }  
  } else if (PA > PC || QA > QC) {	/* spread operations */
    LocMat sA;				FDATA *WrsA = NULL, *WcsA = NULL;
    FDATA ScAlpha = *pAlpha;		
    int SpreadToC = 0; 			/*=1 if a spread can go directly into C*/
    DistMat_RVAlloc(1);
    if (PA > PC) {
      int ForceCopy = 0, ncsA;	
      LM_ADR(sA) = NULL;
      if (Beta0 && LM_Inc(1, DM_Loc(C))==1 && QA==QC &&
        (DM_Trans(A) || d_P0(DM_D1(C))==d_P0(DM_D1(A)))) {
      	SpreadToC = 1,			ForceCopy = 1;
        LM_Copy(0, sA, DM_Loc(C));   
      }  
      ncsA = SpreadMat(0, M, N, A, &ScAlpha, QA > QC? DM_D1(A): DM_D1(C),
                       &sA, ForceCopy, 1+d_B(DM_DM(A)), DM_DM(A));
      if (QA > QC) 
        A = Spread2DistMat(1, 0, QA > QC? DM_D1(A): DM_D1(C), &sA); 
      if (ncsA > 0 && !SpreadToC)
        WcsA = LM_ADR(sA);                  
    }	
    if (QA > QC) {
      int ForceCopy = 0, nrsA;	
      LM_ADR(sA) = NULL;			     
      if (Beta0 && LM_Inc(0, DM_Loc(C))==1 && 
          (DM_Trans(A) || d_P0(DM_D0(C))==d_P0(DM_D0(A)))) {
        SpreadToC = 1,                  ForceCopy = 1;
        LM_Copy(0, sA, DM_Loc(C));   
      }  
      nrsA = SpreadMat(1, N, M, A, &ScAlpha, DM_D0(C),
                       &sA, ForceCopy, 1+d_B(DM_DN(A)), DM_DN(A));
      /*LM_Trans(1, sA);                 */
      if (nrsA > 0 && !SpreadToC)
        WrsA = LM_ADR(sA);
    } 
    if (!SpreadToC) 
      (Beta0? MatCopy: MatAdd)(m, n, &sA, &ScAlpha, DM_pLoc(C));
    else 
      MatScale(m, n, &ScAlpha, DM_pLoc(C)); 
    MatFree(WrsA);			
    MatFree(WcsA);                     		    
  } else {				/* update from an aligned spread */
    int B0i=0, B0j=0, Bi=1, Bj=1, si=1, sj=1;
    LocMat lA;
    LM_Copy(DM_Trans(A), lA, DM_Loc(A));
    if (PC > PA) { 
      pLocSubMat(&lA, BlkIxP0d(DM_DM(C)), 0);
      B0i = BlkB0d(0, DM_DM(C)),	Bi = d_B(DM_DM(C));
      si = PC;
    }
    if (QC > QA) {      
      pLocSubMat(&lA, 0, BlkIxP0d(DM_DN(C)));
      B0j = BlkB0d(0, DM_DN(C)),	Bj = d_B(DM_DN(C));
      sj = QC;
    }  
    BlkMatStridePack(1-Beta0, m, n, &lA, pAlpha, DM_pLoc(C), 
    		     B0i, Bi, si, B0j, Bj, sj);
  } 		    
  TRACE_EXIT; 	    
} /* D_ScaleAdd() */
        
