/************************************************************************/
/*	DistMatrix: distributed matrix procedures used by DBLAS	*/
/************************************************************************/
/* SCCS INFO @(#)DistMatrix.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 93/15/01
 */
/* set of higher-level distributed matrix routines used by main DBLAS
 * parallel driver routines.  See FDATACommun.h for description of matrix notations.
 * These use message partitioning along a dimension (eg. M) so that all messages
 * sizes <= OPT_MSG_SIZE and <= 2*OPT_MSG_SIZE space required by system 
 * receive buffers at any time. Generally, these algorithm trades off 
 * algorithm pipelining in order to reduce receive buffers space, but
 * try to maintain load balance as much as  possible.
 * 
 * The LocMat A in these routines generally will have Inc(A) = {n, 1};
 * in this case, A will be a workspace for accumulating the matrix dimension M 
 *
 * compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *		if macro TRACE defined, parameter value tracing enabled 
 */
#ifndef __DistMatrix_INCL__
#define __DistMatrix_INCL__
 
#include "FDATASetUp.h"
#include "LocMat.h"
#include "MatDim.h"

#if FDATATYPE==SINGLE
#define RotMat			DSRotMat
#define GatherMat		DSGatherMat
#define MultSumMat		DSMultSumMat
#define TransMat		DSTransMat
#define TransReSpreadMat 	DSTransReSpreadMat
#elif FDATATYPE==DOUBLE
#define RotMat			DDRotMat
#define GatherMat		DDGatherMat
#define MultSumMat		DDMultSumMat
#define TransMat		DDTransMat
#define TransReSpreadMat 	DDTransReSpreadMat
#elif FDATATYPE==COMPLEX
#define RotMat			DCRotMat
#define GatherMat		DCGatherMat
#define MultSumMat		DCMultSumMat
#define TransMat		DCTransMat
#define TransReSpreadMat 	DCTransReSpreadMat
#else
#define RotMat			DZRotMat
#define GatherMat		DZGatherMat
#define MultSumMat		DZMultSumMat
#define TransMat		DZTransMat
#define TransReSpreadMat 	DZTransReSpreadMat
#endif

    
void RotMat STDC_PARAMS(
	(int RowCol, int dP0, int m, pMatDim D, int n, FDATA A[])
);
/* pre:  RCOK(RowCol) && BlkOK(B0, d_B(D), d_P(D)); Inc(A) = {n, 1}; B0 = BlkB0d(0, D)
 *	 all cells in scope of RowCol are participating  in this call; 
 *	 m, n, D are the same across all these cells; 
 * post: A[m::B0:d_B(D):d_P(D), 0..n-1] is rotated distance DP0 in scope of RowCol
 * notes: generally d_P(D) = GetNumCells(!RowCol) 
 */

void GatherMat STDC_PARAMS(
	(int RowCol, pMatDim D, int M, int n, pLocMat A)
);
/* pre:  RCOK(RowCol); BlkOK(d_B0(D), d_B(D), d_P(D)); 
 *	 LM_CInc(A)=1; LM_RInc(A) = n || (P0==0 && NBlk1d(M,D)) 
 *	 all (d_P(D)) cells in scope of RowCol are participating in this call; 
 *	 M, n amd D are consistent across all these cells; 
 *	 starting from some cell I in the scope, P0=(0,1,..,d_P(D)-1)
 * post: A[0..M-1, 0..n-1] is gathered across cells in scope of RowCol,
 *	 with cell I+i contributing block rows i+k*P, k>=0 
 * notes:if nB > 1, algorithm is self-synchronizing 
 */

void MultSumMat STDC_PARAMS(
	(int RowCol, pMatDim D, int M, int n, pLocMat A)
);
/* pre:  RCOK(RowCol); BlkOK(d_B0(D), d_B(D), d_P(D)); 
 *	 (Inc(A) = {n, 1}) || (MB=1 && d_P0(D)=0), where MB = NBLKSd(M, D), 
 *	 all (d_P(D)) cells in scope of RowCol are participating in this call; 
 *	 M, n and D are consistent across all these cells; 
 *	 starting from some cell I in the scope, P0=(0,1,..,d_P(D)-1)
 * post: in cell (I+i)%d_P(D), the ith row block of A contains the sum of 
 *	 that in all cells in the scope, 0<=i<MB
 */

void TransMat STDC_PARAMS( 
	(int RowCol, int Add, int M, int N, pLocMat A, pMatDim DM, pMatDim DN, 
	 int MBIncA, FDATA *alpha, pLocMat B, int NP0, int MQ0, int MBIncB)
);
/* pre:  RCOK(RowCol), 0 <= MQ0 < d_P(DN); 0<= NP0 < d_P(DM); M, N >= 0; 
 *	 MBIncA, MBIncB > 0
 *       if RowCol==0 (RowCol==1), A' is a global M x N (N x M) matrix 
 *	 distributed across dimensions DM by DN (DN by DM), 
 *	 with local matrix component LocMatTrans(A) (A).  
 *
 *       if RowCol==0 (RowCol==1), B' is a global  N x M (M x N) matrix 
 *	 distributed across dimensions DN' by DM' (DM' by DN), 
 *	 with local matrix component B (LocMatTrans(B)),
 *	 where DM' = {d_B0(DM), d_B(DM), MQ0, d_P(DN)}, 
 *	       DN' = {d_B0(DN), d_B(DN), NP0, d_P(DM)}
 *
 *	 ie. the 0th stride of A (B) corresp. to M;
 *	 a block increment of MBIncA (MBIncB) applies to this stride on A (B). 
 * post: B'  = (Add? B': 0) + *alpha * A'^T   
 * notes:leading cell of A has rel. co-ordinates (d_P0(DM), d_P0(DN));
 *	 that of B' has (NP0, MQ0). 
 *	 Block offsets and sizes are preserved in the operation.
 *	 use MBIncA = MBIncB = 1 for a true transpose operation;
 *	 use MBIncA = d_P(DN)  (MBIncB = d_P(DN)) for A (B) formed from
 *	 a spread operation across dimension DN.
 *	 algorithm avoids cells sending messages to themselves.
 *	 Partitioning is applied across the N-dimension.
 */

void TransReSpreadMat STDC_PARAMS( 
	(int RowCol, int K, int N, pLocMat A, pMatDim DN, 
	 FDATA *alpha, pMatDim DQ, pLocMat sA, int NoCopysA)
);
/* pre:  RCOK(RowCol) && 0<=Q0<Q; K,N>=0; P==1 || LM_Inc(RowCol,sA)==1)
 *	 All cells (holding/requiring any data) in scope of RowCol 
 *	 are participating in this call. DQ = {_, _, Q0, Q}
 *	 with consistent values of *alpha, RowCol, K, N, Q and DN, Q0.
 *	 A, sA are  K x N  local matrices
 *	 where A' = {A, {_,_,_,Q}, DN} is a KxN matrix distributed across
 *       the 1 x P cells in scope of RowCol.
 *	 If !NoCopysA,  sufficient space must have been 
 *	 allocated at LM_ADR(sA) for the workspace.
 *	 Otherwise, the conditions of the spread must ensure a workspace
 *	 is not required, ie. a single cell is sole source and the selection
 *	 of N elements in dimension DN is trivial.	
 * post: sA = alpha' * Loc(1, Q0, A'[:][[Q0:N:Q]]), with LM_Inc(!RowCol,sA)=1 
 *
 *       where NB=NBLKS(N, DN.B0, DN.B) 
 *	 and the notation [[Q0:NB:Q]] signifies a subrange if blocks
 *	 of number Q0+i*Q, for i=0..,(NB-Q0)/Q-1 
 *
 *       *alpha * alpha' = *alpha_pre, 	*alpha IN {1.0, *alpha_pre} 
 *	 where alpha' = 1.0 iff source cells for spread did not need a workspace
 *       (in this case LM_ADR(sA) points to A)
 * 
 * notes: N is dimension to spread across, and is aligned to the RowCol'th 
 * processor dim. 
 */

#endif /*__DistMatrix_INCL__*/

