/************************************************************************/
/*	FDCommun: low-level matrix communication procedures for DBLAS	*/
/************************************************************************/
/* SCCS INFO @(#)FDCommun.c 1.10         last modified P. Strazdins, 01/01/09
   first created: P. Strazdins 96/01/15
 */
/* compilation: if macro BLACS defined, Fortran BLACS calls are used 
 *		if macro APLIB* is defined, AP/AP+ specific calls are to be used
 *		if macro APLIB_STD, stardard AP calls are used instead of
 *		modified routines (extra overhead implied).
 *		if none of these defined, communication calls
 *		are omitted; this is suitable for running on a uniprocessor.
 */
/*#define PANEL_SCATTER*/

#include <stdio.h>
#include <stdlib.h>

#include <math.h>                       /* fabs() */
#include "FDATAOps.h"
#include "FDCommun.h"
#include "FDAssert.h"
#include "LocMat.h"

#include "FDMalloc.h"
#include "CellMatCopy.h"
#include "CellVector.h"
#include "CellBlkMatrix.h"

#ifdef VPPLIB_STRIDE
# include "VPPStride.h"
# undef I_SEND
# undef WAIT_SEND
# undef CTG_RECV
#endif

#ifdef BLACS
# include "FtnBLACS.h"
# define COMMUN_ERROR_FLAG 	0
#elif defined(APLIB)
# include <ccell.c7.h>
# if defined(APLIB_LINUX)
  int r_rsend_stride(), x_brd_stride(), y_brd_stride();
# endif  
# define COMMUN_ERROR_FLAG 	cap_errno
# if defined(APLIB_STRIDE)
#   include "CellSplitUnpack.c"
# endif
#elif defined(VPPLIB)
# include "vpp_mplib.h"
# define COMMUN_ERROR_FLAG 	0
# define MSG_HNDLE TMSG *
#elif defined(MPI)
# include "mpi.h"
# define MSG_HNDLE MPI_Request
# define COMMUN_ERROR_FLAG 	0
#else
# define MSG_HNDLE void *		/* safest assumption for a
					   message hanfle type */ 
#define COMMUN_ERROR_FLAG 	0 
#endif


/**************** setup of parameters associated with communication ********/ 
#ifndef APLIB
#undef PANEL_SCATTER		/* need global xy comms */
#endif

#ifdef PANEL_SCATTER
# define SPECTOP SpecTop 
#else 
# define SPECTOP 0
#endif

#if defined(APLIB)
# define GET_GLOBAL_PROC_INFO(ncel,id)	(ncel) = getncel(), (id) = getcid()
#elif defined(BLACS)
# define GET_GLOBAL_PROC_INFO(ncel,id)	FTN_BLACS_PINFO(&(id), &(ncel))
#elif defined(VPPLIB)
# define GET_GLOBAL_PROC_INFO(ncel,id)	{(id) = (int) VPP_MyProc(); if ((id)==0) {(void) VPP_GetPeList(&(ncel)); (ncel)++;} }
#elif defined(MPI)
# define GET_GLOBAL_PROC_INFO(ncel,id)	{MPI_Comm_rank((MPI_Comm) MPIComm, &(id)); MPI_Comm_size((MPI_Comm) MPIComm, &(ncel));}
#else
# define GET_GLOBAL_PROC_INFO(ncel,id)	(ncel) = 1, (id) = 0
#endif


#define STATIC 
STATIC char RC_CHAR[] 		= "CR";
STATIC char BrdTopChar[]	= "  ";
STATIC char CombineTopChar[]	= "  ";
STATIC int  AbsCellId[]		= {0, 0};
STATIC int  NumCells[] 		= {1, 1};	
STATIC int  VNumCells[]		= {1, 1};
STATIC int  NCel		= 1;	/* = NumCells[0]*NumCells[1] */
STATIC int  AbsCId		= 0;	/* 1 dim abs cell id */
STATIC int  BLACSCntxt		= 0;
static int  MPIComm		= 0;

#ifdef PANEL_SCATTER			/* support for panel scattering */
STATIC int  SpecTop		= 0;	/* SpecTop>0 ==>
					   VNumCells[SpecTop']=NCel,
					   VNumCells[1-SpecTop'] = 1 */ 	
static int  PInc 		= 1;	/* = NumCells[SpecTop'] */
STATIC int  LgPInc 		= 0;	/* = lg(PInc) */
int PS_P0 = -1;				/* cell id of special topology */
#endif

FD_BASE MsgMaxAbsVal = 0.0;       /*disable checking of FDATA values in messages*/


#ifdef APLIB
#if (FDATATYPE==SINGLE) || (FDATATYPE==COMPLEX)
# define Y_SUM y_fsum
# define X_SUM x_fsum
# define Y_AMAX y_famax
# define X_AMAX x_famax
# define XY_AMAX xy_famax
# define XY_SUM xy_fsum
#else
# define Y_SUM y_dsum
# define X_SUM x_dsum
# define Y_AMAX y_damax
# define X_AMAX x_damax
# define XY_AMAX xy_damax
# define XY_SUM xy_dsum
#endif 
#if defined(APLIB_STRIDE) || defined(APLIB_PLUS_STRIDE)
#define y_fsum _y_fsum
#define x_fsum _x_fsum
#define x_dsum _x_dsum
#define y_dsum _y_dsum
#define xy_fsum _xy_fsum
#define xy_dsum _xy_dsum
#define x_damax _x_damax
#define y_famax _y_famax
#define x_famax _x_famax
#define y_damax _y_damax
#define x_damax _x_damax
#define xy_famax _xy_famax
#define xy_damax _xy_damax
#endif
int Y_SUM(), XY_SUM(), X_SUM(), XY_AMAX(), Y_AMAX(), X_AMAX(), y_brd(), x_brd(), xybrd();
#endif /*APLIB*/

/******************* primitive communication macros **********************/    

/* REL2ABSCID(): 0 <= RCID < NumCells[RowCol] */
#define REL2ABSCID(RowCol, RCID)					\
	CYCLIC_ADD(AbsCellId[RowCol], RCID, NumCells[RowCol])
/* SRCABSCID(): 0 <= P0 < NumCells[RowCol] */  
#define SRCABSCID(RowCol, P0)						\
	CYCLIC_SUB(AbsCellId[RowCol], P0, NumCells[RowCol])

#define REL2ABS1CID(RCIDI, RCIDJ)					\
	(CYCLIC_ADD(AbsCellId[0], RCIDI, NumCells[0])*NumCells[1] +	\
	 CYCLIC_ADD(AbsCellId[1], RCIDJ, NumCells[1]))	


#if !defined(GLOB_BLK_SEND)
# define CTG_BLK_SEND(RDEST, CDEST, N, A, M_TYPE) 			\
	SEND(RDEST, CDEST, 1, N, A, MAX(N,1), M_TYPE)
# define CTG_RECV(RSRC, CSRC, N, A, M_TYPE, r) 				\
	RECV(RSRC, CSRC, 1, N, A, MAX(N,1), M_TYPE, r)
# define WAIT_SEND(hndle)	
#endif


#if defined(BLACS)
# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE) 			\
	{ int aRDEST_ = REL2ABSCID(0, RDEST), aCDEST_ = REL2ABSCID(1, CDEST),\
	      m_ = (M), n_ = (N), inc_ = (IncA);			\
	  FTN_GESD2D(&BLACSCntxt, &n_, &m_, A, &inc_, &aRDEST_, &aCDEST_);\
	}  
	
# define RECV(RSRC, CSRC, M, N, A, IncA, M_TYPE, r) 			\
	{ int aRSRC_ = REL2ABSCID(0, RSRC), aCSRC_ = REL2ABSCID(1, CSRC),\
	      m_ = (M), n_ = (N), inc_ = (IncA);			\
	  FTN_GERV2D(&BLACSCntxt, &n_, &m_, A, &inc_, &aRSRC_, &aCSRC_);\
	  r += FAST_MUL1(m_, n_);					\
	}  

# define RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r) 	\
	{ int aRSRC_ = REL2ABSCID(0, RSRC), aCSRC_ = REL2ABSCID(1, CSRC),\
	      m_ = (M), n_ = (N), mn = FAST_MUL1(m_, n_); 		\
	      FDATA *A_; LocMat lA_, lB_;				\
          MatMalloc_FT("RECV_ADD", 1, &A_, mn);  			\
    	  FTN_GERV2D(&BLACSCntxt, &n_, &m_, A_, &n_, &aRSRC_, &aCSRC_);	\
	  r += mn;							\
	  MatAdd(m_, n_, pMkLocMat(&lA_, A_, n_, 1), alpha,  pMkLocMat(&lB_, A, RIncA, CIncA));\
	  MatFree(A_);							\
	}  

# define RECV_ADD_(Add, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)  \
	if (!Add)		/* assert alpha==CIncA==1 */		\
	  RECV(RSRC, CSRC, M, N, A, RIncA, M_TYPE, r) 			\
	else								\
	  RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r);
	
# define BRD(RowCol, SRCP0, M, N, A, IncA, r)				\
	{ int m_ = (M), n_ = (N), inc_ = (IncA);			\
	  int aRSRC_ = (RowCol)==0? SRCABSCID(0, SRCP0): AbsCellId[0],	\
	      aCSRC_ = (RowCol)==1? SRCABSCID(1, SRCP0): AbsCellId[1];	\
	  if (SRCP0==0) 						\
	    FTN_GEBS2D(&BLACSCntxt, C2F_CHAR(&RC_CHAR[RowCol]), 	\
	    	       C2F_CHAR(&BrdTopChar[RowCol]),			\
		         &n_, &m_, A, &inc_);				\
	   else								\
	    FTN_GEBR2D(&BLACSCntxt, C2F_CHAR(&RC_CHAR[RowCol]), 	\
	    		C2F_CHAR(&BrdTopChar[RowCol]),			\
		         &n_, &m_, A, &inc_, &aRSRC_, &aCSRC_);		\
	  *(r) += FAST_MUL1(m_, n_);					\
	}	                        

# define SUM(RowCol, SRCP0, M, N, A, IncA, r)				\
	{ int m_ = (M), n_ = (N), inc_ = (IncA), 			\
	      rdst_ = (RowCol)==0? SRCABSCID(0, SRCP0): AbsCellId[0],	\
	      cdst_ = (RowCol)==1? SRCABSCID(1, SRCP0): AbsCellId[1];	\
	  FTN_GSUM2D(&BLACSCntxt, C2F_CHAR(&RC_CHAR[RowCol]), 		\
	  	 C2F_CHAR(&CombineTopChar[RowCol]),			\
	  	 &n_, &m_, A, &inc_, &rdst_, &cdst_);			\
	}	                        
	
# define AMAX(RowCol, A, indexA)					\
	{ int n_ = 1, src_[2], dst_[2]; FDATA DindexA = (FDATA) *indexA;\
	  dst_[RowCol] = -1, dst_[1-RowCol] = AbsCellId[1-RowCol];	\
	  FTN_GAMX2D(&BLACSCntxt, C2F_CHAR(&RC_CHAR[RowCol]), 		\
	  		C2F_CHAR(&CombineTopChar[RowCol]), &n_, &n_, A, \
	  	 	&n_, &src_[0], &src_[1], &n_, &dst_[0], &dst_[1]);\
	  if (src_[RowCol]==AbsCellId[RowCol]) 				\
	    FTN_GEBS2D(&BLACSCntxt, C2F_CHAR(&RC_CHAR[RowCol]_, 	\
	    		 C2F_CHAR(&BrdTopChar[RowCol]),			\
		         &n_, &n_, &DindexA, &n_);			\
	   else								\
	    FTN_GEBR2D(&BLACSCntxt, C2F_CHAR(&RC_CHAR[RowCol]), 	\
	    		 C2F_CHAR(&BrdTopChar[RowCol]),			\
		         &n_, &n_, &DindexA, &n_, &src_[0], &src_[1]);	\
	  *indexA = (int) DindexA;					\
	}  

#elif defined(VPPLIB) || defined(MPI)

# if defined (VPPLIB)
#   define WAIT_SEND(hndle)	 VPP_SendDone((TMSG *) (hndle))

#   define I_SEND(RDST, CDST, N, A, M_TYPE, HNDLE)                     \
        (HNDLE) = (MSG_HNDLE) (VPP_Send((PROC) REL2ABS1CID(RDST, CDST), M_TYPE,\
                            (ADDRP) (A), (size_t) (N)*sizeof(FDATA)))
                             
#   define CTG_RECV(RSRC, CSRC, sz, A, M_TYPE, r)			\
	{ int t_, sz__;							\
	  VPP_RecvDone(							\
	  	VPP_Recv((PROC) REL2ABS1CID(RSRC, CSRC), M_TYPE, (ADDRP) (A),\
	  	(size_t) (sz)*sizeof(FDATA)), (PROC*) &t_, &t_, (size_t *) &sz__);\
	  (r) += sz__/sizeof(FDATA);					\
	}	
	

# elif defined (MPI)

#   define WAIT_SEND(hndle)	 					\
	{ MPI_Status status_; MPI_Wait((MPI_Request *) &(hndle), &status_); }

#   define I_SEND(RDEST, CDEST, N, A, M_TYPE, HNDLE)                    \
        MPI_Isend((void *) (A), (N)*sizeof(FDATA), MPI_BYTE,		\
                  REL2ABS1CID(RDEST, CDEST), M_TYPE, (MPI_Comm) MPIComm,\
                  (MPI_Request *) &(HNDLE))                             

#   define CTG_RECV(RSRC, CSRC, sz, A, M_TYPE, r)			\
	{ MPI_Status status_; 						\
	  MPI_Recv((void *) (A), (sz)*sizeof(FDATA), MPI_BYTE,		\
	  	   REL2ABS1CID(RSRC, CSRC), M_TYPE, (MPI_Comm) MPIComm, &status_);\
	  (r) += (sz);							\
	}		
#endif

# define CTG_BLK_SEND(RDEST, CDEST, N, A, M_TYPE) 			\
	{ MSG_HNDLE Hndle_; 						\
	  I_SEND(RDEST, CDEST, N, A, M_TYPE, Hndle_); WAIT_SEND(Hndle_);}


# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE) 			\
	{ int sz_ = FAST_MUL1(M, N);					\
	  FDATA *A_; MSG_HNDLE Hndl_;					\
	  if (sz_<=(N) || (IncA)==(N)) A_ = (A);			\
	  else { LocMat lA_, lA__;					\
	         MatMalloc_F("SEND", 1, &A_, sz_);			\
	         MatCopy(M, N,  pMkLocMat(&lA_, A, IncA, 1), NULL, 	\
	         		pMkLocMat(&lA__, A_, N, 1)); 		\
	         if (NoSendWait) WaitHandleBuf = A_;	}		\
	  I_SEND(RDEST, CDEST, sz_, A_, M_TYPE, Hndl_);  		\
	  if (NoSendWait==0) {						\
	    WAIT_SEND(Hndl_); if (A_!=(A)) MatFree(A_);			\
	  } else AddWaitHandle(Hndl_);				\
	}  

# define RECV_ADD_(Add, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r) \
	{ int sz_ = FAST_MUL1(M, N);					\
	  FDATA *A_;							\
	  if (!Add && (sz_<=(N) || (RIncA)==(N))) A_ = (A);		\
	  else MatMalloc_FT("RECV", 1, &A_, sz_);			\
	  CTG_RECV(RSRC, CSRC, sz_, A_, M_TYPE, r);			\
	  if ((A)!=A_) { LocMat lA__, lA_;				\
	    pMkLocMat(&lA__, A_, N, 1); pMkLocMat(&lA_,  A, RIncA, CIncA);\
	    (Add? MatAdd: MatCopy)(M, N, &lA__, alpha, &lA_);		\
	    MatFree(A_); }						\
	}

# define RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r) 	\
	RECV_ADD_(1, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)

# define RECV(RSRC, CSRC, M, N, A, IncA, M_TYPE, r) 	\
	RECV_ADD_(0, RSRC, CSRC, M, N, NULL, A, IncA, 1, M_TYPE, r)
	
#elif defined(APLIB_LINUX)

# define MX_STRIDE_SIZE 128 

# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE) 			\
	{ int i_, inc_ = (IncA)*sizeof(FDATA);				\
	  if ((M)==1) 							\
	    r_rsend(CDEST, RDEST, gettid(), M_TYPE, (MSGP) (A), 	\
	    	    (N)*sizeof(FDATA));					\
	  else 								\
	    for (i_=0; i_ < (N); i_+=MX_STRIDE_SIZE) {			\
	      int nb = MIN((N)-i_, MX_STRIDE_SIZE)*sizeof(FDATA);	\
	      r_rsend_stride(CDEST, RDEST, gettid(), M_TYPE, (u_long *) ((A)+i_), nb, inc_, M);\
	  }								\
	}								

# define RECV(RSRC, CSRC, M, N, A, IncA, M_TYPE, r)                   	\
	RECV_ADD_(0, RSRC, CSRC, M, N, NULL, A, IncA, 1, M_TYPE, r)

# define RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)  \
	RECV_ADD_(1, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)

# define RECV_ADD_(Add, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r) \
	{ int i_ = 0;		LocMat lA_, lB_;			\
	  do {								\
	    FDATA *A_1 = (FDATA *) r_rrecv(CSRC, RSRC, gettid(), M_TYPE);\
            int nb = getmsize()/sizeof(FDATA); (r) += nb; 		\
            if ((M)==1) 						\
              (Add? VecAdd: VecCopy)(nb, A_1, 1, (A)+i_*(CIncA), CIncA);\
            else {							\
              if ((M)> 1) nb /= (M);					\
	      (Add? MatAdd: MatCopy)					\
	      (M, nb,  pMkLocMat(&lA_, A_1, nb, 1), alpha, pMkLocMat(&lB_, (A)+i_*(CIncA), RIncA, CIncA ));\
	    }								\
	    i_ += nb;							\
	  } while (i_ < (N));						\
	}
	  
# define BRD(RowCol, SRCP0, M, N, A, IncA, r) 				\
	{ int i_, inc_ = (IncA)*sizeof(FDATA), I0_ = SRCABSCID(RowCol, SRCP0);\
	  if ((M)==1)							\
	    *(r) += (RowCol==0? y_brd: x_brd)(I0_, (MSGP) (A), (N)*sizeof(FDATA))/sizeof(FDATA);\
	  else								\
	    for (i_=0; i_ < (N); i_+=MX_STRIDE_SIZE) {			\
	      int nb = MIN((N)-i_, MX_STRIDE_SIZE)*sizeof(FDATA);	\
	      *(r) += (RowCol==0? y_brd_stride: x_brd_stride)( I0_, (u_long *) ((A)+i_), nb, inc_, M)/sizeof(FDATA);\
	    }								\
	}

#elif defined (APLIB_PLUS_STRIDE)

# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE)                      \
	xy_stride_send(CDEST, RDEST, M_TYPE, M, (N)*sizeof(FDATA), 	\
		       A, (IncA)*sizeof(FDATA))
	
# define RECV(RSRC, CSRC, M, N, A, IncA, M_TYPE, r)                   	\
	RECV_ADD_(0, RSRC, CSRC, M, N, NULL, A, IncA, 1, M_TYPE, r)

# define RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)  \
	RECV_ADD_(1, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)

# define RECV_ADD_(Add, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r) \
	{ int i_ = 0;							\
	  int r_, r1, s_; FDATA *b1, *b2; int nb; LocMat lB_, lA_; 	\
	  do {								\
	    r_ = xy_stride_recv(CSRC, RSRC, M_TYPE, M, (N)*sizeof(FDATA),\
	  		        &s_, &r1, &b1, &b2);			\
 	    r_ /= sizeof(FDATA); nb = r_;				\
	    if ((M)==1)							\
	      (Add? VecAdd: VecCopy)(nb, b1, 1, (A)+i_*(CIncA), CIncA);	\
	    else {							\
	      if ((M) < 1) break; nb /= (M);				\
	      (Add? MatAdd: MatCopy)(M, nb, pMkLocMat(&lB_, b1, nb, 1), \
	      	    NULL, pMkLocMat(&lA_, (A)+i_*(CIncA), RIncA, CIncA));\
	    }  	    							\
	    i_ += nb; (r) += r_; 					\
	  } while (i_<(N) && FAST_MOD(r_, M)==0);			\
	}

# define BRD(RowCol, SRCP0, M, N, A, IncA, r) 				\
	{ int r_, r1, s_, I0_ = SRCABSCID(RowCol, SRCP0); FDATA *b1, *b2;\
	  if ((SRCP0)==0)  						\
	    r_ = brd_stride_send(RowCol, M, (N)*sizeof(FDATA),		\
	    			 A, (IncA)*sizeof(FDATA)); 		\
	  else { int i_ = 0, nb; LocMat lB_, lA_; 			\
	    do {							\
	      r_ = brd_stride_recv(RowCol, I0_, M, (N)*sizeof(FDATA), 	\
	    			   &s_, &r1, &b1, &b2);  		\
	      r_ /= sizeof(FDATA); nb = r_;				\
	      if ((M)==1)						\
	        VecCopy(nb, b1, 1, (A)+i_, 1);				\
	      else {							\
	        if (M<1) break; nb /= (M);				\
	        MatCopy(M, nb, pMkLocMat(&lB_, b1, nb, 1), NULL, 	\
	      	        pMkLocMat(&lA_, (A)+i_, IncA, 1));		\
	      }	        						\
	      i_ += nb; *(r) += r_; 					\
	    } while (i_<(N) && FAST_MOD(r_, M)==0);			\
	  }								\
	}  

#elif defined(APLIB_STRIDE)

/* this optimization is to reduce software latencies for vectors */	
#define SplitMatUnpack_f(Add, N, r, s, b1, r1, b2, alpha, A, RIncA, CIncA)\
	if ((r)==(r1)*sizeof(FD_BASE)/sizeof(FDATA) && (r)==(N) && (alpha==NULL)) 			\
	  ((Add)? VecAdd: VecCopy)(N, b1, 1, A, CIncA);			\
	else 								\
	  SplitMatUnpack(Add, N, r, s, b1, r1, b2, alpha, A, RIncA, CIncA);

#define SplitMatUnpack_	SplitMatUnpack_f		

# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE)                      \
	xy_stride_send(CDEST, RDEST, M_TYPE, M, (N)*sizeof(FDATA), 	\
		       A, (IncA)*sizeof(FDATA))
	
# define RECV(RSRC, CSRC, M, N, A, IncA, M_TYPE, r)                   	\
	RECV_ADD_(0, RSRC, CSRC, M, N, NULL, A, IncA, 1, M_TYPE, r)

# define RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)  \
	RECV_ADD_(1, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)

# define RECV_ADD_(Add, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r) \
	{ int r_, r1, s_; FDATA *b1, *b2;  				\
	  r_ = xy_stride_recv(CSRC, RSRC, M_TYPE, M, (N)*sizeof(FDATA), \
	  		      &s_, &r1, &b1, &b2);			\
	  r_ /= sizeof(FDATA); r1 /= sizeof(FD_BASE); s_ /= sizeof(FD_BASE);\
	  SplitMatUnpack_(Add, N, r_, s_, b1, r1, b2, alpha, A, RIncA, CIncA);\
	  (r) += r_; 							\
	}

# define BRD(RowCol, SRCP0, M, N, A, IncA, r) 				\
	{ int r_, r1, s_, I0_ = SRCABSCID(RowCol, SRCP0); FDATA *b1, *b2;\
	  if ((SRCP0)==0)  						\
	    r_ = brd_stride_send(RowCol, M, (N)*sizeof(FDATA),		\
	    			 A, (IncA)*sizeof(FDATA)); 		\
	  else { 							\
	     r_ = brd_stride_recv(RowCol, I0_, M, (N)*sizeof(FDATA), 	\
	    			   &s_, &r1, &b1, &b2);  		\
	     r_ /= sizeof(FDATA); r1 /= sizeof(FD_BASE); s_ /= sizeof(FD_BASE);\
	     SplitMatUnpack_(0, N, r_, s_, b1, r1, b2, NULL, A, IncA, 1);\
	     *(r) += r_; 						\
	  }								\
	}


#elif defined(APLIB_STD)

# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE) 			\
	{ int i; FDATA *Ax_ = (A); int n_ = (N), m_ = (M);		\
	  if ((IncA) == n_ && m_!=1) { n_ *= m_; m_ = 1;}		\
          do {                                                          \
            r_rsend(CDEST, RDEST, gettid(), M_TYPE, Ax_, n_*sizeof(FDATA));\
            m_--; Ax_ += (IncA);                                        \
          } while (m_ > 0);                                             \
 	}

# define RECV(RSRC, CSRC, M, N, A, IncA, M_TYPE, r)                   	\
	RECV_ADD_(0, RSRC, CSRC, M, N, NULL, A, IncA, 1, M_TYPE, r)

# define RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)  \
	RECV_ADD_(1, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)


# define RECV_ADD_(Add, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r) \
	{ int M_ = (M), m_, n_; FDATA *Ax_ = (A); LocMat lA_, lB_;	\
          do {                                                          \
            FDATA *A_1 = (FDATA *) r_rrecv(CSRC, RSRC, gettid(), M_TYPE);\
            if (M_==0 || (N)==0) break;                                 \
            n_ = getmsize()/sizeof(FDATA); m_ = FAST_DVDEQ(n_, N); (r) += n_;\
            (Add? MatAdd: MatCopy)    					\
                (m_, N, pMkLocMat(&lA_, A_1, N, 1), alpha, pMkLocMat(&lB_, Ax_,RIncA, CIncA));\
            Ax_ += FAST_MUL(m_, RIncA);         M_ -= m_;               \
          } while (M_ > 0);                                             \
        }
          

# define BRD(RowCol, SRCP0, M, N, A, IncA, r) 				\
        { int nb = (N)*sizeof(FDATA), I0 = SRCABSCID(RowCol, SRCP0);	\
          FDATA *Ax_ = (A);  int m_ = (M);                              \
          if (m_==0 || nb==0) { nb = 0; m_ = 1; }                       \
          do {                                                          \
            *(r) += ((RowCol)==0? y_brd: x_brd)( I0, Ax_, nb )/sizeof(FDATA);\
            Ax_ += (IncA); m_--;                                        \
          } while (m_ > 0);                                             \
        }

#else

# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE)
# define RECV(RDEST, CDEST, M, N, A, IncA, M_TYPE, r)
# define rRECV_ADD(RDEST, CDEST, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)
# define RECV_ADD_(ADD, RDEST, CDEST, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)
# define BRD(RowCol, SRCP0, M, N, A, IncA, r)
# define SUM(RowCol, SRCP0, M, N, A, IncA, r)	
#endif


#ifdef GLOB_VIA_P2P
static void BRD(RowCol, SRCP0, M, N, A, IncA, r)			
  int RowCol, SRCP0, M, N, IncA, *r; FDATA A[];
#ifndef VPPLIB_STRIDE
{ int DelP0 = 1, P = VNumCells[RowCol], id_rc = AbsCellId[RowCol];
  int sz = FAST_MUL1(M, N), r1;
# ifdef NO_STRIDE_COMM
  FDATA *Ax = A; LocMat lAx, lA;
  LM_ADR(lAx) = NULL;
  if ((M)>1 && (IncA)!=(N)) {
    /* optimize by packing message */
    MatMalloc_FT("BRD", 1, &Ax, sz);
    pMkLocMat(&lAx, Ax, N, 1);		pMkLocMat(&lA, A, IncA, 1);
    if ((SRCP0)==0 && P>1)
      MatCopy(M, N, &lA, NULL, &lAx);
  }		       
# endif
  if (BrdTopChar[RowCol] == 'I') {
    r1 = 0;	
    if (SRCP0 != 0) {    	
      int dP0 = RowCol? 0: P-1, dQ0 = RowCol? P-1: 0;
# ifdef NO_STRIDE_COMM      
      CTG_RECV(dP0, dQ0, sz, Ax, BRDMSGTYPE, r1);      
# else
      RECV(dP0, dQ0, M, N, A, IncA, BRDMSGTYPE, r1);
#endif          
    }
    if (SRCP0 != P-1) {
      int dP0 = RowCol? 0: 1, dQ0 = RowCol? 1: 0;
# ifdef NO_STRIDE_COMM 
      CTG_BLK_SEND(dP0, dQ0, sz, Ax, BRDMSGTYPE); 
# else
      SEND(dP0, dQ0, M, N, A, IncA, BRDMSGTYPE);
#endif      	           	
    }
  } else {
    while (DelP0 < P) DelP0*=2;  		/* smallest power of 2>= P*/
    r1 = sz;
    while (DelP0 > 1) {
      int dP0 = RowCol? 0: DelP0/2, dQ0 = RowCol? DelP0/2: 0;
      if ((SRCP0 & (DelP0-1)) == 0 && SRCP0+DelP0/2 < P) {
/*printf("%d: send(%d,%d=%d)\n", id_rc, dP0, dQ0, REL2ABS1CID(dP0, dQ0));*/
# ifdef NO_STRIDE_COMM 
        CTG_BLK_SEND(dP0, dQ0, sz, Ax, BRDMSGTYPE); 
# else
        SEND(dP0, dQ0, M, N, A, IncA, BRDMSGTYPE);
#endif      	      
      } 
    
      if ((SRCP0 & (DelP0-1)) == DelP0/2) { 
        r1 = 0;
        dP0 = NumCells[0] - dP0, 		dQ0 = NumCells[1] - dQ0;
/*printf("%d: recv(%d,%d=%d)\n", id_rc, dP0, dQ0, REL2ABS1CID(dP0, dQ0));*/
# ifdef NO_STRIDE_COMM      
        CTG_RECV(dP0, dQ0, sz, Ax, BRDMSGTYPE, r1);      
# else
        RECV(dP0, dQ0, M, N, A, IncA, BRDMSGTYPE, r1);
#endif      
      }
      DelP0 /= 2;  
    } /* while (DelP0...) */
  } 
# ifdef NO_STRIDE_COMM
  if (LM_ADR(lAx)!=NULL) {
    if (SRCP0!=0)
      MatCopy(M, N, &lAx, NULL, &lA);            	
    MatFree(LM_ADR(lAx));
   }  
# endif   
   (*r) += r1;	                 		    
} /* BRD() */  
#else
{   int DelP0 = 1, 			P = VNumCells[RowCol];
    int Src = -1, 			NDests = 0, Dests[32];
    while (DelP0 < P) DelP0*=2;  		/* smallest power of 2>= P*/
    while (DelP0 > 1) {
      int dP0 = RowCol? 0: DelP0/2, dQ0 = RowCol? DelP0/2: 0;
      if ((SRCP0 & (DelP0-1)) == 0 && SRCP0+DelP0/2 < P) {
        Dests[NDests] = REL2ABS1CID(dP0, dQ0);
        NDests++;	
      }
      if ((SRCP0 & (DelP0-1)) == DelP0/2) { 
        dP0 = NumCells[0] - dP0, 		dQ0 = NumCells[1] - dQ0;
        Src = REL2ABS1CID(dP0, dQ0);
      }
      DelP0 /= 2;  
    } /* while (DelP0...) */ 
    (*r) += VPPTreeStrideBrd(Src, NDests, Dests, M, N, A, IncA);
  } /* BRD() */  
#endif
#endif /* GLOB_VIA_P2P */

/******************* code to manage globally blocking communs *************/
#ifdef GLOB_BLK_SEND 
/* data structures to manage uncompleted send operations */

static int NoSendWait = 0;
struct WaitMsgs {
  MSG_HNDLE Hndle;		/* message handle */
  FDATA *Msg;			/* address of message buffer that needs to be
  				   freed (ony if alloced within this module);
  				   NULL if not alloced here */
} *PendingSends = NULL;		/* array [0..SzPendingSends-1] of WaitMsgs[] */
static int SzPendingSends = 0;	/* ... to be alloced on GridInit() */
static int NumPendingSends = 0;	/* current number of messages */
static int PSndHd = 0, PSndTl = 0; /* head & tail ptrs to PendingSends[];
				      (PSndHd+NumPendingSends)%SzPendingSends == PSndTl */

void SendWaitMark()		/* unblock all SEND()'s after this point */
{	 
  COMMON_ASSERT_TR(
    (TraceStr, "() Num=%d", NumPendingSends), NoSendWait==0, 0, "SendWaitMark", 4, TRACE_COMMUN
  );
  NoSendWait = 1;  	
} /*SendWaitMark()*/  				   

static FDATA *WaitHandleBuf = NULL;

static void AddWaitHandle(hndle)
  MSG_HNDLE hndle;	
{ 
  TRACE_VALS_((TraceStr, "(_), buf=a%d Num=%d, Mx=%d", FD_ADR(WaitHandleBuf), NumPendingSends,
		SzPendingSends));
  CommonAssert(NoSendWait==1 && NumPendingSends < SzPendingSends, 0, 
  	       "AddWaitHandle", 8, TRACE_COMMUN);

  PendingSends[PSndTl].Hndle = hndle; 	       
  PendingSends[PSndTl].Msg = WaitHandleBuf; 	       
  NumPendingSends++; 		PSndTl = CYCLIC_INC(PSndTl, SzPendingSends);
  WaitHandleBuf = NULL;

} /* AddWaitHandle() */

void CompleteSends(NSends)	/* wait till NSends pending sends are completed 
				   and reset data structures */
{ int i;
  COMMON_ASSERT_TR(
    (TraceStr, "(%d) Num=%d", NSends, NumPendingSends), 
    NoSendWait || NumPendingSends==0, 0, 
    "CompleteSends", 4, TRACE_COMMUN
  );
  if (NSends < 0)
    NSends = NumPendingSends;
  else  
    NSends = MIN(NSends, NumPendingSends); 
  for (i=0; i < NSends; i++) {
/*    printf("SzPendingSends=%d NumPendingSends=%d, PSndHd=%d, PSndTl=%d\n",
    	    SzPendingSends, NumPendingSends, PSndHd, PSndTl); 	*/
    WAIT_SEND(PendingSends[PSndHd].Hndle);
    if (PendingSends[PSndHd].Msg!=NULL)
      MatFree(PendingSends[PSndHd].Msg);
     NumPendingSends--;            PSndHd = CYCLIC_INC(PSndHd, SzPendingSends);
  } /* for (i...) */
  if (NumPendingSends==0)	
    NoSendWait = 0;
} /*CompleteSends() */
	  
#endif /*GLOB_BLK_SEND*/

/************************ setup routines **********************************/

void BLACSCntxtSet( BLACSContext )
  int BLACSContext;
{ BLACSCntxt = BLACSContext;
} /* BLACSCntxtSet() */

void MPICommSet( MPICommun )
  int MPICommun;
{ MPIComm = MPICommun;
} /* MPICommSet() */

void GridInit(NCelI, NCelJ) 
  int NCelI, NCelJ;
{  
  NCel = NCelI*NCelJ;

#ifdef MPI_BLACS
  MPI_Init(0, NULL);
#endif      
  GET_GLOBAL_PROC_INFO(NCel, AbsCId);
  NumCells[0] = NCelI, 			NumCells[1] = NCelJ;
  if (NCelJ > 0)  
    AbsCellId[0] = AbsCId / NCelJ, 	AbsCellId[1] = AbsCId % NCelJ;
  AssertCId[0] = AbsCellId[0];	 	AssertCId[1] = AbsCellId[1];

#ifdef GLOB_BLK_SEND
# define PSENDSENDS_OK (NumPendingSends==0)
#else
# defire PSENDSENDS_OK 1
#endif
  TRACE_VALS_((TraceStr, "(%dx%d): cell %d of %d", NCelI, NCelJ, AbsCId, NCel));
  CommonAssert(1<=NCelI && 1<=NCelJ && NCelI*NCelJ<=NCel && PSENDSENDS_OK,
	       0, "GridInit", 2, TRACE_COMMUN);      	
  
  if (NCel > NCelI*NCelJ)		/* this can occur in VPPLIB*/ 
    NCel = NCelI*NCelJ;
#if defined(GLOB_BLK_SEND)
  if (PendingSends!=NULL)		/* free from prev alloced grid */
    free((char *) PendingSends);
  SzPendingSends = 3*NCel;		/* a transpose may have <= NCel
  					   pending messages; a call to 
  					   *BlkMat() may require 3 */
  PendingSends = (struct WaitMsgs *) malloc(SzPendingSends*sizeof(struct WaitMsgs));
  PSndHd = 0, PSndTl = 0;		/* ensure indexes are valid */
#endif     
} /* GridInit() */

	
void CommSetUp(NCelI, NCelJ)
  int NCelI, NCelJ;
{ 
#ifdef BLACS
  if (BLACSCntxt==-1) {		/* illegal context; create new one */
    int BLACSDefHandle=0;
    FTN_BLACS_GET(&BLACSDefHandle, &BLACSDefHandle, &BLACSCntxt);
    FTN_BLACS_GRIDINIT(&BLACSCntxt, C2F_CHAR("Row"), &NCelI, &NCelJ);
  }
  FTN_BLACS_GRIDINFO(&BLACSCntxt, &NumCells[0], &NumCells[1],
  		     &AbsCellId[0], &AbsCellId[1]);
#elif defined(APLIB)
  NumCells[0] = getncely(); 		NumCells[1] = getncelx();
  AbsCellId[0] = getcidy();		AbsCellId[1] = getcidx();      
#elif defined(ONE_DIM_ID) 
  /* assume setup by GridInit() */
#else
  NumCells[0] = 1;			NumCells[1] = 1;
  AbsCellId[0] = 0; 			AbsCellId[1] = 0;
#endif
  
  /* setup for printing of traces / error messages, etc */
  AssertCId[0] = AbsCellId[0];	 	AssertCId[1] = AbsCellId[1];

  VNumCells[0] = NCelI;			VNumCells[1] = NCelJ;
  
#ifdef PANEL_SCATTER
  SpecTop = 0;  
  if (NCelI > NumCells[0]) 
    SpecTop = 1;
  else if (NCelJ > NumCells[1])
    SpecTop = 2;
  if (SpecTop) {	
    int n = NumCells[SpecTop-1];
    PInc = n;
    LgPInc = 0;
    while (n > 0) n /= 2, LgPInc++;
    if (2<<LgPInc != PInc) 
      LgPInc = -1, NCel = NumCells[1]*NumCells[0];
    else
      NCel =  NumCells[2-SpecTop] << LgPInc;
    if (PS_P0 < 0)         
      PS_P0 = AbsCellId[SpecTop-1];  	
  }
  AbsCId = getcid();  	  
#endif
  
#ifdef PANEL_SCATTER
  COMMON_ASSERT_TR(
    (TraceStr, "(%dx%d): (%d,%d) on %dx%d grid, cntxt %d", NCelI, NCelJ, 
     AbsCellId[0], AbsCellId[1], NumCells[0], NumCells[1], BLACSCntxt),  
    SpecTop? (VNumCells[SpecTop-1]==NCel && VNumCells[2-SpecTop]==1):
    (1==NCelI || NCelI==NumCells[0]) && (1==NCelJ || NCelJ==NumCells[1]), 
    0, "CommSetUp", 6, TRACE_COMMUN
  );  
#else
  COMMON_ASSERT_TR(
    (TraceStr, "(%dx%d): (%d,%d) on %dx%d grid, cntxt %d", NCelI, NCelJ, 
     AbsCellId[0], AbsCellId[1], NumCells[0], NumCells[1], BLACSCntxt),  
    (1==NCelI || NCelI==NumCells[0]) && (1==NCelJ || NCelJ==NumCells[1]),
    0, "CommSetUp", 6, TRACE_COMMUN
  );  
#endif    
} /* CommSetUp() */ 

  
void BLACSTopSet( RowCol, Op, Top )
  int RowCol; char *Op, *Top;
{ COMMON_ASSERT_TR(
    (TraceStr, "(%d,%c,%c)", RowCol, *Op, *Top),
    RCOK(RowCol) && (*Op=='B' || *Op=='C'), 0, "BLACSTopSet", 6, 
    TRACE_COMMUN
  );
  if (*Op == 'B')
    BrdTopChar[RowCol] = *Top;
  else  
    CombineTopChar[RowCol] = *Top;
} /*BLACSTopSet() */


/*************** miscellaneous routines **********************************/
 
void Abort( ErrorNum )
  int ErrorNum;
{ PrintLinePrefix();
  printf("DBLAS fatal error encountered, code %d -- aborting\n", ErrorNum); 
#if defined(BLACS)
  FTN_BLACS_ABORT( &BLACSCntxt, &ErrorNum); 
#elif defined(VPPLIB)
  VPP_Abort();  
#elif defined(MPI)
  MPI_Abort((MPI_Comm) MPIComm, ErrorNum);
#endif    	
  exit( ErrorNum );
} /* Abort() */   


int GetAbsCId( RowCol )
  int RowCol;
{ return(AbsCellId[RowCol]); 
} /*GetAbsCId() */


int GetNumCells( RowCol )
  int RowCol;
{ return(VNumCells[RowCol]); 
} /*GetNumCells() */


void BarrierSync( RowCol )
  int RowCol;
{ COMMON_ASSERT_TR(
    (TraceStr, "(%d)", RowCol),
    (RCOK(RowCol) || RowCol==-1), 0, "BarrierSync", 4, TRACE_COMMUN
  );
  if (RowCol > 0 && VNumCells[RowCol]==1)
    return;    
#ifdef BLACS
  { char c = RowCol==-1? 'A': RC_CHAR[RowCol]; 
    FTN_BLACS_BARRIER( &BLACSCntxt, C2F_CHAR(&c) ); 
  }
#elif defined APLIB
  if (RowCol==-1 || SPECTOP-1==RowCol)
    sync(0, 0);  			/* note: global scope only */
  else {
    FD_BASE x = 0.0;
    (RowCol==0? Y_SUM: X_SUM)( x, &x );      	
  }
#elif defined(VPPLIB) || defined(MPI)
  { FDATA x = FD_0; int src;
    AbsMaxElt(ABS(RowCol), &x, &src);
    if (RowCol==-1)
      AbsMaxElt(0, &x, &src);
  }     
#endif  
} /*BarrierSync() */

/******************* matrix routines ***************************/

#define RIJOK(RI, RJ) (ABS(RI)<VNumCells[0] && ABS(RJ)<VNumCells[1]) 
#ifdef PANEL_SCATTER

# define NORMALIZE(RI, RJ)					\
  { if (VNumCells[0]==1) RI = 0;				\
    if (VNumCells[1]==1) RJ = 0;				\
    if (RI < 0) RI += VNumCells[0];				\
    if (RJ < 0) RJ += VNumCells[1];				\
    CONV2DIMCID(RI, RJ);					\
  }  

#define CONV2DIMCID(RI, RJ) /*RI, RJ should be normalized*/	\
    if (SpecTop) {						\
      int ri_ = MAX(RI, RJ), rj_;				\
      if (LgPInc>=0) 						\
      	rj_ = ri_ & (PInc-1),  ri_ = (ri_ + PS_P0) >> LgPInc;	\
      else							\
      	rj_ = ri_ % PInc,      ri_ = (ri_ + PS_P0) / PInc;	\
      if (SpecTop==1)						\
        RJ = ri_, RI = rj_;					\
      else  							\
        RI = ri_, RJ = rj_;					\
    }			
    
static int SRCABSCID1(P0)
 int P0;
{ int RI = -P0, RJ = -P0; int abscid;
  NORMALIZE(RI, RJ);
  RI = CYCLIC_ADD(RI, AbsCellId[0], NumCells[0]);
  RJ = CYCLIC_ADD(RJ, AbsCellId[1], NumCells[1]);  
  abscid = RI*NumCells[1]+RJ;
  return(abscid);
}   	  
/*# define SRCABSCID1(P0)                               		\
         CYCLIC_SUB(AbsCId, P0, NCel)*/
   
#else
# define NORMALIZE(RI, RJ)					\
  { if (VNumCells[0]==1) RI = 0;				\
    if (VNumCells[1]==1) RJ = 0;				\
    if (RI < 0) RI += VNumCells[0];				\
    if (RJ < 0) RJ += VNumCells[1];				\
  }								
#endif
  
void AbsMaxElt(RowCol, A, indexA)
  int RowCol, *indexA; FDATA *A; 
{ COMMON_ASSERT_TR(
    (TraceStr, "(%d,*%s,*%d)", RowCol, pFD2Str(A), *indexA),
    RCOK(RowCol), 0, "AbsMaxElt", 4, TRACE_COMMUN
  );

  if (VNumCells[RowCol]==1)
    return;
#if defined(BLACS) || (defined(APLIB) && FDATATYPE>=COMPLEX) || defined(GLOB_VIA_P2P)
  { /* combine messages; cell 0 accumulates max 1st */
    FDATA MxI[2], RMxI[2]; 
    int P = VNumCells[RowCol],	P0 = AbsCellId[RowCol];
    int DelP0 = 1, lgDelP0=0, r ;
    FD_SET_REAL(MxI[0], (FD_BASE) *indexA);	MxI[1] = *A; 

    while (DelP0 < P) {
      int SendMsg = (P0 & (2*DelP0-1)) == DelP0;
      int DelP0_ = SendMsg? -DelP0: DelP0;	
      int DelIP0 = RowCol==0? DelP0_: 0,  DelJP0 = RowCol==0? 0: DelP0_;
      if (SendMsg) {
        CTG_BLK_SEND(DelIP0, DelJP0, 2, MxI, SUMMSGTYPE);
      } else if ((P0 & (2*DelP0-1))==0 && P0+DelP0 < P) {
        CTG_RECV(DelIP0, DelJP0, 2, RMxI, SUMMSGTYPE, r);
        if (FD_ABS(RMxI[1]) > FD_ABS(MxI[1]))
          MxI[1] = RMxI[1], MxI[0] = RMxI[0];
      } /* if (SendMsg) */ 
      DelP0 *= 2, lgDelP0++;	
    } /* while (DelP0 ...) */ 
    BRD(RowCol, P0, 1, 2, MxI, 2, &r);
    *indexA = (int) FD_REAL(MxI[0]); *A = MxI[1];    
  } 
#elif defined(APLIB)
# ifdef PANEL_SCATTER
  if (SpecTop)
    XY_AMAX(*A, *indexA, A, indexA);
  else  
#endif   
  (RowCol==0? Y_AMAX: X_AMAX)(*A, *indexA, A, indexA);  
#endif         
#if defined(APLIB)
  COMMON_ASSERT(1, COMMUN_ERROR_FLAG, "AbsMaxElt (post)", 6, TRACE_COMMUN);
#endif  
} /* AbsMaxElt() */

#ifndef BLACS
void SendMat(RI, RJ, m, n, A, IncA, MsgType)
  int RI, RJ, m, n, IncA; FDATA A[]; int MsgType;
{ COMMON_ASSERT_TR_E(
    (TraceStr, "(%d %d,%dx%d,a%d %d,%d)", RI, RJ, m, n, FD_ADR(A), IncA, MsgType),
    RIncOK(n, IncA) && RIJOK(RI, RJ), 0, "SendMat", 4, TRACE_COMMUN
  );
  if (m==0 || n==0) {
    TRACE_RETURN;
  }
  NORMALIZE(RI, RJ);
  SEND(RI, RJ, m, n, A, IncA, MsgType);
  TRACE_EXIT;
 } /* SendMat() */


void RecvMat(RI, RJ, m, n, A, IncA, MsgType)
  int RI, RJ, m, n, IncA, MsgType; FDATA A[]; 
{ int r=0;  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d %d,%dx%d,a%d %d,%d)", RI, RJ, m, n, FD_ADR(A), IncA, MsgType),
    RIncOK(n, IncA) && RIJOK(RI, RJ), 0, "RecvMat", 4, TRACE_COMMUN
  );

  if (m==0 || n==0) {
    TRACE_RETURN;
  }
  NORMALIZE(RI, RJ);
  RECV(RI, RJ, m, n, A, IncA, MsgType, r);
 
  COMMON_ASSERT_TR(
    (TraceStr, " recv: exp=%d, act=%d", m*n, r), 
    m*n==r, COMMUN_ERROR_FLAG, "RecvMat", 5, TRACE_COMMUN
  );
  TRACE_EXIT;
} /* RecvMat() */


void BrdMat(RowCol, P0, m, n, A, IncA)
  int RowCol, P0, IncA, m, n; FDATA A[]; 
{ int r=0;  				
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%dx%d,a%d %d)", RowCol, P0, m, n, FD_ADR(A), IncA),
    RCOK(RowCol) && RIncOK(n, IncA) && (!SPECTOP || m<=1 || n==IncA || n==0)
    && IncA < (1<<(8*sizeof(short)-1)), 
    0, "BrdMat", 4, TRACE_COMMUN
  );

  if (VNumCells[RowCol]==1 || m==0 || n==0) {
    TRACE_RETURN;
  } 

#ifdef PANEL_SCATTER
  if (SpecTop) {
    int n_ = n;				
    if (m > 1) /* assert n==IncA */ n_ *= m;    
    r = xy_brd(SRCABSCID1(P0), A, n_*sizeof(FDATA))/sizeof(FDATA);
  } else
#endif    
  BRD(RowCol, P0, m, n, A, IncA, &r);  

  if (P0 != 0) {
    COMMON_ASSERT_TR(
      (TraceStr, "(%d,%d,%d,%d,a%d,%d): recv: exp=%d, act=%d\n",  
       RowCol, P0, m, n, FD_ADR(A), IncA, m*n, r), 
       m*n==r, COMMUN_ERROR_FLAG, "BrdMat", 5, TRACE_COMMUN
    );
  }  
  TRACE_EXIT;
} /* BrdMat() */
#endif 

void SumMat(RowCol, P0, P, m, n, A)
  int RowCol, P0, m, n; pLocMat A; 
{ int r=0, IncA = pLM_Ld(A); FDATA *Ax = pLM_ADR(A);
  LocMat lA;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d %d,%dx%d,a%d %d %d)", 
     RowCol, P0, P, m, n, FD_ADR(Ax), pLM_RInc(A), pLM_CInc(A)),
     RCOK(RowCol) && pLM_IncOKUnit(m, n, A) && P<=VNumCells[RowCol] &&
     (pLM_CInc(A)==1 || P0==0), 0, "SumMat", 4, TRACE_COMMUN
  );

  if (P==1 || m==0 || n==0) {
    TRACE_RETURN;
  }  
#ifdef BLACS
  LM_ADR(lA) = NULL;
  if (pLM_CInc(A)!=1 && n>1) {	/* assert P0==0 */
    if (m>1) {
      /* BLACS will not be able to handle reversed format of A in receiver */
      MatMalloc_FT("SumMat", 1, &Ax, m*n);
      IncA = n;  
      MatCopy(m, n, A, NULL, pMkLocMat(&lA, Ax, IncA, 1));
    } else /* interchange m, n to compensate for that within SUM() */
      m = n, n = 1;
  } 
  SUM(RowCol, P0, m, n, Ax, IncA, r);  
  if (LM_ADR(lA)!=NULL) {
    MatCopy(m, n, &lA, NULL, A);
    MatFree(LM_ADR(lA));
  }
#elif defined(APLIB)
  if ((m == 1) && (n == 1)) {
# ifdef PANEL_SCATTER
    if (SpecTop) 			
      XY_SUM(Ax[0], Ax);     	
    else
# endif    
    (RowCol==0? Y_SUM: X_SUM)( FD_REAL(Ax[0]),  &(FD_REAL(*Ax)) );	
# if (FDATATYPE>=COMPLEX)    
    (RowCol==0? Y_SUM: X_SUM)( FD_IMAG(Ax[0]),  &(FD_IMAG(*Ax)) );	
# endif    
    TRACE_RETURN;
  } else 
#endif
#if defined (VPPLIB_STRIDE)
  { int DelP0 = 1, lgDelP0=0;		int sz = FAST_MUL1(m, n); 	  
    int Dst = -1,               	NSrcs= 0, Srcs[32];
    while (DelP0 < P) {
      int SendMsg = (P0 & (2*DelP0-1)) == DelP0;
      int DelP0_ = SendMsg? P-DelP0: DelP0;	
      int DelIP0 = RowCol==0? DelP0_: 0,  DelJP0 = RowCol==0? 0: DelP0_;
      if (SendMsg) {
      	Dst = REL2ABS1CID(DelIP0, DelJP0);
      } else if ((P0 & (2*DelP0-1))==0 && P0+DelP0 < P) {              	
        Srcs[NSrcs] = REL2ABS1CID(DelIP0, DelJP0);
        NSrcs++;
      } /* if (SendMsg) */ 
    DelP0 *= 2, lgDelP0++;	
    } /* while (DelP0 ...) */ 
    r += VPPTreeStrideSum(Dst, NSrcs, Srcs, m, n, pLM_ADR(A), pLM_RInc(A), pLM_CInc(A));
  } /* if (m... n...) */

#elif defined (APLIB) || defined(GLOB_VIA_P2P)
  { int DelP0 = 1, lgDelP0=0;	int sz = FAST_MUL1(m, n); 	  
# if defined(NO_STRIDE_COMM)	
    LM_Init(lA, NULL, pLM_RInc(A), pLM_CInc(A));
    if (P0!=0 && m>1 && IncA != n) {
      /* optimize by packing message if we are not on dest'n cell */
      MatMalloc_FT("SumMat", 1, &Ax, sz);
      IncA = n;  
      MatCopy(m, n, A, NULL, pMkLocMat(&lA, Ax, IncA, 1));
    }  
# endif    
    if (BrdTopChar[RowCol] == 'I') {
      if (P0 != 1) {
        int dP0 = RowCol? 0: P-1, dQ0 = RowCol? P-1: 0;	      
# ifdef PANEL_SCATTER              
        NORMALIZE(dP0, dQ0);         	
# endif     
        RECV_ADD(dP0, dQ0, m, n, NULL, Ax, LM_RInc(lA), LM_CInc(lA), SUMMSGTYPE, r);
      }
      if (P0 != 0) {
# if !defined(NO_STRIDE_COMM) 
        SEND(1-RowCol, RowCol, m, n, Ax, IncA, SUMMSGTYPE);
# else
	CTG_BLK_SEND(1-RowCol, RowCol, sz, Ax, SUMMSGTYPE);	 
# endif	       
      } 
    } else {            
      while (DelP0 < P) {
        int SendMsg = (P0 & (2*DelP0-1)) == DelP0;
        int DelP0_ = SendMsg? P-DelP0: DelP0;	
        int DelIP0 = RowCol==0? DelP0_: 0,  DelJP0 = RowCol==0? 0: DelP0_;
# ifdef PANEL_SCATTER              
        NORMALIZE(DelIP0, DelJP0);         	
# endif     
        if (SendMsg) {
# if !defined(NO_STRIDE_COMM) 
          SEND(DelIP0, DelJP0, m, n, Ax, IncA, SUMMSGTYPE);
# else
	  CTG_BLK_SEND(DelIP0, DelJP0, sz, Ax, SUMMSGTYPE);	 
# endif	       
        } else if ((P0 & (2*DelP0-1))==0 && P0+DelP0 < P) {              	
          RECV_ADD(DelIP0, DelJP0, m, n, NULL, Ax, LM_RInc(lA), LM_CInc(lA), SUMMSGTYPE, r);
        } /* if (SendMsg) */
      DelP0 *= 2, lgDelP0++;	
      } /* while (DelP0 ...) */
    } /* else <use tree topology> */  
# if defined(NO_STRIDE_COMM)    
    if (LM_ADR(lA)!=NULL) 		/* must not be dest'n cell */
      MatFree(LM_ADR(lA));
# endif      
  } /* if (m... n...) */
#endif
 
  COMMON_ASSERT(1, COMMUN_ERROR_FLAG, "SumMat (post)", 5, TRACE_COMMUN);
  TRACE_EXIT;  
 } /* SumMat() */


/******************* [1] block matrix routines ***************************/
	       
void BrdBlkMat(RowCol, P0, m, B0, B, BInc, n, A, IncA)
  int RowCol, P0, IncA, m, B0, B, BInc, n; FDATA A[]; 
{ int r=0;  
#ifdef ASSERT
  int expsize= m*n;
#endif  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%d[%d %d %d]x%d,a%d %d)",
     RowCol, P0, m, B0, B, BInc, n, FD_ADR(A), IncA),
    RCOK(RowCol) && (m+B0<=B || BInc==1 || IncA==n) && !SPECTOP  &&
    RIncOK(n, IncA) && BlkOK(B0, B, BInc), 0, "BrdBlkMat", 4, TRACE_COMMUN
  );

  if (VNumCells[RowCol]==1 || m==0 || n==0) {
    TRACE_RETURN;
  }  
  if (m+B0<=B || BInc==1) {
    BRD(RowCol, P0, m, n, A, IncA, &r);  
#ifdef APLIB_STRDE			/*from here on, m=1 optim. unimportant*/
# undef SplitMatUnpack_
# define SplitMatUnpack_ SplitMatUnpack
#endif
  } else {
    IncA = FAST_MUL1(B, n*BInc);	
    COMMON_ASSERT_TR((TraceStr, "(%d[%d %d %d]x%d)", m, B0, B, BInc, n), 
    		      !LM_InvalStride(IncA), 0, "BrdBlkMat", 8, TRACE_COMMUN);
    if (B0 > 0) {
      int m1 = MIN(B-B0, m);		
      BRD(RowCol, P0, 1, m1*n, A, IncA, &r);
      A += (B*BInc-B0)*n;		m -= m1;
    }
    if (m >= B) {
      int m1 = FAST_DVD(m, B);            	
      BRD(RowCol, P0, m1, FAST_MUL2(n, B), A, IncA, &r);
      A += m1*IncA;      		m = FAST_MOD(m, B);	
    }
    if (m > 0)
      BRD(RowCol, P0, 1, (m%B)*n, A, IncA, &r);
  }   

  if (P0 != 0) {
    COMMON_ASSERT_TR(
      (TraceStr, " recv: exp=%d, act=%d", expsize, r), 
       expsize==r, COMMUN_ERROR_FLAG, "BrdBlkMat", 5, TRACE_COMMUN
    );
  }  
  TRACE_EXIT;
} /* BrdBlkMat() */


void SendBlkMat(RI, RJ, m, B0, B, BInc, n, A, MsgType)
  int RI, RJ, m, B0, B, BInc, n; FDATA A[]; int MsgType;
{ COMMON_ASSERT_TR_E(
    (TraceStr, "(%d %d,%d[%d %d %d]x%d,a%d,%d)",
     RI, RJ, m, B0, B, BInc, n, FD_ADR(A), MsgType),
    RIJOK(RI, RJ) && BlkOK(B0, B, BInc), 0, "SendBlkMat", 4, TRACE_COMMUN
  );

  if (m==0 || n==0) {
    TRACE_RETURN;
  }
  NORMALIZE(RI, RJ);
  if (m+B0<=B || BInc==1) {
    SEND(RI, RJ, m, n, A, n, MsgType);  
  } else {
    int IncA = FAST_MUL1(B, n*BInc);	
    if (B0 > 0) {
      int m1 = MIN(B-B0, m);
      SEND(RI, RJ, 1, m1*n, A, IncA, MsgType);
      A += (B*BInc-B0)*n;		m -= m1;
    }
    if (m >= B) {
      int m1 = FAST_DVD(m, B);	
      SEND(RI, RJ, m1, FAST_MUL2(n, B), A, IncA, MsgType);
      A += m1*IncA;      		m = FAST_MOD(m, B);	
    }
    if (m > 0)
      SEND(RI, RJ, 1, m*n, A, IncA, MsgType);   
  }    
  TRACE_EXIT;
} /* SendBlkMat() */


void RecvBlkMat(RI, RJ, Add, m, B0, B, BInc, n, alpha, A, MsgType)
  int RI, RJ, Add, m, B0, B, BInc, n; FDATA  *alpha, A[]; int MsgType; 
{ int r=0;				
#ifdef ASSERT
  int expsize= m*n;
#endif  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d %d,%d,%d[%d %d %d]x%d,%s,a%d,%d)",
     RI, RJ, Add, m, B0, B, BInc, n, pFD2Str(alpha), FD_ADR(A), MsgType),
    RIJOK(RI, RJ) && BlkOK(B0, B, BInc), 0, "RecvBlkMat", 4, TRACE_COMMUN
  );

  if (m==0 || n==0) {
    TRACE_RETURN;
  }  
  NORMALIZE(RI, RJ);
  if (m+B0<=B || BInc==1) {
#ifdef APLIB_STRDE			/*here, m=1 optim. important*/
# undef SplitMatUnpack_
# define SplitMatUnpack_ SplitMatUnpack_f
#endif
    RECV_ADD_(Add, RI, RJ, m, n, alpha, A, n, 1, MsgType, r);
#ifdef APLIB_STRDE			/*from here on, m=1 optim. unimportant*/
# undef SplitMatUnpack_
# define SplitMatUnpack_ SplitMatUnpack
#endif
  } else {
    int IncA = FAST_MUL1(B, n*BInc);        	  	
    COMMON_ASSERT_TR((TraceStr, " %d[%d %d %d]x%d IncA=%d", IncA, m, B0, B, BInc, n), 
    		      !LM_InvalStride(IncA), 0, "RecvBlkMat", 8, TRACE_COMMUN);
    if (B0 > 0) {
      int m1 = MIN(B-B0, m);
      RECV_ADD_(Add, RI, RJ, 1, m1*n, alpha, A, IncA, 1, MsgType, r); 
      A += (B*BInc-B0)*n;		m -= m1;
    }
    if (m >= B) {
      int m1 = FAST_DVD(m, B);            	
      RECV_ADD_(Add, RI, RJ, m1, FAST_MUL2(n, B), alpha, A, IncA, 1, MsgType, r);  
      A += m1*IncA;      		m = FAST_MOD(m, B);	
    }
    if (m > 0)
      RECV_ADD_(Add, RI, RJ, 1, m*n, alpha, A, IncA, 1, MsgType, r);   
  }  
  COMMON_ASSERT_TR_X(
    (TraceStr, " recv: exp=%d, act=%d", expsize, r),
     expsize==r, COMMUN_ERROR_FLAG, "RecvBlkMat", 5, TRACE_COMMUN
  );
} /* RecvBlkMat() */
	       
/*********************** 2-dim block matrix routines ************************/
	       
void Send2BlkMat(RI, RJ, M, N, alpha, A, RB0, RB, RBInc, CB0, CB, CBInc)
  int RJ, RI, RB0, RB, RBInc, CB0, CB, CBInc; 
  FDATA *alpha; pLocMat A;
{ int r=FAST_MUL1(M, N); 
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d %d,%dx%d,%s,a%d %d %d,%d %d %d,%d %d %d)", RI, RJ, M, N, 
     pFD2Str(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A), RB0, RB, RBInc, CB0, CB, CBInc),
    RIJOK(RI, RJ) && pLM_BlkIncOK(M, N, A, RB0, RB, RBInc, CB0, CB, CBInc),
    0, "Send2BlkMat", 4, TRACE_COMMUN
  );	

  if (M==0 || N==0) {
    TRACE_RETURN;
  }  
  NORMALIZE(RI, RJ);
  if (pFD_UNITY(alpha) && ((pLM_CInc(A)==1) && (N+CB0<=CB || CBInc==1) && 
      (M==1 || (N==pLM_RInc(A) && (M+RB0<=RB || RBInc==1))))) {
    SEND(RI, RJ, 1, r, pLM_ADR(A), r, TRANSMSGTYPE);
  } else {
    FDATA *Bx; 				LocMat B;
#if !defined(GLOB_BLK_SEND)
    MatMalloc_FT("Send2BlkMat", 1, &Bx, r);
#else    		
    MatMalloc("Send2BlkMat", 1, &Bx, r);
    /* there may be <= NCel of these allocs, so dont use up quickly allocatable 
       memory, which can be used for the corresp. calls to Recv2BlkMat() */
    if (NoSendWait) 
      WaitHandleBuf = Bx;		/* to be freed when send completes*/   
#endif    		
    BlkMatStridePack(0, M, N, A, alpha, pMkLocMat(&B, Bx, N, 1), 
                     RB0, RB, RBInc, CB0, CB, CBInc);              
    SEND(RI, RJ, 1, r, Bx, r, TRANSMSGTYPE);
#if defined(GLOB_BLK_SEND)
    if (!NoSendWait)
#endif
      MatFree(Bx);
  }  
  TRACE_EXIT;
} /* Send2BlkMat */


void Recv2BlkMat(RI, RJ, Add, M, N, B, RB0, RB, RBInc, CB0, CB, CBInc)
  int RJ, RI, Add, M, N, RB0, RB, RBInc, CB0, CB, CBInc; 
  pLocMat B; 
{ int r=0;   		    		
  FDATA *Ax;				LocMat A;
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d %d,%d,%dx%d,a%d %d %d,%d %d %d,%d %d %d)",
     RI, RJ, Add, M, N, FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), 
     RB0, RB, RBInc, CB0, CB, CBInc),
    RIJOK(RI, RJ) && pLM_BlkIncOK(M, N, B, RB0, RB, RBInc, CB0, CB, CBInc),
    0, "Recv2BlkMat", 4, TRACE_COMMUN
  );	

  if (M==0 || N==0) {
    TRACE_RETURN;
  } 
  NORMALIZE(RI, RJ);
  if (N+CB0<=CB) CBInc==1;
  if (M+RB0<=RB) RBInc==1;
#if defined(BLACS) || defined(VPPLIB) || defined(MPI)
  if (!Add && (pLM_CInc(B)==1) && (CBInc==1) && (RBInc==1)) {
    RECV(RI, RJ, M, N, pLM_ADR(B), pLM_RInc(B), TRANSMSGTYPE, r);
  } else {			
    int mn = FAST_MUL1(M, N);	
    MatMalloc_FT("Recv2BlkMat", 1, &Ax, mn);		
    RECV(RI, RJ, 1, mn, Ax, mn, TRANSMSGTYPE, r);
    BlkMatStrideUnpack(Add, FAST_DVDEQ(r ,N), N, pMkLocMat(&A, Ax, N, 1), NULL, B,
                       RB0, RB, RBInc, CB0, CB, CBInc);              
    MatFree(Ax);
  }
#elif defined(APLIB_STD) || defined(APLIB_LINUX) || defined(APLIB_PLUS_STRIDE) 
# if defined(APLIB_PLUS_STRIDE)
  { int r1, s; FDATA *b2;
    r = FAST_MUL1(M, N*sizeof(FDATA));
    r = xy_stride_recv(RJ, RI, TRANSMSGTYPE, 1, r, &s, &r1, &Ax, &b2);
    r /= sizeof(FDATA);
  }
# else    
  Ax = (FDATA *) r_rrecv(RJ, RI, gettid(), TRANSMSGTYPE);
  r = getmsize()/sizeof(FDATA);
# endif
  pMkLocMat(&A, Ax, N, 1);
  if (RBInc==1 && CBInc==1) 
    (Add? MatAdd: MatCopy)(FAST_DVDEQ(r, N), N, &A, NULL, B);
  else
    BlkMatStrideUnpack(Add, FAST_DVDEQ(r, N), N, &A,
                       NULL, B, RB0, RB, RBInc, CB0, CB, CBInc);              
#elif defined(APLIB_STRIDE) 
  { FDATA *b2; int r1; int s;
    r = FAST_MUL1(M, N*sizeof(FDATA));
    r = xy_stride_recv(RJ, RI, TRANSMSGTYPE, 1, r, &s, &r1, &Ax, &b2)/sizeof(FDATA);
    /*assert s=0*/
    if (r1/sizeof(FDATA)==r && RBInc==1 && CBInc==1) 
      (Add? MatAdd: MatCopy)(FAST_DVDEQ(r, N), N, pMkLocMat(&A, Ax, N, 1), NULL, B);
    else   				/*message was split (AP ring buffer)*/
      SplitBlkMatStrideUnpack(Add, N, r, Ax, r1/sizeof(FD_BASE),
  	                      b2, N, NULL, B, RB0, RB, RBInc, CB0, CB, CBInc); 
  }
#else 
  /*this function should not have been called & following assertion will fail*/;
#endif  
 
  COMMON_ASSERT_TR_X(
    (TraceStr, " recv: exp=%d, act=%d", M*N, r), 
    r==M*N, COMMUN_ERROR_FLAG, "Recv2BlkMat", 5, TRACE_COMMUN
  );
} /* Recv2BlkMat */


void RecvBrd2BlkMat(RowCol, P0, M, N, B, RB0, RB, RBInc, CB0, CB, CBInc)
  int RowCol, P0, M, N, RB0, RB, RBInc, CB0, CB, CBInc; 
  pLocMat B;
{ int r=0;
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%dx%d,a%d %d %d,%d %d %d,%d %d %d)", RowCol, P0, M, N, 
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), RB0, RB, RBInc, CB0, CB, CBInc),
    RCOK(RowCol) && pLM_BlkIncOK(M, N, B, RB0, RB, RBInc, CB0, CB, CBInc) &&
    VNumCells[RowCol]>1 && P0!=0 && (!SPECTOP || (M<=1 || N==pLM_RInc(B)) &&
    (pLM_CInc(B)==1) && (N+CB0<=CB || CBInc==1) && (M+RB0<=RB || RBInc==1)), 
    0, "RecvBrd2BlkMat", 4, TRACE_COMMUN
  );	
  
  if (M==0 || N==0) {
    TRACE_RETURN;
  }  
  
  if ((pLM_CInc(B)==1) && (N+CB0<=CB || CBInc==1) && (M+RB0<=RB || RBInc==1)) {
#ifdef PANEL_SCATTER
    if (SpecTop) {
      int n_ = N;				
      if (M > 1) /* assert n==IncA */ n_ *= M;			
      r = xy_brd(SRCABSCID1(P0), pLM_ADR(B), n_*sizeof(FDATA))/sizeof(FDATA);
    } else
#endif    
    BRD(RowCol, P0, M, N, pLM_ADR(B), pLM_Ld(B), &r);
  } else {
#if defined(BLACS) || defined(APLIB_STD) || defined(APLIB_LINUX) || defined(VPPLIB) || defined(MPI)
    FDATA *Ax;				LocMat lA_;
    MatMalloc_FT("RecvBrd2BlkMat", 1, &Ax, FAST_MUL1(M, N));		
    BRD(RowCol, P0, M, N, Ax, N, &r);
    BlkMatStrideUnpack(0, FAST_DVDEQ(r, N), N,  pMkLocMat(&lA_, Ax, N, 1), 
                       NULL, B, RB0, RB, RBInc, CB0, CB, CBInc);              
    MatFree(Ax);
#elif defined(APLIB_STRIDE)
    FDATA *b1, *b2; int r1; int s;
    r = brd_stride_recv(RowCol, SRCABSCID(RowCol, P0), M, N*sizeof(FDATA), 
    			&s, &r1, &b1, &b2) / sizeof(FDATA);	
    /* assert s=0; also the omission of partitioning over N on AP+ assumes
       that send sender sent from a contiguous message */			
    SplitBlkMatStrideUnpack(0, N, r, b1, r1/sizeof(FD_BASE), b2, N,
  	                    NULL, B, RB0, RB, RBInc, CB0, CB, CBInc);
#elif defined(APLIB_PLUS_STRIDE)
    FDATA *b1, *b2; int r_, r1; int s;
    int i = 0, nb, I0_ = SRCABSCID(RowCol, P0); LocMat lB_, lA_;
    LM_Copy(0, lB_, *B);
    do {
      r_ = brd_stride_recv(RowCol, I0_, M, (N)*sizeof(FDATA),&s, &r1, &b1, &b2);
      r_ /= sizeof(FDATA); nb = FAST_DVD(r_, M);		
      BlkMatStrideUnpack(0, M, nb, pMkLocMat(&lA_, b1, nb, 1),			
                         NULL, &lB_, RB0, RB, RBInc, CB0, CB, CBInc);  
      if (i < (N)) {                  
        pLocSubMat(&lB_, 0, nb);
        if (CBInc > 1) 
          CB0 = (CB0 + nb) % CBInc;
      }                  
      i += nb; r += r_; 				
    } while ((i < N)  && FAST_MOD(r_, M)==0);		
#else
  /*this function should not have been called & following assertion will fail*/;
#endif  
  } /* else (non-trivial stride in B required */
 
  COMMON_ASSERT_TR_X(
    (TraceStr, " recv: exp=%d, act=%d", M*N, r),
    r==M*N, COMMUN_ERROR_FLAG, "RecvBrd2BlkMat", 5, TRACE_COMMUN
  );
} /* RecvBrd2BlkMat */


#if 0

#if FDATATYPE<COMPLEX
# define SplitMatUnpack_c SplitMatUnpack
#else
# define SplitMatUnpack_c(Add, N, r, s, b1, r1, b2, alpha, A, RIncA, CIncA) \
	{ int IsCmplxAlpha = (alpha)!=NULL && FD_IMAG(*alpha)!=0.0); LocMat lA;\
	  FD_BASE *ReAlpha = IsCmplxAlpha? NULL: (FD_BASE *) (alpha);	\
	  SplitMatUnpack(Add, 2*(N), 2*(r), s, (FD_BASE *) b1, 2*(r1), 	\
	  	        (FD_BASE *) b2, ReAlpha, (FD_BASE *) A, RIncA, 2*(CIncA));\
	  if ((N)>0 && IsCmplxAlpha) 					\
	    MatScale(r/(N), (N), alpha,  pLocSubMat_(&lA, A, RIncA, CIncA));\
	}
#endif
#if (FDATATYPE<COMPLEX)
#define SplitBlkMatStrideUnpack(Add, N, r, Ax, r1, b2, N, B, RB0, RB, RBInc, CB0, CB, CBInc) \
      SplitBlkMatStrideUnpack(Add, N, r, Ax, r1, b2, N, NULL,		\
  	                      B, RB0, RB, RBInc, CB0, CB, CBInc); 
#else
#define SplitBlkMatStrideUnpack(Add, N, r, Ax, r1, b2, N, B, RB0, RB, RBInc, CB0, CB, CBInc) \
      SplitBlkMatStrideUnpack(Add, N, r, Ax, r1, b2, N, NULL,		\
  	                      B, RB0, RB, RBInc, CB0, CB, CBInc); 
#endif

#endif
