 /************************************************************************/
/*	FDCommun: low-level matrix communication procedures for DBLAS	*/
/************************************************************************/
/* SCCS INFO @(#)FDCommun.h 1.6         last modified P. Strazdins, 99/06/09
   first created: P. Strazdins 93/15/01
 */
/* set of matrix communication procedures convenient for DBLAS.
 * design of _BlkMat procedures is dependend on the internal representation for
 * the DBLAS buffers to accumulate a common global matrix dimension (M).
 * Using a block-cyclic distribution of M, we thus have associated:
 *	m	: a local matrix dimension corresponding to M
 *	B0	: block offset of 0th block
 *	B	: width of blocks
 *	BInc	: stride of blocks
 *	n	: length of a row in each block, local dim. corresp to n
 * Thus the number of blocks is (m+B0+B-1)/B, and a block (of rows) spans B*n 
 * contiguous memory locations. The gap between consecutive blocks is B*n*BInc.	
 * Such a segment of a local matrix/workspace is the unit of communication
 * for this module. 
 * The notation m::B0:B:BInc indicates the 1st m indices for blocks of
 * widtf B, block-stride BInc and zeroth block offset B0. 
 * 
 * Apart from the issue of the 1st & last block may be odd-sized, 
 * communication required is a stride one of size n*B, with stride n*B*BInc.
 * Here consistent values of m, n, B0, B, BInc is required by all cells
 * in the scope of the call.
 *
 * Exceptions occur in MatBlkBrd() in which the argument can be 
 * m x n matrices {A, IncA, 1}. Here IncA != n requires (m+B0 <= B || BInc==1).
 * This in turn means that the index range [m::B0:B:BInc] reduces to [0..m-1].
 * Note that an n x m matrix {A, 1, Inc} can be passed as m x n {A, IncA, 1}. 
 * In this case, it is required that only m and n are consistent across all 
 * cells in the scope of the call. This is also the convention for SumMat(),
 * SendMat() and RecvMat().
 *
 * Thus, the routines for this module are not much more than an
 * interface to BLACS or a communication library of similar functionality.
 * While such an module may add yet another layer of overhead on 
 * communication calls, for modularity reasons it is desirable to have
 * these in a single stand-alone module. 
 *
 * This library also has a concept of a processor (sub-)grid over which
 * the current DBLAS computation applies over. This P'xQ' sub-grid is a 
 * sub-grid of the actual PxQ process grid. 
 * Currently P'= 1 or P, and Q' = 1 or Q are supported. 
 * 
 */

#ifndef __FDCommun_INCL__
#define __FDCommun_INCL__

#include "FDATASetUp.h"
#include "LocMat.h"

/****** secondary macros definng characteristics of underlying system *****/

#if defined(APLIB_STRIDE) || defined(APLIB_LINUX) || defined(APLIB_STD) || defined (APLIB_PLUS_STRIDE)
# define APLIB
#endif

#if defined(VPPLIB) || defined (MPI)
# define ONE_DIM_ID             /* interface has only 1-dim cell ids*/        
#endif
 
#if defined(VPPLIB) || defined (MPI) /* for prototype  MPI impl. only */
# define GLOB_VIA_P2P           /* interface provides no b/c reduce opns*/      
				
#endif
 
#if defined(VPPLIB) || defined (MPI) /* for prototype  MPI impl. only */
# define NO_STRIDE_COMM         /* interface system provides no stride commun.*/
#endif
 
#if defined(VPPLIB) || defined (MPI) 
# define GLOB_BLK_SEND           /* interface doesnt guarantee locally blocking send*/ 
#endif
 
#if FDATATYPE==SINGLE
#define GridInit	DSGridInit
#define BLACSCntxtSet	DSBLACSCntxtSet
#define MPICommSet	DSMPICommSet
#define CommSetUp	DSCommSetUp
#define BLACSTopSet	DSBLACSTopSet
#define Abort		DSAbort
#define GetAbsCId	DSGetAbsCId
#define	BarrierSync	DSBarrierSync

#define	AbsMaxElt	DSAbsMaxElt
#define SendMat		DSSendMat
#define RecvMat		DSRecvMat
#define BrdMat		DSBrdMat
#define SumMat		DSSumMat
#define BrdBlkMat	DSBrdBlkMat
#define SendBlkMat	DSSendBlkMat
#define RecvBlkMat	DSRecvBlkMat
#define RecvSumBlkMat	DSRecvSumBlkMat
#define Send2BlkMat	DSSend2BlkMat
#define Recv2BlkMat	DSRecv2BlkMat
#define RecvBrd2BlkMat	DSRecvBrd2BlkMat

#define	MsgMaxAbsVal    DSMsgMaxAbsVal
#define	BrdTopChar    	DSBrdTopChar
#define	CombineTopChar  DSCombineTopChar
#define	BLACSCntxt	DSBLACSCntxt
#define	AbsCellId	DSAbsCellId
#define NumCells	DSNumCells
#define	VNumCells	DSVNumCells
#define	RC_CHAR		DSRC_CHAR

#elif FDATATYPE==DOUBLE
#define GridInit	DDGridInit
#define BLACSCntxtSet	DDBLACSCntxtSet
#define MPICommSet	DDMPICommSet
#define CommSetUp	DDCommSetUp
#define BLACSTopSet	DDBLACSTopSet
#define Abort		DDAbort
#define GetAbsCId	DDGetAbsCId
#define	BarrierSync	DDBarrierSync

#define	AbsMaxElt	DDAbsMaxElt
#define SendMat		DDSendMat
#define RecvMat		DDRecvMat
#define BrdMat		DDBrdMat
#define SumMat		DDSumMat
#define BrdBlkMat	DDBrdBlkMat
#define SendBlkMat	DDSendBlkMat
#define RecvBlkMat	DDRecvBlkMat
#define RecvSumBlkMat	DDRecvSumBlkMat
#define Send2BlkMat	DDSend2BlkMat
#define Recv2BlkMat	DDRecv2BlkMat
#define RecvBrd2BlkMat	DDRecvBrd2BlkMat
#define	MsgMaxAbsVal    DDMsgMaxAbsVal

#define	BrdTopChar    	DDBrdTopChar
#define	CombineTopChar  DDCombineTopChar
#define	BLACSCntxt	DDBLACSCntxt
#define	AbsCellId	DDAbsCellId
#define NumCells	DDNumCells
#define	VNumCells	DDVNumCells
#define	RC_CHAR		DDRC_CHAR

#elif FDATATYPE==COMPLEX
#define GridInit	DCGridInit
#define BLACSCntxtSet	DCBLACSCntxtSet
#define MPICommSet	DCMPICommSet
#define CommSetUp	DCCommSetUp
#define BLACSTopSet	DCBLACSTopSet
#define Abort		DCAbort
#define GetAbsCId	DCGetAbsCId
#define	BarrierSync	DCBarrierSync

#define	AbsMaxElt	DCAbsMaxElt
#define SendMat		DCSendMat
#define RecvMat		DCRecvMat
#define BrdMat		DCBrdMat
#define SumMat		DCSumMat
#define BrdBlkMat	DCBrdBlkMat
#define SendBlkMat	DCSendBlkMat
#define RecvBlkMat	DCRecvBlkMat
#define RecvSumBlkMat	DCRecvSumBlkMat
#define Send2BlkMat	DCSend2BlkMat
#define Recv2BlkMat	DCRecv2BlkMat
#define RecvBrd2BlkMat	DCRecvBrd2BlkMat
#define	MsgMaxAbsVal    DCMsgMaxAbsVal

#define	BrdTopChar    	DCBrdTopChar
#define	CombineTopChar  DCCombineTopChar
#define	BLACSCntxt	DCBLACSCntxt
#define	AbsCellId	DCAbsCellId
#define NumCells	DCNumCells
#define	VNumCells	DCVNumCells
#define	RC_CHAR		DCRC_CHAR

#else
#define GridInit	DZGridInit
#define BLACSCntxtSet	DZBLACSCntxtSet
#define MPICommSet	DZMPICommSet
#define CommSetUp	DZCommSetUp
#define BLACSTopSet	DZBLACSTopSet
#define Abort		DZAbort
#define GetAbsCId	DZGetAbsCId
#define	BarrierSync	DZBarrierSync

#define	AbsMaxElt	DZAbsMaxElt
#define SendMat		DZSendMat
#define RecvMat		DZRecvMat
#define BrdMat		DZBrdMat
#define SumMat		DZSumMat
#define BrdBlkMat	DZBrdBlkMat
#define SendBlkMat	DZSendBlkMat
#define RecvBlkMat	DZRecvBlkMat
#define RecvSumBlkMat	DZRecvSumBlkMat
#define Send2BlkMat	DZSend2BlkMat
#define Recv2BlkMat	DZRecv2BlkMat
#define RecvBrd2BlkMat	DZRecvBrd2BlkMat
#define	MsgMaxAbsVal    DZMsgMaxAbsVal

#define	BrdTopChar    	DZBrdTopChar
#define	CombineTopChar  DZCombineTopChar
#define	BLACSCntxt	DZBLACSCntxt
#define	AbsCellId	DZAbsCellId
#define NumCells	DZNumCells
#define	VNumCells	DZVNumCells
#define	RC_CHAR		DZRC_CHAR
#endif /*FDATATYPE*/

#ifdef  GLOB_BLK_SEND
# if FDATATYPE==SINGLE
# define SendWaitMark	DSSendWaitMark
# define CompleteSends	DSCompleteSends
# elif FDATATYPE==DOUBLE
# define SendWaitMark	DDSendWaitMark
# define CompleteSends	DDCompleteSends
# elif FDATATYPE==COMPLEX
# define SendWaitMark	DCSendWaitMark
# define CompleteSends	DCCompleteSends
# elif FDATATYPE==DOUBLE_COMPLEX
# define SendWaitMark	DZSendWaitMark
# define CompleteSends	DZCompleteSends
# endif
#else
# define SendWaitMark()
# define CompleteSends(NSends)
#endif

/************************ constants associated with communication ***********/

#define COL 		0	/* for communication in I or Y dimension */#define ROW 		1	/* for communication in J or X dimension */


/* differentiate message types for different operations, so that messages
   are not confused */

#define TRANSMSGTYPE 	1120	
#define ROTMSGTYPE   	1125
#define SUMMSGTYPE   	1130
#define GENMSGTTYPE	1135
#define BRDMSGTYPE	1140

/**************** assertion checking macros ********************************/
#define BlkOK(B0, B, BInc)  (0<=(B0) && (B0)<(B) && (BInc)>=1)
#define RCOK(RowCol)	    (0<=(RowCol) && (RowCol)<=1)
#define RIncOK(n, Inc)	    ((Inc) >= (n) && (Inc) > 0)
#define P0OK(RowCol, P0)    (0 <= (P0) && (P0) < GetNumCells(RowCol))
/*
#define RIJOK(RI, RJ) (ABS(RI) < NumCells[0] && ABS(RJ < NumCells[1])
*/

extern FD_BASE MsgMaxAbsVal; 	/* if >0, can be used check for `bad' data 
				in received messages */

/**************** setup of parameters associated with communication ********/

extern char RC_CHAR[2], BrdTopChar[2], CombineTopChar[2];
extern int AbsCellId[2], NumCells[2], VNumCells[2], BLACSCntxt;

void BLACSCntxtSet STDC_PARAMS( ( int BLACSContext ) );
/* posts: sets DBLAS internal BLACS context to BLACSContext
 * note:  in BLACS versions of DBLAS, this must be called before CommSetUp()
 */ 

void MPICommSet STDC_PARAMS( ( int MPIComm ) );
/* posts: sets DBLAS internal MPI Communicator to MPIComm
 * note:  in MPI versions of DBLAS, this must be called before GridInit()
 */ 

void CommSetUp STDC_PARAMS( (int NCelI, int NCelJ) );
/* pre:	 NCelI (NcelJ) = 1 or the number of cells in I (J) dimension of the 
 *	 communication context (set by last call to BLACSCntxtSet() for BLACS 
 *	 version, implicitly defined for APLIB version, 1x1 configuration for 
 *	 CELLBLAS version
 * post: sets up the absolute cell ids, and the actual processor configuration
 *	 according to the context. Sets the virtual processor configuration
 *	 to NCelI x NCelJ
 * note: for BLACS versions, a previously set value of the BLACS context of -1
 *	 causes a BLACS context to be created using NCelI, NCelJ and the default
 *	 system handle.
 * note: this MUST be called before other calls to any of these routines
 *	 that perform actual communication.
 */

extern int PS_P0;
#define PS_OrigSet(P0)			PS_P0 = (P0)

void BLACSTopSet STDC_PARAMS( ( int RowCol, char *Op, char *Top) );
/* pre:  RCOK(RowCol), *Op = 'B" (b/c) or 'C' (combine)
 * post: sets BLACS topology in scope of RowCol corresp. to *Op to *Top
 */

/*************** miscellaneous routines **********************************/

void Abort STDC_PARAMS( ( int ErrorNum ) );
/* post: abort this process, printnig out a message with ErrorNum
 */

int GetAbsCId STDC_PARAMS( ( int RowCol ) );
/* pre:  RCOK(RowCol)
 * post: returns the absolute cell id of this cell in dimension RowCol
 */
 
int GetNumCells STDC_PARAMS( ( int RowCol ) );
/* pre:  RCOK(RowCol)
 * post: returns the current # cells in dimension RowCol of (virtual) process
 *    	 grid
 */
 
void BarrierSync STDC_PARAMS( ( int RowCol ) );
/* pre:  ABS(RowCol) <= 1; all cells in scope of RowCol particiate in this call
 * post: provide a barrier sychronization in scope of RowCol 
 *	  (-1 signifies `all')
 * notes: with APLIB, the only safe scope is -1
 */

/******************* matrix routines ***************************/

void AbsMaxElt STDC_PARAMS( (int RowCol, FDATA *A, int *indexA) );
/* pre:   RCOK(RowCol)
 *        all cells in scope of RowCol are participating in this call 
 * post: *A contains the abs max of *A_pre in all cells; 
 *	 *indexA contains the value of *indexA_pre in the 1st cell having the max 
 */

#ifndef BLACS
void SendMat STDC_PARAMS( 
	(int RI, int RJ, int m, int n, FDATA A[], int IncA, int MsgType)
);	
/* pre:  RIncOK(n, Inc) && RIJOK(RI, RJ); Inc(A) = {IncA, 1}
 *	 m, n must be consistent for sender & receiver;
 * post: A[0..m-1,0..n-1] is sent from this cell to relative cell (RI, RJ)
 *	 with tag MsgType
 */
#else
#include "FtnBLACS.h"
#define REL2ABSCID(RowCol, RCID)                                        \
        CYCLIC_ADD(AbsCellId[RowCol], RCID, NumCells[RowCol])

#define NORMALIZE(RI, RJ)                                       \
  { if (VNumCells[0]==1) RI = 0;                                \
    if (VNumCells[1]==1) RJ = 0;                                \
    if (RI < 0) RI += NumCells[0];                              \
    if (RJ < 0) RJ += NumCells[1];                              \
  }
          
# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE)                      \
        { int aRDEST_ = REL2ABSCID(0, RDEST), aCDEST_ = REL2ABSCID(1, CDEST),\
              m_ = (M), n_ = (N), inc_ = (IncA);                        \
          FTN_GESD2D(&BLACSCntxt, &n_, &m_, A, &inc_, &aRDEST_, &aCDEST_);\
        }

#undef SendMat
static inline 
void SendMat (int RI, int RJ, int m, int n, FDATA A[], int IncA, int MsgType)
{
  if (m==0 || n==0) 
    return;
  NORMALIZE(RI, RJ);
  SEND(RI, RJ, m, n, A, IncA, MsgType);	
} /* SendMat() */
#endif

#ifndef BLACS
void RecvMat STDC_PARAMS(
        (int RI, int RJ, int m, int n, FDATA A[], int IncA, int MsgType)
);
/* pre:  RIncOK(n, Inc) && RIJOK(RI, RJ); Inc(A) = {IncA, 1}
 *	 m, n must be consistent for sender & receiver;
 * post: m x n message A' is received from relative cell (RI, RJ) with tag 
 *	 MsgType; A[0..m-1,0..n-1] = A'
 */
#else
# define RECV_(RSRC, CSRC, M, N, A, IncA, M_TYPE)                     \
        { int aRSRC_ = REL2ABSCID(0, RSRC), aCSRC_ = REL2ABSCID(1, CSRC),\
              m_ = (M), n_ = (N), inc_ = (IncA);                        \
          FTN_GERV2D(&BLACSCntxt, &n_, &m_, A, &inc_, &aRSRC_, &aCSRC_);\
        } 

#undef RecvMat
static inline 
void RecvMat (int RI, int RJ, int m, int n, FDATA A[], int IncA, int MsgType)
{ 
  if (m==0 || n==0) 
    return;
  NORMALIZE(RI, RJ);
  RECV_(RI, RJ, m, n, A, IncA, MsgType);  	
} /* RecvMat() */
#endif 

#undef NORMALIZE
#undef REL2ABSCID

void SumMat STDC_PARAMS( 
	(int RowCol, int P0, int P, int m, int n, pLocMat A)
);
/* pre:  RCOK(RowCol) LM_IncOKUnit(m, n, A) && LM_CInc(A)==1 || P0==0
 *	 all (P) cells in scope of RowCol are participating in this call; 
 *	 starting from some cell I in the scope, P0=(0,1,..,P-1)
 *	 m, n must be consistent across cells in scope;
 * post: the sum of A[0..m-1, 0..n-1] across all cells is left in cell I;
 *	 A[0..m-1, 0..n-1] in other cells in scope of RowCol may be overwritten
 *	 by partial sums. 
 * notes:a multi-block version is not required here as for multiple blocks,
 *	 ring summnation is superior.
 */

#ifndef BLACS
void BrdMat STDC_PARAMS(
	(int RowCol, int P0, int m, int n, FDATA A[], int IncA)
);
/* pre:  RCOK(RowCol) && RIncOK(n, Inc), Inc(A) = (IncA, 1)
 *	 all cells in scope of RowCol are participating in this call; 
 *	 P0 = 0 in only one cell, cell I; 
 *	 m & n must be consistent across cells in scope;
 * post: A[0..m-1, 0..n-1] is b/c from cell I to all other cells in the 
 *	 scope if RowCol. 
 */
#else
/* SRCABSCID(): 0 <= P0 < NumCells[RowCol] */
#define SRCABSCID(RowCol, P0)                                           \
        CYCLIC_SUB(AbsCellId[RowCol], P0, NumCells[RowCol])              
# define BRD_(RowCol, SRCP0, M, N, A, IncA)                           	\
        { int m_ = (M), n_ = (N), inc_ = (IncA);                        \
          if (SRCP0==0)                                                 \
            FTN_GEBS2D(&BLACSCntxt, C2F_CHAR(&RC_CHAR[RowCol]), 	\
                         C2F_CHAR(&BrdTopChar[RowCol]), &n_, &m_, A, &inc_);\
           else {                                                         \
            int aRSRC_ = (RowCol)==0? SRCABSCID(0, SRCP0): AbsCellId[0],  \
                aCSRC_ = (RowCol)==1? SRCABSCID(1, SRCP0): AbsCellId[1];  \
            FTN_GEBR2D(&BLACSCntxt, C2F_CHAR(&RC_CHAR[RowCol]), 	\
            		 C2F_CHAR(&BrdTopChar[RowCol]),			\
                         &n_, &m_, A, &inc_, &aRSRC_, &aCSRC_);         \
           }								\
        }

#undef BrdMat
static inline 
void BrdMat (int RowCol, int P0, int m, int n, FDATA A[], int IncA)
{
  if (VNumCells[RowCol]==1 || m==0 || n==0) 
    return;
  BRD_(RowCol, P0, m, n, A, IncA);	
} /* BrdMat() */
#endif

/******************* [1] block matrix routines ***************************/
	       
void BrdBlkMat STDC_PARAMS(
	(int RowCol, int P0, int m, int B0, int B, int BInc, int n, 
	 FDATA A[], int IncA)
);
/* pre:  RCOK(RowCol) && (m+B0 <= B || BInc==1 || IncA==N) && RIncOK(n, Inc)
 *	 && BlkOK(B0, B, BInc), Inc(A) = (IncA, 1)
 *	 all cells in scope of RowCol are participating in this call; 
 *	 P0 = 0 in only one cell, cell I; 
 *	 if (m+B0<=B || BInc==1) m, n must be consistent across cells in scope;
 *	 otherwise IncA==n and  m, B0, B, BInc, n must be consistent.
 * post: A[m::B0:B:BInc, 0..n-1] is b/c from cell I to all other cells in the 
 *	 scope if RowCol. 
 */

void SendBlkMat STDC_PARAMS(
        (int RI, int RJ, int m, int B0, int B, int BInc, int n,
         FDATA A[], int MsgType)
);
/* pre:  RIJOK(RI, RJ) && BlkOK(B0, B, BInc); Inc(A) = (n, 1)
 *	 if (m+B0<=B || BInc==1) m,n must be consistent for sender & receiver;
 *	 otherwise m, B0, B, BInc, n must be consistent.
 * post: A[m::B0:B:BInc, 0..n-1] is sent from this cell to relative cell (RI, RJ)
 */
  
void RecvBlkMat STDC_PARAMS(
        (int RI, int RJ, int Add, int m, int B0, int B, int BInc, 
         int n, FDATA *alpha, FDATA A[], int MsgType)
);
/* pre:  RIJOK(RI, RJ) && BlkOK(B0, B, BInc); Inc(A) = (n, 1)
 *	 if (m+B0<=B || BInc==1) m,n must be consistent for sender & receiver;
 *	 otherwise m, B0, B, BInc, n must be consistent.
 * post: message A'[m::B0:B:BInc, 0..n-1] is received from relative cell (RI,RJ)
 *       with tag MsgType; 
 *       A[m::B0:B:BInc, 0..n-1] = (Add? A[m::B0:B:BInc, 0..n-1]: 0) + 
 *	 			   (Add? *alpha: 1.0)*A'[m::B0:B:BInc, 0..n-1] 
 */

/*********************** 2-dim block matrix routines ************************/
	       
void Send2BlkMat STDC_PARAMS(
        (int RI, int RJ, int M, int N, FDATA *alpha, pLocMat A,
	 int RB0, int RB, int RBInc, int CB0, int CB, int CBInc)
);
/* pre:	 RIJOK(RI, RJ) 
 * post: M x N B = {Bx, N, 1} is sent to cell of relative co-ords (RI, RJ)
 *       where B is established by BlkMatStridePack(0, M, N, A, alpha, B,
 *					  RB0, RB, RBInc, CB0, CB, CBInc)
 */  

void Recv2BlkMat STDC_PARAMS(
        (int RI, int RJ, int Add, int M, int N, pLocMat B,
	 int RB0, int RB, int RBInc, int CB0, int CB, int CBInc)
);
/* pre:	 RIJOK(RI, RJ);
 *	 M x N A = {Ax, N, 1} is received from cell of relative co-ords (RI, RJ)
 * post: as for BlkMatStrideUnpack(Add, M, N, A, &FD_1, B, 
 *				   RB0, RB, RBInc, CB0, CB, CBInc); 
 */  

void RecvBrd2BlkMat STDC_PARAMS(
        (int RowCol, int P0, int M, int N, pLocMat B,
	 int RB0, int RB, int RBInc, int CB0, int CB, int CBInc)
);
/* pre:	 RCOK(RowCol); P0 != 0
 *	 M x N A = {Ax, N, 1} is received from a cell of relative address -P0 
 *	 in scope of RowCol
 * post: as for BlkMatStrideUnpack(0, M, N, A, &FD_1, B, 
 *				   RB0, RB, RBInc, CB0, CB, CBInc); 
 */  

#if defined(GLOB_BLK_SEND)
void SendWaitMark STDC_PARAMS( (void) );
/* pre:  GridInit() has been called; there are no outstandings sends
 *	 from last call to SendWaitMark().
 * post: future send operations will be initiated in non-blocking mode
 *	 until pending sends are completed by call(s) to CompleteSends().
 */
  
void CompleteSends STDC_PARAMS( (int NSends) );
/* pre:  SendWaitMark() has been called exactly once since last call to 
 *	 GridInit(), or after all pending sends were completed by a
 *	 previous call to CompleteSends();
 *	 There are at NPending <= NCel outstanding send calls.
 *	 The buffers associated with each send call have not been changed
 *	 since the send.
 *	  
 * post: wait until NSends' sends have been completed, and release any 
 *	 associated resources, where NSends' = NPending, if NSends<0 or
 *	 NSends > NPending, and NSends' = NSends, otherwise.
 * note: a typical usage is  SendWaitMark(); <commun pattern>; ... 
 *	 CompleteSends(-1);, where <commun pattern> might for example be
 *	 a `ring-combine' or transpose operation (that satisifies
 *	 the above precondition); note that no other communication may
 *	 occur in this time. This can be used to prevent either
 *	 excessive message buffering or deadlock.
 */
#endif

#endif /*__FDCommun_INCL__*/
