/************************************************************************/
/*              FDMalloc: FDATA workspace memory allocation           */
/************************************************************************/
/* SCCS INFO @(#)FDMalloc.c 1.4         last modified P. Strazdins, 00/08/01
   first created: P. Strazdins 97/02/23 
*/
/* compilation: DEBUG_MALLOC can be set to test for memory leaks,
 *		buffer overwrite, and freeing of invalid pointers.
 */
  
/*#define DEBUG */
#include <stdlib.h>			/* memalign(), free() */
#define memalign(align, size) 	malloc(size)
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDMalloc.h"
#include "FDAssert.h"
#include "MachineParams.h"

#ifdef ASSERT
#define DEBUG_MALLOC
#endif

#ifdef DEBUG_MALLOC
#define SQRT_RSIZE

#define MAX_NODES 16

typedef struct malloc_node {
  FDATA *val; int size; int rsize; int call;   	
} mallocinfo;

static int mcalls=0;

static int mlen=0, fhd=0, ftl=0;
static mallocinfo mnode[MAX_NODES], fnode[MAX_NODES]; 

#define MIN_RAMPSIZE 8
#define RAMP_INIT 1.5
#define RAMP_INC 1.0

static int getRampSize(size)
{ int rsize=0; int size_=1, lgsize=0;
  while (size_ <= size) 
    size_ *= 2, lgsize++;
#ifdef SQRT_RSIZE
  lgsize/=2;
#else
  lgsize-=3;
#endif 
  size_=1;
  while (lgsize > 0)
    size_ *=2, lgsize--;       
  return(MAX(MIN_RAMPSIZE, size_));  		
}

static void SetRamp(A, n)
 FDATA *A; int n;
{ int i; int di = n>0? 1: -1;
  FD_BASE v = RAMP_INIT;  
  for (i=0; i!=n; i+=di) {
    FD_SET_REAL(A[i], v);
    FD_SET_IMAG(A[i], v);
    v+= RAMP_INC;
  }  
} /* SetRamp() */

static int ChkRamp(A, n, fsterr)
 FDATA *A; int n; int *fsterr;
{ int i; int di= n>0? 1: -1; int nerr=0;
  int an = ABS(n);
  FD_BASE v = RAMP_INIT;
#ifdef QUICK_RAMP_CHK			/* actually appears to be slower */  
  FD_BASE ChkSum = an*RAMP_INIT + an*(an+1)/2*RAMP_INC, Sum=0; 
#endif  
  *fsterr = -1;  
#ifdef QUICK_RAMP_CHK
  for (i=0; i!=n; i+=di) 
    Sum+=FD_REAL(A[i]) + FD_IMAG(A[i]);
  if (FDATATYPE >= COMPLEX)
    ChkSum *= 2.0; 			 
  if (Sum==ChkSum)      	
    return(0);
#endif    
  for (i=0; i!=n; i+=di) {
    if ((FD_REAL(A[i])!=v) || (FDATATYPE>=COMPLEX && FD_IMAG(A[i])!=v)) {
      nerr++;
      if (*fsterr==-1)
        *fsterr=i;
    }
    v += RAMP_INC;
  }
  return(nerr);
} /* ChkRamp() */

void DumpMalloc() {
  int i;
  for (i=0; i<mlen;i++)
    printf(" %7d", FD_ADR(mnode[i].val));
  printf("\n");      
  for (i=0; i<mlen;i++)
    printf(" %7d", mnode[i].size);
  printf("\n");      
  for (i=0; i<mlen;i++)
    printf(" %7d", mnode[i].rsize);
  printf("\n");      
  for (i=0; i<mlen;i++)
    printf(" %7d", mnode[i].call);
  printf("\n");      
}	
#else
void DumpMalloc() {
}	
#endif


/* Motivation: 
 * W is an Abstract Data Object to enable allocation of
 * small matrices or `blocks' (eg. Size<=FastWSSize, or possibly Size<SizeW)
 * with minimal software overheads. Larger blocks will be allocated
 * via the standard malloc library (and with greater overheads), and will be
 * outside W.
 *
 * It assumes a `stack-like' pattern of MatMalloc()/MatFree() calls
 * for this efficiency (other patterns will result in greater overheads
 * and/or poor utilization of W). However, if the bottom block is freed,
 * a pointer is kept to the next unfreed block.  
 * Such patterns occurs within level 1 and 2 DBLAS operations.
 *
 * Implementation:
 * A block of size S is allocated in W from words Wa..Wa+S+1;
 * the last word (the `size field') contains the value of S. 
 * This is sufficient to track backwards all blocks in W from Wa. 
 * Thus, only Wa and the word containing the size in W need be changed
 * upon a memory allocation or free.
 *	nb. this scheme assumes sizeof(int) <= sizeof(FDATA)
 *
 * Defensive Programming:
 * 2 types of error can occur when using dynamic memory allocation
 * in an application:
 *	1. blocks are over-written, resulting in a corrupted W
 *	   (ie. the size fileds becomes corrupted)
 *	2. a bad pointer is supplied with MatFree().
 * Using standard malloc library, both kinds of error can cause highly
 * unpredicatable results, and can be extremely difficult to find.
 * In this module, both types of error can be detected within MatFree()
 * provided the block was allocated from within W.
 */
 	
static struct {
  FDATA * W0,			/* start of W */
  	* W0_u,			/* bottom of unfreed space in W */
  	* W1,			/* end of W */
	* Wa;			/* points to next available space in W */
  int 	Size,			/* size of W */
  	FastWSSize;		/* max size of blocks in W */
} W = { 
  NULL, NULL, NULL, NULL, 4*FastWSSize_DEF, FastWSSize_DEF 
};		
/* invariants: 	W0 <= W0_u <= Wa <= W1; 0 <= FastWSSize <= Size;
 *		if W0!=NULL, FastWSSize=W1-W0
 */
 


void MatMallocInit (ParamType, ParamVal) 
  int ParamType, ParamVal;
{ 
#ifndef DEBUG_MALLOC
  if (ParamType==0) {
    if (W.W0==W.Wa) {		/* W is not currently in use */
      MatFreeBuf();
      if (ParamVal >= 0)
        W.Size = ParamVal;
    }
  } else {
    if (ParamVal>=0 && ParamVal<=W.Size)
      W.FastWSSize = ParamVal;
  }
#endif  
} /* MatMallocInit() */

void MatFreeBuf()
{ 
#ifndef DEBUG_MALLOC
  if (W.W0!=NULL) {		/* free W for latter allocation */ 
    free ((void *) W.W0);
    W.W0 = W.W0_u = W.W1 = W.Wa = NULL;
  }	
#endif  
} /* MatFreeBuf() */


int MatMalloc_f(Where, CrucialAlloc, A, Size, UseFastMalloc, ImmedFree) 
  char *Where; FDATA **A; int CrucialAlloc, Size, UseFastMalloc, ImmedFree;
{ FDATA *W1 = W.W1;			FDATA *Wa;
  int FastWSSize;
  if (Size <= 0) {
    *A = NULL;
    return 0;
  }    	
  
#ifndef DEBUG_MALLOC 
  if (W1==NULL && UseFastMalloc==FD_ALLOC_TRUE) {
    if (MatMalloc("MatMalloc_f", 0, &W.W0, W.Size)) {
      W.Wa = W.W0_u = W.W0;		W1 = W.W1 = W.W0 + W.Size;
    }  
  }  

  Wa = W.Wa;				FastWSSize = W.FastWSSize;
  if ((Size+1 <= ((W1-Wa)&UseFastMalloc)) && 
      ((Size&ImmedFree) <= FastWSSize)) {	/* use W for fast malloc */
     *A = Wa;
      Wa += Size+1;
      *((int *) (Wa-1)) = Size+1; 
      W.Wa = Wa;
  } else {
    *A = (FDATA *) memalign(MAX(sizeof(FDATA), CACHE_LINE_SIZE), 
    			    Size*sizeof(FDATA));
  }


# ifdef DEBUG
printf("%d.%d: MatMalloc(%s,%d,a%d,%d,%d,%d)\n", 
	getcidy(), getcidx(), Where, CrucialAlloc, FD_ADR(*A), Size, UseFastMalloc, ImmedFree);
# endif

#else
  {int rsize=getRampSize(Size);
  *A = (FDATA *) memalign(MAX(sizeof(FDATA), CACHE_LINE_SIZE),
                          (Size+2*rsize)*sizeof(FDATA));
  mcalls++;                        
  if (*A!=NULL) {
    if (mlen>=MAX_NODES) {	
      DumpMalloc();
      TRACE_VALS_((TraceStr, "(%s,%d,a%d,%d,%d,%d) call %d", Where, CrucialAlloc,
                   FD_ADR(*A), Size, UseFastMalloc, ImmedFree, mcalls));
      CommonAssert(0, 0, "MatMalloc_f(mem. leak?)", 4, TRACE_MALLOC); 
    } 
    (*A) += rsize;
    SetRamp((*A)-1, -rsize);		
    SetRamp((*A)+Size, rsize);     
    mnode[mlen].val = *A, mnode[mlen].size = Size, 
    mnode[mlen].rsize = rsize, mnode[mlen].call = mcalls;
    mlen++;
  }
  }  
#endif

  COMMON_ASSERT_TR(
    (TraceStr, "(%s,%d,a%d,%d,%d,%d)", Where, CrucialAlloc,
                 FD_ADR(*A), Size, UseFastMalloc, ImmedFree),
    (UseFastMalloc==FD_ALLOC_FALSE ||  UseFastMalloc==FD_ALLOC_TRUE) &&
    (ImmedFree==FD_ALLOC_FALSE ||  ImmedFree==FD_ALLOC_TRUE), 0,
    "MatMalloc_f", 4, TRACE_MALLOC
  );
    
  if (*A==NULL && CrucialAlloc) {
    CommonAssert(1, Size, "MatMalloc_f", 4, TRACE_MALLOC);
  } 	

  return (*A != NULL);  
} /* MatMalloc() */

void MatFree(A) 
  FDATA *A;
{ 
  COMMON_ASSERT_TR(
    (TraceStr, "(a%d)", FD_ADR(A)), 1, 0, "MatFree", 4, TRACE_MALLOC
  );

#ifndef DEBUG_MALLOC 
   
#ifdef DEBUG
printf("%d.%d: MatFree(a%d)\n", getcidy(), getcidx(), FD_ADR(A)); 
#endif
  if (A != NULL) { 
    FDATA *W0_u = W.W0_u; 		 FDATA *Wa_ = W.Wa;
    int HeapSize = Wa_ - W0_u;
    if (W.W0_u<=A && A<W.W1) {		/* A is within (unfreed blocks of) W */
      int Size = *((int *) (Wa_-1));
      Wa_ = Wa_ - Size;
      if (A==Wa_) {			/* A is top block in W */
        if (Wa_ == W0_u)		/* A is also only unfreed block in W */ 
          W.Wa = W.W0_u = W.W0;
        else  
          W.Wa = Wa_;  
#ifdef DEBUG
printf("%d.%d: Wa_=a%d W0_u=a%d\n", getcidy(), getcidx(), FD_ADR(W.Wa), FD_ADR(W.W0_u));         
#endif
      } else {  	
        FDATA *Wa_l; 
        do {				/* search back from Wa_ for A */
#ifdef DEBUG        
printf("%d.%d: HeapSize=%d, Size=%d, Wa_=a%d W0_u=a%d\n", getcidy(), getcidx(), 
	HeapSize, Size, FD_ADR(Wa_), FD_ADR(W0_u));    
#endif 
	  if (Size<=1 || Size>HeapSize)
            CommonAssert(0, 0, "MatFree: corrupted heap", 4, TRACE_MALLOC);
          if (A > Wa_)
            CommonAssert(0, 0, "MatFree: bad pointer", 4, TRACE_MALLOC);
          if (A == Wa_) break;   
          Wa_l = Wa_;
          HeapSize -= Size;
          Size = *((int *) (Wa_-1));   
          Wa_ -= Size;
        } while (1);
        if (A == W0_u)			/* A is bottom unfreed block in W */
          W.W0_u += Size;
        else {  			/*merge block at A with one below*/
          Size += *((int *) (Wa_-1));
          *((int *) (Wa_l-1)) = Size; 
        }
#ifdef DEBUG
printf("%d.%d: W0_u=a%d *(a%d-1)=%d\n", getcidy(), getcidx(), FD_ADR(W.W0_u), FD_ADR(Wa_l), Size);
#endif
      }
    } else if (W.W0<=A && A<W.W0_u)
      CommonAssert(0, 0, "MatFree: bad pointer", 4, TRACE_MALLOC);      
    else				/* A is outside W */
      free ((char *) A);
  }            
#else
  { int i, j, rsize, Size, chk, nerrs;

  if (A==NULL) return;
  i = mlen-1;  
  while (i > 0 && mnode[i].val!=A) i--;
  if (mnode[i].val!=A) {
     i = ftl;
     while (i!=fhd && fnode[i].val!=A) {
       i--; if (i<0) i=MAX_NODES-1;
     } 
     DumpMalloc();
     if (i!=fhd) 
       printf("MatFree: tried to re-free pointer of size %d at call %d\n",
              fnode[i].size, fnode[i].call);
     TRACE_VALS_((TraceStr, "(a%d) now malloc call %d", FD_ADR(A), mcalls)); 
     CommonAssert(0, 0, "MatFree(bad ptr)", 4, TRACE_MALLOC);
  }

  Size = mnode[i].size; 
  rsize = mnode[i].rsize;
  ftl = (ftl+1) % MAX_NODES;  		
  if (ftl==fhd) fhd = (fhd+1) % MAX_NODES;
  fnode[ftl] = mnode[i];
  for (j=i; j<mlen-1; j++)
    mnode[j] = mnode[j+1];
  mlen--;
 
  nerrs = ChkRamp(A+Size, rsize, &chk);
  if (nerrs) {
    DumpMalloc();	
    TRACE_VALS_((TraceStr, "(a%d) call %d size %d, 1st elt %d, #elts=%d", 
       		 FD_ADR(A), fnode[ftl].call, Size, chk, nerrs)); 
    CommonAssert(0, 0, "MatFree(post-overwrite)", 4, TRACE_MALLOC);
  }
  nerrs = ChkRamp(A-1, -rsize, &chk);
  if (nerrs) {
    DumpMalloc();
    TRACE_VALS_((TraceStr, "(a%d) call %d size %d, 1st elt %d, #elts=%d", 
       		 FD_ADR(A), fnode[ftl].call, Size, chk, nerrs)); 
    CommonAssert(0, 0, "MatFree(pre-overwrite)", 4, TRACE_MALLOC);
  }
  
  A -= rsize;

  free ((char *) A);
  }  
#endif  
} /* MatFree() */

