/************************************************************************/
/* GenMult: DBLAS General Matrix Multiplication procedures		*/
/************************************************************************/
/* SCCS INFO @(#)GenMult.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 93/15/01
*/

#include "GenMult.h"

#include "BLASOptions.h"
#include "FDATAOps.h"
#include "FDAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "MachineParams.h"
#include "CellMatrix.h"
#include "SpreadMatrix.h"


#define CDM DM_DM(C)
#define CDN DM_DN(C)
#define ADM DM_DM(A)
#define ADK DM_DN(A)
#define BDK DM_DM(B)
#define BDN DM_DN(B)  

#define lA DM_Loc(A)
#define lB DM_Loc(B)
#define lC DM_Loc(C)

void DistRankKMult(M, N, K, alpha, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta; 
  DistMat A, B, C; 
{ int m, n;				/* local sizes of C */
  int k, MAXdK; 			/* max size of spread across K-dim */
  DistMatStore AK, BK;
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%dx%d,%s,%d a%d %d %d,%d a%d %d %d,%s,%d a%d %d %d)",
     M, N, K, pFD2Str(alpha), DM_Trans(A), FD_ADR(LM_ADR(lA)), LM_RInc(lA), LM_CInc(lA),
     DM_Trans(B), FD_ADR(LM_ADR(lB)), LM_RInc(lB), LM_CInc(lB), pFD2Str(beta),
     DM_Trans(C), FD_ADR(LM_ADR(lC)), LM_RInc(lC), LM_CInc(lC)),
    M>=0 && N>=0 && K>=0, 0, "DistRankKMult", 1, TRACE_ALL
  );

  if (DM_Trans(C)) {
    SWAP(int, M, N);	
    DM_Transpose(A);		DM_Transpose(B);	DM_Transpose(C);
    SWAP(DistMat, A, B);
  }

  if (!pFD_UNITY(beta)) {
    m = G2Ld(M, DM_D0(C)); 		n = G2Ld(N, DM_D1(C));
    MatScale(m, n, beta, &lC);
  }

  if (K==0 || (FD_EQ(*alpha, FD_0)) || M==0 || N==0) {
    TRACE_RETURN;  
  }
  
  MAXdK=MIN(PAR_INNER_VECSIZE, K);

  k=0;  
  while (k < K) { 			/* C += alpha A[:,k..+dK] B[k..+dK,:] */
    FDATA ScAlpha = *alpha;
    int dK = MIN(MAXdK, K-k);
    LocMat AsK, BsK;			/* spread workspaces of A, B */
    int AsKn, BsKn;

    LM_ADR(AsK) = LM_ADR(BsK) = NULL; 
    if (M<N)				/* scale smaller matrix 1st, if can */
      AsKn = SpreadMat(1, dK, M, A, &ScAlpha, CDM, &AsK, 0, N, CDN);
    BsKn = SpreadMat(0, dK, N, B, &ScAlpha, CDN, &BsK, 0, M, CDM);    
    if (M>=N)
      AsKn = SpreadMat(1, dK, M, A, &ScAlpha, CDM, &AsK, 0, N, CDN);

    m = ABS(AsKn);			n = ABS(BsKn);
#if defined(BLAS) || defined(BLAS_MM)
    if (dK==1)
      R1Mult(m, n, &ScAlpha, &AsK, &BsK, &lC);      
    else
#endif        
    MatPartMult(m, n, dK, &ScAlpha, &AsK, &BsK, NULL, &lC);

    if (BsKn>0)  
      MatFree(LM_ADR(BsK));  
    if (AsKn>0) 
      MatFree(LM_ADR(AsK));
    if (PrintRes)
      MatPrint("C", MIN(PrintRes, m), MIN(PrintRes, n), lC); 	
    k += dK;
    if (k < K) {
      A = SubMat_(&AK, A, 0, dK);		B = SubMat_(&BK, B, dK, 0);
    }  	     	     
  } /* while (k...) */

  TRACE_EXIT;
} /*DistRankKMult*/


void DistVecMatMult(M, N, K, alpha, A, B, beta, C)
  int M, N, K;
  FDATA *alpha, *beta; DistMat A, B, C; 
{ int m, MAXdM;                         /* max size of spread across M-dim */
  int RowCol = DM_Trans(B);
  DistMatStore AM, CM;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%dx%d,%s,%d a%d %d %d,%d a%d %d %d,%s,%d a%d %d %d)",
     M, N, K, pFD2Str(alpha), DM_Trans(A), FD_ADR(LM_ADR(lA)), LM_RInc(lA), LM_CInc(lA),
     DM_Trans(B), FD_ADR(LM_ADR(lB)), LM_RInc(lB), LM_CInc(lB), pFD2Str(beta),
     DM_Trans(C), FD_ADR(LM_ADR(lC)), LM_RInc(lC), LM_CInc(lC)),
    M>=0 && N>=0 && K>=0, 0, "DistVecMatMult", 1, TRACE_ALL
  ); 
 
  if (M==0 || N==0) {
    TRACE_RETURN;
  }
  if (K==0 || FD_EQ(*alpha, FD_0)) {
    LocMat lCt;				LM_Copy(DM_Trans(C), lCt, lC);	
    MatScale(G2Ld(M, CDM), G2Ld(N, CDN), beta, &lCt);
    TRACE_RETURN;
  }

  MAXdM=MIN(PAR_OUTER_VECSIZE, M);
  m=0;
  while (m < M) {
    int nB /*= G2Ld(N, BDN)*/, 		kB /*= G2Ld(K, BDK)*/;
      FDATA ScAlpha = *alpha;
    int dM = MIN(MAXdM, M-m);
    DistMat Ct = RowCol==1? Trans(C): C, At = RowCol==0? Trans(A): A;	
    					 /* Ct is M x N, At is K x M */
    LocMat AsM, CsM, lBt;                /* spread workspaces of A, C */
    int AsMn, CsMn;

    LM_ADR(AsM) = NULL;
    AsMn = SpreadMat(1-RowCol, dM, K, At, &ScAlpha, BDK, &AsM, 0, N, BDN);
    kB = ABS(AsMn);			
    LM_Trans(1-RowCol, AsM);		/* ensure AsM is dM x kB */

    LM_ADR(CsM) = NULL;
    CsMn = InitReduceWS(RowCol, dM, N, Ct, beta, BDN, &CsM, 0, K, BDK);    
    nB = ABS(CsMn);
    LM_Trans(RowCol, CsM)		/* ensure CsM is dM x nB */
    LM_Copy(RowCol, lBt, lB);		/* ensure lBt is kB x nB */
#if defined(BLAS) || defined(BLAS_MM)
    if (dM==1)
      VMMult(nB, kB, &ScAlpha, &AsM, &lBt, NULL, &CsM);      
    else
#endif        
    MatPartMult(dM, nB, kB, &ScAlpha, &AsM, &lBt, NULL, &CsM); 

    LM_Trans(RowCol, CsM)		/* restore CsM */
    ReduceMat(RowCol, dM, N, Ct, beta, BDN, &CsM, K, BDK);
    if (CsMn>0)
      MatFree(LM_ADR(CsM));
    if (AsMn>0)
      MatFree(LM_ADR(AsM));
      
    m += dM;
    if (m < M) {
      C = SubMat_(&CM, C, dM, 0);	A = SubMat_(&AM, A, dM, 0);
    }  
  } /* while (m...) */
  TRACE_EXIT;
} /*DistVecMatMult()*/
