/************************************************************************/
/* GenMult: DBLAS General Matrix Multiplication procedures		*/
/************************************************************************/
/* SCCS INFO @(#)GenMult.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 93/15/01
*/
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *              if macro TRACE defined, parameter value tracing enabled
 *		if macro PRINTMAT define, matrices are printed for debugging
 */
#ifndef __GenMult_INCL__
#define __GenMult_INCL__

#include "FDATASetUp.h"
#include "DistMat.h"

#if FDATATYPE==SINGLE
#define DistGenMult	DSDistGenMult
#define DistRankKMult	DSDistRankKMult
#define DistVecMatMult	DSDistVecMatMult
#elif FDATATYPE==DOUBLE
#define DistGenMult	DDDistGenMult
#define DistRankKMult	DDDistRankKMult
#define DistVecMatMult	DDDistVecMatMult
#elif FDATATYPE==COMPLEX
#define DistGenMult	DCDistGenMult
#define DistRankKMult	DCDistRankKMult
#define DistVecMatMult	DCDistVecMatMult
#else
#define DistGenMult	DZDistGenMult
#define DistRankKMult	DZDistRankKMult
#define DistVecMatMult	DZDistVecMatMult
#endif

void DistGenMult STDC_PARAMS( 
	(int M, int N, int K, FDATA *alpha, DistMat A, DistMat B, 
	 FDATA *beta, DistMat C)
);
/* pre:  all cells are executing DistGenMult() with 
 *	 M, N, K, *alpha, *beta,  FormC parameters and (D[0..1], Trans) 
 * 	 fields for A, B & C consistent across all cells;
 *	 M, N, K >= 0; 
 *	 if FormC != Rectangular, C is a triangular matrix of form FormC;
 * post: performs global operation  C = (*beta)*C + (*alpha)*AB,
 * notes:chooses between DistRankKMult() & the 2 variants of DistVecMatMult();
 *	 this corresponds to spreading across the dimension K, M, or N 
 *	 on the basis of (in priority): 
 *		1) block incompatability: spread across the (at most 1)
 *		   dimension that is not block aligned; this includes a dim.
 *		   of A or B that has been (non-trivially) `pre-spread'.
 *		   This is because the spread will hide the misalignment.
 *		2) the greatest load imbalance, as spreading across this
 *		   dimension will spread the imbalance also.
 *		3) the smallest (parallel) communication volume.
 *		   If an explicit transpose is required with a spread operation,
 *		   this (rather crudely) is taken to effectively double the
 *		   volume of that spread. 
 *		   Currently, rotation requirements for spreads are ignored
 *		   (as generally smaller than the spread proper).
 *		   Also, the `weighting' of all spread communication, whether 
 *		   a broadcast or a summnation, is taken to be unity;
 *		   this is a simplification, as these can be different or
 *		   vary with matrix sizes.	
 */


void DistRankKMult STDC_PARAMS( 
	(int M, int N, int K, FDATA *alpha, DistMat A, DistMat B, 
	 FDATA *beta, DistMat C)
);
/* pre:  as for DistGenMult();
 * post: as for DistGenMult() 
 * notes:efficient for K << M, N.
 */

void DistVecMatMult STDC_PARAMS( 
	(int M, int N, int K, FDATA *alpha, DistMat A, DistMat B, 
	 FDATA *beta, DistMat C)
);
/* pre:  as for DistGenMult();
 * post: as for DistGenMult() 
 * notes:efficient for M << K, N; 
 * 	 !DM_Trans(B) is the `normal' variant for this compuation 
 *	 (vector-matrix multiply). To achieve a matrix-vector multiply, 
 *	 transpose all DistMat parameters, and interchange A,B and M,N.
 */

#endif /*__GenMult_INCL__*/

