/************************************************************************/
/*                     	IntOps.h		                        */
/************************************************************************/
/* SCCS INFO @(#)IntOps.h 1.1         last modified P. Strazdins, 98/12/24
   first created: P. Strazdins 97/01
*/
/* header file for integer operations; this is motivated by performance
 * issues, under the assumptions that integer multiply, divide and remainder
 * operations are relatively expensive.
 * In certain instances, they can be computed by 1-2 comparisons and
 * integer add/sub or logical instructions.
 * 
 * However, on architectures with very fast implementations of these operations,
 * these can be replaced by normal definitions
 *
 * Also contains a GCD function
 */
 
#ifndef __IntOps_INCL__
#define __IntOps_INCL__

/* for when integer multiply/divide is expensive (eg. SPARC V7), and 
   there is a fair chance one of the arguments is unity */
#define FAST_MUL(a, b)	((a)==1? (b): (b)==1? (a): (a)*(b))
#define FAST_MUL1(a, b)	((a)==1? (b): (a)*(b))
#define FAST_MUL2(a, b)	((b)==1? (a): (a)*(b))
#define FAST_DVD(a, b)	((b)==1? (a): (a)/(b))
#define FAST_MOD(a, b)	((b)==1? 0: (a)%(b))

#define FAST_MULZ(a, b)	((a)==0? 0: (a)*(b))
#define FAST_DVDEQ(a, b)((a)==(b)? 1: (a)/(b))
#define FAST_DIVLT(a, b)((a)<(b)?  0: (a)/(b))

#define IS_P2(b)	(((b)&((b)-1))==0)	
#define MOD_P2(a, b)	((a)&((b)-1))
#define FAST_MODEQ(a, b)((a)==(b)? 0: (a)%(b))
#define FAST_MODLT(a, b)((a)<(b)? (a): (a)%(b))
#define FAST_MODP2(a, b) (IS_P2(b)? MOD_P2(a, b): (a)%(b))

/* for the following, assumes that 0 <= i,n < p */
#define CYCLIC_DEC(i, p)        CYCLIC_SUB(i, 1, p)
#define CYCLIC_SUB(i, n, p)     (((i)>=(n))? ((i)-(n)): ((p)+(i)-(n)))
				/* = (((P)+(i)-(n))%(P)) */

#define CYCLIC_INC(i, p)        CYCLIC_ADD(i, 1, p)
#define CYCLIC_ADD(i, n, p)     (((i)+(n)<(p))? ((i)+(n)): ((i)+(n)-(p)))
				/* = (((P)+(i)+(n))%(P)) */

/* DIV_MOD() is a combined division-remainder decomposition.
 * this is a potenitially expensive operation, which may lend itself to 
 * machine- specific optimizations.
 */   
/* pre:  a, b > 0. 
 * post: q*b + r = a, q>=0, 0<=q<r 
 * note: q may safely on a
 */
#ifdef COMPILER_COMB_DIVMOD	/* compiler can optimize DIV-MOD */
# define DIV_MOD(a, b, q, r)						\
	{ int a_ = (a); (q) = a_ / (b); (r) = a_ % (b); }
#else				/* replace % by * ; gnerally less expensive */ 
# define DIV_MOD(a, b, q, r)						\
	{ int a_ = (a); (q) = a_ / (b); (r) = a_ - (q)*(b); }
#endif 
 
	

/********************* exported symbols  ***********************************/

#define GCD             GGCD

int GCD STDC_PARAMS( ( int a, int b) );
/* pre:  a, b > 0
 * post: returns greatest common divisor of a & b
 */

#endif /*__IntOps_INCL__*/
