/************************************************************************/
/*	LocMat: definitions for cell matrix data structures		*/
/************************************************************************/
/* SCCS INFO @(#)LocMat.h 1.3         last modified P. Strazdins, 99/07/06
   first created: P. Strazdins 96/01/15
*/
/* this module facilitiates easy local matrix manipulations
 * by generalized element row/column increments. 
 *  
 * Generally, Inc[0] (Inc[1]) will be associated with a row (col) incremement.
 * Thus the address of A[i,j] = LM_ADR(A) + i*LM_RInc(A) + j*LM_CInc(A).  
 *
 * Typically, MIN(Inc[0..1]) = 1; this is required for matrix multiply at least.
 *
 * notes: a problem with the defnition of LocMat is that LM_ADR() field
 *	  is less `stable' than the others, which means it is likely
 *	  to vary over a computation. For this reason  LocMat SubLocMat()
 *	  must be a proper function.
 *  
 *  compilation: macro FDATATYPE must be defined;
 */

#ifndef __LocMat_INCL__

#define __LocMat_INCL__

#include "FDATASetUp.h"

#if FDATATYPE==SINGLE
#define LocSubMat	SLocSubMat
#define MkLocMat	SMkLocMat
#define LocMatTrans	SLocMatTrans
#define MatPrint        SMatPrint
#elif FDATATYPE==DOUBLE
#define LocSubMat	DLocSubMat
#define MkLocMat	DMkLocMat
#define LocMatTrans	DLocMatTrans
#define MatPrint        DMatPrint
#elif FDATATYPE==COMPEX
#define LocSubMat	CLocSubMat
#define MkLocMat	CMkLocMat
#define LocMatTrans	CLocMatTrans
#define MatPrint        CMatPrint
#else
#define LocSubMat	ZLocSubMat
#define MkLocMat	ZMkLocMat
#define LocMatTrans	ZLocMatTrans
#define MatPrint        ZMatPrint
#endif

typedef /*unsigned short*/ int StrideSize; 
typedef StrideSize IPair[2];	/* 2-d integer params, ie. (I,J), (RInc,CInc) */

#define LM_InvalStride(Inc) 	((Inc) > (unsigned)(1<<(sizeof(StrideSize)*8-1)))
				/* slightly conservative, but avoids overflow */
				
typedef struct LocMatStruct {
  FDATA 	*x;		/* points to local element [0,0] in cell */
  IPair		Inc;		/* row/column strides for elements */
} LocMat;

typedef LocMat *pLocMat;

#define IncOK(M, N, RInc, CInc)  /*ensures no 2 elts of matrix map to 1 addr*/ \
        (ABS(RInc) > ((N)-1)*ABS(CInc) || ABS(CInc) > ((M)-1)*ABS(RInc))
         
#define IncOKUnit(M, N, RInc, CInc)  \
        ((MIN(RInc, CInc) == 1) && IncOK(M, N, RInc, CInc))
         
#define LM_IncOK(M, N, A)       IncOK(M, N, LM_RInc(A), LM_CInc(A))
 
#define LM_IncOKUnit(M, N, A)   IncOKUnit(M, N, LM_RInc(A), LM_CInc(A))


#define LM_ADR(M) 	((M).x)
#define LM_RInc(M)	((M).Inc[0])
#define LM_CInc(M)	((M).Inc[1])
#define LM_Inc(t, M)	((M).Inc[t])
#define LM_Ld(M)	(MAX(LM_RInc(M), LM_CInc(M)))
#define LM_Eq(L1, L2)	((L1).x==(L2).x && 		\
			 (L1).Inc[0]==(L2).Inc[0] && (L1).Inc[1]==(L2).Inc[1])
	 
#define LM_Index(M, i, j) 	\
	(LM_ADR(M)+FAST_MUL2(i, LM_RInc(M))+ FAST_MUL2(j, LM_CInc(M)))

#define LM_Init( M, x_, Inc0, Inc1 ) \
	(M).x = x_, (M).Inc[0] = Inc0, (M).Inc[1] = Inc1 

#define LM_Copy( Tr, MT, M ) \
	{ int t=(M).Inc[1-(Tr)]; (MT).x = (M).x; \
	  (MT).Inc[0] = (M).Inc[Tr]; (MT).Inc[1] = t;}

#define LM_Trans( Tr, M ) \
	if (Tr) SWAP(int, (M).Inc[0], (M).Inc[1]);

#define LM_SubMat(A, i, j)	LM_ADR(A) = LM_Index(A, i, j)

LocMat MkLocMat STDC_PARAMS( (FDATA *A, int RInc, int CInc) );
/* post: returns {A, RInc, CInc};
 */

LocMat LocSubMat STDC_PARAMS( (LocMat M, int i, int j) );
/* post: returns {LM_Index(M, i, j), LM_RInc(M), LM_CInc(M)};
 */

LocMat LocMatTrans STDC_PARAMS( (LocMat M) );
/* post: returns {LM_ADR(M), LM_CInc(M), LM_RInc(M)};
 */

#define LocMatTr(Tr, M)		(!Tr? M: LocMatTrans(M))

#define LocSubMatTr( Tr, M, i, j ) LocSubMat(M, (Tr)? (j):(i), (Tr)? (i):(j))  

void MatPrint STDC_PARAMS(
        (char *Name, int M, int N, LocMat A)
 );
/* pre:   LM_IncOK(M, N, A)
 * post:  prints A[0..M-1][0..N-1] in row-major form to stdout
 */

#define MatOverlap( MA, NA, A, MB, NB, B) \
        0
/*        ((MA)>0 && (NA)>0 && (MB)>0 && (NB)>0 &&                      \
        (((LM_ADR(A) < LM_ADR(B)) &&                                    \
          (LM_ADR(A)+((MA)-1)*LM_RInc(A)+((NA)-1)*LM_CInc(A) >= LM_ADR(B)))||\
         ((LM_ADR(B) < LM_ADR(A)) &&                                    \
          (LM_ADR(B)+((MB)-1)*LM_RInc(B)+((NB)-1)*LM_CInc(B) >= LM_ADR(A)))))
*/
/* pre:  MIN(RIncA, CIncA) = MIN(RIncB, CIncB) = 1
 * post: returns TRUE iff A[0..MA-1,0..NA-1] & B[0..MB-1,0..NB-1] overlap
 */

/******************** macros with pLocMat values *******************/ 

#define pLM_ADR(pM) 	LM_ADR(*(pM))
#define pLM_RInc(pM)	LM_RInc(*(pM))
#define pLM_CInc(pM)	LM_CInc(*(pM))
#define pLM_Inc(t, pM)	LM_Inc(t, *(pM))
#define pLM_Ld(pM)	(MAX(pLM_RInc(pM), pLM_CInc(pM)))

#define pLM_IncOK(M, N, pA)  LM_IncOK(M, N, *(pA))
 
#define pLM_IncOKUnit(M, N, pA)   LM_IncOKUnit(M, N, *(pA))

#define pLM_Index(pM, i, j)	LM_Index(*(pM), i, j) 	

#define	pLM_Trans(A)	SWAP(int, pLM_RInc(A), pLM_CInc(A))


#define pLocSubMat(pA, di, dj)		\
	(pLM_ADR(pA) = pLM_Index(pA, di, dj), (pA))

#define pLocSubMat_x(pA, A, i, j) 	\
	(pLM_ADR(pA) = pLM_Index(A, i, j), (pA))

#define pLocSubMat_(pA, A, i, j) 	\
	pMkLocMat(pA, pLM_Index(A, i, j), pLM_RInc(A), pLM_CInc(A))

#define pMkLocMat( pA, Ax, RInc, CInc )	\
	((pA)->x=(Ax), (pA)->Inc[0] = (RInc), (pA)->Inc[1] = (CInc), (pA))

#define pLocSubMatTr(Tr, pA, di, dj)          \
	pLocSubMat(pA, Tr? (dj): (di), r? (di): (dj))
        
#define pLocSubMat_Tr(Tr, pA, A, i, j) 	\
	pLocSubMat_(pA, A, Tr? (dj): (di), r? (di): (dj))
	
#endif /*__LocMat_INCL__*/
