/************************************************************************/
/* SpreadMatrix: distributed matrix spread procedures used by DBLAS	*/
/************************************************************************/
/* SCCS INFO @(#)SpreadMatrix.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 93/15/01
 */
/* high-level level distributed matrix `spread' and `de-spread' operations,
 * which are at the heart of high-level parallel matrix operations.
 *
 * A spread of an K x N  matrix B along dimension K puts in a workspace sB
 * the cell's portion of B in the N-dimension, and the global portion of B
 * in the K-dimension (thus typically K << N). Often, eg. if B is a 
 * block-vector, no workspace is needed, with sB pointing within DM_Loc(B).
 * 
 * A `reduce' into B requires the workspace sB to be initialized
 * along all cells in dimension K (typically from the result of local cell
 * matrix multiply operations). Each cells receives in its part of sB
 * which corresponds to B the sum of those parts of sB in all cells in 
 * dimension K. This then is added to the cell's portion of B.
 * Thus a `reduce' operation can be thought of the complement of the spread,
 * and shares many similarities.
 *
 * compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *		if macro TRACE defined, parameter value tracing enabled 
 *		if macro PRINTMAT defined, printing of results performed
 */
#ifndef __SpreadMatrix_INCL__
#define __SpreadMatrix_INCL__
 
#include "FDATASetUp.h"
#include "LocMat.h"
#include "DistMat.h"

#if FDATATYPE==SINGLE
#define SizeSpreadWS	DSSizeSpreadWS
#define InitSpreadWS	DSInitSpreadWS
#define SpreadMat	DSSpreadMat
#define SizeReduceWS	DSSizeReduceWS
#define ReduceMat	DSReduceMat
#define InitReduceWS	DSInitReduceWS
#elif FDATATYPE==DOUBLE
#define SizeSpreadWS	DDSizeSpreadWS
#define InitSpreadWS	DDInitSpreadWS
#define SpreadMat	DDSpreadMat
#define SizeReduceWS	DDSizeReduceWS
#define ReduceMat	DDReduceMat
#define InitReduceWS	DDInitReduceWS
#elif FDATATYPE==COMPLEX
#define SizeSpreadWS	DCSizeSpreadWS
#define InitSpreadWS	DCInitSpreadWS
#define SpreadMat	DCSpreadMat
#define SizeReduceWS	DCSizeReduceWS
#define ReduceMat	DCReduceMat
#define InitReduceWS	DCInitReduceWS
#else
#define SizeSpreadWS	DZSizeSpreadWS
#define InitSpreadWS	DZInitSpreadWS
#define SpreadMat	DZSpreadMat
#define SizeReduceWS	DZSizeReduceWS
#define ReduceMat	DZReduceMat
#define InitReduceWS	DZInitReduceWS
#endif

/***************** high-level `spread' operations ********************/
     
int SizeSpreadWS STDC_PARAMS(
	(int RowCol, int K, int N, DistMat B, 
	 pMatDim DN, int ForceWS, int M, pMatDim DM)
);
/* pre:	 
 * post: returns the size of the workspace required for 
 *		SpreadMat(RowCol, K, N, B, _, DN, ForceWS, M, DM)
 *       if none is needed, returns -1 (in general, this means that this cell 
 *       is the sole source of the required data for the spread).
 */

int InitSpreadWS STDC_PARAMS(
	(int RowCol, int K, int N, DistMat B, 
	 pMatDim DN, pLocMat sB, int ForceWS, int M, pMatDim DM)
);
/* post: initializes *sB as would the call:
 *		SpreadMat(RowCol, K, N, B, sB, DN, ForceWS, M, DM)
 *	 returns WS*n , where WS = +/- 1 (-1 if no workspace needed, ie
 *	 sB can point within B)
 *	 and n = G2Ld(N, DN) (the local length of sB).
 * note: if WS=1, K*G2Ld(N, DN) will be the size of the workspace.
 *	 the return value can be used to obviate the caller from calculating n
 *	 (which has potentially significant overheads)
 */

int SpreadMat STDC_PARAMS(
	(int RowCol, int K, int N, DistMat B, 
	 FDATA *alpha, pMatDim DN, pLocMat sB, int ForceWS, int M, pMatDim DM)
);
/* pre:  RCOK(RowCol) && CompatBlk(BDN,DN) && K,N,M>=0
 *	 All cells (holding/requiring any data) are participating in this call
 *	 with consistent values of *alpha, RowCol, K, N, B, *DN, ForceWS, M, 
 * 	 & *DM.
 *	 Along processor dimension RowCol specified by DM, 
 *	 the result of the spread is applied to a matrix of size M.
 *	 If LM_ADR(*sB) != NULL and a workspace is required for the spread,
 *	 sufficient space must be allocated at LM_ADR(*sB) for the workspace
 * post: *sB = alpha' * (RowCol==0? B'[0..K-1, :DN] : B'[:DN, 0..K-1])
 *
 *       where the notation :D signifies the selection of a dimension by D.
 *		(ie. select local sub-matrix of cell d_P0(D), 
 * 		     block offset d_B0(D) and block size d_B(B))
 *	 B' = DM_Trans(B)? B^T: B; 
 * 
 *       *alpha * alpha' = *alpha_pre, 	*alpha IN {1.0, *alpha_pre} 
 *	 where alpha' = 1.0 iff source cells for spread did not need a workspace
 *       and !ForceWS (in this case LM_ADR(*sB) points to DM_Loc(A))
 * 
 *	 if LM_ADR(*sB_pre)!=NULL, LM_ADR(*sB) is unchanged and is 
 *	 used for the workspace. if LM_Inc(0..1, *sB_pre) = 0,
 *	 LM_Inc(0..1, *sB) will be set to approproriate values 
 * 
 *	 the cells requiring the spread are determined by M and DM 
 *	 (ie. d_P0(DM)..d_P0(DM)+Blk1d(M,DM)-1). if Blk1d(M,DM)-1>1, 
 *	 then all cells across dimension RowCol receive the spread.
 *
 *	 return value is as for InitSpreadWS()
 * notes: B' is a (RowCol==0? K x N: N x K) global matrix 
 * 	K is dimension to spread across, and is aligned to the RowCol'th 
 * 	processor dimension. Typical calls are: 
 * 	SpreadMat(0, K, N, B, alpha, DM_Dim(1, C), &sB, 0, N, DM_D0(C));
 *		K x n *sB, Inc(*sB) = {Ld, 1}	Ld >= n
 * 	SpreadMat(1, K, M, A, alpha, DM_Dim(0, C), &sA, 0, M, DM_D1(C));
 *		m x K *sA, Inc(*sA) = {1, Ld}	Ld >= m	
 * generally G2Ld(K, BDK) << n = G2Ld(N, DN)
 */

/***************** high-level `de-spread' operations ********************/
     
int SizeReduceWS STDC_PARAMS(
	(int Source, int RowCol, int K, int N, DistMat B, 
	 pMatDim DN, int ForceWS, int M, pMatDim DM)
);
/* pre/post: as for SizeSpreadWS(), except returns the size of the workspace 
 *	 required for 
 *		ReduceMat(RowCol, K, N, B, _, DN, ForceWS, M, DM)
 *	 The difference is that the strides of DM_Loc(B) are no longer
 *       relevant (as the de-spread is received into the owning cells, 
 *	 not sent from), and special optimizations for B being pre-spread
 *       over the K-dimension do not apply here.
 */

int InitReduceWS STDC_PARAMS(
	(int RowCol, int K, int N, DistMat B, FDATA *beta,
	 pMatDim DN, pLocMat sB, int ForceWS, int M, pMatDim DM)
);
/* post: initializes *sB appropriately for the call:
 *		ReduceMat(RowCol, K, N, B, sB, DN, M, DM)
 *	 essentially, this means that *sB points into DM_Loc(B)  
 *	 if this cell is the sole recipient of the reduce (and !ForceWS).
 *	 In this case, sB will be scaled by beta.
 *	 If this is not the case, sB will be a separate workspace that will be 
 *	 allocated in this procedure, if LM_ADR(*sB_pre)==NULL, or has been 
 *	 previously allocated, otherwise. For the latter, *sB is also zeroed.
 *	 Return value is as for InitSpreadWS().
 * note: *sB is (RowCol==0? K x n: n x K)	
 */

void ReduceMat STDC_PARAMS(
	(int RowCol, int K, int N, DistMat A, 
	 FDATA *beta, pMatDim DN, pLocMat sB, int M, pMatDim DM)
);
/* pre:  RCOK(RowCol) && CompatBlk(BDN,DN) && K,N,M>=0
 *	 All cells (holding/requiring any data) are participating in this call
 *	 with consistent values of *beta, RowCol, K, N, *A, *DN, ForceWS, M, 
 *	 & *DM.
 *	 Along processor dimension RowCol specified by DM, 
 *	 the result of spread operation was applied from a matrix of size M
 *	 and the local result is stored in sB
 * post: B' = beta'*B' + (RowCol==0? {Bs, BDK, DN}: {Bs, DN, BDK})
 *
 *       where Bs = Sum across processor dimension RC of selection of sB
 *	            from block d_P0(DN):K, block-stride d_P(DN)
 *	       BDK = DM_Dim(RowCol=!Trans(B), B)
 *	       B' = DM_Trans(B)? B^T: B;
 *	 beta' = *Beta if LM_ADR(DM_Loc(B))!=LM_ADR(sB)) (=> these are disjoint)
 *	       = 1.0 otherwise	(in this case, scaling by *beta should have been
 *				 performed by InitReduceWS())
 * note: B' is a (RowCol==0? K x N: N x K) global matrix 
 */


#endif /*__SpreadMatrix_INCL__*/

