/************************************************************************/
/* 	LDLT: indefinite factorization of a global NxN Matrix A		*/
/************************************************************************/
/* SCCS INFO @(#)LDLT.c 3.3         last modified P. Strazdins, 00/10/18
   first created: P. Strazdins 94/09
*/

/* reduced startup version supporting algorithmic blocking 
 * (externally-defined) cpp FLAGs:
 *      PRINTMAT        print out intermediate matrices
 *	MULT1X1PIVOTS	apply series of 1x1 pivots from current run of 
 *			diagonals, provided element growth is within bounds
 *			since last interchange.
 *      BOUNDMULT1X1    variant of above; include in series only if
 *                      current max / diag within a constant (> alpha).
 *                      This can reduce the growth bound on L.
 *	PIVOT_STATS	take detailed statistics on distribution of pivots
 */
/*#define PIVOT_STATS*/

#if 0				/*fastest*/ 
# define MULT1X1PIVOTS
# define BOUNDMULT1X1           /* recommended - improves accuracy */ 
#else				/*slowest*/

#endif

#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#ifdef ASSERT
# define PRINTMAT
#endif
#include "GetBlockFactor.h"
#include "MatDim.h"
#include "DistMat.h"
#include "D_BLAS.h"

#include "SpreadMatrix.h"
#include "CellTriMatrix.h"
#include "BLASOptions.h"
#include "FDMalloc.h"
#include "FDCommun.h"
#include "CellMatrix.h"
#include "CellVector.h"

#include "D_LAPACK.h"

double sqrt STDC_PARAMS( (double) );

/******* merged communication routines *************************************/

static void D_BKSWP(Direction, M, N, A, P, P0)
  char *Direction; int M, N; DistMat A; int P[], P0;
{ int j, dj, Lj, *LaPiv;

  COMMON_ASSERT_TR_E(  
   (TraceStr, "(%c,%dx%d,%s,a%d %d)", *Direction, M, N, DM_TRACE(0,A),
   ((int) P)/sizeof(int), P0), P!=NULL && DistMatOK(M, N, A), 0, "BKSWP", 0,
   TRACE_OUTER);
  
   if (M<=0) {
     TRACE_RETURN;
   }

   LaPiv = (int *) malloc(N*sizeof(int));
   j=0;
   while (j<N) {
     if (P[j] < 0) 
       LaPiv[j] = j, LaPiv[j+1] = -P[j] - P0, j++;
     else
       LaPiv[j] = P[j] - P0;
     j++;
   }
   D_LASWP(Direction, M, N, A, LaPiv);     
   free((char *) LaPiv);
   TRACE_EXIT;
} /*D_BKSWP*/ 


static void D_SWAP2(M, Aj, Wj, Ai, Wi)
  int M; DistMat Aj, Wj, Ai, Wi;
/* pre: DM_DM(Aj) = DM_DM(Wj) = DM_DM(Ai) = DM_DM(Wi);  
 *      DM_DN(Aj) = DM_DN(Wj);  DM_DN(Ai) = DM_DN(Wi)
 */
{ CommSetUp(d_P(DM_D0(Aj)), d_P(DM_D1(Aj)));  
  if (InColScope(Aj) == InColScope(Ai)) {	
    /* no perf. advantage possible */
    D_SWAP(M, Aj, Ai); 			D_SWAP(M, Wj, Wi);
  } else {  	
    int m = G2Ld(M, DM_DM(Aj));
    FDATA *AW, *pA, *pW; 		int IncA, IncW;
    int DI = 0,        			DJ = -d_P0(DM_DN(Ai))-d_P0(DM_DN(Aj));
    if (InColScope(Ai))
      IncA = LM_Inc(DM_Trans(Ai), DM_Loc(Ai)), 
      IncW = LM_Inc(DM_Trans(Wi), DM_Loc(Wi)),
      pA = LM_ADR(DM_Loc(Ai)), 		pW = LM_ADR(DM_Loc(Wi));
    else
      IncA = LM_Inc(DM_Trans(Aj), DM_Loc(Aj)), 
      IncW = LM_Inc(DM_Trans(Wj), DM_Loc(Wj)),
      pA = LM_ADR(DM_Loc(Aj)), 		pW = LM_ADR(DM_Loc(Wj));
    if (DM_Trans(Aj))
      SWAP(int, DI, DJ);
    if (m > 0) {  
      MatMalloc_FT("D_SWAP2", 1, &AW, 2*m);
      VecCopy(m, pA, IncA, AW, 1);	VecCopy(m, pW, IncW, AW+m, 1);
      SendMat(DI, DJ, 1, 2*m, AW, 2*m, GENMSGTTYPE);  
      RecvMat(DI, DJ, 1, 2*m, AW, 2*m, GENMSGTTYPE);  
      VecCopy(m, AW, 1, pA, IncA);	VecCopy(m, AW+m, 1, pW, IncW);
      MatFree(AW);
    }  
  }     
} /* D_SWAP2() */

static int D_AMAX2(N, X, Max, AuxX, AuxV)
 int N; DistMat X, AuxX; FDATA *Max, *AuxV;
/* pre:  N>= 0; DistMatOK(N, 1, X); DM_DN(X) = DM_DN(AuxX)
 *	 if InEltScope(AuxX), v = *AuxV
 * post: as fpr D_AXAX(N, X, Max), and *AuxV = v 
 * note: the motivation of this function is combine the broadcast of v
 *       in the broadcasts within the D_AMAX() computation.
 */
{ FDATA amax[3]; 	
  CommSetUp(d_P(DM_D0(X)), d_P(DM_D1(X)));  
  COMMON_ASSERT_TR_E( 
    (TraceStr, "(%d,%s,_%s,_)", N, DM_TRACE(0, X), DM_TRACE(1, AuxX)),
     N>=0 && DistMatOK(N, 1, X) && Dim_Eq(DM_DN(X), DM_DN(AuxX)),
     0, "AMAX2", 1, TRACE_OUTER
  );
  
  if (InEltScope(AuxX)) 
    amax[2] = *AuxV;
  if (d_P0(DM_DN(X))==0) {
    int nx = G2Ld(N, DM_DM(X));         int IncX = LM_Inc(DM_Trans(X), DM_Loc(X));
    int imax = VecAbsMaxPos(nx, LM_ADR(DM_Loc(X)), IncX);
    if (imax==-1) {                     /* assert nx = 0 */
      amax[0] = FD_0;                   imax = 0;
    } else
      amax[0] = *(LM_ADR(DM_Loc(X)) + FAST_MUL2(imax, IncX));
/*
{ int i;
  for (i=0; i < nx; i++)
    printf("%s ", pFD2Str(LM_ADR(DM_Loc(X))+IncX*i));
  printf("\n");
} */
/*printf("AMAX2: loc: %d (g %d) %s\n", imax, L2Gd(imax, DM_DM(X)),  pFD2Str(&amax[0]));*/

    imax = L2Gd(imax, DM_DM(X));        /* convert to a global index */
    FD_SET_REAL(amax[1], (FD_BASE) imax);
    if (d_P(DM_DM(X)) > 1) {
      int P = d_P(DM_DM(X)),   		P0 = d_P0(DM_DM(AuxX)),
          DelP0 = 1, lgDelP0=0,		RowCol = DM_Trans(X);
      while (DelP0 < P) {
	FDATA RMxI[2];
        int SendMsg = (P0 & (2*DelP0-1)) == DelP0;
        int DelP0_ = SendMsg? -DelP0: DelP0;
        int DelIP0 = RowCol==0? DelP0_: 0,  DelJP0 = RowCol==0? 0: DelP0_;
        if (SendMsg) {
          SendMat(DelIP0, DelJP0, 1, 2, amax, 2, SUMMSGTYPE);
        } else if ((P0 & (2*DelP0-1))==0 && P0+DelP0 < P) {
          RecvMat(DelIP0, DelJP0, 1, 2, RMxI, 2, SUMMSGTYPE);
        if (FD_ABS(RMxI[0]) > FD_ABS(amax[0]))
          amax[1] = RMxI[1], amax[0] = RMxI[0];
        } /* if (SendMsg) */
        DelP0 *= 2, lgDelP0++;
      } /* while (DelP0 ...) */
      BrdMat(DM_Trans(X), P0, 1, 3, amax, 3);
    }
    if (FD_EQ(amax[0], FD_0))           /* take top element as being the max */
      imax = 0, amax[1] = FD_0;
  }
  if (d_P(DM_DN(X)) > 1)
    BrdMat(1-DM_Trans(X), d_P0(DM_DN(X)), 1, 3, amax, 3);
  *Max = amax[0], *AuxV = amax[2];
/*printf("AMAX2(%d): %d %s aux=%s\n", N, (int) FD_REAL(amax[1]), pFD2Str(Max), pFD2Str(AuxV));*/
  TRACE_RETURN((int) FD_REAL(amax[1]));         	
} /* AMAX2() */

/*************** replicteed diagonal matrix handling routines ***************/

static FDATA *Dii, *Di1i1, *Di1i;
static DistMatStore A00_;
static DistMat A00 = NULL;

static SetUpDInv(N, A, DInv)
  int N; DistMat A; FDATA DInv[];
{ int n = G2Ld(N, DM_DM(A));
  Dii = DInv, Di1i1 = Dii + n, Di1i = Di1i1 + n; 
  A00 = SubMat_(&A00_, A, 0, 0);
}

static SetDInv1(j, Ajj, D00) 
  int j; DistMat Ajj; FDATA *D00;
{ int j_loc = G2Ld(j, DM_DM(A00)); 
  if (InRowScope(Ajj))
    Dii[j_loc] = *D00;
}   

static SetDInv(j, Ajj, Aj1j, D00, D11, D10) 
  int j; DistMat Ajj, Aj1j; FDATA *D00, *D11, *D10;
{ int j_loc = G2Ld(j, DM_DM(A00)); 
  if (InRowScope(Ajj) || (Aj1j!=NULL && InRowScope(Aj1j)))
    Dii[j_loc] = *D00, Di1i1[j_loc] = *D11, Di1i[j_loc] = *D10;
}   

static FDATA GetDInv1(j) 
  int j; 
{ int j_loc = G2Ld(j, DM_DM(A00)); 
  return(Dii[j_loc]);
}   

static GetDInv3(j, D00, D11, D10) 
  int j; FDATA *D00, *D11, *D10;
{ int j_loc = G2Ld(j, DM_DM(A00)); 
  *D00 = Dii[j_loc], *D11 = Di1i1[j_loc] , *D10 = Di1i[j_loc] ;
}   

/*************** pivoting statistics code ***********************************/

#ifdef PIVOT_STATS

static int RunMax = 0, RunCnt, RunD2Cnt;
static int *RunHist = NULL, 		/* histogram of runs of case D1, */
	   *RunD2Hist = NULL;		/* and D1&D2 combined. (within blocks)*/
static double *WRunHist = NULL;		/* RunHist weighted for the amount of work */

#define PIVOT_STATS_SetUp(w)						\
  { RunMax = ABS(w)+2;							\
    WRunHist = (double *)						\
    calloc(RunMax*(2*sizeof(int)+sizeof(double)), 1);			\
    RunHist = (int *) ((double *) WRunHist + RunMax);			\
    RunD2Hist = RunHist +  RunMax;        				\
    RunCnt = RunD2Cnt = 0;						\
  }									

#define PIVOT_STATS_IncRun()						\
    { /*printf(" *%d", RunCnt);*/ RunCnt++, RunD2Cnt++; }	

#define PIVOT_STATS_EndRun(CaseD2, n)					\
    { CommonAssert(RunCnt<RunMax && RunD2Cnt<RunMax, 0, "PIVOT_STATS", 15, TRACE_ALL);\
      /*printf(" (%d,%d)", CaseD2, n);*/				\
      if (CaseD2) RunD2Cnt++;						\
      else RunD2Hist[RunD2Cnt]++, RunD2Cnt=0;				\
      RunHist[RunCnt]++; WRunHist[RunCnt]+=(n), RunCnt=0; 		\
    }

#if 0
     printf("\n	D1Runs:   ");						\
     for (i=1; i < RunMax; i++)						\
         printf("%d ", RunHist[i]);					\
     printf("\n	D1D2Runs: ");						\
     for (i=1; i < RunMax; i++)						\
         printf("%d ", RunD2Hist[i] );					\
     printf("\n	wD1Runs:  ");						\
     for (i=1; i < RunMax; i++)						\
         printf("%3.2f ", WRunHist[i]);					\

#endif
    
#define PIVOT_STATS_Dump(N)						\
   { int i; extern int AbsCId; /* FDCommun.h */				\
     if (AbsCId==0) {							\
     printf("\n	D1Runs:   ");						\
     for (i=1; i < RunMax; i++)						\
       if (RunHist[i] > 0)						\
         printf("%d:%0.2f ", i, (double) i*RunHist[i] / (double) (N));	\
     printf("\n	D1D2Runs: ");						\
     for (i=1; i < RunMax; i++)						\
       if (RunD2Hist[i] > 0)						\
         printf("%d:%0.2f ", i, (double) i*RunD2Hist[i] / (double) (N));\
     printf("\n	wD1Runs:  ");						\
     for (i=1; i < RunMax; i++)						\
       if (WRunHist[i] > 0.0)						\
         printf("%d:%0.2f ", i, i*WRunHist[i] / (((double) (N)*(N)+1)/2));\
     printf("\n"); free(RunHist); RunHist = RunD2Hist = NULL, WRunHist = NULL;\
     }									\
   }      

#else

#define PIVOT_STATS_SetUp(w)

#define PIVOT_STATS_IncRun()

#define PIVOT_STATS_EndRun(CaseD2, n)

#define PIVOT_STATS_Dump(N)

#endif /* PIVOT_STATS */

/*************** main routines **********************************************/

int Case2Cnt = 0;			/* used to determine # times the 2nd 
  					   case of B-K alg. occured. */

int NColSearches = 0;

void D_SYTRF(N, A, DInv, Piv, info, MaxW)
  int N, Piv[], *info, MaxW; DistMat A; FDATA *DInv;
{ int j;
  char svBrdTopChar[2];  
  FD_BASE alpha = (1.0 + sqrt(17.0))/8.0;
#ifdef MULT1X1PIVOTS
# ifdef BOUNDMULT1X1
  FD_BASE Palpha = alpha*alpha*alpha*alpha*alpha;
  /* A[j,j] >= Palpha lambda[j] to be included in multiple 1x1 pivot */
# endif
  FD_BASE Mu = 1.0, MaxGrowth = 1.0;		
  /* estimate & bounds for element growth since last interchange */
# define MAX_1X1PIV 64	
  FD_BASE Mu_k[MAX_1X1PIV];		/* circular queue of recent Mu values */
  int mu_k_hd = 0, mu_k_tl = 0;		
#endif
  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%s,_,%d)", N, DM_TRACE(0,A), MaxW),
    DistMatOK(N, N, A), 0, "SYTRF", -1, TRACE_OUTER
  );

  PIVOT_STATS_SetUp(MaxW);
  
  svBrdTopChar[0] = BrdTopChar[0],      svBrdTopChar[1] = BrdTopChar[1];
  BrdTopChar[DM_Trans(A)] = ' ',        BrdTopChar[1-DM_Trans(A)] = ' ';
 
  *info = 0;
  SetUpDInv(N, A, DInv);
  Case2Cnt = NColSearches = 0; 
  
  j=0;
  while (j < N) {
    DistMat_RVAlloc(4);
    DistMat Ajj = SubMat(A, j, j);
    int minW = MAX(d_B(DM_DN(A)), MAX(MIN_BLOCKING_FACTOR, -MaxW));
    int maxW = MIN(ABS(MaxW), N-j);
    int w = GenGetW(N-(j+1), 1.0, minW, maxW, DM_DN(Ajj), j, 1);
    DistMat W = Trans(NewMat(w+1, N-j, Trans(Ajj), NULL)); 
    int k = 0, n = N-j;
	       
    while (k < w) {
      DistMat_RVAlloc(9);	 
      DistMat Ak0 = SubMat(Ajj, k, 0),		Akk = SubMat(Ak0, 0, k),
      	      Ak1k = SubMat(Akk, 1, 0),		Ak1k1 = SubMat(Ak1k, 0, 1),
      	      Akk1 =  SubMat(Akk, 0, 1),
       	      Wk0 = SubMat(W, k, 0), 		Wkk = SubMat(Wk0, 0, k),
       	      Wk1k = SubMat(Wkk, 1, 0),		Wkk1 = SubMat(Wkk, 0, 1);
      FDATA eWkk, eWik, eWii;
      FD_BASE AbsWkk, AbsWik; 
      int i, dk;
       	
      D_COPY(n-k, Akk, Wkk);
      D_GEMV(n-k, k, FD_M1, Ak0, Trans(Wk0), FD_1, Wkk);
                 	
      if (InEltScope(Wkk))
	eWkk = DM_Elt00(Wkk);	
      if (j+k+1 < N) {
	i = k+1 + D_AMAX2(n-(k+1), Wk1k, &eWik, Wkk, &eWkk);
        /* eWkk = D_GetElt(Wkk) */
        AbsWik = FD_ABS(eWik);
        NColSearches++;
      } else {					/* prevent 2x2 pivot col N-1*/
        AbsWik = 0.0;
        eWkk = D_GetElt(Wkk);
      }  
      AbsWkk = FD_ABS(eWkk);

#if defined(MULT1X1PIVOTS)
      if (AbsWkk > 0.0) {
        Mu *= (Mu_k[mu_k_tl] = 1.0 + AbsWik / AbsWkk);
        mu_k_tl = (mu_k_tl + 1) % MAX_1X1PIV;
        if (mu_k_tl == mu_k_hd) {			/* circ queue full */
          Mu /= Mu_k[mu_k_hd], mu_k_hd = (mu_k_hd + 1) % MAX_1X1PIV;
        } else
          MaxGrowth *= (1.0 + 1.0/alpha);
      }
#endif

      dk = 1; Piv[j+k] = j+k;
      if (MAX(AbsWik, AbsWkk) == 0.0) {
       	/* a zero column; L^-1 does not exist, but LDL^T fact'n can exist */
       	if (*info==0)
       	  *info = j+k+1;
      } else {	
#if !defined(MULT1X1PIVOTS)
        if (AbsWkk >= alpha * AbsWik) 
#else 
# ifndef BOUNDMULT1X1
          if (Mu <= MaxGrowth || i==k)
# else
          if ((Mu<=MaxGrowth && AbsWkk >= Palpha * AbsWik) || i==k)
# endif
#endif        
        { /* no pivoting needed */
          PIVOT_STATS_IncRun();
	} else {
	  DistMat_RVAlloc(6);	
	  DistMat Wi0 = SubMat(W, i, 0), 	Wik1 = SubMat(Wi0, 0, k+1), 	
	  	  Ai0 = SubMat(Ajj, i, 0),
	  	  Aik = SubMat(Ai0, 0, k),	Aii = SubMat(Ai0, 0, i),
	  	  Aik1 = SubMat(Aik, 0, 1);	
	  FDATA eWij; FD_BASE AbsWij; int j_;
	  FDATA eAkk1;

#if defined(MULT1X1PIVOTS)
/*        printf("j=%3d, pj=%-3d, i=%3d, AbsWkk=%1.2e, AbsWik=%1.2e Mu=%1.2e, G=%1.2e\n", j+k, Piv[j+k], i, AbsWik, AbsWkk, Mu, MaxGrowth);*/
	  /* case D1 not applied; reset growth estimate and bound */
	  Mu = MaxGrowth = 1.0; mu_k_hd = mu_k_tl;
#endif	  

	  /* copy column i to W[:,k+1] and update it to determine pivot;
	     note that A[i,i] is placed at  W[i,k+1] */	
	  D_COPY(i-k, Trans(Aik), Wkk1);

	  D_COPY(n-i, Aii, Wik1);
	  D_GEMV(n-k, k, FD_M1, Ak0, Trans(Wi0), FD_1, Wkk1);

	  if (InEltScope(Wik1))
	    eWii = DM_Elt00(Wik1);	
	  D_SetElt(Wik1, FD_0);			/* exclude from D_AMAX() */
	  j_ = i + D_AMAX2(n-(k+1), Wkk1, &eWij, Wik1, &eWii);
	  NColSearches++;
	  
	  AbsWij = FD_ABS(eWij);
	  D_SetElt(Wik1, eWii); 

	  if (AbsWkk >= alpha * AbsWik * (AbsWik / AbsWij)) 
	    /* 2x2 pivot may be unstable; may as well pivot from row k  */
	    Case2Cnt++;
	  else if (FD_ABS(eWii) >= alpha * AbsWij) {
	    /* 1x1 pivot from row i will be stable */ 	
	    D_COPY(n-k, Wkk1, Wkk);
	     
	    D_COPY(i-k, Akk, Trans(Aik));	
	    D_COPY(n-i, Aik, Aii);
	     
	    /* merge these 2 calls to reduce startups */
	    D_SWAP2(k+1, Trans(Ak0), Trans(Wk0), Trans(Ai0), Trans(Wi0));

	    eWkk = eWii;
	    Piv[j+k] = j+i;
	  } else {
	    /* 2x2 pivot will be stable */	
	    dk = 2;	

  	    if (InEltScope(Akk1)) { /* set A[k,k+1] to A[k+1,k] for 2x2 pivot */
	      eAkk1 = *LM_ADR(DM_Loc(Akk1));
	      *LM_ADR(DM_Loc(Akk1)) = *LM_ADR(DM_Loc(Wkk1));
	    }  	

	    if (i!=k+1) { /* interchange cols k+1 and i */
	      DistMat_RVAlloc(2);
	      DistMat Ak10 = SubMat(Ak0, 1, 0),	Wk10 = SubMat(Wk0, 1, 0);
	      
	      D_COPY(i-k, Akk1, Trans(Aik));
	      D_COPY(n-i, Aik1, Aii);
	     
              D_SWAP2(k+2, Trans(Ak10), Trans(Wk10), Trans(Ai0), Trans(Wi0));
	    }
	    if (InEltScope(Akk1))  /* restore A[k,k+1] to orig value */
	      *LM_ADR(DM_Loc(Akk1)) = eAkk1;

	    Piv[j+k+1] = Piv[j+k] = -(j+i);
	    
	  }
	  PIVOT_STATS_EndRun(Piv[j+k]==j+k, N-(j+k));
	}


	if (dk==1) {
	  /* D = W[k;k]; update A[k+1:N-1, k] by D^-1 */	
	  D_COPY(n-k, Wkk, Akk);
	  if (/*k < n-1*/ 1) {
	    FDATA InvWkk;		FD_RECIP(InvWkk, eWkk);	
	    D_SCAL(n-(k+1), InvWkk, Ak1k);
	    SetDInv1(j+k, Akk, &InvWkk);
	  }  	
	} else {
	  /* D = W[k:k+1, k:k+1]; update A[k+2:N-1, k:k+1] by D^-1 */	
	  /* note: now eWkk = W[k,k], eWii = W[k+1,k+1], eWik = W[k+1,k] */
	  DistMat_RVAlloc(4);
	  DistMat Wk2k = SubMat(Wkk, 2, 0),	Wk2k1 = SubMat(Wk2k, 0, 1),
	          Ak2k = SubMat(Akk, 2, 0),	Ak2k1 = SubMat(Ak2k, 0, 1);
	  FDATA D10 = eWik, D00, D11, T = FD_M1; /* explicit inv of 2x2 pivot */
	  FD_DVDSET(D00, eWii, D10);		FD_DVDSET(D11, eWkk, D10); 	
	  FD_MULSET(T, +=, D00, D11);		FD_RECIP(T, T);
	  FD_DVDSET(D10, T, D10);

	  SetDInv(j+k, Akk, Ak1k, &D11, &D00, &D10);
	  
          if (d_P0(DM_DN(Ak2k))==d_P0(DM_DN(Ak2k1))) {
	    /* can optimize: 2 columns on the same cell column */
            if (d_P0(DM_DN(Ak2k1))==0) {
              int l, l_max = G2Ld(n-k-2, DM_DM(Ak2k));
              FDATA *pAk=LM_ADR(DM_Loc(Ak2k)), *pAk1 = LM_ADR(DM_Loc(Ak2k1)), 	
		    *pWk=LM_ADR(DM_Loc(Wk2k)), *pWk1 = LM_ADR(DM_Loc(Wk2k1));
	      int RIncA = LM_Inc(DM_Trans(Ak2k), DM_Loc(Ak2k)),
	      	  RIncW = LM_Inc(DM_Trans(Ak2k), DM_Loc(Wk2k));	      	    
	      for (l=0; l < l_max; l++) {
	        FD_MULSET(*pAk, =, D00, *pWk);	FD_MULSET(*pAk1, =, D11, *pWk1);
	        FD_ASSIGN(*pAk, -=, *pWk1);	FD_ASSIGN(*pAk1, -=, *pWk);
	        pWk += RIncW,			pWk1 += RIncW;
	        FD_MULSET(*pAk, =, *pAk, D10);	FD_MULSET(*pAk1, =, *pAk1, D10);
	        pAk += RIncA,			pAk1 += RIncA;
	      }
	    }  	
	  } else {	  
	    D_COPY(n-k-2, Wk2k, Ak2k);	   	D_COPY(n-k-2, Wk2k1, Ak2k1);
	    D_ScaleAdd(n-k-2, 1, FD_M1, Wk2k1, D00, Ak2k);
	    D_ScaleAdd(n-k-2, 1, FD_M1, Wk2k,  D11, Ak2k1);
	    D_SCAL(n-k-2, D10, Ak2k);	  	D_SCAL(n-k-2, D10, Ak2k1);
	  }
	   
	  D_SetElt(Akk, eWkk);			D_SetElt(Ak1k1, eWii);
	  D_SetElt(Ak1k, eWik);
	  
	  DM_PRINT("Ak", n-k, 2, Akk);
	 }
       }                    
       k += dk;	              
    } /* while (k..) */	

    PIVOT_STATS_EndRun(0, N-(j+k));
    w = k;   
    if (n-k > 0) {
      /* update trailing tri sub-matrix Ajkjk -= Ajkj Wk0^T */	
      DistMat_RVAlloc(3);
      DistMat Ajkj = SubMat(Ajj, k, 0),		
              Ajkjk = SubMat(Ajkj, 0, k), 	
 	      Wk0 = SubMat(W, k, 0);     	
      LocMat sAk, sWk;				int ms, ns;
      FDATA ScW = FD_M1;
       
      LM_ADR(sAk) = NULL;
      ms = SpreadMat(1, k, n-k, Ajkj, &FD_1, DM_DM(Ajkjk), &sAk, 0, 
      		     n-k, DM_DN(Ajkjk));
      LM_ADR(sWk) = NULL;
      ns = SpreadMat(0, k, n-k, Trans(Wk0), &ScW, DM_DN(Ajkjk), &sWk, 0, 
      		     n-k, DM_DM(Ajkjk));
      TriMatPartMult(ABS(ms), ABS(ns), k, &ScW, &sAk, &sWk, &FD_1, Lower, Ajkjk);
      if (ns > 0)
        MatFree(LM_ADR(sWk));
      if (ms > 0)
        MatFree(LM_ADR(sAk));
    }
    DisposeMat(W);  
    
    /* complete row interchanges on A */
    D_BKSWP("Fwd", j, w, Trans(SubMat(A, j, 0)), &Piv[j], j);
    j += w;        
  } /* while (j...) */

  BrdTopChar[0] = svBrdTopChar[0],      BrdTopChar[1] = svBrdTopChar[1];
  PIVOT_STATS_Dump(N);
  TRACE_EXIT;   
} /* D_SYTRF() */ 

static FDATA *SubDiag = NULL;

static ClrSubDiag(N, A, Piv)
  int N, Piv[]; DistMat A;
{ int n = G2Ld(N, DM_DM(A));
  int j, Cpy = SubDiag==NULL;
  if (Cpy)
    MatMalloc("CpSubDiag", 1, &SubDiag, n); 
  j=0;
  while (j < N) {
    if (Piv[j] < 0) {
      DistMat_RVAlloc(1);
      DistMat Aj1j = SubMat(A, j+1, j);
      if (InEltScope(Aj1j)) {
        if (Cpy) {
 	  int j_loc = G2Ld(j+1, DM_DM(A));
	  SubDiag[j_loc] = DM_Elt00(Aj1j);
        }
        DM_Elt00(Aj1j) = FD_0;
      }
      j++;
    } 
    j++;
  } /* while (j...) */
} /* CpSubDiag() */

static ResetSubDiag(N, A, Piv)
  int N, Piv[]; DistMat A;
{ int j;
  j=0;
  while (j < N) {
    if (Piv[j] < 0) {
      DistMat_RVAlloc(1);
      DistMat Aj1j = SubMat(A, j+1, j);
      if (InEltScope(Aj1j)) {
        int j_loc = G2Ld(j+1, DM_DM(A));
        DM_Elt00(Aj1j) = SubDiag[j_loc];
      }  
      j++;
    }
    j++;
  } /* while (j...) */
} /*ResetSubDiag()*/


void D_SYTRS(N, K, A, DInv, Piv, B, info)
  int K, N;
  DistMat A, B; FDATA *DInv; int Piv[], *info; 
{ int j;	
  char svBrdTopChar[2];  

  COMMON_ASSERT_TR_E(
     (TraceStr, "(%d,%d,%s,_,%s_)", N, K,  DM_TRACE(0, A), DM_TRACE(1, B)),
      DistMatOK(N, N, A) && DistMatOK(N, K, B), 0, "SYTRS", -1, TRACE_OUTER
  );
  svBrdTopChar[0] = BrdTopChar[0],      svBrdTopChar[1] = BrdTopChar[1];

  D_BKSWP("Fwd", K, N, Trans(B), Piv, 0);

  BrdTopChar[0] = 'I',        		BrdTopChar[1] = 'I';    
  ClrSubDiag(N, A, Piv);    
  D_TRSM("Lower", "Unit", N, K, FD_1, A, B);

  j=0;
  while (j < N) {
    DistMat_RVAlloc(4);
    DistMat Bj = SubMat(B, j, 0),	Bj1 = SubMat(Bj, 1, 0),
            Ajj = SubMat(A, j, j),	Aj1j = SubMat(Ajj, 1, 0);

    if (Piv[j] >= 0) {			/* 1 x 1 pivot */
      DistMat_RVAlloc(2);
      if (InRowScope(Ajj)) {		/* multiply by inverse of 1x1 block */
        FDATA eAjj; 
	eAjj = GetDInv1(j);
        D_SCAL(K, eAjj, Trans(Bj));	
      }
      j++;
    } else {				/* 2 x 2 pivot */
      DistMat_RVAlloc(8);
      DistMat Bj2 = SubMat(Bj1, 1, 0), 
              Aj2j = SubMat(Aj1j, 1, 0), 	Aj1j1 = SubMat(Aj1j, 0, 1),
              Aj2j1 = SubMat(Aj2j, 0, 1);
      int Pj = -Piv[j];

      if (d_P0(DM_DM(Bj))==d_P0(DM_DM(Bj1))) {
      	/* can optimize: 2 columns on the same cell row */
      	if (d_P0(DM_DM(Bj))==0) {
      	  int l, l_max = G2Ld(K, DM_DN(Bj));
          FDATA *pBj=LM_ADR(DM_Loc(Bj)), 	*pBj1 = LM_ADR(DM_Loc(Bj1)); 	
          int CIncB = LM_Inc(1-DM_Trans(Bj), DM_Loc(Bj));          
      	  FDATA D00, D10, D11, T;

	  GetDInv3(j, &D00, &D11, &D10);	

      	  for (l=0; l < l_max; l++) { 	
      	     FDATA eBj = *pBj, eBj_,		eBj1 = *pBj1, eBj1_;
      	     FD_MULSET(eBj_, =, eBj, D11);	FD_MULSET(eBj1_, =, eBj1, D00); 
	     FD_ASSIGN(eBj_, -=, eBj1);		FD_ASSIGN(eBj1_, -=, eBj);
	     FD_MULSET(*pBj, =, eBj_, D10);	FD_MULSET(*pBj1, =, eBj1_, D10);
      	     pBj += CIncB,			pBj1 += CIncB;	    
      	  }    
      	}
      } else {
      	DistMat Bj_ = NewMat(1, K, Bj, NULL);
      	FDATA D00, D10, D11, T;

	GetDInv3(j, &D00, &D11, &D10);

	D_COPY(K, Trans(Bj), Trans(Bj_));      	
      	D_ScaleAdd(1, K, FD_M1, Bj1, D11, Bj);
      	D_ScaleAdd(1, K, FD_M1, Bj_, D00, Bj1);
      	D_SCAL(K, D10, Trans(Bj));		D_SCAL(K, D10, Trans(Bj1));
      	DisposeMat(Bj_);
      }	    
      j+=2;    	
    }
  } /* while (j...) */

  BrdTopChar[0] = ' ',        		BrdTopChar[1] = ' ';      
  D_TRSM("Upper", "Unit", N, K, FD_1, Trans(A), B);
  ResetSubDiag(N, A, Piv);  
  MatFree(SubDiag);				SubDiag = NULL;

  D_BKSWP("Bwd", K, N, Trans(B), Piv, 0);
  BrdTopChar[0] = svBrdTopChar[0],      BrdTopChar[1] = svBrdTopChar[1];
    
  TRACE_EXIT;
} /* D_SYTRS() */

