/************************************************************************/
/* 	LDLT: indefinite factorization of a global NxN Matrix A		*/
/************************************************************************/
/* SCCS INFO %Z%%M% %I%         last modified P. Strazdins, %E%
   first created: P. Strazdins 94/09
*/

/* reduced startup version supporting algorithmic blocking 
 * (externally-defined) cpp FLAGs:
 *      PRINTMAT        print out intermediate matrices
 *	MULT1X1PIVOTS	apply series of 1x1 pivots from current run of 
 *			diagonals, provided element growth is within bounds
 *			since last interchange.
 *	BOUNDMULT1X1	variant of above; include in series only if
 *			current max / diag within a constant (> alpha).
 *			This can reduce the growth bound on L.	
 *	PIVOT_STATS	take detailed statistics on distribution of pivots
 *	INERTIA_STATS	record statistics on inertia of matrix (for non-
 *			complex or hermitean matrices)
 *	RESPREAD_W	recompute the row spread of the lower panel of L
 *			from that of W; this enables the column spread of W^T
 *			to be computed without explicit transposition, at the
 *			cost of having to redundantly recompute row spread
 *			of L. 
 *	LOOKAHEAD	implement lookahead over the columns in the
 *			current panel, so (faster) modified Cholesky
 *			level 2 factorizations can occur over sub-panels
 *	QUAD_LOOKAHEAD	a variant of LOOKAHEAD using a `quadratic' a priori
 *			bound (non-optimal)
 *	POST_CALC_GROWTH apply actual growth calculations over sub-panels
 *			factored by modified Cholesky. While this requires
 * 			extra overhead, it results in a less conservative
 *			element growth estimation
 *	LA_STOP_GROWTH_EXCEED
 *			when looking ahead over a panel, stop a  block pivot
 *			on any column wherew the a priori growth exceeds its
 *			bounds. This should achieve the same pivoting behavior
 *			if LOOKAHEAD was not defined.
 *			Otherwise, apply the less conservative strategy
 *			of keep looking ahead in case the growth again comes
 *			within bounds. 
 */
#define PIVOT_STATS
#define INERTIA_STATS
#define LOOKAHEAD
#define LA_STOP_GROWTH_EXCEED

#if 1				/*fastest*/ 
# define MULT1X1PIVOTS
# define BOUNDMULT1X1 		/* recommended - improves accuracy */
# define RESPREAD_W
#else				/*slowest*/

#endif

#ifdef LOOKAHEAD
# define MULT1X1PIVOTS		/*necessary*/
# define RESPREAD_W		/*recommended; temporarily necessary */
# define POST_CALC_GROWTH	/*recommended*/

#endif

#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#ifdef ASSERT
# define PRINTMAT
#endif
#include "GetBlockFactor.h"
#include "MatDim.h"
#include "DistMat.h"
#include "D_BLAS.h"

#include "SpreadMatrix.h"
#include "CellTriMatrix.h"
#include "TRSM.h"
#include "BLASOptions.h"
#include "FDMalloc.h"
#include "FDCommun.h"
#include "DistMatrix.h"
#include "CellMatrix.h"
#include "CellVector.h"

#include "D_LAPACK.h"

double sqrt STDC_PARAMS( (double) );
extern int AbsCId; /* FDCommun */

/******* merged communication routines *************************************/

static void D_BKSWP(Direction, M, N, A, P, P0)
  char *Direction; int M, N; DistMat A; int P[], P0;
{ int j, dj, Lj, *LaPiv;

  COMMON_ASSERT_TR_E(  
   (TraceStr, "(%c,%dx%d,%s,a%d %d)", *Direction, M, N, DM_TRACE(0,A),
   ((int) P)/sizeof(int), P0), P!=NULL && DistMatOK(M, N, A), 0, "BKSWP", 0,
   TRACE_OUTER);
  
   if (M<=0) {
     TRACE_RETURN;
   }

   LaPiv = (int *) malloc(N*sizeof(int));
   j=0;
   while (j<N) {
     if (P[j] < 0) 
       LaPiv[j] = j, LaPiv[j+1] = -P[j] - P0, j++;
     else
       LaPiv[j] = P[j] - P0;
     j++;
   }
   D_LASWP(Direction, M, N, A, LaPiv);     
   free((char *) LaPiv);
   TRACE_EXIT;
} /*D_BKSWP*/ 


static void D_SWAP2(M, Aj, Wj, Ai, Wi)
  int M; DistMat Aj, Wj, Ai, Wi;
/* pre: DM_DM(Aj) = DM_DM(Wj) = DM_DM(Ai) = DM_DM(Wi);  
 *      DM_DN(Aj) = DM_DN(Wj);  DM_DN(Ai) = DM_DN(Wi)
 */
{ CommSetUp(d_P(DM_D0(Aj)), d_P(DM_D1(Aj)));  
  if (InColScope(Aj) == InColScope(Ai)) {	
    /* no perf. advantage possible */
    D_SWAP(M, Aj, Ai); 			D_SWAP(M, Wj, Wi);
  } else {  	
    int m = G2Ld(M, DM_DM(Aj));
    FDATA *AW, *pA, *pW; 		int IncA, IncW;
    int DI = 0,        			DJ = -d_P0(DM_DN(Ai))-d_P0(DM_DN(Aj));
    if (InColScope(Ai))
      IncA = LM_Inc(DM_Trans(Ai), DM_Loc(Ai)), 
      IncW = LM_Inc(DM_Trans(Wi), DM_Loc(Wi)),
      pA = LM_ADR(DM_Loc(Ai)), 		pW = LM_ADR(DM_Loc(Wi));
    else
      IncA = LM_Inc(DM_Trans(Aj), DM_Loc(Aj)), 
      IncW = LM_Inc(DM_Trans(Wj), DM_Loc(Wj)),
      pA = LM_ADR(DM_Loc(Aj)), 		pW = LM_ADR(DM_Loc(Wj));
    if (DM_Trans(Aj))
      SWAP(int, DI, DJ);
    if (m > 0) {  
      MatMalloc_FT("D_SWAP2", 1, &AW, 2*m);
      VecCopy(m, pA, IncA, AW, 1);	VecCopy(m, pW, IncW, AW+m, 1);
      SendMat(DI, DJ, 1, 2*m, AW, 2*m, GENMSGTTYPE);  
      RecvMat(DI, DJ, 1, 2*m, AW, 2*m, GENMSGTTYPE);  
      VecCopy(m, AW, 1, pA, IncA);	VecCopy(m, AW+m, 1, pW, IncW);
      MatFree(AW);
    }  
  }     
} /* D_SWAP2() */

static int D_AMAX2(N, X, Max, AuxX, AuxV)
 int N; DistMat X, AuxX; FDATA *Max, *AuxV;
/* pre:  N>= 0; DistMatOK(N, 1, X); DM_DN(X) = DM_DN(AuxX)
 *	 if InEltScope(AuxX), v = *AuxV
 * post: as fpr D_AXAX(N, X, Max), and *AuxV = v 
 * note: the motivation of this function is combine the broadcast of v
 *       in the broadcasts within the D_AMAX() computation.
 */
{ FDATA amax[3]; 	
  CommSetUp(d_P(DM_D0(X)), d_P(DM_D1(X)));  
  COMMON_ASSERT_TR_E( 
    (TraceStr, "(%d,%s,_%s,_)", N, DM_TRACE(0, X), DM_TRACE(1, AuxX)),
     N>=0 && DistMatOK(N, 1, X) && Dim_Eq(DM_DN(X), DM_DN(AuxX)),
     0, "AMAX2", 3, TRACE_OUTER
  );
  
  if (InEltScope(AuxX)) 
    amax[2] = *AuxV;
  if (d_P0(DM_DN(X))==0) {
    int nx = G2Ld(N, DM_DM(X));         int IncX = LM_Inc(DM_Trans(X), DM_Loc(X));
    int imax = VecAbsMaxPos(nx, LM_ADR(DM_Loc(X)), IncX);
    if (imax==-1) {                     /* assert nx = 0 */
      amax[0] = FD_0;                   imax = 0;
    } else
      amax[0] = *(LM_ADR(DM_Loc(X)) + FAST_MUL2(imax, IncX));

    imax = L2Gd(imax, DM_DM(X));        /* convert to a global index */
    FD_SET_REAL(amax[1], (FD_BASE) imax);
    if (d_P(DM_DM(X)) > 1) {
      int P = d_P(DM_DM(X)),   		P0 = d_P0(DM_DM(AuxX)),
          DelP0 = 1, lgDelP0=0,		RowCol = DM_Trans(X);
      while (DelP0 < P) {
	FDATA RMxI[2];
        int SendMsg = (P0 & (2*DelP0-1)) == DelP0;
        int DelP0_ = SendMsg? -DelP0: DelP0;
        int DelIP0 = RowCol==0? DelP0_: 0,  DelJP0 = RowCol==0? 0: DelP0_;
        if (SendMsg) {
          SendMat(DelIP0, DelJP0, 1, 2, amax, 2, SUMMSGTYPE);
        } else if ((P0 & (2*DelP0-1))==0 && P0+DelP0 < P) {
          RecvMat(DelIP0, DelJP0, 1, 2, RMxI, 2, SUMMSGTYPE);
        if (FD_ABS(RMxI[0]) > FD_ABS(amax[0]))
          amax[1] = RMxI[1], amax[0] = RMxI[0];
        } /* if (SendMsg) */
        DelP0 *= 2, lgDelP0++;
      } /* while (DelP0 ...) */
      BrdMat(DM_Trans(X), P0, 1, 3, amax, 3);
    }
    if (FD_EQ(amax[0], FD_0))           /* take top element as being the max */
      imax = 0, amax[1] = FD_0;
  }
  if (d_P(DM_DN(X)) > 1)
    BrdMat(1-DM_Trans(X), d_P0(DM_DN(X)), 1, 3, amax, 3);
  *Max = amax[0], *AuxV = amax[2];
/*printf("AMAX2(%d): %d %s aux=%s\n", N, (int) FD_REAL(amax[1]), pFD2Str(Max), pFD2Str(AuxV));*/
  TRACE_RETURN((int) FD_REAL(amax[1]));         	
} /* AMAX2() */

/*************** replicated diagonal matrix handling routines ***************/

static FDATA *Dii, *Di1i1, *Di1i;
static DistMatStore A00_;
static DistMat A00 = NULL;

static void SetUpDInv(N, A, DInv)
/* post: sets up column-replicated 3-vector DInv[] for storing
 *       D^-1 in LDL^T decomposition, for an N x N matrix A 
 */
  int N; DistMat A; FDATA DInv[];
{ int n = G2Ld(N, DM_DM(A));
  Dii = DInv, Di1i1 = Dii + n, Di1i = Di1i1 + n; 
  A00 = SubMat_(&A00_, A, 0, 0);
} /*SetUpDInv()*/

static void SetDInv1(j, Ajj, D00) 
  int j; DistMat Ajj; FDATA *D00;
/* post: add 1x1 block of D^-1 from column j to DInv[] 
 */  
{ int j_loc = G2Ld(j, DM_DM(A00)); 
  if (InRowScope(Ajj))
    Dii[j_loc] = *D00;
}   

static SetDInv(j, Ajj, Aj1j, D00, D11, D10) 
  int j; DistMat Ajj, Aj1j; FDATA *D00, *D11, *D10;
/* post: add 2x2 block of D^-1 from column j to DInv[] 
 */  
{ int j_loc = G2Ld(j, DM_DM(A00)); 
  if (InRowScope(Ajj) || (Aj1j!=NULL && InRowScope(Aj1j)))
    Dii[j_loc] = *D00, Di1i1[j_loc] = *D11, Di1i[j_loc] = *D10;
}   

static FDATA GetDInv1(j) 
  int j; 
/* returns 1x1 block of D^-1 from column j
 */  
{ int j_loc = G2Ld(j, DM_DM(A00)); 
  return(Dii[j_loc]);
}   

static GetDInv3(j, D00, D11, D10) 
  int j; FDATA *D00, *D11, *D10;
/* 2 x 2 block of D^-1 from columns j & j+1 as (D00, D10; D10, D11)
 */   
{ int j_loc = G2Ld(j, DM_DM(A00)); 
  *D00 = Dii[j_loc], *D11 = Di1i1[j_loc] , *D10 = Di1i[j_loc] ;
}   

static void ApplyD2Inv(D00, D11, D10, n, Wj, Wj1, IncW, Aj, Aj1, IncA)
/* post: (Aj[0..n-1] )  =   (Wj[0..n-1] )  * (D11, -1) / D10 
 *	 (Aj1[0..n-1])      (Wj1[0..n-1])    (-1, D00) 
 * notes: this applies a plane-like rotation; 
 * 	  Aj=Wj and Aj1=Wj1 is allowed
 */
  FDATA *D00, *D11, *D10, Wj[], Wj1[], Aj[], Aj1[]; int n,  IncW, IncA;	   
{
  COMMON_ASSERT_TR( 
    (TraceStr, "(%s,%s,%s,%d,a%d a%d %d,a%d a%d %d)", pFD2Str(D00), 
     pFD2Str(D11), pFD2Str(D10), n, FD_ADR(Wj), FD_ADR(Wj1), IncW,
     FD_ADR(Aj), FD_ADR(Aj1), IncA), 1, 0, "ApplyD2Inv", 6, TRACE_COMPUTE
  );
  for (; n > 0; n--) { 	
    FDATA eAj = *Wj, eAj_,		eAj1 = *Wj1, eAj1_;
    FD_MULSET(eAj_, =, eAj, *D11);	FD_MULSET(eAj1_, =, eAj1, *D00); 
    Wj += IncW,				Wj1 += IncW;
    FD_ASSIGN(eAj_, -=, eAj1);		FD_ASSIGN(eAj1_, -=, eAj);
    FD_MULSET(*Aj, =, eAj_, *D10);	FD_MULSET(*Aj1, =, eAj1_, *D10);
    Aj += IncA,				Aj1 += IncA;	    
  }    
} /*ApplyD2Inv()*/

/*************** pivoting statistics code ***********************************/
#ifdef INERTIA_STATS
static int PosEigCnt = 0, NegEigCnt = 0;/* counts +ve & -ve eigenvalues */
# define INERTIA_STATS_RESET		PosEigCnt = 0, NegEigCnt = 0

# define INERTIA_STATS_ADD(Diag)					\
	{ if (FD_REAL(Diag) > 0.0) PosEigCnt++; 			\
	   else if (FD_REAL(Diag) < 0.0) NegEigCnt++; }
	   
# define INERTIA_STATS_DUMP						\
	printf("	inertia: +ve %d, -ve %d\n", PosEigCnt, NegEigCnt)
#else

# define INERTIA_STATS_RESET

# define INERTIA_STATS_ADD(Diag)

# define INERTIA_STATS_DUMP

#endif


#ifdef PIVOT_STATS

static int RunMax = 0, RunCnt, RunD2Cnt;

static int *RunHist = NULL, 		/* histogram of runs of case D1, */
	   *RunD2Hist = NULL;		/* and D1&D2 combined. (within blocks)*/
static double *WRunHist = NULL;		/* RunHist weighted for the amount of work */

# ifdef LOOKAHEAD
static int *LARunHist = NULL; 		/* histogram of runs of case D1 via lookaeahd */

#   define LARunHist_ALLOC_		1

#   define LA_SetUp_			LARunHist = RunD2Hist + RunMax;
	
#   define LA_Dump_							\
     { int i, n=0, av=0.0; printf("\n	LAD1Runs: ");     		\
       for (i=0; i < RunMax; i++)	{				\
         if (LARunHist[i] > 0)						\
           printf("%d:%d ", i, /*(double) i* */ LARunHist[i] /*/ (double) (N)*/);\
         av += i*LARunHist[i], n+= LARunHist[i];	}		\
       printf("  av: %d %d", av / n, n>LARunHist[0]? av / (n-LARunHist[0]): -1);					\
     }

#   define PIVOT_STATS_LAEnter(w1)					\
      LARunHist[w1]++

#   if defined(LA_STOP_GROWTH_EXCEED)

#   define LARunHist_ALLOC 	LARunHist_ALLOC_
#   define LA_SetUp 		LA_SetUp_
#   define LA_Dump 		LA_Dump_

#   else

static int *LAExGrowth = NULL;

#   define LARunHist_ALLOC 	(LARunHist_ALLOC_+1)
#   define LA_SetUp 		LA_SetUp_; LAExGrowth = LARunHist + RunMax 
#   define LA_Dump 							\
     { int i; LA_Dump_;						 	\
       printf("\n	LAExGr:   ");					\
       for (i=1; i < RunMax; i++)					\
         if (LAExGrowth[i] > 0)						\
           printf("%d:%d ", i, LAExGrowth[i]); 				\
     }      		

#   define PIVOT_STATS_LAExGrowthEnter(dw1)				\
	LAExGrowth[dw1]++	

#   endif /*LA_STOP_GROWTH_EXCEED*/


# else

#   define LARunHist_ALLOC		0

#   define LA_SetUp

#   define LA_Dump

# endif

#define PIVOT_STATS_SetUp(w)						\
  { RunMax = ABS(w)+2;							\
    WRunHist = (double *)						\
    calloc(RunMax*((2+LARunHist_ALLOC)*sizeof(int)+sizeof(double)), 1);	\
    RunHist = (int *) ((double *) WRunHist + RunMax);			\
    RunD2Hist = RunHist +  RunMax;        				\
    LA_SetUp;								\
    RunCnt = RunD2Cnt = 0;						\
  }									

#define PIVOT_STATS_Dump(N)						\
   { int i; int av=0, n=0; 						\
     if (AbsCId==0) {							\
     printf("\n	D1Runs:   ");						\
     for (i=1; i < RunMax; i++)	{					\
       if (RunHist[i] > 0)						\
         printf("%d:%0.2f ", i, (double) i*RunHist[i] / (double) (N));	\
       av += i*RunHist[i], n+= RunHist[i]; }				\
     printf("  av: %d/%d = %d", av, n, av / n);					\
     LA_Dump;								\
     printf("\n	D1D2Runs: ");						\
     for (i=1; i < RunMax; i++)						\
       if (RunD2Hist[i] > 0)						\
         printf("%d:%0.2f ", i, (double) i*RunD2Hist[i] / (double) (N));\
     printf("\n	wD1Runs:  ");						\
     for (i=1; i < RunMax; i++)						\
       if (WRunHist[i] > 0.0)						\
         printf("%d:%0.2f ", i, i*WRunHist[i] / (((double) (N)*(N)+1)/2));\
     printf("\n"); free(RunHist); RunHist = RunD2Hist = NULL, WRunHist = NULL;\
     }									\
     free(WRunHist);							\
   }      

#define PIVOT_STATS_IncRun()						\
    { /*printf(" *%d", RunCnt);*/ RunCnt++, RunD2Cnt++; }	

#define PIVOT_STATS_EndRun(CaseD1, CaseD2, n)				\
    { CommonAssert(RunCnt<RunMax && RunD2Cnt<RunMax, 0, "PIVOT_STATS", 15, TRACE_ALL);\
      /*printf(" (%d,%d)", CaseD2, n);*/				\
      if (CaseD2) RunD2Cnt++;						\
      else RunD2Hist[RunD2Cnt]++, RunD2Cnt=0;				\
      RunHist[RunCnt]++; WRunHist[RunCnt]+=(n), RunCnt=0; 		\
      if (AbsCId==0 &&!CaseD1) printf("%c%d", (CaseD2)? '*': ' ',  n);	\
    }

#else

#define PIVOT_STATS_SetUp(w)

#define PIVOT_STATS_IncRun()

#define PIVOT_STATS_LAEnter(w1) 

#define PIVOT_STATS_EndRun(CaseD1, CaseD2, n)

#define PIVOT_STATS_Dump(N)

#endif /* PIVOT_STATS */

/****************** multiple 1x1 pivots data structures & macros ********/

#if defined(MULT1X1PIVOTS)
# define MAX_1X1PIV 128	
static FD_BASE Mu_k[MAX_1X1PIV];	/* circular queue of recent Mu values */
static int mu_k_hd = 0, mu_k_tl = 0;		

#define ACC_NEW_MU_VAL(Mu, MaxGrowth, val, alpha)			\
	{ (Mu) *= Mu_k[mu_k_tl] = (val);				\
	  mu_k_tl = (mu_k_tl + 1) % MAX_1X1PIV;				\
	  if (mu_k_tl == mu_k_hd) {                       /* circ queue full */\
	    (Mu) /= Mu_k[mu_k_hd], mu_k_hd = (mu_k_hd + 1) % MAX_1X1PIV;\
  	  } else							\
	    (MaxGrowth) *= (1.0 + 1.0/(alpha));				\
	}  

#define RESET_MU_VALS(Mu, MaxGrowth)	                   		\
	 (Mu) = (MaxGrowth) = 1.0, mu_k_hd = mu_k_tl;
	 
#define RE_CALC(Mu)							\
    { int j = mu_k_hd; (Mu) = 1.0;					\
      while (j!=mu_k_tl) {						\
        (Mu) *= Mu_k[j], j = (j + 1) % MAX_1X1PIV;} 			\
    }

#else

#define ACC_NEW_MU_VAL(Mu, MaxGrowth, val, alpha) 			\
	{ (Mu) *= (val);						\
	  (MaxGrowth) *= (1.0 + 1.0/(*alpha));				\
	} 

#define RESET_MU_VALS(Mu, MaxGrowth)	                   		\
	 (Mu) = (MaxGrowth) = 1.0;	 

#define RE_CALC(Mu)							
   
#endif
 

/************************ lookahead code **********************************/
static int D_MultAMAX(N, w, A, Max)
 int N, w; DistMat A; FDATA Max[];
/* pre:  N>= 0; DistMatOK(N, w, A)
 *	 Max[] is a local vector of w elements
 * post: Max[i] = eleemnt chich is abs max. of A[i:N-1,i]
 * note: this is used to obtain the maximums of the next panel of A
 *       for a priori bound calculations
 */
{ int i, w_loc = 0, 			n_loc;
  LocMat lMax;    
  DistMatStore Ai1i_;			DistMat Ai1i = &Ai1i_;
  int CIncA = LM_Inc(DM_Trans(A), DM_Loc(A));

  COMMON_ASSERT_TR_E( 
    (TraceStr, "(%d,%d,%s,_)", N, w, DM_TRACE(0, A)),
     N>=0 && DistMatOK(N, w, A), 0, "MultAMAX", 2, TRACE_OUTER
  );
  
  Ai1i = SubMat_(Ai1i, A, 1, 0);	n_loc = G2Ld(N-1, DM_DM(Ai1i));
  for (i=0; i < w; i++) {
    if (d_P0(DM_DN(Ai1i))==0) {
      int imax = VecAbsMaxPos(n_loc, LM_ADR(DM_Loc(Ai1i)), CIncA);
      Max[w_loc] = *(LM_ADR(DM_Loc(Ai1i)) +  FAST_MUL2(imax, CIncA));  
printf("i=%d, imax=%d Max[%d]=%f\n", i, imax, w_loc, Max[w_loc]);        
      w_loc++;
    }    
    if (d_P0(DM_DM(Ai1i))==0)
      n_loc--;
    Ai1i = SubMat_(Ai1i, Ai1i, 1, 1);
  } /* for (i...) */
  
  if (w_loc > 0 && d_P(DM_DM(A)) > 1) {
    int P = d_P(DM_DM(A)),   		P0 = d_P0(DM_DM(A)),
        DelP0 = 1, lgDelP0=0,		RowCol = DM_Trans(A);
    FDATA *RMax;
    MatMalloc_FT("D_MultAMAX", 1, &RMax, w_loc);
    CommSetUp(d_P(DM_D0(A)), d_P(DM_D1(A)));
    while (DelP0 < P) {
      int SendMsg = (P0 & (2*DelP0-1)) == DelP0;
      int DelP0_ = SendMsg? -DelP0: DelP0;
      int DelIP0 = RowCol==0? DelP0_: 0,  DelJP0 = RowCol==0? 0: DelP0_;
      if (SendMsg) {
        SendMat(DelIP0, DelJP0, 1, w_loc, Max, w_loc, SUMMSGTYPE);
      } else if ((P0 & (2*DelP0-1))==0 && P0+DelP0 < P) {
        RecvMat(DelIP0, DelJP0, 1, w_loc, RMax, w_loc, SUMMSGTYPE);
        for (i=0; i < w_loc; i++) {
printf("%.3f ? %.3f ", RMax[i], Max[i]);        	 
          if (FD_ABS(RMax[i]) > FD_ABS(Max[i]))
            Max[i] = RMax[i];        	
        }
printf("\n");           
      } /* if (SendMsg) */
      DelP0 *= 2, lgDelP0++;
    } /* while (DelP0 ...) */
    MatFree(RMax);
    BrdMat(DM_Trans(A), P0, 1, w_loc, Max, w_loc);
  }

  LM_Init(lMax, Max, 1, 1);		/* horiz. replicate Max[] */
MatPrint("L", 1, w, lMax);
  
  if (w_loc < w) {			/* A[:,0:w-1] over several cell columns*/
    int i_loc = 0;	
    Ai1i = SubMat_(Ai1i, A, 1, 0);
    for (i=0; i < w; i++) {		/* puts this col's Max[] in global order */
      if (DM_DN(Ai1i)==0)	
        Max[i] = Max[i_loc], i_loc++;
      Ai1i = SubMat_(Ai1i, Ai1i, 0, 1);
    }
  } 

  GatherMat(1-DM_Trans(A), DM_DN(A), w, 1, &lMax);
#if 1
  MatPrint("L", 1, w, lMax);
#endif     
  TRACE_EXIT;
} /* D_MultAMAX() */

#ifdef LOOKAHEAD 


static DistMat D_DETF2(N, A)
  int N; DistMat A;
/* pre:  a global N x N symmetric matrix is stored in lower tri. portion of A
 *       DistMatOK(N, N, A), Blk1d(N, DM_DM(A)), 
 * post: A is overwritten by L\D, where L is stored in the proper lower
 *       triagular portion of A, D is stored in the diagonal of A  and:
 *              A_pre = L W^T, LD = W
 *	 returns W if this cell owns a portion of A (ow. returns NULL);
 *	 W is a fully-replicated if !Blk1d(N, DM_DN(A)), and is
 *	 aligned with A otherwise.
 * notes:we will have L[i+1:,i] / L[i,i] = W[i+1:,i], W[i,i] = L[i,i].
 *       If a zero is left in D, then A is not definite, and the
 *	 factorization from that point onwards should be discarded.
 *	 Currently, this procedure assumes Blk1d(N, DM_DM(A)),
 *	 as this is sufficient for D_SYTRF() with lookahead.
 *	 If W != NULL, its storage must bwe released by the caller
 *	 before the next call to D_DETF2().
 */
{ int j;
  LocMat lAj0, lAjj, lWj0, lWjj;
  static DistMatStore W_;
  DistMat W = &W_;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%s)", N, DM_TRACE(0,A)),
    DistMatOK(N, N, A) && Blk1d(N, DM_DM(A)), 0, 
    "DETF2", 0, TRACE_OUTER
  );
 
  if (!Blk1d(N, DM_DN(A))) {
    FDATA *WS;				/* scratch for lAj1j */
    MatMalloc("DETF2", 1, &WS, N); 
    W = Trans(NewMat_(W, N, N, ReduceDimMat_(W, -1, Trans(A)), NULL)); 
    for (j=0; j < N; j++) {
      DistMat_RVAlloc(5);
      DistMat Ajj = SubMat(A, j, j);
      DistMat Aj1j = SubMat(Ajj, 1, 0),	Aj1j1 = SubMat(Aj1j, 0, 1);
      DistMat Wj = SubMat(W, j, j),	Wj1 = SubMat(Wj, 1, 0);
      FDATA Ajje;
      LocMat lWj1T, lAj1j;
      D_COPY(N-j, Ajj, Wj);
      Ajje = DM_Elt00(Wj);
      LM_Copy(1-DM_Trans(Wj), lWj1T, DM_Loc(Wj1));
      if (InColScope(Ajj)) {
        LM_Copy(DM_Trans(Ajj), lAj1j, DM_Loc(Aj1j));
      } else {
        LM_Init(lAj1j, WS, 1, N);
      }
      if (!FD_EQ(Ajje, FD_0)) {
        FD_RECIP(Ajje, Ajje);
	MatScale(N-(j+1), 1, &Ajje, &lAj1j);
      }
      TriMatMult(N-(j+1), G2Ld(N-(j+1), DM_DN(Aj1j1)), 1,
	         &FD_M1, &lAj1j, &lWj1T, &FD_1, Lower, Aj1j1); 
    } /* for (j...) */
    MatFree(WS);
    TRACE_EXIT;
    return(W);
  }
        
  if (!D_CellHolds(N, N, A)) {
    TRACE_EXIT;
    return(NULL);        
  }
       
  W = Trans(NewMat_(W, N, N, Trans(A), NULL)); 
  LM_Copy(DM_Trans(A), lAj0, DM_Loc(A));
  LM_Copy(0, lAjj, lAj0);
  LM_Copy(1-DM_Trans(W), lWj0, DM_Loc(W));
  LM_Copy(DM_Trans(W), lWjj, DM_Loc(W));
  for (j=0; j < N; j++) {  /* use faster left-looking version if A in single cell */
    FDATA Ajj;
    MatPartMult(N-j, 1, j, &FD_M1, &lAj0, &lWj0, &FD_1, &lAjj);
/*MatPrint("Aj0", N-j, j+1, lAj0);*/
    VecCopy(N-j, LM_ADR(lAjj), LM_RInc(lAjj), LM_ADR(lWjj), LM_RInc(lWjj));
    Ajj = *LM_ADR(lAjj);
    pLocSubMat(&lAjj, 1, 0);
    if (!FD_EQ(Ajj, FD_0)) {          	   
      FD_RECIP(Ajj,  Ajj);
      MatScale(N-j-1, 1, &Ajj, &lAjj);
    } 
    pLocSubMat(&lAjj, 0, 1);
    pLocSubMat(&lAj0, 1, 0);
    pLocSubMat(&lWj0, 0, 1);
    pLocSubMat(&lWjj, 1, 1);
  } /* for (j...) */
  TRACE_EXIT;
  return(W);
} /* D_DETF2 () */


#define HUGE_LAMBDADIAG 1.0e+12		/* a huge guard value for LambdaDiag([] */ 

static void  GetDiagRatios(k, N, Lambda, Tk0, LambdaDiag)
  int N; FDATA Lambda[]; DistMat Tk0; FD_BASE LambdaDiag[];
{ int i, j;  
  LocMat lT; 	
#if !defined(QUAD_LOOKAHEAD)
  LM_Copy(DM_Trans(Tk0), lT, DM_Loc(Tk0));
  for (i=0; i < N; i++) {
    FD_BASE Lambda_ = 0.0;
    FDATA *xT = LM_ADR(lT);		int CIncT = LM_CInc(lT);
    for (j=0; j < k; j++) {
      Lambda_ += FD_ABS(Lambda[j])*FD_ABS(*xT);
      xT += CIncT;            
    } 	
    for (j=0; j < i; j++) {
      Lambda_ += LambdaDiag[j]*FD_ABS(*xT);
      xT += CIncT;
    }  
    LambdaDiag[i] = FD_ABS(Lambda[i+k]) + Lambda_;
    pLocSubMat(&lT, 1, 0);
  } /* for (i..) */
#endif  
  LM_Copy(DM_Trans(Tk0), lT, DM_Loc(Tk0));
  for (i=0; i < N; i++) {
    FD_BASE Diag = FD_ABS(*LM_Index(lT, i+k, i));	
    if (Diag > 0.0)
#if defined(QUAD_LOOKAHEAD)
      LambdaDiag[i] = FD_ABS(Lambda[i+k]) / Diag; 
#else
      LambdaDiag[i] /= Diag; 
#endif      
    else
      LambdaDiag[i] = HUGE_LAMBDADIAG;  
  } /* for (i..) */
} /* GetDiagRatios() */ 

#define W_MIN 1 /*16*/	
/* smallest blocking factor for modifed Cholesky decomposition
 * to yield a worthwhile speedup
 */

static int CalcNxtPart(alpha, N, LambdaDiag, Mu, MaxGrowth, kP)
  FD_BASE *alpha, *Mu, *MaxGrowth, LambdaDiag[]; int N, *kP;
/* pre:  LambdaDiag[0..N-1] is ration of the column maximums to the diagonal 
 *	 elements.
 *	 *alpha is the stability constant for the LDLT algorithm; 
 *	 *Mu and *MaxGrowth are the accumulated element growth
 *	 estimate and bounds for the computation previous to 
 *	 the factorization of submatrix A. 
 * post: returns w1, 0 <= w1 <= N, where columns 0..w1-1 can be stably
 *	 eliminated via case D1 of the LDLT algorithm (ie. can be solved
 *	 by a faster, modifed Cholesky decomposition).
 *	 *Mu, *MaxGrowth are updated to reflect processing up to
 *	 column w1 (using an a priori grwoth bound).
 */
{ int w1 = 0, i = 0;		
  FD_BASE lMu, lMaxGrowth, Mu0;
  int FstGrowthExceed = -1;
  i=0, lMu = *Mu, lMaxGrowth = *MaxGrowth, Mu0 = 1.0;  
/*printf("Mu=%.2e MaxGrowth=%.2e\n", lMu, lMaxGrowth);*/
  while (i < N && lMu < HUGE_LAMBDADIAG * lMaxGrowth) {
    /* determine last column whose accumulated growth is within the
       a priori bounds */	
#if defined(QUAD_LOOKAHEAD) 
    lMu *= 1.0 + Mu0 * LambdaDiag[i]; lMaxGrowth *= (1.0 + 1.0 / *alpha);  
    Mu0 *= 1.0 + Mu0 * LambdaDiag[i];   
#else
    lMu *= 1.0 + LambdaDiag[i],	      lMaxGrowth *= (1.0 + 1.0 / *alpha); 
#endif       
/*printf("LambdaDiag[%d]=%.2e Mu0=%.2e Mu=%.2e MaxGrowth=%.2e\n", i, LambdaDiag[i], Mu0, lMu, lMaxGrowth);  */
    if (lMu <= lMaxGrowth)	
      w1 = i+1;
    else
#if defined(LA_STOP_GROWTH_EXCEED)
       break;
#else
       if (FstGrowthExceed < 0) FstGrowthExceed = i+1;       
#endif       
    i++;  
  } /* while (i...) */
/*printf("i=%d lMu=%.2e lMaxGrowth=%.2e\n", i, lMu, lMaxGrowth);*/
  if (w1 < W_MIN)
    w1 = 0, *kP = MIN(N, W_MIN);
  else
    *kP = 1;  
#if !defined(LA_STOP_GROWTH_EXCEED)
  if (w1 > FstGrowthExceed && FstGrowthExceed >=0)
    PIVOT_STATS_LAExGrowthEnter(w1 - FstGrowthExceed);  
#endif 
#if !defined(POST_CALC_GROWTH)
  lMu = *Mu, lMaxGrowth = *MaxGrowth, Mu0 = 1.0; 
  for (i=0; i < w1; i++) {	/* recalc to ensure integrity of Mu_k[] */  	
# if defined(QUAD_LOOKAHEAD)
    ACC_NEW_MU_VAL(lMu, lMaxGrowth, 1.0 + Mu0 * LambdaDiag[i], *alpha);
    Mu0 *= 1.0 + lMu * LambdaDiag[i];   
# else
    ACC_NEW_MU_VAL(lMu, lMaxGrowth, 1.0 + LambdaDiag[i], *alpha);    
# endif    
  }  	
  *Mu = lMu, *MaxGrowth = lMaxGrowth;  
#endif  
  PIVOT_STATS_LAEnter(w1);
  /*printf("w1=%d, Mu=%.2e MaxGrowth=%.2e\n", w1, *Mu, *MaxGrowth);*/
  return(w1);	
} /* CalcNxtPart() */ 

int UseLookAhead = 1;			/* flog to turn on or off lookahead */

#else

#define UseLookAhead 0

#endif
  
/*************** main routines **********************************************/

int Case2Cnt = 0;			/* used to determine # times the 2nd 
  					   case of B-K alg. occurred. */

void D_SYTRF(N, A, DInv, Piv, info, MaxW)
  int N, Piv[], *info, MaxW; DistMat A; FDATA *DInv;
{ int j;
  FD_BASE alpha = (1.0 + sqrt(17.0))/8.0;
#ifdef MULT1X1PIVOTS
# ifdef BOUNDMULT1X1
  FD_BASE Palpha = alpha*alpha*alpha*alpha*alpha; 
  /* A[j,j] >= Palpha lambda[j] to be included in multiple 1x1 pivot */ 
# endif   
  FD_BASE Mu, MaxGrowth;   
  /* estimate & bounds for element growth since last interchange */
  RESET_MU_VALS(Mu, MaxGrowth);		

#endif
  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%s,_,%d)", N, DM_TRACE(0,A), MaxW),
    DistMatOK(N, N, A), 0, "SYTRF", -1, TRACE_OUTER
  );

  PIVOT_STATS_SetUp(MaxW);
  INERTIA_STATS_RESET;
  
  BrdTopChar[DM_Trans(A)] = ' ',        BrdTopChar[1-DM_Trans(A)] = ' ';
 
  *info = 0;
  SetUpDInv(N, A, DInv);
  Case2Cnt = 0; 
  
  j=0;
  while (j < N) {
    DistMat_RVAlloc(5);
    DistMat Ajj = SubMat(A, j, j);
    int minW = MAX(d_B(DM_DN(A)), MAX(MIN_BLOCKING_FACTOR, -MaxW));
    int maxW = MIN(ABS(MaxW), N-j);
    int w = GenGetW(N-(j+1), 1.0, minW, maxW, DM_DN(Ajj), j, 1);
    DistMat W = Trans(NewMat(w+1, N-j, Trans(Ajj), NULL)); 
    int k = 0, n = N-j;
    int LstCaseD3D4 = 0;		/* records if cases D3-4 since last lookahead*/
#if defined(LOOKAHEAD)    
    FDATA *Lambda = NULL;		/* a priori maximums in panel */    
    FD_BASE *
    LambdaDiag = NULL;		/* ratio of diag elements to maximums */
    DistMat T = NULL;			/* replicated tri. factor in current panel */
    DistMat TW = NULL;			/* unscaled tri factor from last ...*/
    int kTW;				/* call DETF2(); kTW records value of k*/			
#endif    
#if defined(RESPREAD_W)
    FDATA *DInvf = NULL, *SDInvf = NULL;/* record fully replicated D^-1 */
    MatMalloc("SYTRF", 1, &DInvf, 2*(w+1)); SDInvf = DInvf + (w+1);
#endif            
#if defined(MULT1X1PIVOTS)
/*if (AbsCId==0)
  printf("recalc: j=%d, Mu=%1.2e ", j, Mu);*/
    RE_CALC(Mu);    			/* avoid accumulated * & / errors */
/*if (AbsCId==0) 
  printf("%1.2e, G=%1.2e\n", Mu, MaxGrowth);*/
#endif

#if defined(LOOKAHEAD)    
    if (UseLookAhead) { /* get maximuns of next w columns for a priori
    			   bounds growth calculations */
      MatMalloc("SYTRF", 1, &Lambda, 2*w);	
      LambdaDiag = (FD_BASE *) (((FDATA *) &Lambda[w])); 
      D_MultAMAX(n, w, Ajj, Lambda);
      
      T = Trans(NewMat(w, w, ReduceDimMat(-1, Trans(Ajj)), NULL));
    }  
#else
     /* temp code to compare growth */	
     { FDATA *Lambda; MatMalloc("SYTRF", 1, &Lambda, w);	
      D_MultAMAX(n, w, Ajj, Lambda);
      MatFree(Lambda);
     }     
#endif

    while (k < w) {
      int BlockT = Blk1d(w, DM_DM(Ajj)) && Blk1d(w, DM_DN(Ajj));    
      DistMat_RVAlloc(8);
      DistMat Ak0 = SubMat(Ajj, k, 0),	Akk = SubMat(Ak0, 0, k);
#if defined(LOOKAHEAD)
      DistMat Wk0 = SubMat(W, k, 0),	
      	      Tk0 = SubMat(T, k, 0),	Tkk = SubMat(Tk0, 0, k),
              Wk0_ = NULL,		/* replicated copy of Wk0 */
              TWkk;			/* replicated tri factor corresp Wkk */
#endif              
      int w1 = 0, kP = 1;		/* w1 is blocking factor for lookahead*/
      int CaseD3D4 = 0; 		/* records if cases D3,D4 have occured*/

#if defined(LOOKAHEAD)  
      if (UseLookAhead) {
        if (k > 0) {			/* make fully replicated copy of Wk0 */
      	  Wk0_ = Trans(NewMat(k, w-k, ReduceDimMat(-1, Trans(Wk0)),
      	     	       NULL));
      	  D_ScaleAdd(w-k, k, FD_M1, Wk0, FD_0, Wk0_);
        }  	
      	if (LstCaseD3D4 || k==0) { 	/* Tkk invalid; needs to be recalc. */
           D_ScaleAdd(w-k, w, FD_1, Ak0, FD_0, Tk0);  
      	   if (k > 0) {			/* apply updates from cols to left */ 	
      	     LM_Trans(DM_Trans(Tk0), DM_Loc(Tk0));
      	     LM_Trans(1-DM_Trans(Wk0_), DM_Loc(Wk0_));
      	     TriMatMult(w-k, w-k, k, &FD_1, DM_pLoc(Tk0), DM_pLoc(Wk0_), 
			&FD_1, Lower, Tkk); 
      	     LM_Trans(1-DM_Trans(Wk0_), DM_Loc(Wk0_));
      	   }
      	   if (TW!=NULL)			
      	     DisposeMat(TW);
           TWkk = TW = D_DETF2(w-k, Tkk); 
           kTW = k;     	   
           DM_PRINT("T", w-k, w-k, Tkk);DM_PRINT("TW", w-k, w-k, TW);    
      	} else 				/* reuse Tkk,TW of last call D_DETF2()*/
      	  TWkk = SubMat(TW, k-kTW, k-kTW);

      	GetDiagRatios(k, w-k, Lambda, Tk0, &LambdaDiag[k]);
        w1 = CalcNxtPart(&alpha, w-k, &LambdaDiag[k], &Mu, &MaxGrowth, &kP);
      }  
      if (w1 > 0) {			/* eliminate next w1 block columns */ 
      	int i;	
      	DistMat_RVAlloc(5);		
        DistMat Akw1k = SubMat(Akk, w1, 0),
        	Akw1j = SubMat(Ajj, k+w1, 0),
		Wkw1k = SubMat(W, k+w1, k);
      	DistMat WkkT = NewMat(w1, w1, Trans(ReduceDimMat(DM_Trans(Ajj), Akk)), 
      			      NULL);	/* col-replicated */
	LocMat lWk0;			
        D_ScaleAdd(w1, w1, FD_1, Tkk, FD_0, Akk); /* copy back to A */
        DM_PRINT("Akw1k", n-(k+w1), w1, Akw1k);

        if (k > 0) {			/* apply updates of 1st k cols in panel*/
	  DistMat_RVAlloc(2);		/* reuse Wk0_ to avoid commun. */
          DistMat cWk0T = NewMat(k, w1, Trans(ReduceDimMat(DM_Trans(Wk0), Wk0)), NULL);
          D_ScaleAdd(w1, k, FD_1, Wk0_, FD_0, Trans(cWk0T));
          /* there may be communication, if !BlockT, of Akw1j */	
          D_GEMM(N-(j+k+w1), w1, k, FD_1, Akw1j, cWk0T, FD_1, Akw1k);
          DisposeMat(cWk0T);
        }
        D_ScaleAdd(w1, w1, FD_1, TWkk, FD_0, WkkT);
        LM_Copy(1-DM_Trans(W), lWk0, DM_Loc(Wkw1k));
        TRSMsBUnScal = &lWk0;		/* save unscaled Akw1k in W */
        TR_M(1, (DM_Trans(WkkT)^DM_Trans(Akw1k))? Lower: Upper, NonUnit, w1, 
              n-(k+w1), &FD_1, WkkT, Trans(Akw1k));
        /* same as the following, but only TR_M() can implement TRSMsBUnScal:
           D_TRSM("Lower", "NonUnit", w1, n-(+k+w1), FD_1, WkkT
      	       Trans(Akw1k));
      	 */      
DM_PRINT("Akw1k", n-(k+w1), w1, Akw1k);
DM_PRINT("Wkw1k", n-(k+w1), w1, Wkw1k);
      	TRSMsBUnScal = NULL;
      	DisposeMat(WkkT);

#ifdef POST_CALC_GROWTH		/* get actual values rather than estimates */
        D_MultAMAX(n-k, w-k, Akk, &Lambda[k]);	  
#endif
        for (i=0; i < w1; i++) {
          DistMat_RVAlloc(1);
          DistMat Akiki = SubMat(Ajj, k, k);	
          FDATA Diag = *LM_Index(DM_Loc(Tkk), i, i);
	  INERTIA_STATS_ADD(Diag);
#ifdef POST_CALC_GROWTH
      	  ACC_NEW_MU_VAL(Mu, MaxGrowth, 1.0 + FD_ABS(Lambda[k]) / FD_ABS(Diag), alpha);
#endif
          FD_RECIP(Diag, Diag);	
	  SetDInv1(j+k, Akiki, &Diag);
          Piv[j+k] = j+k; 			k++;
          PIVOT_STATS_IncRun(); 
        } /* for (i...) */ 
      } /* if (w1 > 0) */   
      if (Wk0_!=NULL)
        DisposeMat(Wk0_);
DM_PRINT("Ak", n-k, 2, Akk);
#endif /* defined(LOOKAHEAD) */

      kP += k;
      while (k < kP && kP <= w) {
      	/* eliminate columns by normal algorithm */
        DistMat_RVAlloc(9);	 
        DistMat Ak0 = SubMat(Ajj, k, 0),	Akk = SubMat(Ak0, 0, k),
      	        Ak1k = SubMat(Akk, 1, 0),	Ak1k1 = SubMat(Ak1k, 0, 1),
      	        Akk1 =  SubMat(Akk, 0, 1),
       	        Wk0 = SubMat(W, k, 0), 		Wkk = SubMat(Wk0, 0, k),
       	        Wk1k = SubMat(Wkk, 1, 0),	Wkk1 = SubMat(Wkk, 0, 1);
        FDATA eWkk, eWik, eWii;
        FD_BASE AbsWkk, AbsWik; 
        int i, dk;
       	
        D_COPY(n-k, Akk, Wkk);
        D_GEMV(n-k, k, FD_M1, Ak0, Trans(Wk0), FD_1, Wkk);
DM_PRINT("Ak", n-k, 2, Akk);
                 	
        if (InEltScope(Wkk))
	  eWkk = DM_Elt00(Wkk);	
        if (j+k+1 < N) {
	  i = k+1 + D_AMAX2(n-(k+1), Wk1k, &eWik, Wkk, &eWkk);
          /* eWkk = D_GetElt(Wkk) */
          AbsWik = FD_ABS(eWik);
        } else {			/* prevent 2x2 pivot on col N-1*/
          AbsWik = 0.0; i = k;
          eWkk = D_GetElt(Wkk);
        }  
        AbsWkk = FD_ABS(eWkk);

#if defined(MULT1X1PIVOTS)
        if (AbsWkk > 0.0) {
      	  ACC_NEW_MU_VAL(Mu, MaxGrowth, 1.0 + AbsWik / AbsWkk, alpha);
        }	
#endif
#ifdef POST_CALC_GROWTH
	FD_SET_REAL(Lambda[k], AbsWik);	FD_SET_IMAG(Lambda[k], 0.0);		
#endif
        dk = 1; Piv[j+k] = j+k;
        if (MAX(AbsWik, AbsWkk) == 0.0) {
       	  /* a zero column; L^-1 does not exist, but LDL^T fact'n can exist */
       	  if (*info==0)
       	    *info = j+k+1;
#if defined(MULT1X1PIVOTS)	/* reset growth estimate and bound */
	  RESET_MU_VALS(Mu, MaxGrowth);
#endif	  
	  SetDInv1(j+k, Akk, &FD_0);	                
#if defined(RESPREAD_W) 
	  DInvf[k] = FD_0;
#endif	     
        } else {	
#if !defined(MULT1X1PIVOTS)
          if (AbsWkk >= alpha * AbsWik) { 
            PIVOT_STATS_IncRun(); 
          }  
#else 
# ifndef BOUNDMULT1X1
	  if (Mu <= MaxGrowth || i==k) 
# else
	  if ((Mu<=MaxGrowth && AbsWkk >= Palpha * AbsWik) || i==k)
# endif
	  {
            /* no pivoting needed */
/*if (AbsCId==0)       	  
printf("j=%d k=%d i=%d D1: AbsWkk=%1.2e AbsWik=%1.2e Mu=%1.2e G=%1.2e\n", j, k, i, AbsWik, AbsWkk, Mu, MaxGrowth);*/
            PIVOT_STATS_IncRun(); 
          }  
#endif          
	  else {
	    DistMat_RVAlloc(6);	
	    DistMat Wi0 = SubMat(W, i, 0), 	Wik1 = SubMat(Wi0, 0, k+1), 	
	  	    Ai0 = SubMat(Ajj, i, 0),
	  	    Aik = SubMat(Ai0, 0, k),	Aii = SubMat(Ai0, 0, i),
	  	    Aik1 = SubMat(Aik, 0, 1);	
	    FDATA eWij; FD_BASE AbsWij; int j_;

	    /* copy column i to W[:,k+1] and update it to determine pivot;
	       note that A[i,i] is placed at  W[i,k+1] */	
	    D_COPY(i-k, Trans(Aik), Wkk1);

	    D_COPY(n-i, Aii, Wik1);
	    D_GEMV(n-k, k, FD_M1, Ak0, Trans(Wi0), FD_1, Wkk1);

	    if (InEltScope(Wik1))
	      eWii = DM_Elt00(Wik1);	
	    D_SetElt(Wik1, FD_0);		/* exclude from D_AMAX() */
	    j_ = i + D_AMAX2(n-(k+1), Wkk1, &eWij, Wik1, &eWii);

	    AbsWij = FD_ABS(eWij);
	    D_SetElt(Wik1, eWii); 

	    if (AbsWkk >= alpha * AbsWik * (AbsWik / AbsWij)) 
	      /* 2x2 pivot may be unstable; may as well pivot from row k  */
	      Case2Cnt++;
	    else if (FD_ABS(eWii) >= alpha * AbsWij) {
	      /* 1x1 pivot from row i will be stable */ 	
	      D_COPY(n-k, Wkk1, Wkk);
	     
	      D_COPY(i-k, Akk, Trans(Aik));	
	      D_COPY(n-i, Aik, Aii);
	     
	      /* merge these 2 calls to reduce startups */
	      D_SWAP2(k+1, Trans(Ak0), Trans(Wk0), Trans(Ai0), Trans(Wi0));

	      eWkk = eWii;
	      Piv[j+k] = j+i;
	      CaseD3D4 = 1;
	    } else {
	      /* 2x2 pivot will be stable */	
	      dk = 2;	
	      if (i!=k+1) { /* interchange cols k+1 and i */
	        DistMat_RVAlloc(2);
	        DistMat Ak10 = SubMat(Ak0, 1, 0), Wk10 = SubMat(Wk0, 1, 0);
	        FDATA eAkk1;
	      
	        if (InEltScope(Akk1)) { /* A[k,k+1] := A[k,k+1]; simply next copy */
	          eAkk1 = *LM_ADR(DM_Loc(Akk1));
	          *LM_ADR(DM_Loc(Akk1)) = *LM_ADR(DM_Loc(Wkk1));
	        }  	
	        D_COPY(i-k, Akk1, Trans(Aik));
	        if (InEltScope(Akk1))  /* restore A[k,k+1] to orig. value */
	          *LM_ADR(DM_Loc(Akk1)) = eAkk1;
	        D_COPY(n-i, Aik1, Aii);
	     
                D_SWAP2(k+2, Trans(Ak10), Trans(Wk10), Trans(Ai0), Trans(Wi0));
	      }

	      Piv[j+k+1] = Piv[j+k] = -(j+i);
	      CaseD3D4 = 1;	    
	    }
	    PIVOT_STATS_EndRun(0, Piv[j+k]==j+k, N-(j+k));
/*if (AbsCId==0)       	  
printf("j=%d k=%d p=%d i=%d AbsWkk=%1.2e AbsWik=%1.2e Mu=%1.2e G=%1.2e\n", j, k, Piv[j+k], i, AbsWik, AbsWkk, Mu, MaxGrowth);*/
#if defined(MULT1X1PIVOTS)
	    /* case D1 not applied; reset growth estimate and bound */
	    RESET_MU_VALS(Mu, MaxGrowth);
#endif	  
	  }

	  INERTIA_STATS_ADD(eWkk);
          if (dk==2) INERTIA_STATS_ADD(eWii);

	  if (dk==1) {
	    /* D = W[k;k]; update A[k+1:N-1, k] by D^-1 */	
	    FDATA InvWkk;		FD_RECIP(InvWkk, eWkk);	
	    D_COPY(n-k, Wkk, Akk);
	    D_SCAL(n-(k+1), InvWkk, Ak1k);
	    SetDInv1(j+k, Akk, &InvWkk);
#if defined(RESPREAD_W) 
	    DInvf[k] = InvWkk;
#endif	     
	    DM_PRINT("Ak", n-k, 2, Akk);
	  } else {
	    /* D = W[k:k+1, k:k+1]; update A[k+2:N-1, k:k+1] by D^-1 */	
	    /* note: now eWkk = W[k,k], eWii = W[k+1,k+1], eWik = W[k+1,k] */
	    DistMat_RVAlloc(4);
	    DistMat Wk2k = SubMat(Wkk, 2, 0),	Wk2k1 = SubMat(Wk2k, 0, 1),
	            Ak2k = SubMat(Akk, 2, 0),	Ak2k1 = SubMat(Ak2k, 0, 1);
	    FDATA D10 = eWik, D00, D11, T = FD_M1; /* explicit inv of 2x2 pivot */
	    FD_DVDSET(D00, eWii, D10);	FD_DVDSET(D11, eWkk, D10); 	
	    FD_MULSET(T, +=, D00, D11);	FD_RECIP(T, T);
	    FD_DVDSET(D10, T, D10);

	    SetDInv(j+k, Akk, Ak1k, &D11, &D00, &D10);
#if defined(RESPREAD_W) 
	    DInvf[k] = D11, DInvf[k+1] = D00, SDInvf[k] = D10;
#endif	     
	  
            if (d_P0(DM_DN(Ak2k))==d_P0(DM_DN(Ak2k1))) {
	      /* can optimize: 2 columns on the same cell column */
              if (d_P0(DM_DN(Ak2k1))==0) 
	        ApplyD2Inv(&D11, &D00, &D10, G2Ld(n-k-2, DM_DM(Ak2k)), 
	                   LM_ADR(DM_Loc(Wk2k)), LM_ADR(DM_Loc(Wk2k1)), 
	                   LM_Inc(DM_Trans(Wk2k), DM_Loc(Wk2k)),	   
	                   LM_ADR(DM_Loc(Ak2k)), LM_ADR(DM_Loc(Ak2k1)), 
	                   LM_Inc(DM_Trans(Ak2k), DM_Loc(Ak2k)));

	    } else {	  
	      D_COPY(n-k-2, Wk2k, Ak2k);D_COPY(n-k-2, Wk2k1, Ak2k1);
	      D_ScaleAdd(n-k-2, 1, FD_M1, Wk2k1, D00, Ak2k);
	      D_ScaleAdd(n-k-2, 1, FD_M1, Wk2k,  D11, Ak2k1);
	      D_SCAL(n-k-2, D10, Ak2k);	D_SCAL(n-k-2, D10, Ak2k1);
	    }
	   
	    D_SetElt(Akk, eWkk);	D_SetElt(Ak1k1, eWii);
	    D_SetElt(Ak1k, eWik);
	  
	    DM_PRINT("Ak", n-k, 2, Akk);
	   }
         }                    
         k += dk;
#if 0
         if (CaseD3D4)			/* do not try lookahead again in this panel */
           kP = k + 1;	              
#endif           
      } /* while (k<kP) */
      LstCaseD3D4 = CaseD3D4;
    } /* while (k<w) */

    DM_PRINT("Ajj", n, w, Ajj);	      	DM_PRINT("W", n, w, W);	
#if defined(LOOKAHEAD)
    if (TW != NULL) 
      DisposeMat(TW);
    if (T != NULL) 
      DisposeMat(T);
    if (Lambda!=NULL) 
      MatFree(Lambda);
#endif
    PIVOT_STATS_EndRun(1, 0, N-(j+k));
    w = k;   
    if (n-k > 0) {
      /* update trailing tri sub-matrix Ajkjk -= Ajkj Wk0^T */	
      DistMat_RVAlloc(4);
      DistMat Ajkj = SubMat(Ajj, k, 0),		
              Ajkjk = SubMat(Ajkj, 0, k), 	
 	      Wk0 = SubMat(W, k, 0),		sWk0;
 	           	
      LocMat sAk, sWk;				int ms, ns;
      FDATA ScW = FD_M1;
       
      LM_ADR(sAk) = NULL;
#ifndef RESPREAD_W
      ms = SpreadMat(1, k, n-k, Ajkj, &FD_1, DM_DM(Ajkjk), &sAk, 0, 
      		     n-k, DM_DN(Ajkjk));
      		     
      LM_ADR(sWk) = NULL;
      ns = SpreadMat(0, k, n-k, Trans(Wk0), &ScW, DM_DN(Ajkjk), &sWk, 0, 
      		     n-k, DM_DM(Ajkjk));
      	
#else
      ms = SpreadMat(1, k, n-k, Wk0, &FD_1, DM_DM(Ajkjk), &sAk, 0, 
      		     n-k, DM_DN(Ajkjk));
      sWk0 = Spread2DistMat(DM_Trans(Wk0), 0, DM_DM(Wk0), &sAk);
      LM_ADR(sWk) = NULL;
      ns = SpreadMat(0, k, n-k, Trans(sWk0), &ScW, DM_DN(Ajkjk), &sWk, ms>0, 
      		     n-k, DM_DM(Ajkjk));
      		    /* here, if a workspace was created for sAk,
      		       sWk needs a new workspace, as we may later modify sAk */
      if (ms < 0 || (d_P0(DM_DN(Ajkj))==0 && Blk1d(w, DM_DN(Ajkj)))) {
      	/* this cell column was the sole source; there is no need to recompute 
      	   Ajkj; if ms<0 and ns<0, recomputing sAk will lead to an error */
        if (ms > 0 && ns >= 0) 
          MatFree(LM_ADR(sAk));
        LM_Copy(DM_Trans(Ajkj), sAk, DM_Loc(Ajkj));
        ms = -ABS(ms);			/* prevent refreeing */
      }	else {     /* apply sAk <- sAk * D^-1 */		
        int i=0;			int Neg = FD_EQ(ScW, FD_M1);
        LocMat sAki;			LM_Copy(0, sAki, sAk);
        while (i < w) {
	  if (Piv[j+i] >= 0) {
	    if (Neg) FD_NEG(DInvf[i], DInvf[i]);	
	    MatScale(ABS(ms), 1, &DInvf[i], &sAki);      
	    i++;			pLocSubMat(&sAki, 0, 1);
	  } else {
	    FDATA *psAki = LM_ADR(sAki), *psAki1 = psAki + LM_CInc(sAki);
	    int Inc = LM_RInc(sAki);	
	    if (Neg) FD_NEG(SDInvf[i], SDInvf[i]);	
	    ApplyD2Inv(&DInvf[i], &DInvf[i+1], &SDInvf[i], 
	    	       ABS(ms), psAki, psAki1, Inc, psAki, psAki1, Inc);
	    i+=2;			pLocSubMat(&sAki, 0, 2);	
	  }  	
        } /* while (i < w) */
        if (Neg) ScW = FD_1;		/* have applied scaling by -1 */
      }  
#endif /*RESPREAD_W*/      		     
      TriMatPartMult(ABS(ms), ABS(ns), k, &ScW, &sAk, &sWk, &FD_1, Lower, Ajkjk);
/*MatPrint("A'", ABS(ms), ABS(ns), DM_Loc(Ajkjk));      */
      if (ns > 0)
        MatFree(LM_ADR(sWk));
      if (ms > 0)
        MatFree(LM_ADR(sAk));
    }

#ifdef RESPREAD_W    
    MatFree(DInvf);
#endif 
    DisposeMat(W);  
    
    /* complete row interchanges on A */
    D_BKSWP("Fwd", j, w, Trans(SubMat(A, j, 0)), &Piv[j], j);
    j += w;        
  } /* while (j...) */

  PIVOT_STATS_Dump(N);
  INERTIA_STATS_DUMP;
  TRACE_EXIT;   
} /* D_SYTRF() */ 

static FDATA *SubDiag = NULL;

static ClrSubDiag(N, A, Piv)
  int N, Piv[]; DistMat A;
{ int n = G2Ld(N, DM_DM(A));
  int j, Cpy = SubDiag==NULL;
  if (Cpy)
    MatMalloc("CpSubDiag", 1, &SubDiag, n); 
  j=0;
  while (j < N) {
    if (Piv[j] < 0) {
      DistMat_RVAlloc(1);
      DistMat Aj1j = SubMat(A, j+1, j);
      if (InEltScope(Aj1j)) {
        if (Cpy) {
 	  int j_loc = G2Ld(j+1, DM_DM(A));
	  SubDiag[j_loc] = DM_Elt00(Aj1j);
        }
        DM_Elt00(Aj1j) = FD_0;
      }
      j++;
    } 
    j++;
  } /* while (j...) */
} /* CpSubDiag() */

static ResetSubDiag(N, A, Piv)
  int N, Piv[]; DistMat A;
{ int j;
  j=0;
  while (j < N) {
    if (Piv[j] < 0) {
      DistMat_RVAlloc(1);
      DistMat Aj1j = SubMat(A, j+1, j);
      if (InEltScope(Aj1j)) {
        int j_loc = G2Ld(j+1, DM_DM(A));
        DM_Elt00(Aj1j) = SubDiag[j_loc];
      }  
      j++;
    }
    j++;
  } /* while (j...) */
} /*ResetSubDiag()*/


void D_SYTRS(N, K, A, DInv, Piv, B, info)
  int K, N;
  DistMat A, B; FDATA *DInv; int Piv[], *info; 
{ int j;	

  COMMON_ASSERT_TR_E(
     (TraceStr, "(%d,%d,%s,_,%s_)", N, K,  DM_TRACE(0, A), DM_TRACE(1, B)),
      DistMatOK(N, N, A) && DistMatOK(N, K, B), 0, "SYTRS", -1, TRACE_OUTER
  );

  D_BKSWP("Fwd", K, N, Trans(B), Piv, 0);

  ClrSubDiag(N, A, Piv);  
  D_TRSM("Lower", "Unit", N, K, FD_1, A, B);

  j=0;
  while (j < N) {
    DistMat_RVAlloc(4);
    DistMat Bj = SubMat(B, j, 0),	Bj1 = SubMat(Bj, 1, 0),
            Ajj = SubMat(A, j, j),	Aj1j = SubMat(Ajj, 1, 0);

    if (Piv[j] >= 0) {			/* 1 x 1 pivot */
      DistMat_RVAlloc(2);
      if (InRowScope(Ajj)) {		/* multiply by inverse of 1x1 block */
        FDATA eAjj; 
	eAjj = GetDInv1(j);
        D_SCAL(K, eAjj, Trans(Bj));	
      }
      j++;
    } else {				/* 2 x 2 pivot */
      DistMat_RVAlloc(8);
      DistMat Bj2 = SubMat(Bj1, 1, 0), 
              Aj2j = SubMat(Aj1j, 1, 0), 	Aj1j1 = SubMat(Aj1j, 0, 1),
              Aj2j1 = SubMat(Aj2j, 0, 1);
      int Pj = -Piv[j];

      if (d_P0(DM_DM(Bj))==d_P0(DM_DM(Bj1))) {
      	/* can optimize: 2 columns on the same cell row */
      	if (d_P0(DM_DM(Bj))==0) {
      	  FDATA D00, D10, D11;
      	  FDATA *pBj = LM_ADR(DM_Loc(Bj)), 	*pBj1 = LM_ADR(DM_Loc(Bj1));
      	  int CIncB = LM_Inc(1-DM_Trans(Bj), DM_Loc(Bj));
	  GetDInv3(j, &D00, &D11, &D10);	
	  ApplyD2Inv(&D00, &D11, &D10, G2Ld(K, DM_DN(Bj)), 
	  	     pBj, pBj1, CIncB, pBj, pBj1, CIncB);
        }	  	     
      } else {
      	DistMat Bj_ = NewMat(1, K, Bj, NULL);
      	FDATA D00, D10, D11, T;

	GetDInv3(j, &D00, &D11, &D10);

	D_COPY(K, Trans(Bj), Trans(Bj_));      	
      	D_ScaleAdd(1, K, FD_M1, Bj1, D11, Bj);
      	D_ScaleAdd(1, K, FD_M1, Bj_, D00, Bj1);
      	D_SCAL(K, D10, Trans(Bj));		D_SCAL(K, D10, Trans(Bj1));
      	DisposeMat(Bj_);
      }	    
      j+=2;    	
    }
  } /* while (j...) */

  D_TRSM("Upper", "Unit", N, K, FD_1, Trans(A), B);
  ResetSubDiag(N, A, Piv);  
  MatFree(SubDiag);				SubDiag = NULL;

  D_BKSWP("Bwd", K, N, Trans(B), Piv, 0);
  TRACE_EXIT;
} /* D_SYTRS() */



