/************************************************************************/
/* 	LDLT: indefinite factorization of a global NxN Matrix A		*/
/************************************************************************/
/* SCCS INFO @(#)LDLT.new.c 1.7         last modified P. Strazdins, 00/10/18
   first created: P. Strazdins 00/05
*/

/* reduced startup version supporting algorithmic blocking (#ifndef BLOCK_SEARCH)
 * based on bounded Bunch-Kaufman algorithm; stability of all variants
 * is proven.
 * (externally-defined) cpp FLAGs:
 *      PRINTMAT        print out intermediate matrices
 *	PIVOT_STATS	take detailed statistics on distribution of pivots
 *      RIGHT_LOOKING   level 2 factorization is right-looking on 1st w cols of W
 *      BLOCK_SEARCH    perform exhaustive search in remainder of
 *                      current block for a stable pivot
 *	BS_LIMIT_MOVEMENT limit the number of interchanges for any column
 *			in the block search
 */
/*#define PIVOT_STATS*/
#define BLOCK_SEARCH
/*#define RIGHT_LOOKING*/
/*#define BS_LIMIT_MOVEMENT*/

#ifdef BLOCK_SEARCH
# define RIGHT_LOOKING  /* necessary for impl.; ow. inherently inefficient */

# define BS_SearchSB 4 	/* block column searches in units lg_2(P)*BS_SearchSB
			   to reduce commun. startups */ 
# define BS_WIDTH 16	/* max block search width */
# define M2x2	4	/* search for 2x2 pivots for M2X2+1 cols before
			   searching for any 1x1 pivots */

# ifdef BS_LIMIT_MOVEMENT
#   define BS_MAX_COL_SEARCHES 8 /* max number of times a column in any
			            block will be searched */
#   define BS_MAX_1X1	2	 /* ... be moved by a 1x1 interchange 
			   	    searching for any M1x1 pivots */
# endif
#endif

#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#ifdef ASSERT
# define PRINTMAT
#endif
#include "GetBlockFactor.h"
#include "MatDim.h"
#include "DistMat.h"
#include "D_BLAS.h"

#include "SpreadMatrix.h"
#include "CellTriMatrix.h"
#include "BLASOptions.h"
#include "FDMalloc.h"
#include "FDCommun.h"
extern int AbsCId; /* FDCommun.h */
#include "CellMatrix.h"
#include "CellVector.h"

#include "D_LAPACK.h"

double sqrt STDC_PARAMS( (double) );

/******* merged communication routines *************************************/

static void D_BKSWP(Direction, M, N, A, P, P0)
  char *Direction; int M, N; DistMat A; int P[], P0;
{ int j, dj, Lj, *LaPiv;

  COMMON_ASSERT_TR_E(  
   (TraceStr, "(%c,%dx%d,%s,a%d %d)", *Direction, M, N, DM_TRACE(0,A),
   ((int) P)/sizeof(int), P0), P!=NULL && DistMatOK(M, N, A), 0, "BKSWP", 0,
   TRACE_OUTER);
  
   if (M<=0) {
     TRACE_RETURN;
   }

   LaPiv = (int *) malloc(N*sizeof(int));
   j=0;
   while (j<N) {
     if (P[j] < 0) 
       LaPiv[j] = -P[j] - P0 /*, LaPiv[j+1] = -P[j] - P0, j++*/;
     else
       LaPiv[j] = P[j] - P0;
     j++;
   }
   D_LASWP(Direction, M, N, A, LaPiv);     
   free((char *) LaPiv);
   TRACE_EXIT;
} /*D_BKSWP*/ 


static void D_SWAP2(MA, MW, Aj, Wj, Ai, Wi)
  int MA, MW; DistMat Aj, Wj, Ai, Wi;
/* pre: DM_DM(Aj) = DM_DM(Wj) = DM_DM(Ai) = DM_DM(Wi);  
 *      DM_DN(Aj) = DM_DN(Wj);  DM_DN(Ai) = DM_DN(Wi)
 */
{ CommSetUp(d_P(DM_D0(Aj)), d_P(DM_D1(Aj)));  
  if (InColScope(Aj) == InColScope(Ai)) {	
    /* no perf. advantage possible */
    D_SWAP(MA, Aj, Ai); 		D_SWAP(MW, Wj, Wi);
  } else {  	
    int ma = G2Ld(MA, DM_DM(Aj)),	mw = G2Ld(MW, DM_DM(Wj));
    FDATA *AW, *pA, *pW; 		int IncA, IncW;
    int DI = 0,        			DJ = -d_P0(DM_DN(Ai))-d_P0(DM_DN(Aj));
    if (InColScope(Ai))
      IncA = LM_Inc(DM_Trans(Ai), DM_Loc(Ai)), 
      IncW = LM_Inc(DM_Trans(Wi), DM_Loc(Wi)),
      pA = LM_ADR(DM_Loc(Ai)), 		pW = LM_ADR(DM_Loc(Wi));
    else
      IncA = LM_Inc(DM_Trans(Aj), DM_Loc(Aj)), 
      IncW = LM_Inc(DM_Trans(Wj), DM_Loc(Wj)),
      pA = LM_ADR(DM_Loc(Aj)), 		pW = LM_ADR(DM_Loc(Wj));
    if (DM_Trans(Aj))
      SWAP(int, DI, DJ);
    if (ma + mw > 0) {  
      MatMalloc_FT("D_SWAP2", 1, &AW, ma+mw);
      VecCopy(ma, pA, IncA, AW, 1);	VecCopy(mw, pW, IncW, AW+ma, 1);
      SendMat(DI, DJ, 1, ma+mw, AW, ma+mw, GENMSGTTYPE);  
      RecvMat(DI, DJ, 1, ma+mw, AW, ma+mw, GENMSGTTYPE);  
      VecCopy(ma, AW, 1, pA, IncA);	VecCopy(mw, AW+ma, 1, pW, IncW);
      MatFree(AW);
    }  
  }     
} /* D_SWAP2() */

static int D_AMAX2(N, X, Max, AuxX, AuxV)
 int N; DistMat X, AuxX; FDATA *Max, *AuxV;
/* pre:  N>= 0; DistMatOK(N, 1, X); DM_DN(X) = DM_DN(AuxX)
 *	 if InEltScope(AuxX), v = *AuxV
 * post: as fpr D_AXAX(N, X, Max), and *AuxV = v 
 * note: the motivation of this function is combine the broadcast of v
 *       in the broadcasts within the D_AMAX() computation.
 */
{ FDATA amax[3]; 	
  CommSetUp(d_P(DM_D0(X)), d_P(DM_D1(X)));  
  COMMON_ASSERT_TR_E( 
    (TraceStr, "(%d,%s,_%s,_)", N, DM_TRACE(0, X), DM_TRACE(1, AuxX)),
     N>=0 && DistMatOK(N, 1, X) && Dim_Eq(DM_DN(X), DM_DN(AuxX)),
     0, "AMAX2", 1, TRACE_OUTER
  );
  
  if (InEltScope(AuxX)) 
    amax[2] = *AuxV;
  if (d_P0(DM_DN(X))==0) {
    int nx = G2Ld(N, DM_DM(X));         int IncX = LM_Inc(DM_Trans(X), DM_Loc(X));
    int imax = VecAbsMaxPos(nx, LM_ADR(DM_Loc(X)), IncX);
    if (imax==-1) {                     /* assert nx = 0 */
      amax[0] = FD_0;                   imax = 0;
    } else
      amax[0] = *(LM_ADR(DM_Loc(X)) + FAST_MUL2(imax, IncX));

    imax = L2Gd(imax, DM_DM(X));        /* convert to a global index */
    FD_SET_REAL(amax[1], (FD_BASE) imax);
    if (d_P(DM_DM(X)) > 1) {
      int P = d_P(DM_DM(X)),   		P0 = d_P0(DM_DM(AuxX)),
          DelP0 = 1, lgDelP0=0,		RowCol = DM_Trans(X);
      while (DelP0 < P) {
	FDATA RMxI[2];
        int SendMsg = (P0 & (2*DelP0-1)) == DelP0;
        int DelP0_ = SendMsg? -DelP0: DelP0;
        int DelIP0 = RowCol==0? DelP0_: 0,  DelJP0 = RowCol==0? 0: DelP0_;
        if (SendMsg) {
          SendMat(DelIP0, DelJP0, 1, 2, amax, 2, SUMMSGTYPE);
        } else if ((P0 & (2*DelP0-1))==0 && P0+DelP0 < P) {
          RecvMat(DelIP0, DelJP0, 1, 2, RMxI, 2, SUMMSGTYPE);
        if (FD_ABS(RMxI[0]) > FD_ABS(amax[0]))
          amax[1] = RMxI[1], amax[0] = RMxI[0];
        } /* if (SendMsg) */
        DelP0 *= 2, lgDelP0++;
      } /* while (DelP0 ...) */
      BrdMat(DM_Trans(X), P0, 1, 3, amax, 3);
    }
    if (FD_EQ(amax[0], FD_0))           /* take top element as being the max */
      imax = 0, amax[1] = FD_0;
  }
  if (d_P(DM_DN(X)) > 1)
    BrdMat(1-DM_Trans(X), d_P0(DM_DN(X)), 1, 3, amax, 3);
  *Max = amax[0], *AuxV = amax[2];
  TRACE_RETURN((int) FD_REAL(amax[1]));         	
} /* AMAX2() */

/*************** replicteed diagonal matrix handling routines ***************/

static FDATA *Dii, *Di1i1, *Di1i;
static DistMatStore A00_;
static DistMat A00 = NULL;
static int Piv2x2at0 = 0;	/* set if there is a 2x2 pivot in col 0;
				   needed at the Pivot vector convention cannot
				   distinguish this */
static SetUpDInv(N, A, DInv)
  int N; DistMat A; FDATA DInv[];
{ int n = G2Ld(N, DM_DM(A));
  Dii = DInv, Di1i1 = Dii + n, Di1i = Di1i1 + n; 
  A00 = SubMat_(&A00_, A, 0, 0);
}

static SetDInv1(j, Ajj, D00) 
  int j; DistMat Ajj; FDATA *D00;
{ int j_loc = G2Ld(j, DM_DM(A00)); 
  if (InRowScope(Ajj))
    Dii[j_loc] = *D00, Di1i[j_loc] = FD_0;
}   

static SetDInv(j, Ajj, Aj1j, D00, D11, D10) 
  int j; DistMat Ajj, Aj1j; FDATA *D00, *D11, *D10;
{ int j_loc = G2Ld(j, DM_DM(A00)); 
  if (InRowScope(Ajj) || (Aj1j!=NULL && InRowScope(Aj1j)))
    Dii[j_loc] = *D00, Di1i1[j_loc] = *D11, Di1i[j_loc] = *D10;
}   

static FDATA GetDInv1(j) 
  int j; 
{ int j_loc = G2Ld(j, DM_DM(A00)); 
  return(Dii[j_loc]);
}   

static GetDInv3(j, D00, D11, D10) 
  int j; FDATA *D00, *D11, *D10;
{ int j_loc = G2Ld(j, DM_DM(A00)); 
  *D00 = Dii[j_loc], *D11 = Di1i1[j_loc] , *D10 = Di1i[j_loc] ;
}   

static void ApplyD2Inv(D00, D11, D10, n, Wj, Wj1, IncW, Aj, Aj1, IncA)
/* post: (Aj[0..n-1] )  =   (Wj[0..n-1] )  * (D11, -1) / D10
 *       (Aj1[0..n-1])      (Wj1[0..n-1])    (-1, D00)
 * notes: this applies a plane-like rotation;
 *        Aj=Wj and Aj1=Wj1 is allowed
 */
  FDATA *D00, *D11, *D10, Wj[], Wj1[], Aj[], Aj1[]; int n,  IncW, IncA;
{  
  COMMON_ASSERT_TR(
    (TraceStr, "(%s,%s,%s,%d,a%d a%d %d,a%d a%d %d)", pFD2Str(D00),
     pFD2Str(D11), pFD2Str(D10), n, FD_ADR(Wj), FD_ADR(Wj1), IncW,
     FD_ADR(Aj), FD_ADR(Aj1), IncA), 1, 0, "ApplyD2Inv", 6, TRACE_COMPUTE
  );
  for (; n > 0; n--) {
    FDATA eAj = *Wj, eAj_,              eAj1 = *Wj1, eAj1_;
    FD_MULSET(eAj_, =, eAj, *D11);      FD_MULSET(eAj1_, =, eAj1, *D00);
    Wj += IncW,                         Wj1 += IncW;
    FD_ASSIGN(eAj_, -=, eAj1);          FD_ASSIGN(eAj1_, -=, eAj);
    FD_MULSET(*Aj, =, eAj_, *D10);      FD_MULSET(*Aj1, =, eAj1_, *D10);
    Aj += IncA,                         Aj1 += IncA;
  }
} /*ApplyD2Inv()*/
 
      

/*************** pivoting statistics code ***********************************/
#ifdef PIVOT_STATS

#define BS_MaxW 128
static int BS_SearchHist[2][BS_MaxW], BS_w, BS_1x1Piv, BS_2x2Piv, BS_fail;

static void BS_Stats_init(int w) 
{ int i;
  for (i=0; i < w+1; i++)
    BS_SearchHist[0][i]=BS_SearchHist[1][i]=0;
  BS_w = w+1; BS_1x1Piv = BS_2x2Piv = 0;  BS_fail = 0;
}

static void BS_Stats_update(int dK, int nsearches)
{ if (dK==0) { 
    BS_SearchHist[0][nsearches]++; 
    BS_fail++;
  } else {
    if (dK==1) BS_1x1Piv++; else  BS_2x2Piv++;
    BS_SearchHist[1][nsearches]++;
  }
}  

static void BS_Stats_dump()
{ int i;
  printf("Blocksearch: found %d (%d,%d), histogram: ", 
          BS_1x1Piv+BS_2x2Piv, BS_1x1Piv, BS_2x2Piv);
  for (i=0; i < BS_w; i++)
    if (BS_SearchHist[1][i]) printf("%d:%d ", i, BS_SearchHist[1][i]);
  printf("\n");
  printf("Blocksearch: not found %d, histogram: ", BS_fail);
  for (i=0; i < BS_w; i++)
    if (BS_SearchHist[0][i]) printf("%d:%d ", i, BS_SearchHist[0][i]);
  printf("\n");
}


#define PIVOT_STATS_SetUp(w)	BS_Stats_init(w)

#define PIVOT_STATS_Update(dK, nsearches)	BS_Stats_update(dK, nsearches)

#define PIVOT_STATS_Dump(N) 						\
   { int i; extern int AbsCId; /* FDCommun.h */				\
     if (AbsCId==0) BS_Stats_dump(); }
   
#else

#define PIVOT_STATS_SetUp(w)

#define PIVOT_STATS_Update(dK, nsearches)

#define PIVOT_STATS_Dump(N)

#endif /* PIVOT_STATS */

/******* block search procedures ****************************/
static FD_BASE alpha;

int NColSearches = 0;

#ifdef BLOCK_SEARCH

static int D_MultAMAX(N, w, A, Max)
 int N, w; DistMat A; FDATA Max[];
/* pre:  N>= 0; DistMatOK(N, w, A); Blk1d(w, DM_DN(A), InColScope(A)
 *       Max[] is a local vector of w elements
 * post: in cell holding 1st block of A, 
 *	 Max[i] = element which is abs max. of A[0:N-1,i]
 */
{ int i,		     		n_loc = G2Ld(N, DM_DM(A));
  int RIncA = LM_Inc(  DM_Trans(A), DM_Loc(A)),
      CIncA = LM_Inc(1-DM_Trans(A), DM_Loc(A));
  FDATA *Ax = &DM_Elt00(A);       

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%s,_)", N, w, DM_TRACE(0, A)),
     N>=0 && DistMatOK(N, w, A) && Blk1d(w, DM_DN(A)) && InColScope(A), 
     0, "MultAMAX", 2, TRACE_OUTER
  );
      
  for (i=0; i < w; i++) {		/* put local max. of col i in Max[i] */ 
    int imax = VecAbsMaxPos(n_loc, Ax, RIncA);
    Max[i] = n_loc==0? FD_0: *(Ax + FAST_MUL2(imax, RIncA));
    Ax += CIncA;
  } /* for (i...) */
      
  if (d_P(DM_DM(A)) > 1) {
    int P = d_P(DM_DM(A)),              P0 = d_P0(DM_DM(A)),
        DelP0 = 1, lgDelP0=0,           RowCol = DM_Trans(A);
    FDATA *RMax;
    MatMalloc_FT("D_MultAMAX", 1, &RMax, w);
    CommSetUp(d_P(DM_D0(A)), d_P(DM_D1(A)));
    while (DelP0 < P) {
      int SendMsg = (P0 & (2*DelP0-1)) == DelP0;
      int DelP0_ = SendMsg? -DelP0: DelP0;
      int DelIP0 = RowCol==0? DelP0_: 0,  DelJP0 = RowCol==0? 0: DelP0_;
      if (SendMsg) {
        SendMat(DelIP0, DelJP0, 1, w, Max, w, SUMMSGTYPE);
      } else if ((P0 & (2*DelP0-1))==0 && P0+DelP0 < P) {
        RecvMat(DelIP0, DelJP0, 1, w, RMax, w, SUMMSGTYPE);
        for (i=0; i < w; i++) 
          if (FD_ABS(RMax[i]) > FD_ABS(Max[i]))
            Max[i] = RMax[i];
      } /* if (SendMsg) */
      DelP0 *= 2, lgDelP0++;
    } /* while (DelP0 ...) */
    MatFree(RMax);
  }
  TRACE_EXIT;
} /* D_MultAMAX() */


int SearchSubBlkSize = BS_SearchSB;	/* width for sub-block max col search */

#ifdef BS_LIMIT_MOVEMENT
# define BS_MAXW  128                    /* these are used to limit number of col
                                           searches on each column */
static int ColSearches[BS_MAXW],        /* number of col searches performed */
           Col1x1Pivs[BS_MAXW];         /* # of 1x1 pivots from this column */

# define BlockSearchSetup(w)             \
        {int i; for (i=0; i < (w); i++) ColSearches[i] = Col1x1Pivs[i] = 0;}

#else

#define BlockSearchSetup(w)

#endif


int BlockSearch(int n, int w, DistMat W, FDATA *eW00, FDATA *eW11, 
		FDATA *eW10, int k0, int *pi, int *pk, 
		int *w_s, DistMat Al, DistMat Wl)
/* pre:  W is a n x w matrix; Blkd(w, DM_DN(W))
 *	 i' is position of max in W[0:n-1,0], 
 *	 *eW00 = W[0,0], *eW10 = W[i',0] 
 * inv:	 0 <= w_s <= w; *w_s columns of W have been formed;
 *	 column i can be formed from Al[k0+i:n-i,0..k0+i](Wl[k0+i,0..k0+i])^T+
 *				     Al[k0+i:n-i,k0+i]	 
 * post: performs exhaustive search for stable pivot in current block of W
 *	 returns dK, having values:
 *		0: no stable pivot was found; 
 *		   W, *pi, *eW00, *eW10, and *eW11 are unchanged
 *		1: a 1x1 pivot was found at i, 0 <= i < w
 *		   perform a transpose & column interchange 0<->i in W  
 *		2; a 2x2 pivot was found at i,k , 0 <= i < k < w
 *		   perform transpose & column interchanges 0<->i 1<->k in W  
 * the interchanges are full except row swaps (to be performed after in caller);
 * if dK, after these row swaps we will have:
 *			*eW00 = W[0,0], *eW10=W[1,0]; if DK==2: eW11=W[1,1]
 * where i = *pi - k0, k = *k - k0
 */
{ int i, k; int dK = 0, nsearches=0; 
  int CIncW = LM_Inc(1-DM_Trans(W), DM_Loc(W)), 
      RIncW = LM_Inc(  DM_Trans(W), DM_Loc(W));

#define BS_ResultSz 8
  FDATA results[BS_ResultSz];
#define BS_SetResults							\
  { FD_SET_REAL(results[0], (FD_BASE) dK);				\
    FD_SET_REAL(results[6], (FD_BASE) nsearches);			\
    FD_SET_REAL(results[7], (FD_BASE) *w_s);				\
    if (dK) FD_SET_REAL(results[1], (FD_BASE) *pi),			\
      FD_SET_REAL(results[2], (FD_BASE) *pk), 				\
      results[3] = *eW00, results[4] = *eW10, results[5] = *eW11; } 
#define BS_GetResults   						\
  { dK = (int) FD_REAL(results[0]);					\
    nsearches = (int) FD_REAL(results[6]);				\
    *w_s = (int) FD_REAL(results[7]);					\
    if (dK) { *pi = (int) FD_REAL(results[1]),				\
      *pk = (int) FD_REAL(results[2]),					\
      *eW00 = results[3], *eW10 = results[4], *eW11 = results[5];} }

  if (w <= 0)				/* quick exit */
    return(0);

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%s,_,_,_,%d,%d,_,%d,_,_)", n, w, DM_TRACE(0, W), k0, *pi, *w_s),
     DistMatOK(n, w, W) && Blk1d(w,DM_DN(W)) && k0 <= d_B0(DM_DN(W)), 0, "BlockSearch", 1, TRACE_OUTER
  );

  if (InColScope(W)) {
    /* only cell column involved in search need participate until result known*/
    int LenGamma=1;
    FDATA *GammaElt, *Diag; FD_BASE *Gamma; 

#ifdef BS_LIMIT_MOVEMENT
    int RejectSearch = (ColSearches[k0] > BS_MAX_COL_SEARCHES ||
			Col1x1Pivs[k0] > BS_MAX_1X1);
#else
    int RejectSearch = 0;
#endif
   
/*    if (RejectSearch) printf("%d%c ", k0, Col1x1Pivs[k0] > BS_MAX_1X1? '*': ' ');*/

    MatMalloc("BlockSearch", 1, &GammaElt, 3*w);
    Diag = &GammaElt[w], Gamma = (FD_BASE *) (&Diag[w]);
    GammaElt[0] = *eW10, Diag[0] = *eW00, Gamma[0] = FD_ABS(*eW10); 

    if (w <= 1)	 /*search will fail (dK=0) as k loop not entered */
      BS_SetResults; 		/* thus must set result vector now */ 

    for (k = 1; k < w && !RejectSearch; k++) {
      /* try to find pivot with column k: any column to left (or itself) */
      /* Here, maximums are got in chunks of SearchSubBlkSize to amortize 
         start up costs;  tradeoff is in extra column scans if max found early*/
      DistMat_RVAlloc(3); 	
      DistMat W0k = SubMat(W, 0, k),	Wkk = SubMat(W0k, k, 0),
	      Wk0 = SubMat(W, k, 0); 		

      if (LenGamma <= k) {		/* compute max in next sub-block */
        int GetGamma = MIN(w-k, SearchSubBlkSize);
        int ds = (k+GetGamma) - *w_s;	/* ds cols need to be brought up-to-date*/
        if (ds > 0) {			/* bring all required cols up-to-date*/
	  int s = *w_s;
          DistMat_RVAlloc(4);
	  DistMat Wss = SubMat(W, s, s),	Ass = SubMat(Al, s, k0+s),
		  As0 = SubMat(Al, s, 0),	Ws0 = SubMat(Wl, s, 0);
	  D_ScaleAdd(n-s, ds, FD_1, Ass, FD_0, Wss);
          D_GEMM(n-s, ds, k0, -1.0, As0, Trans(Ws0), 1.0, Wss);

	  CommSetUp(d_P(DM_D0(W)), d_P(DM_D1(W))); 
	  /* need to restore grid size from above DBLAS call with collapsed 
	     matrix dimensions  for next BrdMat() */	
          *w_s += ds;
        }

        if (InRowScope(W)) {
          /* copy rows into tops of columns to simplify max finding  */
 	  FDATA *pWi0 = &DM_Elt00(Wk0),	*pW0i = &DM_Elt00(W0k),
		*pWii = &DM_Elt00(Wkk); 
          for (i=k; i < k+GetGamma; i++) {
            VecCopy(i, pWi0, CIncW, pW0i, RIncW);          
            Diag[i] = *pWii, *pWii = FD_0;/* exclude diag from maximum */
            pWi0 += RIncW, pW0i += CIncW, pWii += RIncW+CIncW;
          }
	}
        /* do block half-reduce into cell holding diagonal blocks */
        D_MultAMAX(n, GetGamma, W0k, &GammaElt[k]);

        if (InRowScope(W)) {		/* restore diagonals */
 	  FDATA *pWii = &DM_Elt00(Wkk);
          for (i=k; i < k+GetGamma; i++) 
            *pWii = Diag[i], pWii += RIncW+CIncW; 
        }
	LenGamma += GetGamma;
      }
/*printf("k=%d, w=%d, LenGamma=%d\n", k, w, LenGamma);*/

      if (InRowScope(W) && !dK) {	/* perform search for col k */
	FDATA *pWki = &DM_Elt00(Wk0);	
        FDATA  eWkk = Diag[k];		FD_BASE AbsWkk = FD_ABS(eWkk);
        Gamma[k] = FD_ABS(GammaElt[k]);
    
        for (i=0; i < k-1 && !dK; i++) {
          /* try to find a suitable 2x2 pivot */ 	
	  FDATA eWki = *pWki;
          FD_BASE AbsWki = FD_ABS(eWki),	AbsWii = FD_ABS(Diag[i]);
          if (MAX(AbsWkk*Gamma[i] + AbsWki*Gamma[k], 	
	          AbsWii*Gamma[k] + AbsWki*Gamma[i])*alpha <
	      FD_ABS(eWkk*Diag[i] - eWki*eWki)) 
	    dK = 2, *pi = i+k0, *pk = k+k0, 
            *eW00 = Diag[i], *eW10 = eWki, *eW11 = Diag[k];
          pWki+=CIncW;	      
        } /* for (i...) */

        if (!dK && k==M2x2) 		/* now search for 1x1 pivots */
          for (i=1; i<=k && !dK; i++) 
            if (FD_ABS(Diag[i]) > alpha * Gamma[i])	
	      dK = 1, *pi = i+k0, *eW00 = Diag[i];
        if (!dK && k>M2x2 && AbsWkk > alpha * Gamma[k])
          /* AbsWkk > alpha * Gamma[k] ensures that search will fail on a zero column*/
          dK = 1, *pi = k+k0, *eW00 = eWkk;

        nsearches++;
      } /* if (InRowScope(W) && !dK) */

      if (k+1==LenGamma) {
        /* b/c result to cell column */
        if (InRowScope(W)) 
          BS_SetResults;
        BrdMat(DM_Trans(W), d_P0(DM_DM(W)), 1, BS_ResultSz, results, BS_ResultSz);
        if (!InRowScope(W))
          BS_GetResults;
        if (dK) break;
      }
    } /* for (k...) */

    MatFree(GammaElt);

#ifdef BS_LIMIT_MOVEMENT
    if (dK==1) k = *pi - k0; else if (dK==2) k = *pk - k0;
    for (; k>0; k--)			/* record which cols max were searched on*/
      ColSearches[k0+k]++;		
    if (dK==1) 
      Col1x1Pivs[*pi]++, ColSearches[*pi] = ColSearches[k0];
    if (dK==2) 
      ColSearches[*pi] = ColSearches[k0], ColSearches[*pk] = ColSearches[k0+1];
#endif

    if (dK) { 
      /* perform interchanges except for row swaps; this requires:
	 W[0,0] = W_pre[i,0], W[i,0] = W_pre[i,i]; 
         W[1,1] = W_pre[k,1], W[k,1] = W_pre[k,k]
	 if dK=2, W[0,1] = W_pre[k,0] = W[i,1]
       */  
      i = *pi - k0;				k = *pk - k0;
      if (i!=0) {	
	DistMat_RVAlloc(2);	
	DistMat Wi0 = SubMat(W, i, 0),		Wii = SubMat(Wi0, 0, i);
        /*D_SWAP(i, W, Trans(Wi0));*/ 
        if (InRowScope(W))
 	  VecSwap(i, &DM_Elt00(W), RIncW, &DM_Elt00(Wi0), CIncW);  
        D_SWAP(n-i, Wi0, Wii);
      }
      if (dK==2 && k!=1) {
        /* note that by now, W[k,0] = W_pre[k,i], so that after row swap 1<->k,
	   W[1,0] = W_pre[k,i] */
	DistMat_RVAlloc(3);			DistMat W11 = SubMat(W, 1, 1);
	DistMat Wk1 = SubMat(W, k, 1),		Wkk = SubMat(Wk1, 0, k-1);
        /*D_SWAP(k-1, W11, Trans(Wk1));*/
        if (InRowScope(W))
 	  VecSwap(k-1, &DM_Elt00(W11), RIncW, &DM_Elt00(Wk1), CIncW);  
        D_SWAP(n-k, Wk1, Wkk);        
        if (InRowScope(W) && i!=0) 	
	  /* W[0,1] = W[i,1]; needed for compatability with BBK implementation*/
          *(LM_ADR(DM_Loc(W))+CIncW) = *(LM_ADR(DM_Loc(W11))+(i-1)*RIncW); 
      }
      DM_PRINT("W", n, w, W);
    } /* if (dK) */    
  } /* if (InColScope(W)) */

  /* overall result is now b/c across cell rows */

  BrdMat(1-DM_Trans(W), d_P0(DM_DN(W)), 1, BS_ResultSz, results, BS_ResultSz);
  if (!InColScope(W)) 
    BS_GetResults;

  PIVOT_STATS_Update(dK, nsearches);
  NColSearches += nsearches;
#if 0
printf("found: k0=%d, dK=%d, pi=%d, pk=%d, w_s=%d %f %f %f\n", k0, dK, *pi, *pk, *w_s, *eW00, *eW10, *eW11);
#endif
  TRACE_RETURN(dK);
} /* BlockSearch() */

#endif /* defined(BLOCK_SEARCH) */
/* pivoting conventions:
   note that algorithm ensures |Piv[k]| >= k (searches are always forward)
   Piv[k],Piv[k+1] < 0:	a 2x2 pivot at k, with k,k+1 <-> |Piv[k]|,|Piv[k+1]|
   Piv[k] >= 0: a 1x1 pivot at k, with k <-> |Piv[k]|
	in either case: k <-> |Piv[k]| - useful for BKSWAP()
	ambiguity: Piv[0]=0: is this a 1x1 pivot, or a 2x2 pivot? 
		* can resolve as a 2x2 if an odd # of -ve values after 
		* inside alg, must have a dK value 
*/

/*************** main routines **********************************************/
int Case2Cnt = 0; 			/* unused; for backward compatibility */

void D_SYTRF(N, A, DInv, Piv, info, MaxW)
  int N, Piv[], *info, MaxW; DistMat A; FDATA *DInv;
{ int j; 
  char svBrdTopChar[2];
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%s,_,%d)", N, DM_TRACE(0,A), MaxW),
    DistMatOK(N, N, A), 0, "SYTRF", -1, TRACE_OUTER
  );

  PIVOT_STATS_SetUp(MaxW);
  svBrdTopChar[0] = BrdTopChar[0],      svBrdTopChar[1] = BrdTopChar[1];  
  BrdTopChar[DM_Trans(A)] = ' ',        BrdTopChar[1-DM_Trans(A)] = ' ';
 
  Case2Cnt = NColSearches = 0;
  *info = 0;
  alpha = 0.5; 		/* 0.5 is optimal in terms of growth in L for BBK */
  SetUpDInv(N, A, DInv);
  
  j=0;
  while (j < N) {
    DistMat_RVAlloc(5);
    DistMat Ajj = SubMat(A, j, j);
    int minW = MAX(d_B(DM_DN(A)), MAX(MIN_BLOCKING_FACTOR, -MaxW));
    int maxW = MIN(ABS(MaxW), N-j);
    int w = GenGetW(N-(j+1), 1.0, minW, maxW, DM_DN(Ajj), j, 1);
    DistMat W = Trans(NewMat(w+1, N-j, Trans(Ajj), NULL)); 
    int k = 0, n = N-j;
#if defined(RIGHT_LOOKING) && !defined(BLOCK_SEARCH)
    int w_s = w;			/* # of cols of W currently up-to-date*/
#else
    int w_s = 0;			/* for left-looking, w_s=k at loop start*/
#endif

    D_ScaleAdd(n, w_s, FD_1, Ajj, FD_0, W);  	/* initialize W; 
						   note: col w not maintained */
    if (w + d_B0(DM_DN(W)) > d_B(DM_DN(W))  && w <= d_B(DM_DN(W)))
      /* correct to keep within a storage block */
      w = d_B(DM_DN(W)) - d_B0(DM_DN(W));
#ifdef BLOCK_SEARCH
    BlockSearchSetup(w);
#endif

    while (k < w) {
      DistMat_RVAlloc(9);	 
      DistMat Ak0 = SubMat(Ajj, k, 0),		Akk = SubMat(Ak0, 0, k),
      	      Ak1k = SubMat(Akk, 1, 0),		Ak1k1 = SubMat(Ak1k, 0, 1),
      	      Akk1 =  SubMat(Akk, 0, 1),
       	      Wk0 = SubMat(W, k, 0), 		Wkk = SubMat(Wk0, 0, k),
       	      Wk1k = SubMat(Wkk, 1, 0),		Wkk1 = SubMat(Wkk, 0, 1);
      int dK=1, i=k, r;			/* for 1x1 (dK=1): k<->i, Piv[j+k]=j+i 
					   for 2x2 (dK=2): k<->i',j-Piv[j+k]=i';
					   	     k+1<->i, j-Piv[j+k+1]=i */
      FDATA eWkk, eWik, eWii;		/* copies taken before interchanges */
					/* interchanges then effectively applied */
      FD_BASE AbsWkk, AbsWik; 		/* stores |Wkk|, |Wik| */


      Piv[j+k] = j+k; 

      if (w_s <= k) {
        D_COPY(n-k, Akk, Wkk);
        D_GEMV(n-k, k, FD_M1, Ak0, Trans(Wk0), FD_1, Wkk);
        w_s++;
	if (w_s != k+1) { printf("w+s=%d too small, j=%d, k=%d\n", w_s, j, k);
			  exit(2); }
      }
      DM_PRINT("W", n, w_s, W);

      if (j+k+1 < N) {
        if (InEltScope(Wkk))
	  eWkk = DM_Elt00(Wkk);	
	i = k+1 + D_AMAX2(n-(k+1), Wk1k, &eWik, Wkk, &eWkk);
        /* assert eWkk == D_GetElt(Wkk) */
        AbsWik = FD_ABS(eWik);
	NColSearches++;
      } else {					/* prevent 2x2 pivot col N-1*/
        AbsWik = 0.0;
        eWkk = D_GetElt(Wkk);
      }  
      AbsWkk = FD_ABS(eWkk);

      dK = 1; 

      if (MAX(AbsWik, AbsWkk) == 0.0) {
       	/* a zero column; L^-1 does not exist, but LDL^T fact'n can exist */
       	if (*info==0)
       	  *info = j+k+1;
        i = k;
      } else {
        if (AbsWkk >= alpha * AbsWik) {
          /* no pivoting needed */
          i = k;
        } else {
#ifdef BLOCK_SEARCH
          int w_sk = w_s - k;		
	  dK = BlockSearch(n-k, MIN(MIN(w-k, BS_WIDTH), 
			            d_B(DM_DN(Wkk))-d_B0(DM_DN(Wkk))-k), 
			   Wkk, &eWkk, &eWii, &eWik, k, &i, &r, 
			   &w_sk, Ak0, Wk0);	
          w_s = w_sk + k;
#else
	  dK = 0;
#endif
	}

	if (dK > 0) {
	  Piv[j+k] = j+i;
	  if (dK==2)	/* a 2x2 pivot was found by BlockSearch() */ 
	    Piv[j+k] = -Piv[j+k], Piv[j+k+1] = -(j+r), i=r;
	  /* note that Blocksearch() has performed the full-interchanges */
	} else {
	  /* use bounded Bunch-Kaufman to find a pivot outside the block */
	  DistMat_RVAlloc(2);	
	  DistMat Wi = Trans(NewMat(1, n-k, Trans(Wkk), NULL));
	  int NSearches=0;
#ifdef RIGHT_LOOKING    /* keep Wk in case i is in current block */
          FDATA eWkk_orig = eWkk;
          DistMat Wk = Trans(NewMat(1, n-(k+1), Trans(Wk1k), NULL));
          D_COPY(n-(k+1), Wk1k, Wk);
#endif
	  while (!dK) {
	    /* perform half-interchanges until a suitable pivot is found */
            DistMat_RVAlloc(6);
	    DistMat Wi_i = SubMat(Wi, i-k, 0), 	Wi0 = SubMat(W, i, 0), 	
	  	    Aik = SubMat(Ajj, i, k),	Aii = SubMat(Ajj, i, i);
            DistMat Wii = SubMat(Wi0, 0, i),    Wik = SubMat(Wi0, 0, k);
	    FDATA eWir; FD_BASE AbsWir; 
	    /* note: if BlockSearch() has failed, i should be outside
		     the current block */		
	    /* copy column i to Wi and update it to determine pivot;
	       note that A[i,i] is placed at Wi[i], instead of Wi[0] */	

            if (i >= w_s) {		/* outside up-to-date cols in block */
	      D_COPY(i-k, Trans(Aik), Wi);
	      D_COPY(n-i, Aii, Wi_i);
	      D_GEMV(n-k, k, FD_M1, Ak0, Trans(Wi0), FD_1, Wi);
	      Case2Cnt++;
            }              
#ifdef RIGHT_LOOKING
            else { 			/* need not re-compute Wi */
	      DistMat_RVAlloc(1);
              D_COPY(i-k, Trans(Wik), Wi);
              if (NSearches > 0)
	      /* in the above, D_GetElt(Wik)!=D_GetElt(SubMat(Wk, i-k-1, 0))
		 as was W[k:n-1,k] was over-written at end of last itn */  
	        D_COPY(1, Trans(SubMat(Wk, i-k-1, 0)), Wi);
              D_COPY(n-i, Wii, Wi_i);
            }
#endif
	    if (InEltScope(Wi_i))
	      eWii = DM_Elt00(Wi_i);	
	    r = k + D_AMAX2(n-k, Wi, &eWir, Wi_i, &eWii);
	    /* note: in Wi, elements 0 and i-k are interchanged */
	    AbsWir = FD_ABS(eWir);
#if 0
printf("j=%d, n=%d, w=%d, w_s=%d, k=%d; dK=%d i=%d, r=%d; aWik=%f aWir=%f diff=%G\n", 
	j, n, w, w_s, k, dK, i, r, AbsWik, AbsWir, AbsWik-AbsWir);
#endif
#define THRESH 0.99
	    if (r==k || AbsWik >= THRESH*AbsWir) {
	      /* eWik is max of current cols k and i; will be a stable 2x2 */
	      /* assert AbsWik==AbsWir => r==k */
#ifdef RIGHT_LOOKING
              { int i = Piv[j+k]-j;
                if (i!=k) { /* complete interchange in W (except row swap) */
                  DistMat_RVAlloc(4);
                  DistMat Wii = SubMat(W, i, i),  Wik1 = SubMat(W, i, k+1);
                  D_COPY(MIN(i, w)-(k+1), Wk, Trans(Wik1));
                  if (i < w_s) {
                    D_COPY(n-(i+1), SubMat(Wk, i-k, 0), SubMat(Wii, 1, 0));
                    D_SetElt(Wii, eWkk_orig);
                  }
                }
              }
              if (i!=k+1) {                     /* copy col k+1 into i in W */
                DistMat_RVAlloc(3);             /* overlap here for W[k+1,k+1]*/
                D_COPY(MIN(i-k, w_s-(k+1)), SubMat(Wkk1, 1, 0),
                       Trans(SubMat(Wik, 0, 1)));  
                if (i < w_s)
                  D_COPY(n-i, SubMat(Wkk1, i-k, 0), Wii);
              }
#endif
	      D_COPY(n-k, Wi, Wkk1);
	      Piv[j+k] = -Piv[j+k], Piv[j+k+1] = -(j+i), dK = 2; 
	      Case2Cnt--;
	    } else {
	      /* column i replaces current column k */
              Piv[j+k] = j+i;		
              if (FD_ABS(eWii) >= alpha * AbsWir) {
	        /* column i is a stable 1x1 pivot */
	        /* as 0 < alpha <= 1.0, r = i implies this */ 
	        dK = 1;
#ifdef RIGHT_LOOKING
                if (i!=k) { /* complete interchange in W (except row swap) */
                  DistMat_RVAlloc(3);
                  D_COPY(MIN(i, w_s)-(k+1), Wk, Trans(SubMat(Wik, 0, 1)));
                  if (i < w_s) {
                    D_COPY(n-(i+1), SubMat(Wk, i-k, 0), SubMat(Wii, 1, 0));
                    D_SetElt(Wii, eWkk_orig);
                  }
                }
#endif
 	      } else
	        /* repeat search assuming interchange from column i */
	        /* assert r != i,k; thus r must be a valid index  */
	        i = r;
	      D_COPY(n-k, Wi, Wkk);
	      eWkk = eWii, eWik = eWir, AbsWik = AbsWir;
	    }
	    NSearches++;
          } /* while */
	  DisposeMat(Wi); 
#ifdef RIGHT_LOOKING
          DisposeMat(Wk);
#endif   
        if (dK==2 && w_s==k+1) w_s++;		/* BBK has added a col to W */
	NColSearches += NSearches;

        } /* else try BBK search */ 

        if (ABS(Piv[j+k]) != j+k) {
	  int i = ABS(Piv[j+k]) - j;		/* note: must be local */
	  DistMat_RVAlloc(4);
	  DistMat Ai0 = SubMat(Ajj, i, 0),	Aik = SubMat(Ai0, 0, k),	
		  Wi0 = SubMat(W, i, 0), 	Aii = SubMat(Ai0, 0, i);
	  /* complete interchanges for k to i in A */ 
	  D_COPY(i-k, Akk, Trans(Aik));
	  D_COPY(n-i, Aik, Aii);

	  /* merge these 2 row swaps in A and W to reduce startups */
	  D_SWAP2(k+1, k+dK, Trans(Ak0), Trans(Wk0), Trans(Ai0), Trans(Wi0));
        }

        if (j+k==0) {
          if (dK==2) 
            printf("warning degenerate 2x2 pivot repn 0,1 <-> 0,%d\n", Piv[j+k+1]); 
	  Piv2x2at0 = (dK==2);
        }

        
        if (dK==2 && i!=k+1) { 
	  DistMat_RVAlloc(7);
	  DistMat Ai0 = SubMat(Ajj, i, 0),	Aik = SubMat(Ai0, 0, k),	
		  Wi0 = SubMat(W, i, 0), 	Aii = SubMat(Ai0, 0, i),
		  Aik1 = SubMat(Aik, 0, 1);
	  DistMat Ak10 = SubMat(Ak0, 1, 0),	Wk10 = SubMat(Wk0, 1, 0);
	  FDATA eAkk1;

  	  if (InEltScope(Akk1)) { /* set A[k,k+1] to A[k+1,k] for 2x2 pivot */
	    eAkk1 = *LM_ADR(DM_Loc(Akk1));
	    *LM_ADR(DM_Loc(Akk1)) = *LM_ADR(DM_Loc(Wkk1));
	  }  	
	      
	  D_COPY(i-k, Akk1, Trans(Aik));
	  D_COPY(n-i, Aik1, Aii);
	     
          D_SWAP2(k+2, k+2, Trans(Ak10), Trans(Wk10), Trans(Ai0), Trans(Wi0));
	  
	  if (InEltScope(Akk1))  /* restore A[k,k+1] to orig value */
	    *LM_ADR(DM_Loc(Akk1)) = eAkk1;
	} 

	if (dK==1) {
	  /* D = W[k;k]; update A[k+1:N-1, k] by D^-1 */	
	  D_COPY(n-k, Wkk, Akk);
	  if (/*k < n-1*/ 1) {
	    FDATA InvWkk;		FD_RECIP(InvWkk, eWkk);	
	    D_SCAL(n-(k+1), InvWkk, Ak1k);
	    SetDInv1(j+k, Akk, &InvWkk);
	  }  	
	} else {
	  /* D = W[k:k+1, k:k+1]; update A[k+2:N-1, k:k+1] by D^-1 */	
	  /* note: now eWkk = W[k,k], eWii = W[k+1,k+1], eWik = W[k+1,k] */
	  DistMat_RVAlloc(4);
	  DistMat Wk2k = SubMat(Wkk, 2, 0),	Wk2k1 = SubMat(Wk2k, 0, 1),
	          Ak2k = SubMat(Akk, 2, 0),	Ak2k1 = SubMat(Ak2k, 0, 1);
	  FDATA D10 = eWik, D00, D11, T = FD_M1; /* explicit inv of 2x2 pivot */
	  FD_DVDSET(D00, eWii, D10);		FD_DVDSET(D11, eWkk, D10); 	
	  FD_MULSET(T, +=, D00, D11);		FD_RECIP(T, T);
	  FD_DVDSET(D10, T, D10);

	  SetDInv(j+k, Akk, Ak1k, &D11, &D00, &D10);
	  
          if (d_P0(DM_DN(Ak2k))==d_P0(DM_DN(Ak2k1))) {
	    /* can optimize: 2 columns on the same cell column */
            if (d_P0(DM_DN(Ak2k1))==0) {
                ApplyD2Inv(&D11, &D00, &D10, G2Ld(n-k-2, DM_DM(Ak2k)),
                           LM_ADR(DM_Loc(Wk2k)), LM_ADR(DM_Loc(Wk2k1)),
                           LM_Inc(DM_Trans(Wk2k), DM_Loc(Wk2k)),
                           LM_ADR(DM_Loc(Ak2k)), LM_ADR(DM_Loc(Ak2k1)),
                           LM_Inc(DM_Trans(Ak2k), DM_Loc(Ak2k)));
	    }  	
	  } else {	  
	    D_COPY(n-k-2, Wk2k, Ak2k);	   	D_COPY(n-k-2, Wk2k1, Ak2k1);
	    D_ScaleAdd(n-k-2, 1, FD_M1, Wk2k1, D00, Ak2k);
	    D_ScaleAdd(n-k-2, 1, FD_M1, Wk2k,  D11, Ak2k1);
	    D_SCAL(n-k-2, D10, Ak2k);	  	D_SCAL(n-k-2, D10, Ak2k1);
	  }
	   
	  D_SetElt(Akk, eWkk);			D_SetElt(Ak1k1, eWii);
	  D_SetElt(Ak1k, eWik);
	  
	  DM_PRINT("Ak", n-k, 2, Akk);
	} /* else (dk==2) */
#ifdef RIGHT_LOOKING
        if (k+dK < w_s) {  
          DistMat_RVAlloc(3);
          D_GEMM(n-(k+dK), w_s-(k+dK), dK, -1.0, SubMat(Ajj, k+dK, k),
                 Trans(SubMat(W, k+dK, k)), 1.0, SubMat(W, k+dK, k+dK));
        }
#endif
      } /* else: non-zero column */
      k += dK;	              
    } /* while (k..) */	

    w = k;   
    if (n-k > 0) {
      /* update trailing tri sub-matrix Ajkjk -= Ajkj Wk0^T */	
      DistMat_RVAlloc(3);
      DistMat Ajkj = SubMat(Ajj, k, 0),		
              Ajkjk = SubMat(Ajkj, 0, k), 	
 	      Wk0 = SubMat(W, k, 0);     	
      LocMat sAk, sWk;				int ms, ns;
      FDATA ScW = FD_M1;
       
      LM_ADR(sAk) = NULL;
      ms = SpreadMat(1, k, n-k, Ajkj, &FD_1, DM_DM(Ajkjk), &sAk, 0, 
      		     n-k, DM_DN(Ajkjk));
      LM_ADR(sWk) = NULL;
      ns = SpreadMat(0, k, n-k, Trans(Wk0), &ScW, DM_DN(Ajkjk), &sWk, 0, 
      		     n-k, DM_DM(Ajkjk));
      TriMatPartMult(ABS(ms), ABS(ns), k, &ScW, &sAk, &sWk, &FD_1, Lower, Ajkjk);
      if (ns > 0)
        MatFree(LM_ADR(sWk));
      if (ms > 0)
        MatFree(LM_ADR(sAk));
    }
    DisposeMat(W);  
    /* complete row interchanges on A */
    D_BKSWP("Fwd", j, w, Trans(SubMat(A, j, 0)), &Piv[j], j);
    j += w;        
  } /* while (j...) */

  PIVOT_STATS_Dump(N);
  BrdTopChar[0] = svBrdTopChar[0],      BrdTopChar[1] = svBrdTopChar[1];
  TRACE_EXIT;   
}/* D_SYTRF() */ 

static FDATA *SubDiag = NULL;

static ClrSubDiag(N, A, Piv)
  int N, Piv[]; DistMat A;
{ int n = G2Ld(N, DM_DM(A));
  int j, Cpy = SubDiag==NULL;
  if (Cpy)
    MatMalloc("CpSubDiag", 1, &SubDiag, n); 
  j=0;
  while (j < N) {
    if (Piv[j] < 0 || (j==0 && Piv2x2at0)) {
      DistMat_RVAlloc(1);
      DistMat Aj1j = SubMat(A, j+1, j);
      if (InEltScope(Aj1j)) {
        if (Cpy) {
 	  int j_loc = G2Ld(j+1, DM_DM(A));
	  SubDiag[j_loc] = DM_Elt00(Aj1j);
        }
        DM_Elt00(Aj1j) = FD_0;
      }
      j++;
    } 
    j++;
  } /* while (j...) */
} /* CpSubDiag() */

static ResetSubDiag(N, A, Piv)
  int N, Piv[]; DistMat A;
{ int j;
  j=0;
  while (j < N) {
    if (Piv[j] < 0 || (j==0 && Piv2x2at0)) {
      DistMat_RVAlloc(1);
      DistMat Aj1j = SubMat(A, j+1, j);
      if (InEltScope(Aj1j)) {
        int j_loc = G2Ld(j+1, DM_DM(A));
        DM_Elt00(Aj1j) = SubDiag[j_loc];
      }  
      j++;
    }
    j++;
  } /* while (j...) */
} /*ResetSubDiag()*/


void D_SYTRS(N, K, A, DInv, Piv, B, info)
  int K, N;
  DistMat A, B; FDATA *DInv; int Piv[], *info; 
{ int j;	
  char svBrdTopChar[2];
  COMMON_ASSERT_TR_E(
     (TraceStr, "(%d,%d,%s,_,%s_)", N, K,  DM_TRACE(0, A), DM_TRACE(1, B)),
      DistMatOK(N, N, A) && DistMatOK(N, K, B), 0, "SYTRS", -1, TRACE_OUTER
  );

  svBrdTopChar[0] = BrdTopChar[0],      svBrdTopChar[1] = BrdTopChar[1];
  D_BKSWP("Fwd", K, N, Trans(B), Piv, 0);

  BrdTopChar[0] = 'I',               	BrdTopChar[1] = 'I';
  ClrSubDiag(N, A, Piv);  
  D_TRSM("Lower", "Unit", N, K, FD_1, A, B);

  j=0;
  while (j < N) {
    DistMat_RVAlloc(4);
    DistMat Bj = SubMat(B, j, 0),	Bj1 = SubMat(Bj, 1, 0),
            Ajj = SubMat(A, j, j),	Aj1j = SubMat(Ajj, 1, 0);
    if (Piv[j] >= 0 && !(j==0 && Piv2x2at0)) {	/* 1 x 1 pivot */
      DistMat_RVAlloc(2);
      if (InRowScope(Ajj)) {		/* multiply by inverse of 1x1 block */
        FDATA eAjj; 
	eAjj = GetDInv1(j);
        D_SCAL(K, eAjj, Trans(Bj));	
      }
      j++;
    } else {				/* 2 x 2 pivot */
      DistMat_RVAlloc(8);
      DistMat Bj2 = SubMat(Bj1, 1, 0), 
              Aj2j = SubMat(Aj1j, 1, 0), 	Aj1j1 = SubMat(Aj1j, 0, 1),
              Aj2j1 = SubMat(Aj2j, 0, 1);

      if (d_P0(DM_DM(Bj))==d_P0(DM_DM(Bj1))) {
      	/* can optimize: 2 columns on the same cell row */
      	if (d_P0(DM_DM(Bj))==0) {
      	  int l, l_max = G2Ld(K, DM_DN(Bj));
          FDATA *pBj=LM_ADR(DM_Loc(Bj)), 	*pBj1 = LM_ADR(DM_Loc(Bj1)); 	
          int CIncB = LM_Inc(1-DM_Trans(Bj), DM_Loc(Bj));          
      	  FDATA D00, D10, D11, T;

	  GetDInv3(j, &D00, &D11, &D10);	

      	  for (l=0; l < l_max; l++) { 	
      	     FDATA eBj = *pBj, eBj_,		eBj1 = *pBj1, eBj1_;
      	     FD_MULSET(eBj_, =, eBj, D11);	FD_MULSET(eBj1_, =, eBj1, D00); 
	     FD_ASSIGN(eBj_, -=, eBj1);		FD_ASSIGN(eBj1_, -=, eBj);
	     FD_MULSET(*pBj, =, eBj_, D10);	FD_MULSET(*pBj1, =, eBj1_, D10);
      	     pBj += CIncB,			pBj1 += CIncB;	    
      	  }    
      	}
      } else {
      	DistMat Bj_ = NewMat(1, K, Bj, NULL);
      	FDATA D00, D10, D11, T;

        if (d_P0(DM_DM(Bj))==0 || d_P0(DM_DM(Bj1))==0) /* ow access of DInv may be unsafe */
	  GetDInv3(j, &D00, &D11, &D10);					

	D_COPY(K, Trans(Bj), Trans(Bj_));      	
      	D_ScaleAdd(1, K, FD_M1, Bj1, D11, Bj);
      	D_ScaleAdd(1, K, FD_M1, Bj_, D00, Bj1);
      	D_SCAL(K, D10, Trans(Bj));		D_SCAL(K, D10, Trans(Bj1));
      	DisposeMat(Bj_);
      }	    
      j+=2;    	
    }
  } /* while (j...) */

  BrdTopChar[0] = ' ',                  	BrdTopChar[1] = ' ';
  D_TRSM("Upper", "Unit", N, K, FD_1, Trans(A), B);
  ResetSubDiag(N, A, Piv);  
  MatFree(SubDiag);				SubDiag = NULL;

  D_BKSWP("Bwd", K, N, Trans(B), Piv, 0);
  BrdTopChar[0] = svBrdTopChar[0],      	BrdTopChar[1] = svBrdTopChar[1];
  TRACE_EXIT;
} /* D_SYTRS() */
