/************************************************************************/
/*         DBLAS interface to P_LARFT/B() 				*/
/************************************************************************/
/* SCCS INFO @(#)P_LARFTB.c 1.1         last modified P. Strazdins, 99/04/23
   first created: P. Strazdins 96/10
 */
/* note: the DBLAS versions of P_LARFT/B() must both be called together;
 *       each is indiviadually incompatible with the ScaLAPACK version
 *	 of the other routine. This is because T has the wrong orientation in 
 *	 ScaLAPACK for efficient algorithmic blocking to be applied. 
 *	 Hence, these functions are in the same module.
 * compilation: if IGNORE_CALLER_T defined, ignore the callers T as a workspace,
 *	and mallocs a new workspace. This is safer if the caller has not 
 *	left enough space in T for algorithmic blocking.	
 *	If this is not defined, must check that T T given to P_LARFB()
 *	is the same as that for the last call to P_LARFT().
 */

/*#define IGNORE_CALLER_T*/

#include "PBLASCheck.h"
#include "ScaLAPACKMisc.h"

#ifndef ALTNAMES
static int UseDBLAS_P_LARFT=0;		/* to indicate if P_LARFT is linked in*/
static int UseDBLAS_P_LARFB=0;		/* to indicate if P_LARFB is linked in*/
#endif

static FDATA *T1 = NULL;		/* value of T in last call P_LARFT() */
static DistMatStore dT_;
static DistMat dT = &dT_;		/* T matrix used in last call P_LARFT()*/

void P_LARFT( Fdirect, Fstorev, n, k, V, iv, jv, desc_V, Tau, T, Work )
  F_CHAR_T Fdirect, Fstorev;		int *n, *k;
  FDATA V[], Tau[], T[], Work[];  	int *iv, *jv;
  int desc_V[]; 
{ int ParamErr = 0;   
  DistMat dV;
  int tj;
  char *direct = F2C_CHAR(Fdirect), 	*storev = F2C_CHAR(Fstorev);
  char Direct = UpperCase(*direct),     StoreV = UpperCase(*storev);
 
  if (!DirectOK(Direct))           	ParamErr = 1;
  else if (Direct!=Forward)             ParamErr = 1;   /* temporary! */
  else if (!StoreOK(StoreV))            ParamErr = 2;
  else if (!DimOK(*n))                  ParamErr = 3;
  else if (!DimOK(*k))                  ParamErr = 4;
  else {
    int TrV = StoreV==Row;
    ParamErr = PDescChkConv(0, TrV, *n, *k, V, *iv, *jv, desc_V, 8, &dV);
  }

  ERROR_CHECK_TR(
    (TraceStr, "(%c,%c,%d,%d,a%d,%d,%d,_,a%d,a%d,)",
     *direct, *storev, *n, *k, FD_ADR(V), *iv, *jv, FD_ADR(Tau), FD_ADR(T)), 
    -ParamErr, "LARFB", 0
  );


#ifdef IGNORE_CALLER_T
  if (T1!=NULL)				/* release storage of prev. call */
    DisposeMat(dT);
  dT = NewMat_(dT, *k, *k, dV, NULL);  
#else
  dT = NewMat_(dT, *k, *k, dV, T);  
#endif  
  T1 = T;
  
  tj = (LM_ADR(DM_Loc(dV)) - V) / desc_V[LLD_];
  					/* must get corresp local offset to j 
  					   to align with V */ 
  TauRowRep = 1;    			/* use Tau as a row-replicated vector*/
  D_LARFT(&Direct, *n, *k, dV, &Tau[tj], dT);

  TRACE_EXIT;
} /* P_LARFT */

void P_LARFB( Fside, Ftrans, Fdirect, Fstorev, m, n, k, V, iv, jv, desc_V, 
	      T,  C, ic, jc, desc_C, Work )
  F_CHAR_T Fside, Ftrans, Fdirect, Fstorev;	int *m, *n, *k;
  FDATA V[], T[], C[], Work[];  	int *iv, *jv, *ic, *jc;
  int desc_V[], desc_C[]; 
{ int ParamErr = 0;
  char *side = F2C_CHAR(Fside), 	*trans = F2C_CHAR(Ftrans), 
       *direct = F2C_CHAR(Fdirect), 	*storev = F2C_CHAR(Fstorev); 
  char Side = UpperCase(*side),         Trans = UpperCase(*trans),
       Direct = UpperCase(*direct),     StoreV = UpperCase(*storev);       
  int TrC = Side!=Left;
  DistMat dC, dV;
    
  if (!SideOK(Side))                   	ParamErr = 1;
  else if (!TransOK(Trans))             ParamErr = 2;
  else if (!DirectOK(Direct))           ParamErr = 3;
  else if (Direct!=Forward)           	ParamErr = 3;	/* temporary! */
  else if (!StoreOK(StoreV))           	ParamErr = 4;	
  else if (!DimOK(*m))                  ParamErr = 5;
  else if (!DimOK(*n))                  ParamErr = 6;
  else if (!DimOK(*k))                  ParamErr = 7;
  else if (T!=T1)       /* check that T was made by DBLAS version of DLARFT() */
  					ParamErr = 12;
  else {
    int TrV = TrC^(StoreV==Row);	
    ParamErr = PDescChkConv(0, TrC, *m, *n,  C, *ic, *jc, desc_C, 17, &dC);
    if (ParamErr==0)
      ParamErr = PDescChkConv(1, TrV, *m, *k,  V, *iv, *jv, desc_V, 12, &dV);
    if (ParamErr==0) {
      int AlignM = Dim_BCompat(DM_DM(dV), DM_DM(dC));
      if (AlignM==0)			ParamErr = 12;
    }  
  }

  ERROR_CHECK_TR(
    (TraceStr, "(%c,%c,%c,%c,%d,%d,%d,a%d,%d,%d,_,a%d,_,a%d,%d,%d,_,_)",
     *side, *trans, *direct, *storev, *m, *n, *k, FD_ADR(V), *iv, *jv, FD_ADR(T), 
     FD_ADR(C), *ic, *jc),
    -ParamErr, "LARFB", 0
  );

  /* note that the trans and Work parameters are ignored; use dT from
     last call to D_LARFT(); assert dT is fully aligned with dV */

  D_LARFB(&Direct, *m, *n, *k, dV, dT, dC);
  
  TRACE_EXIT;
} /* P_LARFB */

