/************************************************************************/
/*	AuxBlkMatrix: FDATA Auxillary Distributed matrix procedures	*/
/************************************************************************/
/* SCCS INFO %Z%%M% %I%         last modified P. Strazdins, %E%
   first created: P. Strazdins 94/09
*/
/* Auxilliary routines for initializing block-scattered matrices
 * compilation: macro FDATATYPE must be defined
 *	FAST_RAND_GEN:	if defined, try to minimioze calls to srand();
 *			this loses property that matrix will be globally
 *			the same no matter what grid or block size used
 */
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "BLASOptions.h"
#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#define FAST_RAND_GEN

static void GetTriBounds(Trans, A, iG, jG, n, j_mn, j_mx)
  char *Trans; DistMat A; int iG, jG, n, *j_mn, *j_mx;
/* pre:  n is local length of gloabal row iG of distributed matrix A.
 *       The diagonal of A is at local elements where iG==jG.
 * post: *j_mn <= j < *j_mx is a range of local indices sufficient to
 *	 initialize an upper (*Trans=='U'), lower (*Trans=='L')
 *	 or rectangular (otherwise) A
 * note: for the purposes of over-write checking, some elements from
 *	 the other half of a tri. matrix may be included in this range.
 *	 This routine must be called consistently from MakeScattered() and
 *	 CheckScattered().
 */  
#define TRI_GUARD 4	     /* initialize this many elements in the other half,
				so that over-writes can be checked */
{ *j_mn = 0, *j_mx = n;
  if (*Trans=='U') {
    *j_mn = G2Ld(iG - jG, DM_D1(A)) - TRI_GUARD;
    *j_mn = MAX(*j_mn, 0);
  } else if (*Trans=='L') {
    *j_mx = G2Ld(iG+1 - jG, DM_D1(A)) + TRI_GUARD;
    *j_mx = MIN(*j_mx, n);
  }	
  /*printf("GetTriBounds(%c,_,%d %d,%d,%d..%d)\n", *Trans, iG, jG, n, *j_mn, *j_mx);*/
} /* GetTriBounds() */  

int RandSeedOffset = 0;		/* can be varied to get diff random elements */

static void MakeScattered(RandElts, I0, J0, m, n, UnitAdd, Trans, A)
  int RandElts, m, n; FDATA *UnitAdd; char *Trans; DistMat A;
/* post: DM_Loc(A)[i, j] = f(i', j') where i'=I0+L2Gd(i,A.DI), j'=J0+L2Gd(j,A.DJ)
 * 	 	for 0<=i<m, 0<=j<n
 * 	 where  f(i, j) = (RandElts? random(a*i+b*j): a*i+b*j) +
 *			  (i==j? UnitAdd: 0)
 *  		(a,b) = Trans=='S'? (x,x): (Trans==NoTrandpose? (x,y): (y,x))
 *		for some x,y>0.0
 * notes: Trans=='S' is used to generate (square) symmetric matrices.
 *	  Trans=='L'/'U' is for triangular matrices
 *	  0 <= random() <=1 if RandElts>0; else -1 <= random() <=1 
 */
{ int i, j;
  for (i=0; i < m; i++) {
    int iG = I0 + L2Gd(i, DM_D0(A)), j_mn, j_mx;
    GetTriBounds(Trans, A, iG, J0, n, &j_mn, &j_mx);
#ifdef FAST_RAND_GEN
    /* this is as far as srand() can be factored out for CheckScattered() to work */
    if (RandElts && *Trans!='S') {
      int jG = J0 + L2Gd(0, DM_D1(A));	
      if (*Trans==NoTranspose)	       
    	srand(3*iG+jG + RandSeedOffset);
      else
    	srand(3*jG+iG + RandSeedOffset);
    }
#endif    
    for (j=j_mn; j < j_mx; j++) {
      FDATA v;
      int jG = J0 + L2Gd(j, DM_D1(A));
      if (RandElts) {
      	if (*Trans=='S')	     /* nb. dangerous to factor out */
      	  srand(1000*MAX(iG,jG)+MIN(iG,jG) + RandSeedOffset);
#ifndef FAST_RAND_GEN         	   
	else if (*Trans==NoTranspose)	       
    	  srand(3*iG+jG + RandSeedOffset);
  	else
    	  srand(3*jG+iG + RandSeedOffset);
#endif    	  
    	if (RandElts < 0) { 
          FD_SET_REAL(v, (FD_BASE) (10000 - rand() % 20000) * 0.0001 + 0.001);
          FD_SET_IMAG(v, (FD_BASE) (10000 - rand() % 20000) * 0.0001 + 0.001);
        } else {        
          FD_SET_REAL(v, (FD_BASE) (rand() % 10000) * 0.0001 + 0.001);
          FD_SET_IMAG(v, (FD_BASE) (rand() % 10000) * 0.0001 + 0.001);
	}
      } else {
      	FD_BASE vr;
      	if (*Trans=='S')
      	  vr = (FD_BASE) (iG+1) * 0.1 + (FD_BASE) (jG+1) * 0.1;
	else if (*Trans==NoTranspose)
          vr = (FD_BASE) (iG+1) * 0.1 + (FD_BASE) (jG+1) * 0.001;
        else
          vr = (FD_BASE) (iG+1) * 0.001 + (FD_BASE) (jG+1) * 0.1;
        FD_ASSIGN_SC(v, =, vr, -vr/2.0);
      }
      if (iG == jG)
	FD_ASSIGN(v, +=, *UnitAdd);
      *LM_Index(DM_Loc(A), i, j) = v;
    } /* for (j...) */
  } /* for (i...) */
} /*MakeScattered*/


#include <sys/stat.h>
#include "FDCommun.h"
#include "CellBlkMatrix.h"

#define READ_BUFSIZE (1048576 / sizeof(FDATA))	/* efficient size in which to do
						   a block read (ie. 1/4 MB) */
						   
static void ReadScattered(matfile, M, N, SymmetricA, A)
  char *matfile; int M, N; int SymmetricA; DistMat A;
/* pre:  matfile!=NULL is name of a binary data file of an MxN matrix A, with 
 *	 elements in col-major order. 
 *	 if SymmetricA, only lower tri. elements are in matfile.
 * post: global A[i,j] = A'[i,j] 0<=i<M,0<=j<N
 */
{ int i, j;				size_t f_size_exp;
  int i1, w1, bufsize;
  FDATA *readbuf[2];			int r_ix; 
  /* 2 read buffers so that commun & file reading can be overlapped */
  FILE *fp;				struct stat f_stat;
  f_size_exp = (SymmetricA? N*(N+1)/2: M*N) * sizeof(FDATA);
  if (CID==0) {
    stat(matfile, &f_stat);
    fp = fopen(matfile, "r");
    if (fp == NULL) {
      printf("ReadScattered: cant open file %s\n", matfile);
      exit(2);
    }
    if (f_stat.st_size != f_size_exp) {
      printf("ReadScattered: file %s: expected %d bytes, actual %d bytes\n", 
    	      matfile, f_size_exp, f_stat.st_size);
      exit(2);
    }
  }

  MatMalloc("ReadScattered", 1, &readbuf[0], 
  	     MIN(f_size_exp/sizeof(FDATA), READ_BUFSIZE)*2);
  readbuf[1] = readbuf[0] + MIN(f_size_exp, READ_BUFSIZE);	r_ix = 0;
      
  j=0;
  SendWaitMark();			/* make sends non-blocking (if MPI) */
  while (j < N) {        		/* read file in blocks from cell 0;
    					   use pipelined b/c & non-blocking 
    					   sends to minimize load on cell 0 */
    int Mj = SymmetricA? M-j: M;
    int w = MIN(N-j, READ_BUFSIZE / Mj);
    int r_exp = w * Mj - (SymmetricA? w*(w-1)/2: 0); 
    FDATA *rbufp = readbuf[r_ix];      
    if (CID==0) 
      fread(rbufp, sizeof(FDATA), r_exp, fp);
    if (CID > 0)
      RecvMat(-(CID_J==0), -(NP_J>1), 1, r_exp, rbufp, r_exp, 0); 	
    if (CID < NP-1)
      SendMat((CID_J==NP_J-1), (NP_J>1), 1, r_exp, rbufp, r_exp, 0); 	
    while (w > 0) {			/* copy out this cell's portion */
      DistMat_RVAlloc(1);
      DistMat Aj = SubMat(A, SymmetricA? j: 0, j);
      Mj = SymmetricA? M-j: M;	
      if (d_P0(DM_DN(Aj))==0) {		/* this cell owns some of column */
        LocMat lrbufp, lAj;			
        LM_Init(lrbufp, rbufp+L2Gd(0, DM_DM(Aj)), 1, M);
        LM_Copy(DM_Trans(Aj), lAj, DM_Loc(Aj));
        BlkMatStridePack(0, G2Ld(Mj, DM_DM(Aj)), 1, &lrbufp, NULL, &lAj,
          		 BlkB0d(0, DM_DM(Aj)), d_B(DM_DM(Aj)), d_P(DM_DM(Aj)), 
          		 0, 1, 1);
      }  		   
      rbufp += Mj;
      j++; w--;   		   
    } /* while (w...) */
    if (j > w)  
      CompleteSends(1);			/* ensure readbuf[1-r_ix] free for next itn */
    r_ix = 1-r_ix;       
  } /* while (j...) */  	
  CompleteSends(-1);
  if (CID==0)
    fclose(fp);   
  MatFree(readbuf[0]);     
} /*ReadScattered*/


/*static*/ int CheckScattered(RandElts, I0, J0, m, n, UnitAdd, Trans, A,
			  MxDiff, AvDiff, MaxI, MaxJ)
  int RandElts, m, n; FDATA *UnitAdd; char *Trans; DistMat A;
  FD_BASE *MxDiff, *AvDiff; int *MaxI, *MaxJ; 
/* check m x n local matrix of matrix A is same as that generated by: 
 *	 MakeScattered(RandElts, I0, J0, m, n, UnitAdd, Trans, B)
 * 	 where DM_Dim(i, B) = DM_Dim(i, A)
 * post: r.v. & setting of MxDiff, AvDiff, MaxI, MaxJ as per
 *	 MatDiff(m, n, DM_Loc(A), DM_Loc(B), MxDiff, AvDiff, MaxI, MaxJ)
 */
{ int i, NErrs = 0; 			FD_BASE Av, Mx; int MxI, MxJ;
  FDATA *Bx;
  *MxDiff = 0.0, *MaxI = 0, *MaxJ = 0,	*AvDiff = 0.0;
  if (n <= 0)
    return(NErrs);
  if (DM_Trans(A))	/* ensures iG refers to DM_D0(A) in SubMat(A, iG, 0) */
    A = Trans(A);  
  MatMalloc("CheckScattered", 1, &Bx, n);
  for (i = 0; i < m; i++) {
    DistMat_RVAlloc(1);	
    int nej;	
    int j_mn, j_mx;
    int iG = I0 + L2Gd(i, DM_D0(A));	DistMat Bi = SubMat(A, iG, 0);
    LM_Init(DM_Loc(Bi), Bx, n, 1);	/* 1 x n */

    MakeScattered(RandElts, iG, J0, 1, n, UnitAdd, Trans, Bi);
    GetTriBounds(Trans, A, iG, J0, n, &j_mn, &j_mx);
    nej = MatDiff(1, j_mx-j_mn, LocSubMat(DM_Loc(Bi), j_mn, 0), 
    		  LocSubMat(DM_Loc(A), i, j_mn), &Mx, &Av, &MxI, &MxJ); 
    MxJ += j_mn;		  

    if (nej)
      printf("-%d (%d); ", i, nej);
    NErrs += nej;
    if (Mx > *MxDiff) 
      *MxDiff = Mx, *MaxI = i, *MaxJ = MxJ,
    *AvDiff += Av;
  } /* for (i...) */
  if (m > 0) *AvDiff /= ((FD_BASE) m);
  MatFree(Bx);
  if (NErrs)
    printf("\n");
  return (NErrs);
} /* CheckScattered() */


static void TrPreMultDiag(UpLo, Diag, m, n, alpha, beta, C)
  char *UpLo, *Diag; int m, n; FDATA *alpha, *beta; DistMat C;
/* pre:  UpLoOK(*UpLo)
 *       C <- *alpha * C 
 *	 for proper UpLo triangular part of local m x n matrix of C
 *       if Diag!=Unit, diag(C) = *beta
 */
{ int i; 				pMatDim Di, Dj;
  LocMat lC;
  Di = DM_D0(C);			Dj = DM_D1(C);
  lC = DM_Loc(C);
  if (*UpLo==Upper) {
    SWAP(int, m, n); 			SWAP(pMatDim, Di, Dj);
    SWAP(int, LM_RInc(lC), LM_CInc(lC)); 
  }
  for (i=0; i < m; i++) {
    int iG = L2Gd(i, Di), 	ji = G2Ld(iG, Dj), j;
    if (*Diag==Unit && ji<n && SubMatP0d(iG, Di)==0 && SubMatP0d(iG, Dj)==0) 
      *LM_Index(lC, i, ji) = *beta;
    ji = MIN(ji, n);
    for (j=0; j < ji; j++)
      FD_MULSET(*LM_Index(lC, i, j), =, *LM_Index(lC, i, j), *alpha);
  } /* for (i...) */
} /* TrPreMultDiag() */
 

