/************************************************************************/
/*        AuxMatrix: FDATA Auxillary Matrix procedures           	*/
/************************************************************************/
/* SCCS INFO @(#)AuxMatrix.c 1.3         last modified P. Strazdins, 99/03/19
   first created: P. Strazdins 91/12/12
 */
/* set of matrix routines for matrix initialiation and result checking;
 * not called by DBLAS routines. 
 * compilation: macro FDATATYPE must be defined;
 *		if TRACE_ERRS defined, MatDiff() prints indices where error
 *		detected
 */ 
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "LocMat.h"
#include "stdlib.h"

/*#define TRACE_ERRS*/

/****************** matrix initialization & checking procedures *************/

#define F_RAND_MOD	10000
#define F_RAND		(((FD_BASE) ((rand() % F_RAND_MOD) - F_RAND_MOD/2)) \
			/((FD_BASE) F_RAND_MOD/2))	/* values in [-1..1] */

#define INIT_ELT(v, i, IMult, j, JMult, Offs)				\
	{ FD_SET_REAL(v, i*FD_REAL(IMult) + j*FD_REAL(JMult) + FD_REAL(Offs));\
	  FD_SET_IMAG(v, i*FD_IMAG(IMult) + j*FD_IMAG(JMult) + FD_IMAG(Offs));}

static void MatInit(RandElts, IMult, JMult, Offs, M, N, A)
  int RandElts, M, N; FDATA IMult, JMult, Offs; LocMat A;
/* pre:   LM_IncOK(M, N, A)  
 * post:  A[i,j] = (i*IMult + j*JMult + Offs), for 0 <= i < M and 0 <= j < N
 *	  where f(x) = x, !RandElts and f(x) = {srand(x); F_RAND}  ow.
 */
{ int i, j;
  for (i=0; i < M; i++)
    for (j=0; j < N; j++) {
      FDATA v; 
      INIT_ELT(v, i, IMult, j, JMult, Offs);
      if (RandElts) {
      	srand((int) FD_REAL(v) + FD_IMAG(v));
        FD_SET_REAL(*LM_Index(A, i, j), F_RAND);
        FD_SET_IMAG(*LM_Index(A, i, j), F_RAND);
      } else
        *LM_Index(A, i, j) = v;
   }	        
} /* MatInit() */
 

#define EPSILON			1.0E-10	/* machine precision */		
 
static int MatCheck(RandElts, IMult, JMult, Offs, M, N, A)
  int RandElts, M, N; FDATA IMult, JMult, Offs; LocMat A;
/* pre:   LM_IncOK(M, N, A)  
 * post: returns number of elements of M x N A differing from A' where
 *	 A' satisfies postcond. MatInit(RandElts, IMult, JMult, Offs, M, N, A') 
 */
{ int i, j, nerrs=0; FDATA e, v;
  for (i=0; i < M; i++) {
    int nej=0;	
    for (j=0; j < N; j++) {
      INIT_ELT(v, i, IMult, j, JMult, Offs);	
      if (RandElts) {
        srand((int) FD_REAL(v) + FD_IMAG(v));   
        FD_SET_REAL(v, F_RAND);
        FD_SET_IMAG(v, F_RAND);
      }  
      FD_SETOP(e, v, -, *LM_Index(A, i, j));
      if (FD_ABS(e) > EPSILON) {
      	/* a threshold check required since the value of Offs for MatCheck()
      	   may not necessarily be that for MatInit() of the source matrix,
      	   eg. B[i, j] in MatCheck with Offs'=inc+o 
      	   comes from A[i, j+inc] produced by MatInit() with Offs = o
	*/
#ifdef TRACE_ERRS
	printf("(%d,%d: %.1f#%.1f) ", i, j, *LM_Index(A, i, j), v) ;
#endif    	
        nej++;
      } 
    } /* for (j...) */
#ifdef TRACE_ERRS
    if (nej)
      printf("\n");
#endif    
    nerrs += nej;  
  } /* for (i...) */ 
  return (nerrs);       
} /* MatCheck() */
 

static int ScaledDiff = 1;	/* set to 0 for absolute error */

static int MatDiff(M, N, A, B, MxDiff, AvDiff, MaxI, MaxJ)
  int M, N; LocMat A, B; FD_BASE *MxDiff, *AvDiff; int *MaxI, *MaxJ;
/* pre:  LM_IncOK(M, N, A) && LM_IncOK(M, N, B) 
 * post: *MxDiff = Max {0} U {0<=i<M, 0<=j<N: D(i,j)}
 * 	 *AvDiff = Sum {0<=i<M, 0<=j<N: D(i,j)}/M/N  , if M>0 & N>0
 *	  	 = 0					otherwise
 * 	  D(*MaxI,*MaxJ) = *MxDiff		      , if M>0 & N>0
 * 	  where D(i,j) = Abs(A[i,j]-B[i,j])/(Abs(A[i,j])+Abs(B[i,j])+1E-30)
 *	  returns the number of elements st D(i,j) > PRECISION_THRESH 
 */
# define PRECISION_THRESH 5.0E-6                /*set t to 2x precision limit */
{ int i, j, n_err=0; FD_BASE t, d; FDATA del;
  int errjmn;
  *MxDiff = 0.0; *MaxI = 0; *MaxJ = 0;
  *AvDiff = 0.0;
  for (i = 0; i < M; i++) {
#ifdef TRACE_ERRS
    int errjmn=-1, errjmax;
#endif    
    for (j = 0; j < N; j++) {
      FDATA Aij = *LM_Index(A,i,j), Bij = *LM_Index(B,i,j);
      FD_SETOP(del, Aij, -, Bij);
      d = FD_ABS(del);
      t = ScaledDiff? (PRECISION_THRESH + FD_ABS(Aij) + FD_ABS(Bij)): 1.0; 
      d = d / t;
      if (d > *MxDiff) { *MxDiff = d; *MaxI = i; *MaxJ = j;}
      if (d > PRECISION_THRESH) {
        n_err++;
#ifdef TRACE_ERRS
	if (errjmn<0) 
          errjmn = j;
        errjmax = j;  
#endif                        	  
      }
      *AvDiff = *AvDiff + d;
    } /* for (j...) */
#ifdef TRACE_ERRS    
   if (errjmn>0) 
      printf(" e[%d,%d..%d] ", i, errjmn, errjmax);    
#endif
  } /* for (i...) */
  if (M > 0 && N > 0) *AvDiff = *AvDiff / ((FD_BASE) M * N);
  return (n_err);
} /* MatDiff() */
	
