/************************************************************************/
/*                              Commun.Parameters.c                     */
/************************************************************************/
/* SCCS INFO @(#)Commun.Parameters.c 1.1         last modified P. Strazdins, 98/12/24
   first created: P. Strazdins 94/09/01
*/
/* input command line parameters into variables for communication
 * test program suite, with the usage: 					
 *	PROGNAME		
 * followed optionally by following USAGE string:			*/
#define USAGE \
"[-X TReps] [-B rc I0|-S ri rj| -E rc] [-i inc] [-o i0] [-r r] [-s R] 	\n\
	[-p pm pn] [-t tlvl] [-M n v] [-G Ncely Ncelx] m n"
/* with default values:							*/
#define DEFAULTS \
"ri=rj=1, inc=N, TReps=R=r=1, i0=0, pm=pn=0 tlvl=-1"
/* performs for i=0..R-1 (with r internal reps, for timing) distributed
 * stride communications for m x n matrices  with row stride inc and offset i0 
 * from base. Print out pm x pn submatrix of received cells. 
 * The -t option is used for DBLAS trace level setting.
 * The -M option may be repeated with differing n values; all these
 * can set the DBLAS machine parameters before the compuation begins.
 * With: 
 *	-B:  	perform broadcast (column if rc == 1, row ow)
 *		from cell I0,...,I0+r-1
 *	-S:	all cells send to relative cell (ri,rj)
 *	-E:	do ping-pong tests between cells neighbouring cells
 *		in processor dimensions rc
 * source messages will be warm in cache, destination messages will be cold
 * TReps x r*R repetitions are performed, with timings taken over r*R reps
 */

/* command line parameters; see header comment*/
  static int TReps=1, m=0, n=0, I0=0, ri=1, rj=1, i0=0, inc=0, r=1, R=1, pm=0, pn=0;
  static int rc=0, DoBrd=0, DoPingPong=0;
  static int Ncely, Ncelx; 

#include "MachineParams.h"	/* for -M oprtion */
#include "FDAssert.h"		/* for -t option */

static void PrintUsage();

static void GetParameters(argc, argv)
  int argc; char *argv[];
/* post: set Commun command line variables according to argv[0..argc-1] */
{ extern char *optarg;
  extern int optind;
  int c;

  Ncely = NP_I, Ncelx = NP_J;
  while ((c = getopt(argc, argv, "GBSE:i:o:r:X:s:t:Mp")) != -1) /* parse options */
    switch (c) {
      case 'G':
         if (optind + 1 < argc) {
           Ncely = atof(argv[optind]);      Ncelx = atof(argv[optind+1]);
           optind += 2;
         }                                                              break;
      case 'B':
        if (optind + 1 < argc) {
          rc = atof(argv[optind]);      I0 = atof(argv[optind+1]);
          optind += 2;
        } DoBrd=1;
                                                                        break;
      case 'S': 
        if (optind + 1 < argc) {
          ri = atof(argv[optind]);      rj = atof(argv[optind+1]);
	  DoPingPong = DoBrd = 0;                           
          optind += 2;
        }  
									break;
      case 'E':
        rc = atoi(optarg); 
	DoPingPong = 1;                                           	break;
      case 'r':
        r = atoi(optarg);                                           	break;
      case 'X':
        TReps = atoi(optarg);                                           break;
      case 's':
        R = atoi(optarg);                                           	break;
      case 'i':
        inc = atoi(optarg);                                           	break;
      case 'o':
        i0 = atoi(optarg);                                           	break;
      case 't':
        TraceLevel = atoi(optarg);                                    	break;
      case 'p':
        if (optind + 1 < argc) {
          pm = atof(argv[optind]);      pn = atof(argv[optind+1]);
          optind += 2;
        }  								break;
      case 'M':
        if (optind + 1 < argc) {
          MachParamSet(atoi(argv[optind]), atoi(argv[optind+1]));
        }                                                               break;
      case '?':
        PrintUsage();
        exit (2);
    }

  if (optind+1 >= argc || argc<=2) { 
    if (argc >= 2)
      HPRINT((STDOUT, "%s: insufficient parameters m n given.\n", PROGNAME));
    PrintUsage();
    exit (argc >= 2? 2: 0);
  }

  /* get non-optional parameters */
  m = atoi(argv[optind]);
  n = atoi(argv[optind+1]);
  if (inc < n+i0)
    inc = n+i0;
  if (pm > m) pm = m;
  if (pn > inc) pn = inc; 
} /* GetParameters() */

static void PrintParameters();

static void PrintUsage()
/*post: prints program usage to stdout*/
{ HPRINT((STDOUT, "usage : %s %s\n", PROGNAME, USAGE));
  HPRINT((STDOUT, "with default values: %s\n", DEFAULTS));
  PrintParameters();
} /* PrintUsage() */


static void PrintParameters()
/*post: print out current values of variables set by command line to stdout*/
{ HPRINT((STDOUT, "%s: %dx%d mesh stride commun fn %c (%d, %d), inc %d, i0=%d, %d timings %dx%d reps, for %d x %d data\n",
	 PROGNAME, Ncely, Ncelx, DoBrd? 'B': DoPingPong? 'E': 'S', DoBrd? rc: ri, 
	 DoBrd? I0: rj, inc, i0, TReps, R, r, m, n)); 
} /* PrintParameters() */

