/************************************************************************/
/*                              MF.Parameters.c                         */
/************************************************************************/
/* SCCS INFO @(#)MF.Parameters.c 1.3         last modified P. Strazdins, 99/03/19
   first created: P. Strazdins 95/08/17
*/
/* input command line parameters into variables for general serial
 * matrix multiplication  program performing C := alpha*op(A)*op(B)+beta*C
 *        
 * with the usage: 
 *	PROGNAME 
 *  followed optionally by following USAGE string:			*/

#define USAGE \
"[-a T|N ] [-b T|N] [-l LdA LdB] [-i] [-X TReps] [-r NReps] [-e] [-p PrintM PrintN] \n\
	[-t tlvl] [-M n v] [-v AlgVers] [-1] [-f cp|ad|sw|as|mx|sc|] M N [alpha] \n\
   with default values:							\n\
	trlvl=-1 LdA=LdB=N TReps=NReps=1, AlgVers=0, PrintM=PrintN=0; f='cp'; alpha=1.0"

/* and matrix transpose options corresponding to the variables:		
 *	-a T	: TransA = Transpose		, Inc(A) = {LdA, 1}, LdA>=N
 *	-a N	: TransA = NoTranspose (default), Inc(A) = {1, LdA}, LdA>=M
 * (similarly for B); with M x N A and:
 * 	-f cp	: B = alpha A		B - M x N
 *	-f ad	: B = B + alpha A
 *	-f sw	: B <-> A
 *	-f as	: B[i] = { sum j: 0 <= j < N: |A[i,j]|},	0 <= i < M
 *	-f mx	: B[i] = { mx  j: 0 <= j < N: |A[i,j]|},	0 <= i < M
 * with -1 option specified, matrices are forced on odd-word boundary
 * With -e specified, random matrix elts in [-1,1] used.
 * TReps x NReps repetitions are performed, with timings taken over NReps.
 * If -i specified, A & B are identical (gnerally not useful). 
 *
 * The -M option may be repeated with differing n values; all these
 * can set the DBLAS machine parameters before the computation begins.
 * The -t option is used for DBLAS trace level setting.
 * The -v AlgVers option is useful only for -f cp; AlgVers=1 use memcpy()
 * AlgVers=2 use bcopy() 
 */
#include "BLASOptions.h"
#include "MachineParams.h"      /* for -M oprtion */
#include "FDAssert.h"   	/* extern int TraceLevel */

/* command line parameters; see header comment*/
  static int M, N;
  static int TReps, NReps;
  static int TransA, TransB;
  static int PrintM, PrintN, SingleWordBoundary, RandElts;
  static int f;
  static FDATA alpha;
  static int AlgVers=0;
  static int IdenticalAB = 0;		/* for experimentation only */
  static int Ncely=1, Ncelx=1;
  static int LdA, LdB;
  static int Verbose;

  static char *fs[] = { "cp", "ad", "sw", "as", "mx", "sc", ""}; 

static void OptionError(), PrintUsage();

static void GetParameters(argc, argv)
  int argc; char *argv[];
/* post: set GMM command line variables according to argv[0..argc-1] */
{ extern char *optarg;
  extern int optind;
  int c; 

  /* set up default values */
  TransA = TransB = 0; RandElts=0;
  TReps = NReps=1; SingleWordBoundary = 0; 
  PrintM = PrintN=0; f = 0; alpha = FD_1;
  Verbose=0;

  LdA = LdB = 1;
  while ((c = getopt(argc, argv, "1a:b:f:r:X:pt:Mv:eilV")) != -1) /* parse options */
    switch (c) {
      case '1':
	SingleWordBoundary = 1;
									break;
      case 'e':
	RandElts = 1;
									break;
      case 'a': 
	if (optarg[0] == Transpose)
	  TransA = 1;
									break;
      case 'b': 
	if (optarg[0] == Transpose)
	  TransB = 1;
									break;
      case 'f':
	f = 0;
	while ((strcmp(fs[f], optarg) != 0) && (strcmp(fs[f], "") != 0)) {
	  f++;
        }	
	if (strcmp(fs[f], optarg) != 0) 
	  OptionError(c, "invalid function mnemonic");
									break;
      case 'r':
	NReps = atoi(optarg);						break;
      case 'X':
	TReps = atoi(optarg);						break;
      case 'p':
        if (optind + 1 < argc) {
	  PrintM = atoi(argv[optind]); 	PrintN = atoi(argv[optind+1]); 
	  optind += 2;
        } else 
	  OptionError(c, "PrintM or PrintN  parameters are missing");	break;
      case 'l':
        if (optind + 1 < argc) {
	  LdA = atoi(argv[optind]); 	LdB = atoi(argv[optind+1]); 
	  optind += 2;
        } else 
	  OptionError(c, "LdA or LdB  parameters are missing");		break;
      case 'v':
        AlgVers = atoi(optarg);                                    	break;	
      case 't':   
        TraceLevel = atoi(optarg);                                      break;
      case 'M':
        if (optind + 1 < argc) {
          MachParamSet(atoi(argv[optind]), atoi(argv[optind+1]));
          optind += 2;
        } else
          OptionError(c, "n or v parameters are missing");
                                                                        break;
      case 'i':
	IdenticalAB = 1;						break;
      case 'V':
	Verbose = 1;							break;
      case '?':
	OptionError(c, "invalid option given");				break;
    }

  if (optind+1 >= argc || argc<=2) { 
    if (argc >= 2) 
      HPRINT((STDOUT, "%s: insufficient parameters M N given.\n", PROGNAME));
    PrintUsage();
    exit(argc >= 2? 2: 0);
  }

  /* get non-optiinal parameters */
  M = atoi(argv[optind]);	N = atoi(argv[optind+1]);

  if (optind+2 < argc) 
    alpha = Str2FD(argv[optind+2]);

  if (PrintM < 0) PrintM = M;	if (PrintN < 0) PrintN = N;

  LdA = MAX(LdA, TransA? M: N);
  LdB = MAX(LdB, TransB? M: f<=2||f==5? N: 1);

} /* GetParameters() */


static void OptionError(option, msg)
  char *msg, option;
/*post: print out option error message  with msg & option; print usage & exit */
{ HPRINT((STDOUT, "%s: error in parsing option -%c: %s\n", PROGNAME, option, msg));
  PrintUsage();
  exit(2);
}

static void PrintParameters();

static void PrintUsage()
/*post: prints program usage to stdout*/
{ HPRINT((STDOUT, "usage : %s %s\n", PROGNAME, USAGE)); 
  PrintParameters();
}


static void CheckParameters()
/* post: check values of parameter variables, aborting if invalid */
{ if (M < 0 || N < 0) {
    HPRINT((STDOUT, "%s: invalid (-ve) matrix dimensions (M=%d, N=%d)\n",
 	   PROGNAME, M, N));
    PrintUsage();
    exit(2);
  }
} /* CheckParameters() */


static void PrintParameters()
/*post: print out current values of variables set by command line to stdout*/
{ char fss[32];
  if (f==0)
    sprintf(fss, "B = %s A", pFD2Str(&alpha));
  else if (f==1)
    sprintf(fss, "B = B + %s A", pFD2Str(&alpha));	 
  else if (f==2)
    sprintf(fss, "B <--> A");	 
  else if (f==3)
    sprintf(fss, "B = {sum j: |A[:,j]|}"); 
  else if (f==4)
    sprintf(fss, "B = {max j: |A[:,j]}");	 
  else if (f==5)
    sprintf(fss, "A = %s * A", pFD2Str(&alpha));	 
  HPRINT((STDOUT, "%s: %dx%d reps of %s ; A %s%s%s (ld=%d) is %dx%d , B %s (ld=%d) is %dx%d\n",
	 PROGNAME, TReps, NReps, fss, (SingleWordBoundary? "(swb)": ""),
	 RandElts? "(rand)": "", (TransA)? "(T)": "",  LdA, M, N, 
	 (TransB)? "(T)": "", LdB, M, f<=2? N:1));
} /* PrintParameters() */

