/************************************************************************/
/*                              MM.Parameters.c                         */
/************************************************************************/
/* SCCS INFO @(#)MM.Parameters.c 1.3         last modified P. Strazdins, 99/03/19
   first created: P. Strazdins 95/08/17
*/
/* input command line parameters into variables for general serial
 * matrix multiplication  program performing C := alpha*op(A)*op(B)+beta*C
 *        
 * with the usage: 
 *	PROGNAME 
 *  followed optionally by following USAGE string:			*/

#define USAGE \
"[-a T|N ] [-b T|N] [-c T|N|L|U] [-g Mg Ng] [-r NReps] [-X TReps] [-s StepM StepN] [-e]	\n\
	 [-p PrintM PrintN] [-t tlvl] [-M n v] [-v AlgVers] [-1] [-i] [-f mm|pm|mv|r1|rk|ur|Pm] M N K [alpha] \n\
   with default values:							\n\
	TReps=NReps=1, Mg=StepM=M, Ng=StepN=N; PrintM = PrintN=0; f='lm'; alpha=1.0"

/* and matrix transpose options corresponding to the variables:		
 *	-a T	: TransA = Transpose
 *	-a N	: TransA = NoTranspose (default)
 * (similarly for B; C); with:
 * 	-c L|U	: TransC = NoTranspose and FormC = Lower or Upper;
 *		  this forces -i also.
 * with -1 option specified, matrices are forced on odd-word boundary
 * With -i option specified, require A^TB or AB^T combination with m=n;
 * in this case B is identical to A. With -e specified, random matrix elts	
 * in [-1,1] used.
 * TReps x NReps repetitions are performed, with timings taken over NReps.
 * Mg, Ng determine strides of matrices; Mg >= M && Ng >= N 
 *
 * The -M option may be repeated with differing n values; all these
 * can set the DBLAS machine parameters before the computation begins.
 * The -t option is used for DBLAS trace level setting.
 * The -v AlgVers option is used to force an internal version of a DBLAS algorithm.
 */
#include <string.h>

#include "BLASOptions.h"

#include "MachineParams.h"      /* for -M oprtion */
#include "FDAssert.h"        	/* extern int TraceLevel, AlgVers */

/* command line parameters; see header comment*/
  static int M, N, K; static int Mg, Ng;
  static int TReps, NReps, StepM, StepN;
  static int TransA, TransB, TransC, FormC;
  static int PrintM, PrintN, SingleWordBoundary, IdenticalAB, RandElts;
  static int f;
  static FDATA alpha;
  static int Ncely=1, Ncelx=1;

  static char *fs[] = { "mm", "pm", "ur", "mv", "r1", "rk", "Pm", ""}; 

static void OptionError(), PrintUsage();

static void GetParameters(argc, argv)
  int argc; char *argv[];
/* post: set GMM command line variables according to argv[0..argc-1] */
{ extern char *optarg;
  extern int optind;
  int c; int StepOpt=0, GlobDimOpt=0; 

  /* set up default values */
  TransA = TransB = TransC = 0; FormC = 0; RandElts=0;
  TReps = NReps=1; SingleWordBoundary = 0; IdenticalAB=0;
  PrintM = PrintN=0; f = 0; alpha = FD_1;

  while ((c = getopt(argc, argv, "1ia:b:c:f:r:X:sgpt:Mev:")) != -1) /* parse options */
    switch (c) {
      case '1':
	SingleWordBoundary = 1;
									break;
      case 'i':
	IdenticalAB = 1;
									break;
      case 'e':
	RandElts = 1;
									break;
      case 'a': 
	if (optarg[0] == Transpose)
	  TransA = 1;
									break;
      case 'b': 
	if (optarg[0] == Transpose)
	  TransB = 1;
									break;
      case 'c': 
	if (optarg[0] == Transpose)
	  TransC = 1;
	else if (optarg[0] == Lower || optarg[0] == Upper) {
	  FormC = optarg[0];
	  IdenticalAB=1;
        }
									break;
      case 'f':
	f = 0;
	while ((strcmp(fs[f], optarg) != 0) && (strcmp(fs[f], "") != 0)) {
	  f++;
        }	
	if (strcmp(fs[f], optarg) != 0) 
	  OptionError(c, "invalid function mnemonic");
									break;
      case 'r':
	NReps = atoi(optarg);						break;
      case 'X':
	TReps = atoi(optarg);						break;
      case 's':
        if (optind + 1 < argc) {
	  StepM = atoi(argv[optind]); 	StepN = atoi(argv[optind+1]); 
	  optind += 2; 
          StepOpt=1;
        } else 
	  OptionError(c, "StepM, StepN or StepK parameters are missing");
								 	break;
      case 'g':
        if (optind + 1 < argc) {
	  Mg = atoi(argv[optind]); 	Ng = atoi(argv[optind+1]); 
	  optind += 2; 
          GlobDimOpt=1;
        } else 
	  OptionError(c, "StepM, StepN or StepK parameters are missing");
								 	break;
      case 'p':
        if (optind + 1 < argc) {
	  PrintM = atoi(argv[optind]); 	PrintN = atoi(argv[optind+1]); 
	  optind += 2;
        } else 
	  OptionError(c, "PrintM or PrintN  parameters are missing");
								 	break;
      case 't':   
        TraceLevel = atoi(optarg);                                      break;
      case 'M':
        if (optind + 1 < argc) {
          MachParamSet(atoi(argv[optind]), atoi(argv[optind+1]));
          optind += 2;
        } else 
          OptionError(c, "n or v parameters are missing");
                                                                        break;
      case 'v':
        ForceAlgVers = atoi(optarg);                                    break;
      case '?':
	OptionError(c, "invalid option given");				break;
    }

  if (optind+2 >= argc || argc<=3) { 
    if (argc > 1) 
      HPRINT((STDOUT, "%s: insufficient parameters M N K given.\n", PROGNAME));
    PrintUsage();
    exit(argc>1? 2: 0);
  }

  /* get non-optiinal parameters */
  M = atoi(argv[optind]);	N = atoi(argv[optind+1]);
  K = atoi(argv[optind+2]);

  if (optind+3 < argc) 
    alpha = Str2FD(argv[optind+3]);

  if (! StepOpt) {
    StepM = M; StepN = N;
  } 
  if ( ! GlobDimOpt ) {
    Mg = M; Ng = N;
  } else {
    Mg = MAX(Mg, M); Ng = MAX(Ng, N);
  }
} /* GetParameters() */


static void OptionError(option, msg)
  char *msg, option;
/*post: print out option error message  with msg & option; print usage & exit */
{ HPRINT((STDOUT, "%s: error in parsing option -%c: %s\n", PROGNAME, option, msg));
  PrintUsage();
  exit(2);
}

static void PrintParameters();

static void PrintUsage()
/*post: prints program usage to stdout*/
{ HPRINT((STDOUT, "usage : %s %s\n", PROGNAME, USAGE)); 
  PrintParameters();
}


static void CheckParameters()
/* post: check values of parameter variables, aborting if invalid */
{ if (M < 0 || N < 0 || K < 0) {
    HPRINT((STDOUT, "%s: invalid (-ve) matrix dimensions (M=%d, N=%d, K=%d)\n",
 	   PROGNAME, M, N, K));
    PrintUsage();
    exit(2);
  }
  if (StepM <= 0 || StepN <= 0) {
    HPRINT((STDOUT, "%s: invalid (0 or -ve) dimensions step sizes (StepM=%d, StepN=%d)\n",
           PROGNAME, StepM, StepN));
    PrintUsage();
    exit(2);
  }
  if (IdenticalAB && ((TransA==TransB) || (M!=N))) {
    HPRINT((STDOUT, "%s: -i specifies B identical to A; requires A^T B or B^T A compution or (M=%d the same as N=%d)\n",
           PROGNAME, M, N));
    PrintUsage();
    exit(2); 
  }
} /* CheckParameters() */


static void PrintParameters()
/*post: print out current values of variables set by command line to stdout*/
{ HPRINT((STDOUT, "%s: %dx%d reps of C = %s A%s*%s%s + C; %s %s M,N,K = %d(%d:%d),%d(%d:%d),%d, fn=%s\n",
	 PROGNAME, TReps, NReps, pFD2Str(&alpha), (TransA)? "^T": "", IdenticalAB? "A": "B", 
	 (TransB)? "^T": "",(SingleWordBoundary? "(swb)": ""), 
	RandElts? "(rand)": "", M, Mg, StepM, N, Ng, StepN, K, fs[f])); 
} /* PrintParameters() */
