
/************************************************************************/
/*			 	MM.cell.c				*/
/************************************************************************/
/* program for testing performance and correctness of various matrix
 * multiply operations. 
 * performs M x N C = alpha*AB (A is M x K); A, B initialized to regular values,
 * then checks C with simple matrix multiplication algorithm (D = A B);
 *  	matrices A, B, C, D contiguous in memory; thus if procedures 
 *	overwrites memory immediately before or after C, check should fail. 
 *  (externally-defined) cpp FLAGs:
 *	FDATATYPE	SINGLE, DOUBLE
 *
 *	TRACE_BLAS	trace blas calls
 *	GEMM_ONLY	compile out calls to DGER() DGEMV() when only GEMM
 *		  	is available. 
 */
/* SCCS INFO @(#)MM.cell.c 1.5         last modified P. Strazdins, 00/08/28
   first created: P. Strazdins 92/01
 */
/************************************************************************/
#include "Cell.h"
#include "Timer.c"

#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "LocMat.h"
#include "FDAssert.h"
#include "FDMalloc.h"
#include "CellMatrix.h"
#include "FDCommun.h"

#include "DistMat.h"
#include "CellTriMatrix.h"

#define PROGNAME "MM.cell"
#include "MM.Parameters.c"

#include "FDATAOps.h"
#if defined(BLAS) || defined (BLAS_MM)
# include "FtnBLAS.h"
#endif /*BLAS*/
# include "BLASInnerLoops.h"

#include "AuxMatrix.c"

static double Min(a, n)
 double a[]; int n;
{ int i; double m=a[0];
  for (i=1; i < n; i++)
   if (a[i] < m) m = a[i];
  return (m);
}

static double Avg(a, n)
 double a[]; int n;    
{ int i; double m=a[0];
  for (i=1; i < n; i++)
   m += a[i];
  return (m/n);
}

static double var(a, n)
  double a[]; int n;
{ int i; double s=0.0, v=Avg(a, n), sqrt();
  for (i=1; i < n; i++) {
    double d = a[i] - v; 
    s += d*d;
  }
  s = sqrt(s/(n-1));
  return(s);
}

/********************************* MM test procedures ********************/
/* these procedures share the following pre/post conditions:
 * pre:  A' is m x k, B' is k x n, C' is m x n
 *       A' = (TransA? A^T: A), with row-major storages assumed, 
 *	 ie LdA is 2nd dimension of A; similarly for B' & C'
 * post:  C += A' B' 
 */
#define Trans_lm(A) (LM_RInc(A)==1)

#ifndef BLAS             
static void TestURMult(m, n, k, A, B, C)
  int m, n, k;
  LocMat A, B, C;
/* pre:  RECT_X | n && RECT_Y | m
 */
{ register int i, j, di=RECT_Y, dj=RECT_X, tAB;		int bounds[3];
  register FDATA *Ax, *Bx, *Cx;		register int Ainc, Binc, Cinc;
#if FDATATYPE < COMPLEX
  if (Trans_lm(C)) {
    SWAP(LocMat, A, B);                 SWAP(int, LM_RInc(A), LM_CInc(A));
    SWAP(int, LM_RInc(B), LM_CInc(B));  SWAP(int, LM_RInc(C), LM_CInc(C));
    SWAP(int, m, n);                    
  } 
  bounds[0] = LM_Ld(A); bounds[1] = LM_Ld(B); bounds[2] = LM_Ld(C);
  
  tAB = Trans_lm(A) + 2*Trans_lm(B);
  if (tAB==3) {
    SWAP(int, bounds[0], bounds[1]);
    SWAP(int, di, dj);
  }
  Binc = dj*LM_CInc(B);			Cinc = di*LM_RInc(C);
  Ainc = di*LM_RInc(A); 
  Cx = LM_ADR(C);	  		Ax = LM_ADR(A); 
  for (i=0; i < m; i+=di) {
    Bx = LM_ADR(B); 
    for (j=0; j < n; j+=dj) {
      if (tAB == 0)
        UpdateRect(k, 0, bounds, Ax, Bx, Cx+j);
      else if (tAB == 1)
        UpdateRectTr1(k, bounds, Ax, Bx, Cx+j);
      else if (tAB == 2)
        UpdateRectTr2(k, bounds, Ax, Bx, Cx+j);
      else
	UpdateRect(k, 1, bounds, Bx, Ax, Cx+j);
      Bx += Binc;
    }
    Cx += Cinc; 	Ax += Ainc;
  } /* for (i...) */
#endif
} /* TestURMult() */
#endif BLAS

#include "MachineParams.h"
static void MatImpPartMult(M, N, K, alpha, A, B, beta, C)
  int M, N, K;
  pLocMat A, B, C; FDATA *alpha, *beta;
{ int i, j, m, n;  LocMat Aik, Bkj, Cij;
  /* Mw is well tuned for a 8 way set-associative 16K cache */
  int Mw = CACHESIZE/K - 2*MINOR_OUTER_VECSIZE;
  int k;
  LM_Copy(0, *A, Aik); LM_Copy(0, *B, Bkj);	LM_Copy(0, *C, Cij);
  for (k=0; k<K; k+=INNER_VECSIZE) {
    int k0 = MIN(K-k, INNER_VECSIZE);
    for (j=0; j < N; j+= Mw) {
      n = MIN(N-j, Mw);
      for (i=0; i < M; i+= MINOR_OUTER_VECSIZE)  {
        m = MIN(M-i, MINOR_OUTER_VECSIZE);
        MatMult(m, n, k0, alpha, pLocSubMat_x(&Aik, A, i, k), 
	        pLocSubMat_x(&Bkj, B, k, j),
	        beta, pLocSubMat_x(&Cij, C, i, j));
      } /* for (j..) */
    } /* for (k...) */   
  }  
} /* MatImpPartMult() */


static void TestMVMult(StepM, StepN, m, n, k, alpha, A, B, C)
  int m, n, k, StepM, StepN;
  LocMat A, B, C; FDATA *alpha;
{ 
  register int i, j, Ainc, Cinc;	int bounds[3];
  FDATA *ax, *bx, *cx;

  if (Trans_lm(A) && Trans_lm(B)) {
    SWAP(LocMat, A, B);			SWAP(int, LM_RInc(A), LM_CInc(A));
    SWAP(int, LM_RInc(B), LM_CInc(B));  SWAP(int, LM_RInc(C), LM_CInc(C));
    SWAP(int, m, n);  			SWAP(int, StepM, StepN);
  }

  if (!Trans_lm(A)) {
    bounds[0] = LM_RInc(A); bounds[1] = LM_RInc(B); bounds[2] = LM_RInc(C);
    Ainc = StepM*LM_RInc(A);
    Cinc = StepM*LM_RInc(C);

    ax = LM_ADR(A);
    for (i=0; i<m; i+=StepM) {
      cx = LM_ADR(C);     		bx = LM_ADR(B);
      if (StepM > (m-i)) StepM = m-i;
      for (j=0; j < n; j++) {
#ifndef BLAS
# if FDATATYPE < COMPLEX
        MatVecMul(StepM, k, bounds, ax, bx, cx);
# endif
#else
# ifdef TRACE_BLAS
	printf("_GEMV(T,%d,%d,%s,a%d %d,a%d %d,1.0,a%d %d)\n",
		k, StepM, pFD2Str(alpha), FD_ADR(ax), bounds[0], FD_ADR(bx), bounds[1],
	        FD_ADR(cx), bounds[2]);
# endif 
#ifndef GEMM_ONLY
	FTN_GEMV("T", &k, &StepM, alpha, ax, &bounds[0], bx, &bounds[1], 
		 &FD_1, cx, &bounds[2]); 
#endif
#endif
        cx += LM_CInc(C); 		bx += LM_CInc(B);
      }
      LM_ADR(C) += Cinc; 			ax += Ainc;
    } /* for (i...) */
  } else {				/* !Trans_lm(B) && Trans_lm(A) */
    bounds[0] = LM_CInc(A); bounds[1] = LM_RInc(B); bounds[2] = LM_CInc(C);
    Cinc = StepN*LM_CInc(C);
    bx = LM_ADR(B);
    for (j=0; j<n; j+=StepN) {
      cx = LM_ADR(C);      			ax = LM_ADR(A);
      if (StepN > n-j) StepN = n-j;
      for (i=0; i < m; i++) {
#ifndef BLAS 
# if FDATATYPE < COMPLEX
        VecMatMul(k, StepN, bounds, ax, bx, cx);
# endif
#else
# ifdef TRACE_BLAS
	printf("_GEMV(N,%d,%d,%s,a%d %d,a%d %d,1.0,a%d %d)\n",
		StepN, k, pFD2Str(alpha), FD_ADR(bx), bounds[1], FD_ADR(ax), bounds[0],
	        FD_ADR(cx), bounds[2]);
# endif 
#ifndef GEMM_ONLY
	FTN_GEMV("N", &StepN, &k, alpha, bx, &bounds[1], ax, &bounds[0], 
		 &FD_1, cx, &bounds[2]);
#endif
#endif
        cx += LM_RInc(C); 			ax += 1;
      }
      LM_ADR(C) += Cinc; 			bx += StepN;
    } /* for (j...) */
  } /* if (!Trans_lm(A)) */
} /* TestMVMult() */


static void TestR1Mult(StepM, StepN, m, n, k, alpha, A, B, C)
  int m, n, k, StepM, StepN;
  LocMat A, B, C; FDATA *alpha;
{ register int i, j, k1, AInc, BInc; 	int StepMLdA;
  register FDATA *ax, *axk, *bx, *cx; 	int bounds[3];
#ifndef BLAS
  if (Trans_lm(C)) 
#else
  if (!Trans_lm(C))
#endif
  {
    SWAP(LocMat, A, B);                 SWAP(int, LM_RInc(A), LM_CInc(A));
    SWAP(int, LM_RInc(B), LM_CInc(B));  SWAP(int, LM_RInc(C), LM_CInc(C));
    SWAP(int, m, n);                    SWAP(int, StepM, StepN);
  }
  bounds[0] = LM_RInc(A); bounds[1] = LM_CInc(B); bounds[2] = LM_Ld(C);
  AInc = StepM*LM_RInc(A);		BInc = StepN*LM_CInc(B);

  for (j=0; j<n; j+=StepN) {
    cx = LM_ADR(C);     			ax = LM_ADR(A);
    StepMLdA = StepM;
    if (StepN > n-j) StepN = n-j;
    for (i=0; i<m; i+=StepM) {
      if (StepM > m-i) 
	StepMLdA = m-i;
      axk = ax;			      	bx = LM_ADR(B); 
      for (k1=0; k1 < k; k1++) {
#ifndef BLAS
# if FDATATYPE < COMPLEX
        Rank1Update(StepMLdA, StepN, bounds, axk, bx, cx);
# endif
#else
# ifdef TRACE_BLAS
	printf("_GER(%d,%d,%s,a%d %d,a%d %d,a%d %d)\n",
		StepMLdA, StepN, pFD2Str(alpha), FD_ADR(axk), bounds[0], 
		FD_ADR(bx), bounds[1], FD_ADR(cx), bounds[2]);
# endif
#ifndef GEMM_ONLY
	FTN_GER(&StepMLdA, &StepN, alpha, axk, &bounds[0], bx, &bounds[1],
 		cx, &bounds[2]);
#endif
#endif
    	axk += LM_CInc(A); 		bx += LM_RInc(B);
      }
      ax += AInc;		        cx += StepM*LM_RInc(C); 
    } /* for (i...) */	 
    LM_ADR(B) += BInc;			LM_ADR(C) += StepN*LM_CInc(C);
  } /* for (j...) */
} /* TestR1Mult() */


#ifndef BLAS
static void TestRKMult(StepM, StepN, m, n, k, alpha, A, B, C)
  int m, n, k, StepM, StepN;
  LocMat A, B, C; FDATA *alpha;
{ register int i, j, k1, AInc, BInc;
  register FDATA *ax, *axk, *bx, *cx; 	int bounds[6];
  bounds[0] = LM_RInc(A); bounds[1] = LM_RInc(B); bounds[2] = LM_RInc(C);
  bounds[3] = LM_CInc(A); bounds[4] = LM_CInc(B); bounds[5] = LM_CInc(C);

  if (FD_EQ(*alpha, FD_1))  alpha = NULL;

  AInc = StepM*LM_RInc(A);		BInc = StepN*LM_CInc(B);
  for (j=0; j<n; j+=StepN) {
    cx = LM_ADR(C);     			ax = LM_ADR(A);
    if (StepN > n-j) StepN = n-j;
    for (i=0; i<m; i+=StepM) {
      axk = ax;			      	bx = LM_ADR(B); 
      for (k1=0; k1 < k; k1+=RKSIZE) {
# if 0
        RankKUpdate(MIN(StepM, m-i), StepN, bounds, axk, bx, alpha, cx);
# endif
    	axk += RKSIZE*LM_CInc(A); 	bx += RKSIZE*LM_RInc(B);
      }
      ax += AInc;		        cx += StepM*LM_RInc(C); 
    } /* for (i...) */	 
    LM_ADR(B) += BInc;			LM_ADR(C) += StepN*LM_CInc(C);
  } /* for (j...) */
} /* TestRKMult() */
#endif BLAS

#undef Trans
/*********************** general MM checking routine *******************/ 

static void MatMultS(M, N, K, alpha, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta; LocMat A, B, C;
/* pre:  LM_IncOKUnit(M, K, A) && LM_IncOKUnit(K, N, B) &&
 *       LM_IncOKUnit(M, N, C) && 0 <= M, N, K,
 *       !Overlap(M, K, A, M, N, C) && !Overlap(K, N, B, M, N, C)
 * post: C := (*beta)C + (*alpha)A*B;  A is MxK, B is KxN, C is MxN;
 */
{ int i, j, k; FDATA *Ax = LM_ADR(A), *Cx = LM_ADR(C);
  for (i = 0; i < M; i++) {
    FDATA *b = LM_ADR(B), *cx = Cx;
    for (j=0; j < N; j++) {
      FDATA *ax=Ax, *bx=b; FDATA c = FD_0;   
      for (k=0; k < K; k++) {
        FD_MULSET(c, +=, *ax, *bx);
        ax += LM_CInc(A); bx += LM_RInc(B);
      }
      FD_MULSET(*cx, =, *cx, *beta);
      FD_MULSET(*cx, +=, c, *alpha);
      b += LM_CInc(B);  cx += LM_CInc(C);
    } /* for (j...) */
    Cx += LM_RInc(C); Ax+= LM_RInc(A);
  } /* for (i...) */
} /* MatMultS() */


/**************************************** main program ******************/

static FDATA AIMULT = FD_MKVAL(2.0, -1.0),	
	     AJMULT = FD_MKVAL(1.0, -2.0),
	     AIJOFFS= FD_MKVAL(1.0, -1.0);

main(argc,argv) 
  int argc; char *argv[]; 
{
	LocMat A, B, C, D;	     	/* D used for checking */
	DistMat_RVAlloc(3);
	DistMat dC, dD;			/* for triangular operations */
	FDATA FD_alpha, *palpha;
	FD_BASE mxd=0.0, avd=0.0; int NErrs, mxi=0, mxj=0; /* error values */
	int i, j;
        double flops, speed, *tcomp, *utcomp;	/* timing */
	char Date[64];
 
        CELL_INIT;
        STDOUT_SETUP(STDOUT);
        InitArgs(argc, argv);
        GetDate(Date);

  	GetParameters(argc, argv);
	COMMUN_INIT(Ncely, Ncelx);
	for (i=0; i < argc; i++)
	  HPRINT((STDOUT, "%s ", argv[i]));
	HPRINT((STDOUT, "%s", Date));
  	PrintParameters(); 
	CheckParameters();            

	FD_alpha = (FDATA) alpha;
        palpha = FD_EQ(FD_alpha, FD_1)? NULL: &FD_alpha;

# define ASIZE (M*K)
# define BSIZE (IdenticalAB? 0: (K*N))
	LM_Init(A, NULL, (TransA)? 1: K, (!TransA)? 1: M);
	LM_Init(B, NULL, (TransB)? 1: N, (!TransB)? 1: K);
#if 0
if (M==1 && K==1 && N==1)
  LM_CInc(A) = 2, LM_CInc(B)=2; 
#endif
	MatMalloc(PROGNAME, 1, &(LM_ADR(A)), ASIZE+BSIZE+SingleWordBoundary+Mg*Ng);
        LM_ADR(A) += SingleWordBoundary;
	LM_ADR(B) = LM_ADR(A) + (IdenticalAB? 0: ASIZE); 

	LM_Init(C, LM_ADR(A)+ASIZE+BSIZE, 
		(TransC && Mg>1)? 1: Ng, (!TransC && Ng>1)? 1: Mg); 

  	dC = InitDimMat(0, 1, 1, CID_I, 1, CID_J, 1, 0, 0);
	DM_Loc(dC) = C;

	MatZero(M, N, &C); 
	MatInit(RandElts, AIMULT, AJMULT, AIJOFFS, K, N, B);
	MatInit(RandElts, AIMULT, AJMULT, AIJOFFS, M, K, A);
	MatPrint("A", 0, 0, A);		MatPrint("B", 0, 0, B);

	utcomp = (double *) memalign(sizeof(double), sizeof(double)*TReps);
	tcomp = (double *) memalign(sizeof(double), sizeof(double)*TReps);
	for (i=0; i < TReps; i++) {
	start_timer();
	if (f==0) 
	  if (FormC)
	    for (j=NReps; j!=0; j--) 	
	      TriMatMult(M, N, K, palpha, &A, &B, NULL, FormC, dC);
	  else 
	    for (j=NReps; j!=0; j--) 	
	      MatMult(M, N, K, palpha, &A, &B, NULL, &C);
	else if (f==1)
	  if (FormC)
	    for (j=NReps; j!=0; j--) {	
#if 1 /*ndef BLAS*/
	      TriMatPartMult(M, N, K, palpha, &A, &B, NULL, FormC, dC);
#else
# include "FtnBLAS.h"
	      int LdA = LM_Ld(A), 	LdC = LM_Ld(C); 
	      char TrA = ((LM_RInc(A)==1) != (LM_CInc(C)==1))? 
			  NoTranspose: Transpose;
	      char FrmC = (LM_CInc(C)!=1)? FormC: (FormC==Upper? Lower: Upper); 
#ifndef GEMM_ONLY
#if 1
              FTN_SYRK(&FrmC, &TrA, &N, &K, &FD_alpha, LM_ADR(A), &LdA,  
		       &FD_1, LM_ADR(C), &LdC);
#else
	      LdA = LM_RInc(A);
	      FTN_SYR(&FrmC, &N, &FD_alpha, LM_ADR(A), &LdA, LM_ADR(C), &LdC); 	
#endif

#endif
#endif
            }      
	  else
	    for (j=NReps; j!=0; j--) 	
	      MatPartMult(M, N, K, palpha, &A, &B, NULL, &C);
	else if (f==2)
#ifndef BLAS
	  for (j=NReps; j!=0; j--) 	
	    TestURMult(M, N, K, A, B, C);
#else	
	    if (FormC) {
	      int LdA = LM_Ld(A), 	LdC = LM_Ld(C); 
	      char TrA = ((LM_RInc(A)==1) != (LM_CInc(C)==1))? 
			  NoTranspose: Transpose;
	      char FrmC = (LM_CInc(C)!=1)? FormC: (FormC==Upper? Lower: Upper); 
#ifndef GEMM_ONLY
              FTN_SYRK(&FrmC, &TrA, &N, &K, &FD_alpha, LM_ADR(A), &LdA,  
		       &FD_1, LM_ADR(A), &LdC);
#endif
	    } else
	      MatMult(M, N, K, palpha, &A, &B, NULL, &C);
#endif BLAS
	else if (f==3)
	  for (j=NReps; j!=0; j--) 	
	    TestMVMult(StepM, StepN, M, N, K, &alpha, A, B, C);
	else if (f==4)
	  for (j=NReps; j!=0; j--) 	
	    TestR1Mult(StepM, StepN, M, N, K, &alpha, A, B, C);
	else if (f==5)
	  for (j=NReps; j!=0; j--) 	
#ifndef BLAS
	    TestRKMult(StepM, StepN, M, N, K, &alpha, A, B, C);
#else
	  { int k; FDATA *Ax = LM_ADR(A), *Bx = LM_ADR(B);
	    for (k=0; k < K; k+= StepM) {
	      MatMult(M, N, MIN(K-k, StepM), palpha, &A, &B, NULL, &C);
	      pLocSubMat(&A, 0, StepM); pLocSubMat(&B, StepM, 0);
  	    } /* for (k..) */
 	    LM_ADR(A) = Ax; LM_ADR(B) = Bx;
	  }
#endif BLAS
	else if (f==6)
	  for (j=NReps; j!=0; j--)
	    MatImpPartMult(M, N, K, palpha, &A, &B, &FD_1, &C);
	tcomp[i] = end_timer(0);	utcomp[i] = UserEndTimerVal;
	} /* for (i...) */

	flops = (double) M*N*(2*K+(!FD_EQ(FD_alpha, FD_1) && !FD_EQ(FD_alpha, FD_M1)));
	if (FormC) flops /= 2.0;
        speed = NReps * flops / 1.0E6 / Min(tcomp, TReps);

        HPRINT((STDOUT, "Computed %dx%d reps of %dx%dx%d multiply (%d FLOPs) in %Gs (avg=%G) at %G MFLOPs\n",
                TReps, NReps, M, K, N, (int) flops, Min(tcomp, TReps), 
		Avg(tcomp, TReps), speed));

        speed = NReps * flops / 1.0E6 / Min(utcomp, TReps);
        HPRINT((STDOUT, "       user time in %Gs (avg %Gs) at %G MFLOPs\n",
                Min(utcomp, TReps), Avg(utcomp, TReps), speed));
	if (TReps>1) {
	  HPRINT((STDOUT, "%%var in t=%.3f (ut=%f); speed for 1st %d reps: \n", 
		  var(tcomp, TReps)/Avg(tcomp, TReps), 
		  var(utcomp, TReps)/Avg(utcomp, TReps), MIN(10, TReps)));
	  for (i=0; i < MIN(10, TReps); i++)
            HPRINT((STDOUT, "%.2f (%.2f)", NReps * flops / 1.0E6 / tcomp[i],
		    NReps * flops / 1.0E6 / utcomp[i]));
	  HPRINT((STDOUT, "\n"));
        } 

       dD = NewMat(M, N, dC, NULL);          
       D = DM_Loc(dD);
       MatZero(M, N, &D);

	MatMultS(M, N, K, &FD_alpha, A, B, &FD_1, D);
        FD_SET_REAL(FD_alpha, (FD_BASE) NReps*TReps);
        FD_SET_IMAG(FD_alpha, 0.0);
	if (FormC) {
	  TriMatScale(M, N, &FD_alpha, FormC, dD);
	  TriMatScale(M-1, N-1, &FD_0, FormC==Upper? Lower: Upper,
		      SubMat(dD, FormC==Upper, FormC!=Upper));
	} else
          MatScale(M, N, &FD_alpha, &D); /* nb. may introduce numerical diffs */

        NErrs = MatDiff(M, N, C, D, &mxd, &avd, &mxi, &mxj);  
	HPRINT((STDOUT, "NErrs=%d, Max diff is %.G at (%d,%d), av diff is %G\n", 
		NErrs, mxd, mxi, mxj, avd));
	HPRINT((STDOUT, "\n\n"));

	MatPrint("C", MIN(M, PrintM) , MIN(N, PrintN), C);
	MatPrint("D", MIN(M, PrintM) , MIN(N, PrintN), D);

        CELL_FINALIZE;
	exit(NErrs!=0);
} /* main() */

