/************************************************************************/
/*                              MatFact.Parameters.c                    */
/************************************************************************/
/* SCCS INFO %Z%%M% %I%         last modified P. Strazdins, %E%
   first created: P. Strazdins 96/06
*/
/* command line parameters for testing Ncely x Ncelx Distributed 
 * Matrix Factorizations
 * usage:
 *       PROGNAME         # print out options, possible values & defaults
 *       PROGNAME USAGE
 */
#define USAGE \
"[-f F] [-T] [-C] [-l] [-D DiagBias] [-F matfile] [-B BM BN] [-w MaxW] [-2] \n\
	[-M n v] [-t trlvl trtype] [-h] [-v AlgVers] [-E] [-p MFPrintRes] 	 \n\
 	[-r r dN] [-e RndSeed] [-c ChkLvl] [-G Ncely Ncelx] M [N] \n\
with default values: \n\
	MaxW=BN, M=N, r=1,dN=0, BM=BN=1, F='L'('P','Q','D','T'), DiagBias=0.0 RndSeed=0, ChkLvl=1, trlvl=-1 trtype=AlgVers=0 MFPrintRes=0 EMx=0"

				/* values of ChkLvl */
#define CHK_NONE 	0	
#define CHK_NRM  	1	
#define CHK_INV  	2	
#define CHK_TRMM  	3	

static char *ChkLvls[] = {
	"no checking", 
	"via ||Y-Y'||/e*N*||A||*||Y||, X=AY, Y'= A^1 X",
	"recover A via inv. fact'n",
	"recover A via TRMM",
};

/* This specifies a matrix factorization of an M x N global matrix A
 * with block-scattered storage with BMxBN blocks and block& cell offsets (0,0).
 * The -l option specifies A[i,j] = a*i+b*j for some a,b; otherwise
 * A[i,j] is pseudo-random in [0,1) with seed set by a*i+b*j.
 * Diagonal elements of A have DiagBias added onto them; this is useful to
 * ensure +ve definiteness (for Cholesky).
 * If -F is specified, A is read from the binary file matfile (order
 * assumed to be column major and for symmetric, only lower half elements).
 * The -T option set A to be globally transposed, ie. M (N) along cell dim. 1 (0).
 * The -C option forces local part of A to be in column-major storage.
 * The value of F='L','P','Q',D' chooses a LU,Cholesky,QR, LDL', LTL'  
 * factorization of A resp;
 * with a blocking factor of MaxW (unless -2 specified, in which the level-2
 * factorization routine is used (effectively equivalent to MaxW=1).
 * 
 * if -E specified, error checking is suppressed in the DBLAS.
 * 
 * MFPrintRes x MFPrintRes of the local portion of the factored matrices is 
 * printed out. If -n specifed, no checking of result is performed.
 * If -s is specified, (slow) TRMM_based checking is performed (LU & Cholesky only);
 * otherwise checking is perfomed via reversing the factorization alg.
 *
 * The -M option may be repeated with differing n values; all these  
 * can set the DBLAS machine parameters before the computation begins.
 * The -t option is used for DBLAS trace level setting.   
 * Setting EMx > 0  will make DBLAS flag an error
 * if received message buffers contain data of greater absolute value;
 * this is only useful if, given A[i,j] in [0,1), all elements of
 * partially factored A can be within a given absolute value.
 * if -r specified, computations will be repeated for MxN, (M+dN)x(N+dN) .. ,
 * (M+(r-1)*dN)x(N+(r-1)*dN)
 * If -h specified, the host on which each process runs is printed out
 */

/* command line parameters; see header comment*/
static int M, N, MaxW, Ncely, Ncelx;
static int r=1, dN = 0, ChkLvl = CHK_NRM;
static int BM=1, BN=1, RandElts=1, TransA=0, ColMajA=0;
static int HostPrint=0, Level2=0, MFPrintRes=0, SlowCheck=0, NoCheck=0;  
static char FactType='L'; 	
static FDATA DiagBias;;
static char *matfile = NULL;
static int SymmetricA = 0;
#include "FDAssert.h"   	/* extern int TraceLevel, TraceType, ForceAlgVers */
#include "MachineParams.h"      /* for -M oprtion */

void PrintUsage()
/* post: prints program usage to stdout	*/
{ HPRINT((STDOUT, "usage: %s %s\n", PROGNAME, USAGE));
}

void OptionError(option, msg)
  char *msg, option;
/* post: print out option error message  with msg & option; print usage & exit*/
{ HPRINT((STDOUT, "%s: error in parsing option -%c: %s\n", PROGNAME, option, msg));
  PrintUsage();
  exit(1);
}

void GetParameters(argc, argv)
  int argc; char *argv[];
/* post: set MatFact command line variables according to argv[0..argc-1] */
{ extern char *optarg;
  extern int optind;
  extern int RandSeedOffset;
  int c; 				int GotDiag=0;
  MaxW = 0;				DiagBias = FD_0;
  Ncely = NP_I, Ncelx = NP_J;
  while ((c = getopt(argc, argv, "f:F:rTCD:lBw:2Mtp:v:e:sc:EGh")) != -1) /* parse options */
    switch (c) {
      case 'G':
         if (optind + 1 < argc) {
           Ncely = atof(argv[optind]);      Ncelx = atof(argv[optind+1]);
           optind += 2;
         }                                                              break;
      case 'f':
	FactType=optarg[0];				break;
      case 'F':
	matfile=optarg;					break;
      case 'T':
	TransA=1;					break;
      case 'C':
	ColMajA=1;					break;
      case 'D':
	DiagBias=Str2FD(optarg);	GotDiag=1;	break;
      case 'l':
	RandElts=0;					break;
      case 'E':
	NoOuterErrCheck=1;				break;
      case 'h':
	HostPrint=1;					break;
      case '2':
	Level2=1;					break;
      case 'w':
	MaxW=atoi(optarg);				break;       
      case 'B':
        if (optind + 1 < argc) {
          BM = atoi(argv[optind]);   BN = atoi(argv[optind+1]);
          optind += 2; 
        } else 
          OptionError(c, "BM, BN parameters are missing");
                                                                        break;
      case 'r':
        if (optind + 1 < argc) {
          r = atoi(argv[optind]);   dN = atoi(argv[optind+1]);
          optind += 2; 
        } else 
          OptionError(c, "r, dN, parameters are missing");
                                                                        break;
      case 'c':
	ChkLvl = atoi(optarg);					break;
      case 't':
        if (optind + 1 < argc) {
          TraceLevel = atoi(argv[optind]);   TraceType = atoi(argv[optind+1]);
          optind += 2; 
        } else 
          OptionError(c, "trlvl, trtype parameters are missing");
        								break;
      case 'M':
        if (optind + 1 < argc) {
          MachParamSet(atoi(argv[optind]), atoi(argv[optind+1]));
          optind += 2;
        } else
          OptionError(c, "n or v parameters are missing");
                                                                        break;
      case 'p':  
        MFPrintRes = atoi(optarg);   PrintRes = MFPrintRes;
	                                     				break;
      case 'v':
	ForceAlgVers=atoi(optarg);					break;
      case 'e':  
        RandSeedOffset = atoi(optarg);   
									break;
      case '?':
        OptionError(c, "invalid option given");                         break;
    }
        
  if (optind >= argc || argc <= 1) {
    if (argc > 1)
      HPRINT((STDOUT, "%s: insufficient parameters M ... given.\n", PROGNAME));
    PrintUsage();
    exit(1);  
  }
          
  M = atoi(argv[optind]);
  if (argc > optind+1) N=atoi(argv[optind+1]); else N=M;
  if (MaxW==0) MaxW = BN;

  if (ChkLvl < CHK_NONE || ChkLvl > CHK_TRMM)
    ChkLvl = CHK_NRM;

  if (FactType=='P' && !GotDiag)
    FD_SET_REAL(DiagBias, 5.0 + (FD_BASE) (MAX(N, N+(r-1)*dN)/6.0));
  else if (!GotDiag && ChkLvl == CHK_NRM)
    FD_SET_REAL(DiagBias, 0.2);	/* reduce chance of a near-singular matrix */

  SymmetricA = FactType=='P' || FactType=='D'|| FactType=='T';
} /* GetParameters() */


void PrintParameters()
{ HPRINT((STDOUT, "%dx%d config %s factorization (%d reps,dN=%d) ,of %dx%d A%s\n	A is %s-major with %dx%d blocks, A(i,j) is %s in i,j with diag bias %s\n",
	 Ncely, Ncelx, FactType=='L'? "LU": FactType=='P'? "LL'": 
		FactType=='Q'? "QR": FactType=='D'? "LDL'": "LTL'" , 
	 r, dN, M, N, TransA? "^T": "", ColMajA? "col": "row", BM, BN, 
	 matfile!=NULL? matfile: RandElts? "random": "linear", pFD2Str(&DiagBias)));
  if (Level2) {
    HPRINT((STDOUT, "	Level 2 factorization will be used;"));
  } else {
    HPRINT((STDOUT, "	Max blocking factor is %d;", MaxW));
  }
  HPRINT((STDOUT, "checking method: %s.\n", ChkLvls[ChkLvl]));
} /* PrintParameters() */


void CheckParameters()
/* post: check values of parameter variables, aborting if invalid */
{  
  if (Ncely<=0 || Ncely<=0) {
    HPRINT((STDOUT, "%s: invalid cell configuration values (Ncely=%d, Ncelx=%d)\n",
           PROGNAME, Ncely, Ncelx));
    PrintUsage();
    exit(2);
  }
  if (FactType!='L' && FactType!='P' && FactType!='Q' && FactType!='T' &&
      FactType!='D') {
    HPRINT((STDOUT, "%s: invalid factorization type (F=%c)\n", PROGNAME, FactType));
    PrintUsage();
    exit(2);
  }
  if (M < 0 || N < 0) {
    HPRINT((STDOUT, "%s: invalid (-ve) matrix dimensions (M=%d, N=%d)\n",
           PROGNAME, M, N));
    PrintUsage();
    exit(2);
  }
  if (BM < 1 || BN < 1) {
    HPRINT((STDOUT, "%s: invalid block sizes (BM=%d, BN=%d)\n", PROGNAME, BM, BN));
    PrintUsage();
    exit(2);
  }
  if (SymmetricA && (M != N || BM != BN)) {
    HPRINT((STDOUT, "%s: f=%c requires symmetric matrix (M=N,BM=BN) but M=%d N=%d BM=%d BN=%d\n",
	   PROGNAME, FactType, M, N, BM, BN));
    PrintUsage();
    exit(2);
  }
  if (matfile != NULL && dN > 0) {
    HPRINT((STDOUT, "%s: warning: -F %s precludes step size dN=%d > 0, Setting dN to 0...\n",
	   PROGNAME, matfile, dN));
    dN = 0;
  } 
} /* CheckParameters() */

