/************************************************************************/
/*                              MatFact.cell.c                          */
/************************************************************************/
/* SCCS INFO %Z%%M% %I%         last modified P. Strazdins, %E%
   first created: P. Strazdins 96/06
*/

/* cell program for testing Ncely x Ncelx Distributed Matrix Factorizations
 * 	LU, Cholesky or QR
 * on an M x N matrix A with block-scattered storage with BMx BN blocks.
 * block size b and using blocking factor w.
 * If A is specified as a random matrix, A has values in [-1,1],
 * with the exception of the diagonal elements having values in the range [d, d+1)
 * (useful to ensure positive definite),
 * see MatFact.Parameters.c for how these variables can be set from the command line
 * (externally-defined) cpp FLAGs:
 *      FDATATYPE       SINGLE (default), DOUBLE
 *	LAPACK: compile to LAPACK factorization drivers (single cell only)
 * 	PBLAS: compile to ScaLAPACK factorization drivers
 *		* this program will force  col-major, no -T, PivColRep=1 
 *		* in LU.c, must set PivColRep=1
 *      PDESC_CACHE: for anu PBLAS only, put descriptor caching on
 *	
 *	COMB_FACT_SLV: for CheckLevel-CHL_NRM, augment X into A,
 *		to  reduce one of the check stages.
		NOTE: not compataible with PBLAS
 */
/*#define PBLAS*/
/*#define LAPACK*/
#define COMB_FACT_SLV

#if defined(PBLAS) || defined(LAPACK)	/* currently incompatible */
# undef COMB_FACT_SLV
#endif

#ifndef MAX_RESID
# define MAX_RESID 300
#endif

 
#include "FDATASetUp.h"
#include "Cell.h"

#include "Timer.c"
# include "../PBLAS/ScaLAPACKMisc.h"

#include "FDATAOps.h"
#include "BLASOptions.h"
#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDMalloc.h"
#include "CellMatrix.h"
#include "FDCommun.h"
#include "D_BLAS.h"

#include "../DLAPACK/D_LAPACK.h"

#include "HostMsg.c"
#include "AuxMatrix.c"
#include "AuxBlkMatrix.c"
#define PROGNAME "MatFact.cell"
#include "MatFact.Parameters.c"

#ifdef LAPACK
#include "../DLAPACK/LAPACK.h"
#endif

#ifdef PBLAS
# include "../PBLAS/PBLAS.h"

# ifdef PDESC_CACHE
# include "../PBLAS/PBLASCheck.h"
# endif
#endif


/**************************************** main program ******************/
void main (argc, argv)
  int argc; char *argv[];
{ 
  char Date[64], LogFileName[64], ResultMsg[256];
 
  double flops, speed, tcomp, flops_f, tcomp_f; 
  FD_BASE DAv=0.0, DMax=0.0;            /* av. & max. differences */
  int NErr=0, DMaxI=0, DMaxJ=0;         /* # errors; position of max. differences*/
  int i;

  CELL_INIT;
  STDOUT_SETUP(STDOUT);
  InitArgs(argc, argv);
  GetDate(Date);
  
  GetParameters(argc, argv);
  for (i=0; i < argc; i++)
    HPRINT((STDOUT, "%s ",  argv[i]));
  HPRINT((STDOUT, "\n"));

  COMMUN_INIT(Ncely, Ncelx);
  if (HostPrint) {
    printf("%d: cell (%d,%d) of %dx%d\n", CID, CID_I, CID_J, NP_I, NP_J);
    sprintf(ResultMsg, "/bin/echo \"%d:\" `/bin/uname -n`", CID);
    system(ResultMsg);
  }

#if MF_SEL==1
  FactType='L';
#elif MF_SEL==2
  FactType='P';
#elif MF_SEL==3
  FactType='Q';
#endif
  CheckParameters();
#ifdef PBLAS
  ColMajA = 1; TransA = 0;
#endif  
  PrintParameters();
  if (FactType=='Q' && ChkLvl==CHK_TRMM) { 
    HPRINT((STDOUT, "warning: checking method: %s not implemented for QR\n",
		     ChkLvls[ChkLvl]));
    ChkLvl=CHK_NRM;
  }
  if (FactType=='L' && ChkLvl==CHK_NRM && M!=N) { 
    HPRINT((STDOUT, "warning: checking method: %s not implemented for LU if M!=N\n",
		     ChkLvls[ChkLvl]));
    ChkLvl=CHK_INV;
  }
  if (BM!=BN && ChkLvl==CHK_NRM) {  
    HPRINT((STDOUT, "warning: checking method: %s not implemented if BM!=BN\n",
                     ChkLvls[ChkLvl]));
    ChkLvl=CHK_INV;
  }
  if (matfile!=NULL && (ChkLvl==CHK_TRMM || ChkLvl==CHK_INV)) {
    HPRINT((STDOUT, "warning: checking method: %s not implemented for -F %s\n",
                     ChkLvls[ChkLvl], matfile));
    ChkLvl=CHK_NRM;
  }
  if ((FactType=='T' || FactType=='D') && (ChkLvl==CHK_TRMM || ChkLvl==CHK_INV)) {
    HPRINT((STDOUT, "warning: checking method: %s not implemented for -f %d\n",
                     ChkLvls[ChkLvl], FactType));
    ChkLvl=CHK_NRM;
  }
  
  GetLogFileName(argv[0], Date, LogFileName);

  while (r > 0) {			/* repeat for various N */
    FDATA *WS; DistMat A, A1;		/* A1 is used for `slow' checking */
    int WSSize;				/* maximum workspace size*/
    int *Piv=NULL;                      /* LU permutation vectors */
    FDATA *Tau=NULL;			/* QR reflector vector */
    FDATA *DiagInv=NULL;		/* LDLT storage for col-replic D^-1 */
    int ma, na;				/* local matrix storage sizes */
    DistMat X, Y;			/* for norm-based checking*/
    int mx, nx;				/* local dimensions of X, Y */	
    FD_BASE Ynrm_1, Ynrm_inf;		/* for checking */ 
    DistMat_RVAlloc(5); 
#ifdef PBLAS
    int desca[DLEN_], ia=1, ja=1, info=0;
    int descx[DLEN_], ix=1, jx=1;
#define DESC_CONV(desc, M, N, A)                                        \
    desc[DT_] = 1; desc[CTXT_]=BLACSCntxt;  desc[M_]=(M); desc[N_]=(N); \
    desc[MB_] = d_B(DM_D0(A)); desc[NB_] = d_B(DM_D1(A));               \
    desc[RSRC_] = desc[CSRC_] = 0; desc[LLD_] = LM_CInc(DM_Loc(A));
#endif
    double LDLTCase1Frac, LDLTBlkPivFrac, LDLTSearchFrac;	
					/* for assessing LDLT pivoting strategies */
#ifdef COMB_FACT_SLV
    int MA = (FactType=='P')? (M+1): M;
    DistMat XA; 			/* part of A used to store X */
# define NA ((FactType=='D')? N: N+1)
#else
# define MA M
# define NA N
#endif
        
  sprintf(ResultMsg, "\n%.20s # %dx%d ", Date, Ncely, Ncelx);
  LOGFILE_APPEND(LogFileName, ResultMsg);
  Chv2Str(argv, 0, argc, ' ', ResultMsg, i/*dummy*/);
  LOGFILE_APPEND(LogFileName, ResultMsg);
  
  CommSetUp(Ncely, Ncelx);          
  A = InitDimMat(TransA, !TransA? BM: BN, !TransA? BN: BM, 
	         CID_I, Ncely, CID_J, Ncelx, 0, 0);
  ma = G2Ld(!TransA? MA: NA, DM_D0(A));	na = G2Ld(!TransA? NA: MA, DM_D1(A)); 
  if (ChkLvl==CHK_NRM) {
    Y = InitDimMat(TransA, !TransA? BN: 1, TransA? BM: 1,  CID_I, Ncely, CID_J, Ncelx, 0, 0);
    mx = G2Ld(!TransA? M: 1, DM_D0(Y));	nx = G2Ld(TransA? N: 1, DM_D1(Y)); 
  }

  if (N > 0) {
    if (FactType=='L' || FactType=='D' || FactType=='T')
      Piv = (int *) malloc(NA*sizeof(int));
    if (FactType=='Q')
      MatMalloc("MatFact.cell", 1, &Tau, NA); 
    else if (FactType=='D')
      MatMalloc("MatFact.cell", 1, &DiagInv, 3*G2Ld(MA, DM_DM(A))); 
  }

  WSSize =  MAX(ma, 2) * MAX(na, 2);
  if (ChkLvl==CHK_TRMM)
    WSSize *= 2;
  else if (ChkLvl==CHK_NRM) 
    WSSize += 2*mx*nx;
  MatMalloc("MatFact.cell", 1, &WS, WSSize);  
  LM_Init(DM_Loc(A), WS,  ColMajA? 1: MAX(na, 2), !ColMajA? 1: MAX(ma, 2));
  A1 = SubMat(A, 0, 0);
  if (ChkLvl==CHK_TRMM)
    LM_ADR(DM_Loc(A1)) += WSSize/2;
  if (ChkLvl==CHK_NRM) {
    LM_Init(DM_Loc(Y), WS+MAX(ma, 2) * MAX(na, 2), !TransA? 1: MAX(nx, 2), TransA? 1: MAX(mx, 2)); 
    X = SubMat(Y, 0, 0);
    if (mx*nx > 0) 
      LM_ADR(DM_Loc(X)) += mx*nx;
    else /* null out addresses - safer with DBLAS ReduceMat() */	
      LM_ADR(DM_Loc(X)) = LM_ADR(DM_Loc(Y)) = NULL;
  }
  
  if (matfile==NULL)
    MakeScattered(-RandElts, 0, 0, ma, na, &DiagBias, SymmetricA? "Low": "NoTrans", A);
  else
    ReadScattered(matfile, M, N, SymmetricA? 'L': 0, A);  

  if (ChkLvl==CHK_NRM) {
    MakeScattered(-RandElts, 0, 0, mx, nx, &FD_0, "NoTrans", Y); 	

    start_timer();
    if (!SymmetricA)
      D_GEMV(M, N, FD_1, A, Y, FD_0, X);
    else
      D_SYMV("Lower", N, FD_1, A, Y, FD_0, X);
#ifdef COMB_FACT_SLV
    if (FactType!='D') {
      XA = (FactType=='D'|| FactType=='P')? Trans(SubMat(A, N, 0)): 
					    SubMat(A, 0, N);
      D_COPY(N, X, XA);  
      if (FactType=='P') { /* ensure +ve definiteness of A|XA */		
        DistMat_RVAlloc(1);		DistMat ANN = SubMat(A, N, N); 
        D_SetElt(ANN, D_NRM2(N, X) + D_GetElt(ANN)); 
     }
   }      
#endif

    tcomp = end_timer();
    flops = 2.0*M*N;
    speed = flops / (1.0e6*tcomp) /Ncelx/Ncely;
    HPRINT((STDOUT, "Computed X = AY at in %Gs @ %G MFLOPS/cell\n", tcomp, speed));

    MatPrint("Y", MIN(mx, MFPrintRes),  MIN(nx, MFPrintRes), DM_Loc(Y));
    MatPrint("X", MIN(mx, MFPrintRes),  MIN(nx, MFPrintRes), DM_Loc(X));
  }
  MatPrint("A", MIN(ma, MFPrintRes),  MIN(na, MFPrintRes), DM_Loc(A));

  BaseTraceLevel=1;

#ifdef PBLAS
  DESC_CONV(desca, M, N, A);
  DESC_CONV(descx, M, 1, X);
# ifdef PDESC_CACHE
  PDescCacheOn();  
# endif
  MaxW = ABS(MaxW);			/* PBLAS codes cannot use MaxW<0 */
  PivColRep = TauRowRep = 1;
#endif  

  start_timer(); 
#if !defined(MF_SEL) || MF_SEL==1
  if (FactType=='L') {			/* A <- L\U where L*U = P*A */
    if (Level2)
      D_GETF2(M, NA, A, Piv); 		
    else {
# ifdef PBLAS
      setoptw_(&MaxW);
      P_GETRF(&M, &N, LM_ADR(DM_Loc(A)), &ia, &ja, desca, Piv, &info);
# else
#   ifdef COMB_FACT_SLV
      LU_SwapL=0;			/* suppress swaps in L */
#   endif
      D_GETRF(M, NA, A, Piv, MaxW);
# endif
    }	
  } else 
#endif
#if !defined(MF_SEL) || MF_SEL==3
  if (FactType=='Q') {		/* A <- Q\R where H*R = A, H=H(Tau,Q) */
    if (Level2)
      D_GEQF2(M, NA, A, Tau); 
    else {
# ifdef PBLAS
      int P = d_P(DM_DM(A)),		Q = d_P(DM_DN(A));
      int lwork = MaxW * ( (M+P+MaxW)/P + (N+Q+MaxW)/Q + MaxW);
      FDATA *Work;
      MatMalloc("MatFact.cell", 1, &Work, lwork); 
      setoptw_(&MaxW);
      P_GEQRF(&M, &N, LM_ADR(DM_Loc(A)), &ia, &ja, desca, Tau, Work, &lwork, &info);
      MatFree(Work);
# else
      D_GEQRF(M, NA, A, Tau, MaxW); 	
# endif
    }
  } else 
#endif
#if !defined(MF_SEL) || MF_SEL==2
  if (FactType=='P') {				/* A <- L   where L*L^T = A */
    if (Level2)
      D_POTF2(NA, A);		
    else {
# ifdef PBLAS
/*      extern char *TraceBLACS;
      TraceBLACS = "POTRF"; */
      setoptw_(&MaxW);
      P_POTRF(F2C_CHAR("L"), &N, LM_ADR(DM_Loc(A)), &ia, &ja, desca, &info);
/*    TraceBLACS = NULL;*/
# else
      D_POTRF(NA, A, MaxW);
# endif
    }
#endif
  } else if (FactType=='D') {
    int info = 0;
#if defined(LAPACK)
    int LdA = LM_CInc(DM_Loc(A));
    int One = 1, MOne =-1, i, LenWork;
    FDATA *Work;
#if 0					/* for non-standard run-time settable version ileanv()*/
    extern int ilaenvNB;
    ilaenvNB = MaxW;
#endif
    LenWork = ILAENV(&One, "ZSYTRF", "Lower", &N, &MOne, &MOne, &MOne, sizeof("ZSYTRF")-1);
    LenWork *= N;
    MatMalloc("SYTRF", 1, &Work, LenWork);
    _SYTRF(F2C_CHAR("L"), &N, LM_ADR(DM_Loc(A)), &LdA, Piv, Work, &LenWork, &info);
    MatFree(Work);
#elif defined(PBLAS)
    setoptw_(&MaxW);
    P_SYTRF(F2C_CHAR("L"), &N, LM_ADR(DM_Loc(A)), &ia, &ja, desca, 
	    DiagInv, Piv, &info);	
#else    
    D_SYTRF(NA, A, DiagInv, Piv, &info, MaxW);
#endif
#if defined(LAPACK) || defined(PBLAS)	/* convert to C pivot vector convention*/
    for (i=0; i<N; i++)
      if (Piv[i] > 0)
        Piv[i]--;
      else
        Piv[i]++;		 
#endif
    if (info > 0) {
      printf("SYTRF: non-invertible matrix, info=%d\n", info); 
      exit(1);
    }
  }
  tcomp = end_timer(1);

#ifdef PDESC_CACHE
  PDescCacheOff();  
#endif

  if (FactType=='L')
    flops = 1.0*N*N*(MAX(M,N) - MIN(M,N)/3.0);
  else if (FactType=='Q')
    flops = 2.0*N*N*(MAX(M,N) - MIN(M,N)/3.0);
  else
    flops = 1.0/3.0*N*N*N;
#ifdef COMB_FACT_SOLVE
  flops += (FactType=='Q'? 2: 1) * M * N;
#endif
  speed = flops / (1.0e6*tcomp);
  HPRINT((STDOUT, "Computed %dx%d %c matrix factorization (%.3e FLOPs) %Gs (user time %Gs) at %.2f MFLOPs (%.3f/cell)\n",
            M, N, FactType, flops, tcomp, UserEndTimerVal, speed, speed/Ncelx/Ncely));
  sprintf(ResultMsg, ": N=%d #s: %.3f %.3f", N, tcomp, speed/Ncelx/Ncely);
  LOGFILE_APPEND(LogFileName, ResultMsg);

  flops_f = flops; tcomp_f = tcomp;	/* keep for combined  factor-solve speed */

  MatPrint("fA", MIN(ma, MFPrintRes),  MIN(na, MFPrintRes), DM_Loc(A));
  if (FactType=='Q' && MFPrintRes>0) {
    CPRINTF(("tau="));
    for (i=0; i < MIN(N, MFPrintRes); i++)
      CPRINTF((" %1.3f", Tau[i]));
    CPRINTF(("\n"));
  }

#ifdef PBLAS
  if (FactType=='L' && !Level2) {
    PivColRep = 1;		/* checking routine - use same conv pdgetrf() */
    for (i=0; i < G2Ld(MIN(M, N), DM_DM(A)); i++)        
      Piv[i]--;			/* compensate for Ftn indices */
    getoptw_(&i);
    CPRINTF(("blocking factor used was %d\n", i));
  }
#endif
  if (Piv!=NULL && MFPrintRes>0) {
    HPRINT((STDOUT, "P="));
    for (i=0; i < MIN(N, MFPrintRes); i++)
      HPRINT((STDOUT, "%d ", Piv[i]));
    HPRINT((STDOUT, "\n"));
  }

  if (FactType=='D') {
    int NPiv1 = 0, NPiv21=0, NPiv2=0, NPiv22=0, BlkPivCnt=0, NextBlk = 0, i=0;
    extern int Case2Cnt, NColSearches; 		/* freom LDLT*.c */
    while (i < N) {
      if (i>=NextBlk) NextBlk += d_B(DM_D0(A)); 
      if (Piv[i] < 0) {
        NPiv2++; 
        if (ABS(Piv[i]) != i) NPiv21++;
        if (ABS(Piv[i+1]) != i+1) NPiv22++;
        if (ABS(Piv[i]) < NextBlk) BlkPivCnt++;
        if (ABS(Piv[i+1]) < NextBlk) BlkPivCnt++;
	i+=2;
      } else {
        if (Piv[i] != i) 
          NPiv1++;
        if (Piv[i] < NextBlk) BlkPivCnt++;
        i++;
      }
   } 
   HPRINT((STDOUT, "out of %d cols, %d extra col searches, %d 1x1 pivots, %d 2x2 pivots (%d+%d); %d pivots within block; %d searches\n", 
		N, Case2Cnt, NPiv1, NPiv2, NPiv21, NPiv22, BlkPivCnt, NColSearches));
   LDLTCase1Frac = N==0? 0.0: (double) (Case2Cnt/2 + NPiv21 + NPiv1 + NPiv22) / (double) N;  
   LDLTBlkPivFrac = N==0? 0.0: (double) (N-BlkPivCnt) / (double) N;
   LDLTSearchFrac =  N==0? 0.0: (double) NColSearches / (double) N;
  }
   
  if (ChkLvl==CHK_TRMM) {			/* use TRMM() to recover A */		
    BaseTraceLevel=0;
    MatCopy(ma, na, DM_pLoc(A), NULL, DM_pLoc(A1)); 	/* A1 = A */
#if MF_SEL!=2 && MF_SEL!=3
    if (FactType=='L') {  /* set A <- L*U, then A <- P^-1 * A*/
      if (M <= N) { /* make A rectangular by zeroing lower half */
        TrPreMultDiag(!TransA? "Lower": "Upper", "NonUnit", ma, na, &FD_0, &FD_1, A);
        start_timer();
	D_TRMM("Lower", "Unit", M, N, FD_1, A1 /*L*/, A /*U*/);
      } else { /* make A rectangular by zeroing upper half & adding unit diag*/
        TrPreMultDiag(!TransA? "Upper": "Lower", "Unit", ma, na, &FD_0, &FD_1, A);
        start_timer();
        D_TRMM("Lower", "NonUnit", N, M, FD_1, Trans(A1) /*U*/, Trans(A) /*L*/);
      }
      D_LASWP("Backward", N, MIN(M,N), Trans(A), Piv);
    } else 
#endif
#if MF_SEL!=1 && MF_SEL!=3
      if (FactType=='P') { 
      /* make A rectangular by zeroing upper half */
      TrPreMultDiag(!TransA? "Upper": "Lower", "NonUnit", ma, na, &FD_0, &FD_1, A);
      start_timer();
      D_TRMM("Lower", "NonUnit", N, N, FD_1, A1 /*L*/, Trans(A) /*L*/);
    } 
#endif
    tcomp = end_timer(1);
    flops = 1.0*M*M*MIN(M,N);
  } else if (ChkLvl==CHK_INV) {		/*use inv. of factoriz'n to recover A*/
    start_timer();
#if MF_SEL!=2 && MF_SEL!=3
    if (FactType=='L')
      D_LUTRM(M, N, A, Piv, MaxW);
    else 
#endif
#if MF_SEL!=1 && MF_SEL!=2
    if (FactType=='Q')
      D_GEQRM(M, N, A, Tau);
    else
#endif
#if MF_SEL!=1 && MF_SEL!=3
      D_POTRM(N, A, MaxW);
#endif
#if MF_SEL==1 || MF_SEL==2
      ;
#endif
    tcomp = end_timer(1);
    /* flops same as that of fact'n */
  } else if (ChkLvl==CHK_NRM) {
    start_timer();
#ifdef COMB_FACT_SLV
    if (FactType!='D')
      D_COPY(N, XA, X);  
#endif
    if (FactType=='L') {		/* must have M==N */
      int One = 1;
#ifndef PBLAS
# ifndef COMB_FACT_SLV
      D_LASWP("Forward", 1, MIN(M,N), Trans(X), Piv);
      BrdTopChar[0] = 'I',        BrdTopChar[1] = 'I';
      D_TRSV("Lower", "Unit", M, A, X);
# endif
      BrdTopChar[0] = ' ',        BrdTopChar[1] = ' ';
      D_TRSV("Upper", "NonUnit", N, A, X);
#else
      PivColRep = 1;              /* checking routine - use same conv pdgetrf()*/
      for (i=0; i < G2Ld(MIN(M, N), DM_DM(A)); i++)
        Piv[i]++;                 /* convert to Ftn indices */
      P_GETRS("NoTrans", &M, &One, LM_ADR(DM_Loc(A)), &ia, &ja, desca, Piv,
	      LM_ADR(DM_Loc(X)), &ix, &jx, descx,  &info); 
#endif
    } else if (FactType=='P') {
      BrdTopChar[0] = 'I',        BrdTopChar[1] = 'I';
# ifndef COMB_FACT_SLV
      D_TRSV("Lower", "NonUnit", N, A, X);
# endif
      D_TRSV("Upper", "NonUnit", N, Trans(A), X);
    } else if (FactType=='Q') {
# ifndef COMB_FACT_SLV
      BrdTopChar[0] = 'I',        BrdTopChar[1] = 'I';
      D_GEORQM(M, N, 1, A, Tau, X);
# endif
      BrdTopChar[0] = ' ',        BrdTopChar[1] = ' ';
      D_TRSV("Upper", "NonUnit", N, A, X); 
    } else if (FactType=='D') {
      int info = 0;
#if defined(LAPACK)
      int One = 1, LdA = LM_CInc(DM_Loc(A)), LdX =  LM_CInc(DM_Loc(X));
      int i;
      for (i=0; i<N; i++)
        if (Piv[i] >= 0) Piv[i]++; 
        else Piv[i]--;         
      _SYTRS(F2C_CHAR("L"), &N, &One, LM_ADR(DM_Loc(A)), &LdA, Piv, 
	     LM_ADR(DM_Loc(X)),  &LdX, &info);
#elif defined (PBLAS)
      int i; int NRhs=1;
      for (i=0; i<N; i++)		/* convert back to Fortran convention */
        if (Piv[i] >= 0) Piv[i]++; 
        else Piv[i]--;              
      P_SYTRS(F2C_CHAR("L"), &N, &NRhs, LM_ADR(DM_Loc(A)), &ia, &ja, desca, 
	      DiagInv, Piv, LM_ADR(DM_Loc(X)), &ix, &jx, descx, &info);	
#else
      BrdTopChar[0] = 'I',        BrdTopChar[1] = 'I';
      D_SYTRS(N, 1, A, DiagInv, Piv, X, &info);
    }
#endif
    tcomp = end_timer(1);
#ifdef COMB_FACT_SLV 
    flops = M*N;
#else   
    flops = (FactType=='Q'? 3.0: 2.0)*M*N;
#endif
  }

  if (ChkLvl==CHK_INV || ChkLvl==CHK_TRMM)
    MatPrint("A", MIN(ma, MFPrintRes),  MIN(na, MFPrintRes), DM_Loc(A));
  else if (ChkLvl==CHK_NRM) {
    MatPrint("Y'", MIN(mx, MFPrintRes),  MIN(nx, MFPrintRes), DM_Loc(X));
  }

  if (ChkLvl!=CHK_NONE) {
  speed = flops / (1.0e6*tcomp);
  HPRINT((STDOUT, "check (%.3e FLOPs) %Gs at %.2f MFLOPs (%.3f/cell)\n",
         flops, tcomp, speed, speed/Ncelx/Ncely));
  sprintf(ResultMsg, " #c: %.3f %.3f", tcomp, speed/Ncelx/Ncely);
  LOGFILE_APPEND(LogFileName, ResultMsg);
  if (ChkLvl==CHK_NRM) {
    tcomp += tcomp_f, flops += flops_f;
    speed = flops / (1.0e6*tcomp);
    HPRINT((STDOUT, "	total solve: %.2fs at %.1f MFLOPs (%.2f/cell)\n",
            tcomp, speed, speed/Ncelx/Ncely));
    sprintf(ResultMsg, " #s: %.3f %.3f", tcomp, speed/Ncelx/Ncely);
    LOGFILE_APPEND(LogFileName, ResultMsg);
  }

  if (ChkLvl==CHK_INV || ChkLvl==CHK_TRMM) {
    NErr = CheckScattered(RandElts, 0, 0, ma, na, &DiagBias, 
			  SymmetricA? "Lower": "NoTrans", A,
			  &DMax, &DAv, &DMaxI, &DMaxJ);
#define E_MAX LM_Index(DM_Loc(A), DMaxI, DMaxJ)
#ifdef CELL_PRINT
    if (m>0 && n>0 && NErr)
      CPRINTF(("%d errs; Av: %f; Max Error: B[%d,%d] = %s; |error| = %f\n",
             NErr, DAv, DMaxI, DMaxJ, pFD2Str(&E_MAX), DMax));
#endif
#undef E_MAX
  } else {
    int dum1, dum2;
    ScaledDiff = 0;
    NErr = MatDiff(mx, nx, DM_Loc(Y), DM_Loc(X), &DMax, &DAv, &DMaxI, &DMaxJ);
#ifdef CELL_PRINT
    if (mx*nx > 0)
      CPRINTF(("Y-X: #e=%d, av %G mx %G @ (%d, %d)\n", NErr, DAv, DMax, DMaxI, DMaxJ)); */
#endif
    DAv *= mx*nx;
    MatZero(mx, nx, DM_pLoc(X));
    MatDiff(mx, nx, DM_Loc(Y), DM_Loc(X), &Ynrm_inf, &Ynrm_1, &dum1, &dum2);
#ifdef CELL_PRINT
    if (mx*nx > 0)
      CPRINTF(("Y: av %G mx %G @ (%d, %d)\n", Ynrm_1, Ynrm_inf, dum1, dum2)); */
#endif
    Ynrm_1 *= mx*nx;
    ScaledDiff = 1;    
  } 	
      
  CommSetUp(Ncely, Ncelx);
  GatherSend(NErr, sizeof(int));
  GatherSend(DMax, sizeof(FD_BASE));
  GatherSend(DAv,  sizeof(FD_BASE));
  GatherSend(DMaxI, sizeof(int));
  GatherSend(DMaxJ, sizeof(int));
  if (ChkLvl==CHK_NRM) {
    GatherSend(Ynrm_inf, sizeof(FD_BASE));
    GatherSend(Ynrm_1,  sizeof(FD_BASE));
  }

      
  if (CID==0) {                       /* accumulate error results */
    int *NErrs, *MxDiffIs, *MxDiffJs;         FD_BASE *MxDiffs, *AvDiffs;  
    NErrs = (int *) malloc(3*Ncely*Ncelx*sizeof(int));
    MxDiffIs = NErrs + Ncely*Ncelx;   MxDiffJs = MxDiffIs + Ncely*Ncelx;
    MxDiffs = (FD_BASE *) memalign(sizeof(FD_BASE), 2*Ncely*Ncelx*sizeof(FD_BASE));
    AvDiffs = MxDiffs + Ncely*Ncelx;
      
    NErrs[0] = NErr;                  /* needed for non-parallel programs */
    MxDiffs[0] = DMax;                AvDiffs[0] = DAv;
    MxDiffIs[0] = DMaxI;              MxDiffJs[0] = DMaxJ;
    GatherMsg(Ncely, Ncelx,  NErrs, sizeof(int));
    NErr = SumIntMat(Ncely, Ncelx, NErrs);
    GatherMsg(Ncely, Ncelx,  MxDiffs, sizeof(FD_BASE));
    DMax = MaxMatrix(MxDiffs, Ncely, Ncelx);
    GatherMsg(Ncely, Ncelx,  AvDiffs, sizeof(FD_BASE));
    DAv = SumMatrix(AvDiffs, Ncely, Ncelx)/(Ncely*Ncelx);
    GatherMsg(Ncely, Ncelx,  MxDiffIs, sizeof(int));
    GatherMsg(Ncely, Ncelx,  MxDiffJs, sizeof(int));

    if (ChkLvl==CHK_NRM) {
      FD_BASE Anorm, Resid_inf,  Resid_1;
#define EPS 2.0E-16			/* machine precision */
      MxDiffs[0] = Ynrm_inf;
      GatherMsg(Ncely, Ncelx, MxDiffs, sizeof(FD_BASE));
      Ynrm_inf = MaxMatrix(MxDiffs, Ncely, Ncelx);
      AvDiffs[0] = Ynrm_1;
      GatherMsg(Ncely, Ncelx,  AvDiffs, sizeof(FD_BASE));
      Ynrm_1 = SumMatrix(AvDiffs, Ncely, Ncelx)/M;

      /* approximate max row (inf norm) / column (1 norm) sum of abs elements */
      if (RandElts)		/* assume random in range 0...1 */
        Anorm = FD_ABS(DiagBias) + N/2.0 + 4.0 /* safety factor, for small N */;	
      else 			/* assume A(i,j) ~= 0.1*i */
	Anorm = FD_ABS(DiagBias) + 0.1*N*N/2.0; 

      Resid_inf = DMax / (Anorm * Ynrm_inf * N * EPS);
      Resid_1 = DAv / (Anorm * Ynrm_1 * N * EPS);   	
     
      NErr = Resid_inf > MAX_RESID;
      HPRINT((STDOUT, "%s residual check: (inf) %G (1) %G \n", 
	     NErr? "FAILed": "PASSed", Resid_inf, Resid_1));
      if (FactType!='D')
        sprintf(ResultMsg, " #e: %d  %.1f %.1f\n", NErr, Resid_inf, Resid_1);
      else {
        extern int Case2Cnt;
        sprintf(ResultMsg, " #e: %d  %.1f %.1f #p: %d %0.3f:%0.3f:%1.2f\n", 
			    NErr, Resid_inf, Resid_1, Case2Cnt, 
			    LDLTCase1Frac, LDLTBlkPivFrac, LDLTSearchFrac);
      }
      LOGFILE_APPEND(LogFileName, ResultMsg);

      if (FactType=='D') 
        HPRINT((STDOUT, "Diag piv perf (bias, resid, frac): %G %.2e %0.3f:%0.3f:%1.2f\n", 
		FD_ABS(DiagBias), Resid_inf, LDLTCase1Frac,  
		LDLTBlkPivFrac, LDLTSearchFrac));  
    } else {

      HPRINT((STDOUT, "%d global errors, max=%G, av=%G\n", NErr, DMax, DAv));
      sprintf(ResultMsg, " #e: %d %.3f %.3f", NErr, DMax, DAv);
      LOGFILE_APPEND(LogFileName, ResultMsg);
 
      if (NErr) {
        PrintIntMat(Ncely, Ncelx, NErrs);
        PrintMatrix('a', MxDiffs, Ncely, Ncelx, Ncelx);
        PrintMatrix('v', AvDiffs, Ncely, Ncelx, Ncelx);
        PrintCoords(Ncely, Ncelx, MxDiffIs, MxDiffJs);
      }
    }
    free(NErrs);                      free(MxDiffs);
    HPRINT((STDOUT, "\n\n"));
  }
  } else {
    LOGFILE_APPEND(LogFileName, "	#c: -1 -1 #e: -1 -1 -1\n");
    HPRINT((STDOUT, "\n\n"));
    NErr = 0;
  }

  MatFreeBuf();
  MatFree(WS);
  if (Piv!=NULL)
    free((char *) Piv);
  if (Tau!=NULL)
    MatFree(Tau);
  if (DiagInv!=NULL)
    MatFree(DiagInv);

  if (ChkLvl == CHK_NRM) { 
    FDATA FNerr;
    FD_SET_REAL(FNerr, NErr);
    BrdMat(0, CID_I, 1, 1, &FNerr, 1);
    BrdMat(1, CID_J, 1, 1, &FNerr, 1);
    NErr = FD_REAL(FNerr);
  }      	
  if (NErr) break;

  N += dN; M += dN; r--;
  } /* while (r > 0) */
  CELL_FINALIZE;
  exit (NErr>0);
} /* main() */

