/************************************************************************/
/*                              Spread.Parameters.c                        */
/************************************************************************/
/* SCCS INFO @(#)Spread.Parameters.c 1.3         last modified P. Strazdins, 99/04/07
   first created: P. Strazdins 97/01
*/
/* input command line parameters into variables for general distributed
 * matrix spread / reduce program 
 *        
 *  with the usage:
 *	PROGNAME
 *  followed optionally by following USAGE string:                      */
#define USAGE \
"[-f fn RC] [-T] [-C] [-B B[0] B[1]] [-0 dN[0] dN[1]] [-s Rot] [-F] [-m] [-e r[0] r[1]] \n\
	[-w] [-l FastAllocSize] [-p PrintRes] [-M n v] [-t trlvl] [-v AlgVers] \n\
	[-X TReps] [-r NReps] [-G P[0] P[1]] N[0] N[1]  [alpha]"
/*  with default values:						*/
#define DEFAULTS \
"fn='S' dN[i]=0 RC=0 B[i]=1 Rot=0 r[0]=100, r[1]=1; alpha=1.0\n\
	NReps=TReps=1 PrintRes=0 AlgVers=0 trlvl=-1 FastAllocSize=-1"
/* and constraints:							*/
#define CONTRAINTS						\
"fn='S','R','T' RC=0,1 0<=dN[i], 1<=B[i] "

/* fn='S' or 'T' ('R') preforms a spread (reduce) operation scaled by alpha 
 * across dimension RC on a N[RC'] x N[!RC'] distributed matrix A on a 
 * P[0]xP[1] grid.
 * Specifically, it generates the call (see SpreadMatrix.h)
 * 	SpreadMat/ReduceMat(RC, N[RC'], N[!RC'], A, alpha, DN, sB, ForceWS, M, DM)
 * where 
 *	A = SubMat({TransB, lB, D[0..1]}, dN[RC'], dN[!RC'])
 *	D[i] = {0, B[RC'^i], 0, P[i]}
 *	DN = {DNB0, DNB, (DNP0+Rot)%DNP, DNP}, 
 *	{DNB0, DNB, DNP0, DNP} = D[!RC']
 *	RC' = TransB xor RC
 *
 * If -T specified, TransB=1 (otherwise 0).
 * If -m is specified, M and DM are chosen so that the result of the
 * spread is only needed on the cell row or column where the source 
 * elements of B begin (in certain situations (eg. K=1), this results in
 * a trivial spread/reduce operation). 
 * If -F is specified, ForceWS=1. This will force the result of the
 * spread to be copied in a new workspace 
 * (rather than point in lB, which eg can occur when -m specified or P[RC']=1).	
 * If -w is specified, workspace is pre-allocated before timing occurs
 * (this is only useful if fn='S', as it is always the case for fn='T').
 * FastAllocSize can be set to a >= 0 value to reduce the max. matrix size 
 * for fast allocations (of FDATAMalloc);
 * B[i,j] = r[RC']*i + r[1-RC']*j; if -C specified, local storage (lB) 
 * is column-major   
 *
 * NumTiming timings of NRreps repetitions are performed.
 *
 * The -M option may be repeated with differing n values; all these
 * can set the DBLAS machine parameters before the computation begins.
 * 	here -M 3 v (default v=16384) will change the partitoning
 * 	factor for large messages. 
 * The -t option is used for DBLAS trace level setting.
 * The -v AlgVers option is used to force a version of a DBLAS algorithm,
 * is chosse MM alg that `spreads' across the AlgVers'th matrix dimension. 
 * Matrices of a global index (i,j) are initialized to ry*(i%my) + rx*(j%mx).
 * PrintRes>=0 is a debug option to print out matrices for debugging.
 */

/* command line parameters; see header comment */
  static int RC, N[2], B[2], P[2], dN[2],  FastAllocSize,
	     Rot, TransA, ForceWS, PreAllocWS, StdMalloc, MTriv, ColMajA; 
  static int NReps, TReps;
  static FDATA alpha; 
  static FDATA r[2];
  static char fn;

#include "FDAssert.h"   	/* extern int PrintRes, TraceLevel, TraceType, ForceAlgVers*/
#include "MachineParams.h"      /* for -M oprtion */

#define SPREAD_VAL	'S'
#define REDUCE_VAL	'R'
#define TRANSRESPREAD_VAL 'T'
	

static void OptionError(), PrintUsage();

static void GetParameters(argc, argv)
  int argc; char *argv[];
/* post: set Spread command line variables according to argv[0..argc-1] */
{ extern char *optarg;
  extern int optind;
  
  int c; 
  /* set up default values */
  RC=0;  N[0] = N[1] = 1; B[0] = B[1] = 1; dN[0]= dN[1] = 0; 
  P[0] = NP_I, P[1] = NP_J;

  Rot=StdMalloc=PreAllocWS=ForceWS=MTriv=0; fn = SPREAD_VAL;
  TransA = ColMajA = 0;
  FD_ASSIGN_SC(r[0], =, 100.0, -100.0); FD_ASSIGN_SC(r[1], =, 1.0, -1.0);	
  alpha=FD_1;
  NReps = TReps = 1; 	FastAllocSize = -1;

  if (argc==0)
    return;

  while ((c = getopt(argc, argv, "GfB0TCs:Fwl:mep:r:X:t:Mv:")) != -1) 
    /* parse options */
    switch (c) {
      case 'G':
         if (optind + 1 < argc) {
           P[0] = atof(argv[optind]);      P[1] = atof(argv[optind+1]);
           optind += 2;
         }                                                              break;
      case 'f':
        if (optind + 1 < argc) {
	  fn = argv[optind][0];		RC = atoi(argv[optind+1]);
	  optind += 2;	
	} else
	  OptionError(c, "dN[0..1] parameters are missing");		break;
      case 'T':
	TransA = 1;							break;
      case 'C':
	ColMajA = 1; 							break; 
      case 'B':
        if (optind + 1 < argc) {
	  B[0] = atoi(argv[optind]);	B[1] = atoi(argv[optind+1]);
	  optind += 2;	
        } else
	  OptionError(c, "B[0..1] parameters are missing");		break;
      case '0':
        if (optind + 1 < argc) {
	  dN[0] = atoi(argv[optind]);	dN[1] = atoi(argv[optind+1]);
	  optind += 2;	
	} else
	  OptionError(c, "dN[0..1] parameters are missing");		break;
									break;
      case 's':
	Rot = atoi(optarg);						break;
      case 'w':
	PreAllocWS = 1;							break;
      case 'l':
	FastAllocSize = atoi(optarg);					break;
      case 'F':
	ForceWS = 1;							break;
      case 'm':
	 MTriv = 1;							break;
      case 'r':
	NReps = atoi(optarg);						break;
      case 'X':
	TReps = atoi(optarg);					break;
      case 'e':
        if (optind + 1 < argc) {
	  r[0] = Str2FD(argv[optind]); 	r[1] = Str2FD(argv[optind+1]); 
	  optind += 2;
        } else 
	  OptionError(c, "r[0..1] parameters are missing");
								 	break;
      case 't':
        TraceLevel = atoi(optarg);					break;
      case 'v':
	ForceAlgVers = atoi(optarg);					break;
      case 'M':
        if (optind + 1 < argc) {
          MachParamSet(atoi(argv[optind]), atoi(argv[optind+1]));
	  optind += 2;		
        } else 
	  OptionError(c, "n or v parameters are missing");
								 	break;
      case 'p':
	PrintRes = atoi(optarg);					break;
      case '?':
        PrintUsage();
        exit (1);
    } /* switch (c) */

  if (optind+1 >= argc || argc <= 2) { 
    if (argc > 1) 
      HPRINT((STDOUT, "%s: insufficient parameters N[0..1] given.\n", 
	      PROGNAME));
    PrintUsage();
    exit (argc > 1? 2: 0);
  }

  /* get non-optional parameters */
  N[0] = atoi(argv[optind]);		N[1] = atoi(argv[optind+1]);
  if (PrintRes < 0)
    PrintRes = MAX(N[0], N[1]);
  if (optind+2 < argc) 
    alpha = Str2FD(argv[optind+2]);
} /* GetParameters() */


static void OptionError(option, msg)
  char *msg, option;
/*post: print out option error message  with msg & option; print usage & exit */
{ HPRINT((STDOUT, "%s: error in parsing option -%c: %s\n", PROGNAME, option, msg));
  PrintUsage();
  exit(2);
}

static void PrintParameters();

static void PrintUsage()
/*post: prints program usage to stdout*/
{ HPRINT((STDOUT, "usage : %s %s\n", PROGNAME, USAGE));
  HPRINT((STDOUT, "where: %s\n", CONTRAINTS));
  HPRINT((STDOUT, "with default values: %s\n", DEFAULTS));
  PrintParameters();
} /* PrintUsage() */


static void CheckParameters()
/* post: check values of parameter variables, aborting if invalid */
{ if (fn!=SPREAD_VAL && fn!=REDUCE_VAL && fn!=TRANSRESPREAD_VAL) {
    HPRINT((STDOUT, "%s: fn=%c invalid\n", PROGNAME, fn));
    PrintUsage();
    exit (2);
  }
  if (fn==TRANSRESPREAD_VAL && !TransA) {
    HPRINT((STDOUT, "%s: fn=%C requires -T specified\n", PROGNAME, fn));
    PrintUsage();
    exit (2);
  }
  if (!!RC != RC) {
    HPRINT((STDOUT, "%s: RC=%d invalid\n", PROGNAME, RC));
    PrintUsage();
    exit (2);
  }
  if (B[0] < 1 || B[1] < 1) {
    HPRINT((STDOUT, "%s: block sizes B[0..1]=%d,%d invalid\n", 
	   PROGNAME, B[0], B[1]));
    PrintUsage();
    exit (2);
  }
  if (dN[0] < 0 || dN[1] < 0) {
    HPRINT((STDOUT, "%s: index offsets dN[0..1]=%d,%d invalid\n", 
	   PROGNAME, dN[0], dN[1]));
    PrintUsage();
    exit (2);
  }
  if (MTriv && N[0]+dN[0]%B[0] > B[0]) { 
    HPRINT((STDOUT, "%s: -m requires operation across single block only\n", 
	   PROGNAME));
    PrintUsage();
    exit (2);
  }
} /* CheckParameters() */


static void PrintParameters()
/*post: print out current values of variables set by command line to stdout*/
{ HPRINT((STDOUT, "%s: %c (dim %d) operation on %dx%d distributed matrix scale %s, on %dx%d grid, where:\n",
	 PROGNAME, fn, RC, N[0], N[1], pFD2Str(&alpha), P[0], P[1])); 
  HPRINT((STDOUT, "	A has %dx%d blocks, offset (%d,%d); Trans=%d col-major=%d\n", 
	 B[0], B[1], dN[0], dN[1], TransA, ColMajA));
  HPRINT((STDOUT, "	opn has rotation factor %d%,%s workspace: %s%s%s %d\n", 
	 Rot, MTriv? ", trivial dest dimension ": "", ForceWS? " forced" : "",
	 fn==REDUCE_VAL || PreAllocWS? " pre-allocated": "",
	 FastAllocSize>=0? " max fast alloc size=": "", FastAllocSize));
  HPRINT((STDOUT, "	%d timings will be taken over %d repetitions\n",
	 TReps, NReps));
  HPRINT((STDOUT, "Matrices have block-scattered storage with element (i,j) = %s*i + %s*j\n",
         pFD2Str(&r[0]), pFD2Str(&r[1])));
} /* PrintParameters() */

