/************************************************************************/
/*                              Spread.cell.c                              */
/************************************************************************/
/* SCCS INFO @(#)Spread.cell.c 1.2         last modified P. Strazdins, 99/03/19
   first created: P. Strazdins 97/01
*/
/* cell program for testing Ny x Nx Distributed Matrix Spread/Reduce operations 
 * (see Spread.Parameters.c for meaning)
 * (externally-defined) cpp FLAGs:
 *      FDATATYPE       SINGLE (default), DOUBLE
 */
/************************************************************************/

#include "Cell.h"
#include "Timer.c"
#include <math.h>

#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "LocMat.h"
#include "FDMalloc.h"  
#include "MatDim.h"
#include "DistMat.h"
#include "CellMatrix.h"
#include "FDCommun.h"
#include "SpreadMatrix.h"

#include "HostMsg.c"
#define PROGNAME "Spread.cell"
#include "Spread.Parameters.c"
#include "AuxMatrix.c"

/************ scattered matrix initialization /checking routines ************/

static double Max(a, n)
 double a[]; int n;
{ int i; double m=fabs(a[0]);
  for (i=1; i < n; i++)
   if (fabs(a[i]) > m) m = fabs(a[i]);
  return (m);
}
 
static double Min(a, n)  
 double a[]; int n;
{ int i; double m=fabs(a[0]);
  for (i=1; i < n; i++)
   if (fabs(a[i]) < m) m = fabs(a[i]);
  return (m);
}
 
static double Avg(a, n)
 double a[]; int n;
{ int i; double m=fabs(a[0]);
  for (i=1; i < n; i++)
   m += fabs(a[i]);
  return (m/n);
}

static void SetLinearScattered(ry, rx, M, N, beta, A)
  DistMat A; FDATA ry, rx, beta; int M, N;
/* pre:  A is an M x N distributed matrix with block & cell offset (0,0)
 * post: (global) A[i,j] = ry*(i%my) + rx*(j%mx) for 0<=i<M, 0<=j<N
 *       ie. (local) A[i',j'] = ry*(L2Gd(i', DM_Dim(0, A))%my) +
 *		                rx*(L2Gd(j', DM_Dim(1, A))%mx) 
 *	 for 0<=i<G2Ld(M, DM_Dim(0, A)), 0<=j<G2Ld(N, DM_Dim(1, A))%mx)
 */
#define lA DM_Loc(A)
{ FDATA ryb, rxb;
  int i, j; 
  M = G2Ld(M, DM_D0(A)); 		N = G2Ld(N, DM_D1(A));
  COMMON_ASSERT_TR(
	(TraceStr, "(%s,%s,%d,%d,%s,{a%d %d %d,...})", 
	 pFD2Str(&ry), pFD2Str(&rx), M, N, pFD2Str(&beta), FD_ADR(LM_ADR(lA)), LM_RInc(lA), LM_CInc(lA)),
	1, 0, "SetLinearScattered", 2, TRACE_OUTER
  );
  FD_MULSET(ryb, =, ry, beta);		FD_MULSET(rxb, =, rx, beta);
  for (i=0; i < M; i++) {
    int iG=L2Gd(i, DM_D0(A));	 	FDATA riG = ryb;
    FD_ASSIGN_SC(riG, *=, iG, iG); 
    for (j=0; j < N; j++) {
      int jG=L2Gd(j, DM_D1(A));		FDATA v = rxb;
      FD_ASSIGN_SC(v, *=, jG, jG);      FD_ASSIGN(v, +=, riG);
      FD_ASSIGN(*LM_Index(lA, i, j), =, v);
    }
  }      
} /* SetLinearScattered() */
#undef lA

/***************************** main program *******************************/

#ifdef ALLOCA_INIT		/* used for testing alloca() */
#include <alloca.h>
int SpreadMat_(int WSSize, int RC, int N0, int N1, DistMat A, FDATA *alpha_S, 
		pMatDim DN,  pLocMat sA, int ForceWS, int M,  pMatDim DM)
{ if (WSSize > 0)
    pLM_ADR(sA) = (FDATA *) alloca((size_t) WSSize);
  return(SpreadMat(RC, N0, N1, A, alpha_S, DN, sA, ForceWS, M, DM));
} 
#endif /* ALLOCA_INIT */

int main (argc, argv)
  int argc; char *argv[]; 
{
  /* matrices - dynamically allocated using WS*/
  DistMat_RVAlloc(4);
  DistMat A, SA, C;			/* SA is `spread of A'; C is result checking  */
  LocMat lA;	int n[2];
  LocMat sA, rA;	
  int SizeWS;				FDATA alpha_S;
  MatDim DN, DM;
  int M;
  int RCtr;
  double flops, speed, tcomp;
  double *tcomps; int Ti;/* timing */
  char Date[32];
  FD_BASE DAv=0.0, DMax=0.0; 		/* av. & max. differences */ 
  int NErr=0, DMaxI=0, DMaxJ=0;		/* # errors; position of max. difference */
  int i, r1;			

  CELL_INIT; 
  STDOUT_SETUP(STDOUT);
  InitArgs(argc, argv);			

  GetParameters(argc, argv);
  CheckParameters();
  PrintParameters();	

  COMMUN_INIT(P[0], P[1]);
#ifndef BLACS               	/* CommSetUp() is not called in COMMUN_INIT() */
  CommSetUp(P[0], P[1]);
#endif

  if (FastAllocSize >= 0)
    MatMallocInit(0, FastAllocSize);

  /* BK = B[0]  BN = B[1] */
  RCtr=RC^TransA;
  A = InitDimMat(TransA, B[RCtr], B[!RCtr], CID_I, P[0], CID_J, P[1],  
		 dN[RCtr], dN[!RCtr]);
  pInitDim(&DN, d_B0(DM_D(!RCtr, A)), B[1], 
		(d_P0(DM_D(!RC, A)) + Rot)%P[!RC], P[!RC]);
  pInitDim(&DM, d_B0(DM_D(RCtr, A)), B[0], d_P0(DM_D(RC, A)), P[RC]);
  M = MTriv? 1: N[1];
  
  if (fn==TRANSRESPREAD_VAL)
    A = ReduceDimMat(RC, A);
  
  n[0] = G2Ld_Mx(N[RCtr], DM_D0(A));	n[1] = G2Ld_Mx(N[!RCtr], DM_D1(A));
  n[0] = MAX(n[0], 1);			n[1] = MAX(n[1], 1);
  LM_Init(lA, NULL, ColMajA && n[0]>1? 1: n[1],
		    !ColMajA && n[1]>1? 1: n[0]);
  MatMalloc_F(PROGNAME, 1, &LM_ADR(lA), n[0]*n[1]);
  DM_Loc(A) = lA;

  tcomps = (double *) memalign(sizeof(double), sizeof(double)*TReps);

  for (Ti=0; Ti < TReps; Ti++) {
  DistMat_RVAlloc(1); 
  SetLinearScattered(r[RCtr], r[!RCtr], N[RCtr], N[!RCtr], FD_1, A);
  if (Ti==0)
    MatPrint("A", MIN(n[0],PrintRes), MIN(n[1],PrintRes), DM_Loc(A));

  if (fn==REDUCE_VAL) {
    LM_ADR(sA) = NULL;		/* ensure InitReduceWS() mallocs new space */
    SizeWS = InitReduceWS(RC, N[0], N[1], A, &alpha, &DN, &sA, ForceWS, M, &DM);
    SA = Spread2DistMat(1-RC, TransA, &DN, &sA);
    if (SizeWS > -1)	
      SetLinearScattered(r[RC], r[!RC], N[RC], N[!RC], FD_1, SA);
    if (NReps>1 && SizeWS>-1) { 
      /* make a copy of SA to restore it between ReduceMat() calls */
      int n = LM_Inc(RC, sA);
      LM_Copy(0, rA, sA);
      MatMalloc("Spread.cell", 1, &LM_ADR(rA), n*N[0]);
      MatCopy(RC==0? N[0]: n, RC==0? n: N[0], &sA, NULL, &rA);
    }
  } 

  if (PreAllocWS && fn!=REDUCE_VAL) {
    LM_ADR(sA) = NULL;
    InitSpreadWS(RC, N[0], N[1], A, &DN, &sA, ForceWS, M, &DM);
  }

  BarrierSync(-1);
  start_timer();
  for (r1 = 0; r1 < NReps; r1++) {
    if (fn==REDUCE_VAL) {
      FDATA alpha_R = r1==0? alpha: FD_1;
      ReduceMat(RC, N[0], N[1], A, &alpha_R, &DN, &sA, M, &DM);
      if (SizeWS > -1 && r1 < NReps-1) {
	int n = LM_Inc(RC, sA);
	MatCopy(RC==0? N[0]: n, RC==0? n: N[0], &rA, NULL, &sA);
      }
#if 0
      if (N[0]<=B[0] && r1%4==0)
	BarrierSync(RC);
#endif
    } else {
      if (!PreAllocWS)
        LM_ADR(sA) = NULL;
      alpha_S = alpha;
#ifdef ALLOCA_INIT
      SizeWS = SpreadMat_(SizeWS, RC, N[0], N[1], A, &alpha_S, &DN, &sA, ForceWS, M, &DM);
#else
      SizeWS = SpreadMat(RC, N[0], N[1], A, &alpha_S, &DN, &sA, ForceWS, M, &DM);
#endif
      if (!PreAllocWS && r1 < NReps-1 && SizeWS>0)
	MatFree(LM_ADR(sA));
    }
  } /* for (r1...) */

  tcomps[Ti] = end_timer(0);
  if (Ti < TReps-1 && fn==REDUCE_VAL && NReps>1 && SizeWS>-1)
    MatFree(LM_ADR(rA));
  if (Ti < TReps-1 && SizeWS>0)
    MatFree(LM_ADR(sA));

  } /* for (Ti...) */

  if (CID == 0) {             /* cell 0 selects minimum timing */
    tcomp = Min(tcomps, TReps);
    Ti = 0;
    while (tcomps[Ti] > tcomp) Ti++;
  } 
  { FDATA FTi; FD_SET_REAL(FTi, (FD_BASE) (Ti));
    BrdMat(0, CID_I, 1, 1, &FTi, 1);
    BrdMat(1, CID_J, 1, 1, &FTi, 1);
    Ti = (int) FD_REAL(FTi);
  }
  tcomp = tcomps[Ti];

  flops = N[0]*N[1];
#define SPEED_PERCELL(t) (NReps * flops * sizeof(FDATA)/1.0E6/(t)/(P[!RC]))
  CPRINTF(("%d.%d: rep %d: SizeWS=%d time is %G @ %G MOPs/s\n", 
	  CID_I, CID_J, Ti, SizeWS, tcomp, SPEED_PERCELL(tcomp)));
  if (SizeWS <= 0) tcomp = -tcomp; 
  GatherSend(tcomp, sizeof(double));
  if (CID == 0) {
    int i;
    double *T, mint, maxt, avgt;
    T = (double *) memalign(sizeof(double), P[0]*P[1]*sizeof(double));
    T[0] = tcomp;
    GatherMsg(P[0], P[1], T, sizeof(double));
    mint = Min(T, P[0]*P[1]);
    maxt = Max(T, P[0]*P[1]);
    avgt = Avg(T, P[0]*P[1]);
    HPRINT((STDOUT, "rep %d: for %dx%d operation, time: min %f, max %f, avg %f",
	    Ti, N[0], N[1], mint, maxt, avgt));
    HPRINT((STDOUT, "\n	speed (MBs): min %f, max %f, avg %f\n",
	    SPEED_PERCELL(mint), SPEED_PERCELL(maxt), SPEED_PERCELL(avgt)));
    for (i=0; i < P[0]; i++) {
      int j;
      HPRINT((STDOUT, "	"));
      for (j=0; j < P[1]; j++) {
	double t = T[i*P[1]+j];
        HPRINT((STDOUT, "%c%1.2e ", t>=0?' ':'*', SPEED_PERCELL(fabs(t))));
      }
      HPRINT((STDOUT, "\n"));
    }  
    HPRINT((STDOUT, "\n\n"));
    free((char *) T);
  }

  if (fn==REDUCE_VAL) {
    FDATA Ry = alpha, Rx;
    FD_ASSIGN_SC(Ry, +=, (FD_BASE) NReps*(!MTriv? P[RC]-(SizeWS<0): 0), 0.0);
    FD_MULSET(Rx, =, Ry, r[!RCtr]);	FD_MULSET(Ry, =, Ry, r[RCtr]);	
    rA = DM_Loc(A);
    C = NewMat(N[RC], N[!RC], A, NULL);
    SetLinearScattered(Ry, Rx, N[RCtr], N[!RCtr], FD_1, C);
  } else {
    FDATA alpha_scale = FD_EQ(alpha_S, FD_1)? alpha: FD_1;
    LocMat sC;
    rA = sA;
    LM_ADR(sC) = NULL;			/* ensures new workspace created */
    InitSpreadWS(RC, N[0], N[1], A, &DN, &sC, 1, M, &DM); 
    C = Spread2DistMat(1-RC, 0, &DN, &	sC);
    SetLinearScattered(r[RC], r[!RC], N[RC], N[!RC], alpha_scale, C);
  }

  if (fn==REDUCE_VAL) {
    n[0] = G2Ld(N[RCtr], DM_D0(C));	n[1] = G2Ld(N[!RCtr], DM_D1(C));
    if (MTriv && d_P0(&DM)!=0)
      n[RCtr] = 0;
  } else {
    n[0] = G2Ld(N[RC], DM_D0(C));	n[1] = G2Ld(N[!RC], DM_D1(C));
    if (MTriv && d_P0(&DM)!=0)
      n[RC] = 0;
  }
  MatPrint("rA", MIN(n[0],PrintRes), MIN(n[1],PrintRes), rA);
  MatPrint("C", MIN(n[0],PrintRes), MIN(n[1],PrintRes), DM_Loc(C));

  NErr = MatDiff(n[0], n[1], rA, DM_Loc(C), &DMax, &DAv, &DMaxI, &DMaxJ);
  if (n[0]>0 && n[1]>0) 
     CPRINTF(("%d errs; Av: %f; Max Error: C[%d,%d] = %s; error = %f\n", 
	     NErr, DAv, DMaxI, DMaxJ, *LM_Index(rA, DMaxI, DMaxJ),  DMax));
   
  CommSetUp(NP_I, NP_J);
  GatherSend(NErr, sizeof(int));
  GatherSend(DMax, sizeof(FD_BASE));
  GatherSend(DAv,  sizeof(FD_BASE));
  GatherSend(DMaxI, sizeof(int));
  GatherSend(DMaxJ, sizeof(int));

  if (CID==0) {			/* accumulate error results */
      int *NErrs, *MxDiffIs, *MxDiffJs; 	FD_BASE *MxDiffs, *AvDiffs;
      NErrs = (int *) malloc(3*P[0]*P[1]*sizeof(int));
      MxDiffIs = NErrs + P[0]*P[1];	MxDiffJs = MxDiffIs + P[0]*P[1];
      MxDiffs = (FD_BASE *) memalign(sizeof(FD_BASE), 
				     2*P[0]*P[1]*sizeof(FD_BASE));
      AvDiffs = MxDiffs + P[0]*P[1];

      NErrs[0] = NErr; 			/* needed for non-parallel programs */
      MxDiffs[0] = DMax;	
      AvDiffs[0] = DAv;
      MxDiffIs[0] = DMaxI;		MxDiffJs[0] = DMaxJ;

      GatherMsg(P[0], P[1],  NErrs, sizeof(int));
      NErr = SumIntMat(P[0], P[1], NErrs);
      GatherMsg(P[0], P[1],  MxDiffs, sizeof(FD_BASE));
      DMax = MaxMatrix(MxDiffs, P[0], P[1]); 
      GatherMsg(P[0], P[1],  AvDiffs, sizeof(FD_BASE));
      DAv = SumMatrix(AvDiffs, P[0], P[1])/(P[0]*P[1]);
      HPRINT((STDOUT, "%d global errors, max=%G, av=%G\n", NErr, DMax, DAv));
  
      GatherMsg(P[0], P[1],  MxDiffIs, sizeof(int));
      GatherMsg(P[0], P[1],  MxDiffJs, sizeof(int));
      if (NErr) {
        PrintIntMat(P[0], P[1], NErrs);
	PrintMatrix('a', MxDiffs, P[0], P[1], P[1]);
        PrintMatrix('v', AvDiffs, P[0], P[1], P[1]);
        PrintCoords(P[0], P[1], MxDiffIs, MxDiffJs);
      }			
      free(NErrs);			free(MxDiffs);
    }
    HPRINT((STDOUT, "\n\n"));

    CELL_FINALIZE;
    if (NErr)
      exit(1);
} /* main() */


