/************************************************************************/
/*				addfit.cell		 		*/
/************************************************************************/
/* SCCS INFO @(#)addfit.cell.c 1.2         last modified P. Strazdins, 01/01/10
   first created: P. Strazdins 01/01
*/

/* MPI program for testing addfit_solve().
 * Calls addfit_solve(double *a, double *b, int N, double *x, int P, 
 *                  *int Q, int r)
 * on first a PxQ grid amd second on a a pxq grid. 
 * When reduction of the matrix (a|b) is complete, the values of a and b 
 * should be identical on each call. These values will be random in [-1,1]
 * with the value DiagBias added for the diagonmal elements.
 * Before reduction, in each cell, the element corresponding to
 * A[i,j] will either be A[i,j]+(P*Q-1)*v (one cell exactly) or -v;
 * thus these sum to A[i,j]. 
 *
 * Note that in either case, the solution vector (x) should be the same.
 * 
 * example invokations (using a storage block size & blocking factor of 64):
 	addfit.cell -G 3 3 -g 1 1 -B 64 2000
 * compares the result vector x from running on a 3x3 grid on a 1x1 grid
 	addfit.cell -G 3 3 -g 2 2 -B 64 2000	
 * compares the result vector x from running on a 3x3 grid on a 2x2 grid
 */

#define USAGE \
"[-D DiagBias] [-e RndSeed] [-v v] [-t trlvl trttype] [-p PrintRes]\n\
 [-B r] [-G P Q] [-g p q] N \n\
with default values: P=Q=p=q=1, r=1, B=1, DiagBias=0.0 RndSeed=0, trlvl=-1 trtype=0 MFPrintRes=0"

#define PROGNAME "addfit.cell"

/* PrintRes x PrintRes of the local portion of intermediate matrices is 
 * printed out. The -t option is used for DBLAS trace level setting.   
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>


#define MIN(a, b)	((a)<=(b)? (a): (b))

/***************** MPI-dept functions/macros ********************************/
#include "mpi.h"

static int CID=0;

# define HPRINT(PRINTF_ARGS)           {if (CID==0) printf PRINTF_ARGS;}

# define CELL_INIT                      { MPI_Init(&argc, &argv);       \
                                          MPI_Comm_rank(MPI_COMM_WORLD, &CID);}
# define CELL_FINALIZE                    MPI_Finalize()
                                          
# define        SYNCH   MPI_Barrier(MPI_COMM_WORLD)
# define        DTIMER  MPI_Wtime

static double tim[3];
static void start_timer()
{ SYNCH;
  tim[0] = DTIMER();
} /* start_timer() */

static double end_timer()
{ double wt;
  SYNCH;
  tim[1] = DTIMER();
  SYNCH;
  tim[2] = DTIMER();
  wt = tim[1]-tim[0]-(tim[2]-tim[1]);
  return (wt);
} /* end_timer() */

/******************************** command line processing ******************/

static int N=1, MaxW=0, P=1, Q=1, p=1, q=1;
static int r=1;
static double DiagBias=0.0, v=0.0;
static int RandSeed = 0;
  
extern int GPrintRes, GTraceLevel, GTraceType;  /* #include "DBLAS/GenAssert.h" */  	

void PrintUsage()
/* post: prints program usage to stdout	*/
{  HPRINT(( "usage: %s %s\n", PROGNAME, USAGE));
}

void OptionError(option, msg)
  char *msg, option;
/* post: print out option error message  with msg & option; print usage & exit*/
{ HPRINT(( "%s: error in parsing option -%c: %s\n", PROGNAME, option, msg));
  PrintUsage();
  exit(1);
}

void GetParameters(argc, argv)
  int argc; char *argv[];
/* post: set command line variables according to argv[0..argc-1] */
{ extern char *optarg;
  extern int optind;

  int c; 
  while ((c = getopt(argc, argv, "GgtD:w:B:v:e:p:")) != -1) /* parse options */
    switch (c) {
      case 'G':
         if (optind + 1 < argc) {
           P = atof(argv[optind]);      Q = atof(argv[optind+1]);
           optind += 2;
         }                                                              break;
      case 'g':
         if (optind + 1 < argc) {
           p = atof(argv[optind]);      q = atof(argv[optind+1]);
           optind += 2;
         }                                                              break;
      case 'D':
	DiagBias=atof(optarg);						break;
      case 'w':
	MaxW=atoi(optarg);						break;
      case 'B':
        r = atoi(optarg);						break;
      case 't':
        if (optind + 1 < argc) {
          GTraceLevel = atoi(argv[optind]); GTraceType = atoi(argv[optind+1]);
          optind += 2; 
        } else 
          OptionError(c, "trlvl, trtype parameters are missing");       break;
      case 'p':
      	GPrintRes = atoi(optarg);					break;
      case 'v':
	v=atoi(optarg);							break;
      case 'e':  
        RandSeed = atoi(optarg);   
									break;
      case '?':
        OptionError(c, "invalid option given");                         break;
    }
        
  if (optind >= argc || argc <= 1) {
    if (argc > 1)
      HPRINT(( "%s: insufficient parameters M ... given.\n", PROGNAME));
    PrintUsage();
    exit(1);  
  }
          
  N = atoi(argv[optind]);
  if (MaxW==0) MaxW = r;
} /* GetParameters() */


void PrintParameters(int P, int Q, int whichx)
{ HPRINT(( "call addfit_solve(A,b,N=%d,x%d,P=%d,Q=%d,r=%d,w=%d), A: random [-1,1] with diag bias %G, v=%G\n",
	  N, whichx, P, Q, r, MaxW, DiagBias, v)); 
} /* PrintParameters() */


void CheckParameters()
/* post: check values of parameter variables, aborting if invalid */
{  
  if (P<=0 || Q<=0) {
    HPRINT(( "%s: invalid cell configuration values (P=%d, Q=%d)\n",
           PROGNAME, P, Q));
    PrintUsage();
    exit(2);
  }
  if (p<=0 || q<=0) {
    HPRINT(( "%s: invalid aux cell configuration values (p=%d, q=%d)\n",
           PROGNAME, p, q));
    PrintUsage();
    exit(2);
  }
  if (N <= 0) {
    HPRINT(( "%s: invalid -ve) matrix dimensions (N=%d)\n",
           PROGNAME, N));
    PrintUsage();
    exit(2);
  }
  if (r<1) {
    HPRINT(( "%s: invalid block size (r=%d)\n", PROGNAME, r));
    PrintUsage();
    exit(2);
  }
} /* CheckParameters() */

/***************** addfit auxilliary functions ***************************/
#define F_RAND_MOD      10000
#define F_RAND(r)          (((double) ((r % F_RAND_MOD) - F_RAND_MOD/2)) \
                        /((double) F_RAND_MOD/2))      /* values in [-1..1] */


static void addfit_matinit(int P, int Q, int N, double *A, double *b)
/* post: init. the packed tri. N x N matrix A on this cell of PxQ grid s.t. 
 *	 A[i,j] = A'[i,j]+(P*Q-1)*v on exactly one cell and A[i,j]=-v
 *	 on all others. Thus Sum(A[i,j]) = A'[i,j], where
 *	 A'[i,j] will be random in [-1,1] + the value DiagBias added for diagonmal elements;
 *	 its value is independent on P and Q.
 *	 Similarly for b.
 */
{ int i, j, ecnt=CID, ncells=P*Q;		double pqv = (P*Q-1) * v;
  srand(RandSeed);
  for (j=0; j<N; j++)
    for (i=j; i<N+1; i++) {
      int r = rand();	
      /* note all cells must call rand() every time for A'[i,j] to be indpt P,Q */
      double a = F_RAND(r);  
      if (ecnt==0) {
        a = F_RAND(r);
        if (i==j) a += DiagBias;
        a += pqv;
      } else
        a = -v;
      if (i==N) 
        *b = a, b++;
      else
        *A = a, A++;
      ecnt++; if (ecnt==ncells) ecnt = 0;
    }  
} /* addfit_matinit() */
  

static double VecDiff(int N, double *x1, double *x2)
/* post: reurtns ||x1=cx2||_inf
 */
{ int i;				double d=0.0;
  for (i=0; i<N; i++) {
    double di = fabs(x1[i]-x2[i]);
    if (di > d) d = di;
  }
  return (d);
} /* VecDiff() */


static void VecPrint(char *name, int N, double *x)
{ int i;
  if (N > 0)
    printf("%d: %3s: ", CID, name);
  for (i=0; i<N; i++)
    printf("%.3f ", x[i]);
  if (N > 0)
    printf("\n");
}

static void VecClear(int N, double *x)
{ int i;
  for (i=0; i<N; i++)
    x[i] = 0.0;
}

/**************************************** main program ******************/

#include "addfit_solve.h"

void main (argc, argv)
  int argc; char *argv[];
{ 
  int i, rv;
  double *A, *b, *x1, *x2;
  double tcomp;

  CELL_INIT;

  GetParameters(argc, argv);
  for (i=0; i < argc; i++)
    HPRINT(( "%s ",  argv[i]));
  HPRINT(( "\n"));
  CheckParameters();

  A = malloc((N*(N+1)/2 + 3*N)*sizeof(double));
  if (A==NULL) {
    printf("%d: %s: failed alloc of %d doubles\n", PROGNAME, CID, N*(N+1)/2 + 3*N);	
    exit(3);
  } 
  b = A + N*(N+1)/2; x1 = b + N; x2 = x1 + N;

  PrintParameters(P, Q, 1);
  addfit_matinit(P, Q, N, A, b);  
  start_timer(); 
  rv = addfit_solve(A, b, N, x1, P, Q, r, MaxW);
  tcomp = end_timer();
  HPRINT(("solve rv=%d on %dx%d grid in %Gs\n\n", rv, P, Q, tcomp));

  PrintParameters(p, q, 2);
  addfit_matinit(p, q, N, A, b);  
  start_timer(); 
  rv = addfit_solve(A, b, N, x2, p, q, r, MaxW);
  tcomp = end_timer();
  HPRINT(("solve rv=%d on %dx%d grid in %Gs\n\n", rv, p, q, tcomp));
  
  if (CID < p*q && N>0) { 
#define THRESHOLD 1.0e-06
    double zero = 0.0, MxDiff, MxX;
    MxDiff = VecDiff(N, x1, x2); 	
    VecPrint("x1", MIN(N, GPrintRes), x1);
    VecPrint("x2", MIN(N, GPrintRes), x2);
    VecClear(N, x1);
    MxX = VecDiff(N, x1, x2);
    HPRINT(("||x2||=%G ||x1-x2||/||x2|| = %G\n", MxX, MxDiff / MxX)); 
    /* if (MxDiff / MxX > THRESHOLD) */
      printf("%d: rv=%d ||x2||=%G ||x1-x2||/||x2||=%G\n", CID, rv, MxX, MxDiff / MxX);
  }

  free((char *) A);

  CELL_FINALIZE;

  exit (0);
} /* main() */

