/************************************************************************/
/* SCCS INFO @(#)addfit_solve.c 1.2         last modified P. Strazdins, 01/01/10
   first created: P. Strazdins 01/01
 */
   
/* --------------------------------------------------------------------- */
/* addfit_solve - Solve linear system for the AddFit application	 */
/* --------------------------------------------------------------------- */
/* compilation: let $DBLAS = Distributed BLAS root directory, eg. ~peter/dblas
 *		as well as path for mpi.h, add include paths:
 *		-I$DBLAS/DBLAS -I$DBLAS/DLAPACK	
 * linking: 	as well as libraries required by MPI, must also link in
 *		the DLAPACK and DBLAS archives, and a serial BLAS library.
 *		eg. on bunyip:
 *		$DBLAS/LIB/dlapack_d_bunyip.a $DBLAS/LIB/dblas_d_bunyip.a
 *		-lf77blas -latlas
 */ 

#include "addfit_solve.h"

/* tunable parameters */
#define ADDFIT_RESID_CHECK		/* if defined, perform resid. checking*/

#ifndef ADDFIT_MAX_RESID
#define ADDFIT_MAX_RESID 	100	/* max normalized inf-norm residual;
	   				   threshold; see addfit_check() */ 
#endif
#ifndef ADDFIT_PART_N
/* addfit_init() converts the replicated triangular matrix a into a block-cyclic
 * matrix A by considering strpis of width ADDFIT_PART_N at a time.
 * tradeoff between temporary workspace size and communciation costs 
 */	
#define ADDFIT_PART_N 500		
#endif
				

#include "mpi.h"

#define WorldComm MPI_COMM_WORLD
static MPI_Group SolveGroup;
static MPI_Comm SolveComm;
static int SolveRank=MPI_UNDEFINED;
#define IN_SOLVE_GROUP (SolveRank!=MPI_UNDEFINED)

static int CommunSetup(int P, int Q)
/* pre:  P, Q > 0;  
 * post: sets up MPI context SolveComm on SolveRank on 1st P*Q cells of WorldComm
 *	 sets SolveRank to be rank of this cell in SolveComm
 *	 returns 0 on success, SOLVE_COMM_ERROR otherwise 
 */
{ int SolveRange[1][3]; int MPI_rv;    
  MPI_Group WorldGroup; int WorldSize;

  MPI_Comm_group(WorldComm, &WorldGroup);
  MPI_Group_size(WorldGroup, &WorldSize);
  if (WorldSize < P*Q) 
    return(SOLVE_COMM_ERROR);

  SolveRange[0][0] = 0, SolveRange[0][1] = P*Q-1, SolveRange[0][2] = 1;    
  MPI_rv = MPI_Group_range_incl(WorldGroup, 1, SolveRange, &SolveGroup);  
  /*printf("rv=%d SolveGroup=%d\n", MPI_rv, SolveGroup);  */
  MPI_rv = MPI_Comm_create(WorldComm, SolveGroup, &SolveComm);    
  /*printf("rv=%d SolveComm=%d\n", MPI_rv, SolveComm);  */
  SolveRank = MPI_UNDEFINED;
  if (SolveComm!=MPI_COMM_NULL) /* some MPI impl. give MPI_COMM_NULL if
  				   this cell is not in the group */
    MPI_Comm_rank(SolveComm, &SolveRank);
  return(0);
} /* CommunSetup()*/

static void CommunFree(int *ret_val)
/* pre:	 cell 0 of WorldComm is in SolveGroup and has correct global
 *	 value of return value fo addfit_solve()
 * post: ret_val = cell 0's value of ret_val 
 *       dispose of any MPI resources set up by CommunSetup
 */
{ int MPI_rv;
  MPI_Group_free(&SolveGroup);
  MPI_Bcast(ret_val, sizeof(int), MPI_BYTE, 0, WorldComm);
  if (SolveComm!=MPI_COMM_NULL) 
    MPI_Comm_free(&SolveComm);
} /* CommunFree() */


#define FDATATYPE DOUBLE
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#include "MatDim.h"
#include "DistMat.h"
#include "D_BLAS.h"
#include "CellMatrix.h"
#include "SpreadMatrix.h"
#include "FDMalloc.h"  
#include "FDCommun.h"
#define PRINTMAT 			/* enable DM_PRINT() to print */
#include "D_LAPACK.h"


static void addfit_init(int N, FDATA *a, FDATA *b, DistMat A)
/* pre:   A is an N+1 x N+1 matrix (col-major), 
 *	  a is a packed N x N triangular matrix (row-major and upper)
 *	  (= col-major and lower), b is a vector of N elements
 * post:  B = sum of b on all cells, the lower tri part of A' is the sum of a
 *	  on all cells, where A' = A[0:N-1, 0:N-1], B = A[N, 0:N-1]   
 * notes: this process is performed on ADDFIT_PART_N x n sections of A at a 
 *	  time,, where n is width of the section encompassing the lower
 *	  tri. part of A[i:i+ADDFIT_PART_N-1,:]
 *        The value of ADDFIT_PART_N can be adjusted to provide a reasonable 
 *	  balance between minimizing communication startups and
 *	  good load balance (large ADDFIT_PART_N; 
 *	  for latter require ADDFIT_PART_N > r*P)
 *	  and minimizing temporary workspace size and communication volume
 *	  (small ADDFIT_PART_N).
 *	  For best load balance, the reductions are performed
 *	  horizontally first and vertically second. 
 *	  It is assumed that P ~ Q so there is no significant asymmetry.
 */
{ int i; 
  int P = d_P(DM_D0(A)),		Q = d_P(DM_D1(A)),
      r = d_B(DM_D0(A)),		s = d_B(DM_D1(A));
  for (i=0; i < N+1; i+= ADDFIT_PART_N) { 
    DistMat_RVAlloc(5);
    int m = MIN(ADDFIT_PART_N, N+1-i),	n = i + MIN(ADDFIT_PART_N, N-i);
    DistMat Ai = SubMat(A, i, 0);			
    DistMat sA =  P==1  ? Ai: 
    			       NewMat(m, n,       ReduceDimMat(0, Ai),   NULL);
					/* must be row-major if P>1*/ 
    DistMat ssA = P*Q==1? sA: 
    		  	 Trans(NewMat(n, m, Trans(ReduceDimMat(1, sA)),  NULL));
					/* must be column-major if Q>1*/
    FDATA *ssa = LM_Index(DM_Loc(ssA), 0, 0);   
    FDATA *ai = a+i;			
    int ma = m - (i+m==N+1),		dmj = 0, j;

    for (j=0; j < n; j++) {		/* initialize fully-replicated ssA */
      int RIncA = LM_RInc(DM_Loc(ssA)), CIncA = LM_CInc(DM_Loc(ssA));      
      if (j > i) {			/* avoid exceptions */ 
	LocMat lssa;
        MatZero(j-i, 1, pMkLocMat(&lssa, ssa, RIncA, CIncA)); 
	dmj++;
      }   
      VecCopy(ma-dmj, ai, 1, ssa+dmj*RIncA, RIncA);
      ai += N-j; 
      if (i+m==N+1) 
        ssa[ma*RIncA] = b[j];
      ssa += CIncA;
    } /* for (j...) */

    DM_PRINT("ssA", m, n, ssA);
    if (P*Q!=1) 
      ReduceMat(1, n, m, sA, &FD_0, DM_D0(sA), &DM_Loc(ssA), s*Q,  DM_D1(sA));

    if(P!=1)
      ReduceMat(0, m, n, Ai, &FD_0, DM_D1(Ai), &DM_Loc(sA),  r*P,  DM_D0(Ai));
    DM_PRINT("A", m, n, A);

    if (P*Q!=1)
      DisposeMat(ssA);
    if (P!=1)
      DisposeMat(sA);
  } /* for (j...) */
} /* addfit_init() */


#ifdef ADDFIT_RESID_CHECK
static void MatDiff(M, N, A, B, MxDiff)
  int M, N; LocMat A, B; FD_BASE *MxDiff;
/* pre:  LM_IncOK(M, N, A) && LM_IncOK(M, N, B)
 * post: *MxDiff = Max {0} U {0<=i<M, 0<=j<N: D(i,j)}
 *        where D(i,j) = Abs(A[i,j]-B[i,j])
 */
{ int i, j; 				FD_BASE d;  FDATA del;
  *MxDiff = 0.0; 
  for (i = 0; i < M; i++) {
    for (j = 0; j < N; j++) {
      FDATA Aij = *LM_Index(A,i,j), Bij = *LM_Index(B,i,j);
      FD_SETOP(del, Aij, -, Bij);
      d = FD_ABS(del);
      if (d > *MxDiff) 
        *MxDiff = d;
    } /* for (j...) */
  } /* for (i...) */
} /* MatDiff() */


static FDATA addfit_ResidCheck(int P, int Q, int N, DistMat A, DistMat X, DistMat B)
/* pre:   A is an N x N matrix, B is a 1 x N matrix, X is an N x 1 matrix
 * post:  returns the normalized residual ||E|| / (||B||*eps*N),  E = AX-B   
 * notes: more properly, should use ||A||*||X|| instead of ||B||,
 *	  but note that ||B|| ~ ||AX|| <= ||A||*||X||	 
 *	  For well-conditioned A, and an appropriate value of ADDFIT_MAX_RESID
 *	  this should be adequate for checking purposes 
 */
#define EPS 2.0E-16                    /* machine precision */
{ DistMat_RVAlloc(1); 		       
  DistMat AX = NewMat(1, N, B, NULL); 
  FD_BASE Resid, Enrm_inf, Bnrm_inf;
  DM_PRINT("A", N, N, A);
  DM_PRINT("B", 1, N, B);
  DM_PRINT("X", N, 1, X);  
  D_SYMV("Lower", N, FD_1, A, X, FD_0, Trans(AX));
  DM_PRINT("AX", 1, N, AX);  
  CommSetUp(P, Q);
  if (InRowScope(B)) {	      /* this assumes B and AX are aligned */
    int nB = G2Ld(N, DM_D1(B));		int i=0;

    MatDiff(1, nB, DM_Loc(AX), DM_Loc(B), &Enrm_inf);
    AbsMaxElt(1, (FDATA *) &Enrm_inf, &i);      

    MatZero(1, nB, DM_pLoc(AX)); 
    MatDiff(1, nB, DM_Loc(AX), DM_Loc(B), &Bnrm_inf);
    AbsMaxElt(1, (FDATA *) &Bnrm_inf, &i);
  } 
  BrdMat(0, d_P0(DM_D0(B)), 1, 1, (FDATA *) &Enrm_inf, 1);
  BrdMat(0, d_P0(DM_D0(B)), 1, 1, (FDATA *) &Bnrm_inf, 1);
  Resid = Enrm_inf / (Bnrm_inf * N * EPS); 
  DisposeMat(AX);
  return (Resid);
} /* addfit_ResidCheck() */ 
#endif /*ADDFIT_RESID_CHECK*/


static void addfit_bc_result(int N, DistMat X, FDATA *x)
/* pre:  x contains storage for N elements
 * post: copies all elements of X into x on all cells
 */ 
{ DistMat_RVAlloc(2);			
  DistMat sX = Trans(NewMat(1, N, Trans(ReduceDimMat(-1,  X)), x));
  D_ScaleAdd(N, 1, FD_1, X, FD_0, sX);
} /* addfit_bc_result() */


#define I AbsCellId[0]		/* from FDCommun.h*/
#define J AbsCellId[1]

int addfit_solve(FDATA *a, FDATA *b, int N, FDATA *x, int P, int Q, int r, int w)
{ int return_val = 0;
  if (CommunSetup(P, Q) == SOLVE_COMM_ERROR)
    return (SOLVE_COMM_ERROR);

  if (IN_SOLVE_GROUP) {
    DistMat_RVAlloc(5);
    DistMat A, B, X;			/* main block-cyclic matrix descriptors*/
#ifdef ADDFIT_RESID_CHECK
    DistMat cA; 			/* copy of A used for checking */
#endif    
    FDATA *WS=NULL, *cWS=NULL;		/* workspaces for A & X, cA */
    int WSSize;				/* size of workspaces for A & X */
    int ma, na; 			/* local matrix storage sizes */
    int mx, nx; 			/* local dimensions of X */ 
    
    int *Piv=NULL;			/* LDLT permutation vectors */
    FDATA *DiagInv=NULL;		/* LDLT storage for col-replic D^-1 */  
    int info=0;				/* LDLT error code */
    FD_BASE Resid; 			/* for residual checking */

    /* set up DBLAS for communication on these cells */

    MPICommSet((int) SolveComm);
    GridInit(P, Q); 
    CommSetUp(P, Q);

    /* set up main block-cyclic workspaces; append B to last row of A */
    A = InitDimMat(0, r, r, I, P, J, Q, 0, 0);
    ma = G2Ld(N+1, DM_D0(A)); 		na = G2Ld(N+1, DM_D1(A));
    X = InitDimMat(0, r, 1, I, P, J, Q, 0, 0);
    mx = G2Ld(N,   DM_D0(X)); 		nx = G2Ld(1,   DM_D1(X));
    MatMalloc("addfit_solve", 1, &WS, MAX(ma, 2) * MAX(na, 2) + mx*nx);    

    LM_Init(DM_Loc(A), WS, 1, MAX(ma, 2));
    LM_Init(DM_Loc(X), WS+MAX(ma, 2) * MAX(na, 2), 1, MAX(mx, 2));
    B = SubMat(A, N, 0);

    /* initialize A from a and b*/
    addfit_init(N, a, b, A);

#ifdef ADDFIT_RESID_CHECK
    /* create a copy of A for checking */     
    MatMalloc("addfit_solve", 0, &cWS, MAX(ma, 2) * MAX(na, 2));
    if (cWS!=NULL) {
      /* create a copy of A for checking */
      cA = Trans(NewMat(N+1, N+1, Trans(A), cWS));	/* force col-major */
      D_ScaleAdd(N+1, N, FD_1, A, FD_0, cA);
    }
#endif

    /* perform the solve */
    MatMalloc("addfit_solve", 1, &DiagInv, 3*G2Ld(N, DM_DM(A)));    
    Piv = (int *) malloc(N*sizeof(int));
    D_COPY(N, Trans(B), X);
    DM_PRINT("X", N, 1, X);
    D_SYTRF(N, A, DiagInv, Piv, &info, w);
    if (info > 0) { 
      if (SolveRank==0)
        printf("DDSYTRF: non-invertible matrix, info=%d\n", info);
      return_val = NON_INV_A;
    } else
      D_SYTRS(N, 1, A, DiagInv, Piv, X, &info);
    free((char *) Piv);
    MatFree(DiagInv);

#ifdef ADDFIT_RESID_CHECK
    if (cWS==NULL) {
      /* previous alloc for cA has failed; can now re-use WS and re-init */
      cA = Trans(NewMat(N+1, N+1, Trans(A), WS));
      addfit_init(N, a, b, cA);
    }   
    Resid = addfit_ResidCheck(P, Q, N, cA, X, SubMat(cA, N, 0)); 
    if (SolveRank==0)
      printf("addfit_solve: residual=%G\n", Resid);
    MatFree(cWS);
    if (Resid > ADDFIT_MAX_RESID && !return_val)
      return_val = RESID_FAIL; 
#endif

    /* b/c fully replicated solution vector x */
    addfit_bc_result(N, X, x);
    
    MatFree(WS);
  }
  
  CommunFree(&return_val);
  return(return_val);
} /* addfit_solve() */                 
